/*
 *   Copyright (c) 1995 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* This program converts JPEG Movie File Format movies from
 * "old style" to "new style".  That is, movies that were
 * created before Q tables were correctly interpreted by the
 * Parallax hardware library are "old style" movies.
 * This program fixes the movie by halving the Q factor stored
 * in the movie's header.  It turns out that and old style movie
 * recorded at a Q of 100 looks fine if played back on a new server
 * with a Q of 50.
 * The program opens the movie specified on the command line, examines
 * the movie header to determine the Q factor, halves that value, and
 * writes the new header in place.
 * An optional 2nd argument allows the person running the program to
 * specify his/her own Q factor to be written to the movie, thereby
 * overriding the default behavior of this program.
 */
/* $Log: fixMovieQ.c,v $
 * Revision 1.1  1995/10/12  18:55:43  drapeau
 * Initial revision
 * */
static char fixMovieQRcsId[] = "$Id: fixMovieQ.c,v 1.1 1995/10/12 18:55:43 drapeau Exp $";

#include <stdio.h>
#include <movie.h>
#include <errno.h>

int main(int argc, char** argv)
{
  int		bytesRead;
  char*		fileName;
  int		qFactor = 0;
  jpheader	movieHeader;
  int		result;
  char		tempString[32];
  FILE*		theMovie;
  
  if (argc < 2)							    /* Check if enough arguments were specified */
  {
    printf("Usage: %s <movieFileName> [qFactor].\n\n", argv[0]);
    printf("The 'qFactor' argument is optional; if specified, this program\n");
    printf("will set the movie's Q factor to the command-line argument.\n");
    printf("By default, the 'qFactor' argument need not be specified if converting\n");
    printf("an old-style JPEG Movie into a new-style JPEG Movie.\n");
    exit(-1);
  }
  if (argc > 2)							    /* Was the option chosen to force Q factor doubling? */
  {								    /* Yes, set a variable indicating this */
    qFactor = atoi(argv[2]);
    if (qFactor > 1000)
    {
      printf("Warning: the maximum allowable Q factor is 1000; program will proceed with Q of 1000.\n");
      qFactor = 1000;
    }
  }
  fileName = argv[1];
  printf("Fixing movie ::%s::.\n", fileName);
  theMovie = fopen(fileName, "r+");				    /* Attempt to open the movie specified on the command line */
  if ((FILE*)NULL == theMovie)					    /* Did the attempt fail for some reason? */
  {								    /* Yes, print an error message and exit right now */
    printf("Could not open the movie you specified for reading and/or writing.\n");
    exit(1);
  }
  memset(tempString, 0, 10);					    /* Initialize the temporary string */
  bytesRead = fread(tempString, 1, strlen("j_movie"), theMovie);    /* Read in the 1st few bytes of the requested file */
  if (strcmp(tempString, "j_movie") != 0)			    /* Is this a JPEG Movie file?  If not, report the error... */
  {								    /* ...and quit now. */
    printf("The file %s is not a JPEG Movie file.\n", fileName);
    exit(2);
  }
  fseek(theMovie, 0L, 0);					    /* Go back to the beginning of the movie file */
  fread((char*)&movieHeader, sizeof(jpheader), 1, theMovie);	    /* Read in the movie's header information */
  if (0 == qFactor)						    /* Was a Q factor specified on the command line? */
  {								    /* No, use the default method of adjusting Q factor */
    printf("Read in the movie's header.  Q factor was %d, changing to %d.\n",
	   movieHeader.qfactor, movieHeader.qfactor / 2);
  }
  else								    /* Yes, use the command-line-specified Q factor */
  {
    printf("Read in the movie's header.  Q factor was %d, changing to %d.\n",
	   movieHeader.qfactor, qFactor);
  }
  if (qFactor != 0)						    /* Was a Q factor specified on the command line?  */
    movieHeader.qfactor = qFactor;				    /* Yes, override the default, using user-specified Q factor */
  else
    movieHeader.qfactor = movieHeader.qfactor / 2;		    /* Use the default Q factor adjustment */
  fseek(theMovie, 0L, 0);					    /* Go back to the beginning of the movie file */
  result = fwrite(&movieHeader, sizeof(jpheader), 1, theMovie);	    /* Write the modified movie header to disk */
  if (0 == result)						    /* Determine whether updating the header was successful */
  {								    /* A problem occurred; report the error and exit now */
    printf("An error occurred in writing the movie header; the movie is unchanged (errno was %d).\n",
	   errno);
    exit(4);
  }
  fclose(theMovie);
  printf("Modifications complete.  Exiting the program.\n");
  exit(0);
} /* end program */
