/*
 *   Copyright (c) 1993 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* This program creates a video input widget and a JPEG widget.  The purpose of
 * the program is to show how to grab compressed images from a video source at
 * the fastest speed possible.
 * The program takes the following arguments:
 * - howOftenToPrintStatistics: every so many frames, the program will print
 *   performance statistics.  This argument indicates how many frames between
 *   reports.
 * - howManyFramesToGrab: how many frames to grab before exiting the test.
 * - dimensions of the source rectangle to compress (x, y, width, height).
 * - sleepTime: how long to sleep (in microseconds) before getting and putting
 *   the next frame.
 * 
 * Usage: jpegContinuousTest howOftenToPrintStats howManyFramesToGrab x y width height sleepTime
 */
/* $Header: /plx/source/Xclients/generic/RCS/jpegContinuousTest.c,v 1.18 1995/09/28 06:44:05 drapeau Exp $ */
/* $Log: jpegContinuousTest.c,v $
 * Revision 1.18  1995/09/28  06:44:05  drapeau
 * Fixed an error in the code: the XRectangle's used should not be
 * automatic variables, since they cannot afford to go out of scope.  The
 * JPEG widget points to the space allocated by these variables.
 *
 * Revision 1.17  1995/09/07  21:27:42  drapeau
 * Removed extraneous diagnostic messages that only get in the way of
 * performance.
 *
 * Revision 1.16  1994/10/22  23:36:06  drapeau
 * The select() system call is not portable across Unix platforms; the
 * type casts passed in were removed to eliminate compiler complaints on
 * some platforms.
 *
 * Revision 1.15  1994/05/10  03:34:56  drapeau
 * Fixed the test program so that it works with any video signal standard
 * automatically, not just NTSC.
 *
 * Revision 1.14  1993/11/04  21:48:14  drapeau
 * Cleaned up the code so that it no longer stills video before trying
 * to compress it.  This was the original intent of the program anyway.
 *
 * Revision 1.13  1993/10/22  22:04:18  drapeau
 * Rewrote the sample program so that it behaves more in accord with Xt
 * recommended programming practices.  Recommendations say that timer
 * functions should be a brief as possible, so this program was written
 * to be called as often as possible as a WorkProc.  The WorkProc now
 * only gets and puts a single frame, instead of staying the loop
 * continuously.
 *
 * Revision 1.12  1993/10/21  22:41:47  drapeau
 * Made code more portable by replacing a call to usleep() with a call to select().
 *
 * Revision 1.11  1993/10/21  16:26:43  drapeau
 * Added an extra parameter to tell the application how long to sleep after
 * each compression-decompression sequence.
 *
 * Revision 1.10  93/09/10  12:36:58  drapeau
 * First public version of this test program.
 *  */
static char jpegContinuousTestRcsId[] = "$Header: /plx/source/Xclients/generic/RCS/jpegContinuousTest.c,v 1.18 1995/09/28 06:44:05 drapeau Exp $";

#include <unistd.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <XPlxExt.h>
#include <XVideoIn.h>
#include <JPEG.h>

static XrmOptionDescRec options[] = {
	{"-channel",		"*videoIn.channel",			XrmoptionSepArg,	NULL, },
	{"-sourceWin",		"*jpeg.sourceWin",			XrmoptionSepArg,	NULL, },
};

Widget top, videoShell, jpegShell, videoWidget, jpegWidget;
Boolean	TimerFunction(XtPointer);
void	PrintCompressionStatistics(struct timeval*, int);

int	howOftenToPrintStatistics;
int	howManyFramesToGrab;
unsigned int	sleepTime;

void main(int argc, char** argv)
{
  XtAppContext		appContext;
  XtWorkProcId		timerID;
  Window		windowID;
  Dimension		width;
  Dimension		height;
  static XRectangle	compressSourceRect;
  static XRectangle	compressDestRect;
  
  top = XtAppInitialize(&appContext, "JPEGTest", options, XtNumber(options),
			&argc, argv,
			NULL, (ArgList)NULL, 0);
  if (argc > 1)
  {
    printf("Argument was specified: ::%s::.\n", argv[1]);
    howOftenToPrintStatistics = (int)atoi(argv[1]);
  }
  else
    howOftenToPrintStatistics = 10;
  if (argc > 2)
    howManyFramesToGrab = (int)atoi(argv[2]);
  else
    howManyFramesToGrab = 1000000;
  videoWidget = XtVaCreateManagedWidget("videoIn",		    /* Create a VideoIn widget, used for displaying... */
					xVideoInWidgetClass,	    /* ...digitized video on-screen. */
					top,
					NULL);
  XtRealizeWidget(top);
  XtVaGetValues(videoWidget,
		XtNwidth, &width,
		XtNheight, &height,
		NULL);
  if (argc > 7)
  {
    compressSourceRect.x = (int)atoi(argv[3]);
    compressSourceRect.y = (int)atoi(argv[4]);
    compressSourceRect.width = (int)atoi(argv[5]);
    compressSourceRect.height = (int)atoi(argv[6]);
    sleepTime = (int)atoi(argv[7]);
  }
  else
  {
    compressSourceRect.x = 0;
    compressSourceRect.y = 0;
    compressSourceRect.width = width;
    compressSourceRect.height = height;
    sleepTime = 100000;
  }
  jpegShell = XtVaCreateManagedWidget ("jpegShell",		    /* Create a shell to hold the JPEG compression widget */
				       applicationShellWidgetClass, 
				       top,
				       XtNwidth, width,
				       XtNheight, height,
				       XtNx, (Position)width,
				       XtNy, (Position)height,
				       NULL);
  compressDestRect.x = 0;
  compressDestRect.y = 0;
  compressDestRect.width = width;
  compressDestRect.height = height;
  jpegWidget = XtVaCreateManagedWidget("jpeg",			    /* Create a JPEG widget, used for compressing images and... */
				       jpegWidgetClass,		    /* ...displaying compressed images */
				       jpegShell,
				       XtNsourceWin,		    /* Tell the JPEG widget where to get fresh image data */
				       XtWindow(videoWidget),
				       XtNcompressSourceRect,
				       &compressSourceRect,
				       XtNcompressDestRect,
				       &compressDestRect,
				       NULL);
  XtRealizeWidget(jpegShell);
  XtVaSetValues(jpegWidget,
		XtNsourceWin, XtWindow(videoWidget),
		NULL);
  timerID = XtAppAddWorkProc(appContext, TimerFunction,		    /* Start doing continuous compression as soon as possible */
			     (XtPointer)NULL);
  XtAppMainLoop(appContext);
}								    /* end function main */



Boolean TimerFunction(XtPointer clientData)
{
  static int		counter = 0;
  int			result;
  struct timeval	before;
  struct timeval	waitTime;
  XPlxCImage*		jpegData;
  
  if (counter == 0)						    /* Is this the first time the function has been called? */
    result = gettimeofday(&before, (struct timezone*)NULL);	    /* Yes, store current time of day for later stats reporting */
  if (counter < howManyFramesToGrab)				    /* Compress as many times as asked for */
  {
    if (((counter % howOftenToPrintStatistics) == 0) &&		    /* Get stats after every so many compressions */
	(counter > 0))
      PrintCompressionStatistics(&before, howOftenToPrintStatistics);
    XtVaSetValues(jpegWidget,					    /* Tell the JPEG widget to compress another image right now */
		  XtNgetNewImage, True,
		  NULL);
    if (sleepTime > 0)
    {
      waitTime.tv_sec = 0;
      waitTime.tv_usec = sleepTime;
      select(sysconf(_SC_OPEN_MAX),
	     0, 0, 0,
	     &waitTime);
    }
    XtVaSetValues(jpegWidget,					    /* Tell the JPEG widget to compress another image right now */
		  XtNputNewImage, True,				    /* Also tell the JPEG widget to display the new image now */
		  NULL);
    XtVaGetValues(jpegWidget,
		  XtNjpegData, &jpegData,
		  NULL);
  }
  if (counter < howManyFramesToGrab)				    /* Are there still some frames to grab? */
  {
    counter++;							    /* Yes, note that one more frame was grabbed */
    return(False);						    /* Tell the toolkit to call this function again */
  }
  else								    /* No, done grabbing all the frames */
  {
    printf("%d done; I'm outtahere.\n", howManyFramesToGrab);
    XtDestroyWidget(top);
    exit(0);
  }
} /* end function TimerFunction */


long DeltaMSec(struct timeval	*tp1,
	       struct timeval	*tp2)
{
     return (tp2->tv_sec - tp1->tv_sec) * 1000 +
	  (tp2->tv_usec - tp1->tv_usec) / 1000;
}


void PrintCompressionStatistics(struct timeval* before,
				int		numberOfGrabs)
{
  int			result;
  long			grabsPerSecond;
  long			totalTime;
  long			timeForOneGrab;
  struct timeval	after;
  
  result = gettimeofday(&after, (struct timezone*)NULL);
  totalTime = DeltaMSec(before, &after);
  timeForOneGrab = totalTime / numberOfGrabs;
  grabsPerSecond = 1000 / timeForOneGrab;
  printf("Total time taken for %d frames is %d.%d seconds.\n",
	 numberOfGrabs,
	 totalTime / 1000,
	 totalTime % 1000);
  printf("Time taken for 1 frame is %d.%d seconds.\n",
	 timeForOneGrab / 1000,
	 timeForOneGrab % 1000);
  printf("Number of frames per second: %d.\n", grabsPerSecond);
  result = gettimeofday(before, (struct timezone*)NULL);	    /* Store the current time of day for later stats reporting */
  return;
} /* end function PrintCompressionStatistics */
