/*
 *   Copyright (c) 1993, 1995 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* The purpose of this program is to show the ability to digitize images
 * from the screen and save them as Parallax JPEG Movie data on disk.
 * The images are digitized from an incoming video signal, then compressed,
 * then saved to disk.
 * The program takes several arguments:
 * - the number of frames to be recorded and saved to disk
 * - the width and height of the image to be digitized and compressed
 * - the name of the disk file into which the JPEG data will be written.
 * The program creates a VideoIn widget and a JPEG widget, then informs the
 * JPEG widget that it should capture image data from the VideoIn widget (the
 * JPEG widget can point to any window, but in practice only 24-bit windows
 * provide the JPEG widget with workable data; for this example, it seemed
 * reasonable to use video as the best source of 24-bit data).
 * The program then sets up a function to begin work after a short delay.
 * This timer function does all of the work, doing the following:
 * - Before beginning work, it saves the current time-of-day for later
 *   statistics purposes.
 * - For each frame to be saved, the program tells the JPEG widget to digitize
 *   and compress a new frame, then writes that JPEG Movie data to disk.
 * - After compressing and saving all the frames requested, the program
 *   gathers statistics to indicate performance of the widget, then reports
 *   these statistics before exiting.
 *
 * Usage: jpegTimingTest <numberOfFramesToBeSaved> <width> <height> <fileName>
 */
/* $Header: /plx/source/Xclients/generic/RCS/jpegTimingTest.c,v 1.13 1995/08/17 20:10:56 drapeau Exp $ */
/* $Log: jpegTimingTest.c,v $
 * Revision 1.13  1995/08/17  20:10:56  drapeau
 * Added better support for timing of the JPEG frame capture process.
 *
 * Revision 1.12  1995/07/28  15:53:10  drapeau
 * Enhanced the program to produce Parallax JPEG Movie files as opposed to
 * a file of raw JPEG images (the corresponnding program "jpegPlaybackTest"
 * has also been enhanced to accept JPEG Movie files).
 *
 * Revision 1.11  1993/11/17  02:31:37  drapeau
 * No real modifications; only small cosmetic changes.
 *
 * Revision 1.10  93/09/10  12:37:35  drapeau
 * First public version of this test program.
 *  */
static char jpegTimingTestRcsId[] = "$Header: /plx/source/Xclients/generic/RCS/jpegTimingTest.c,v 1.13 1995/08/17 20:10:56 drapeau Exp $";

typedef struct {
	unsigned	sample_rate;		/* samples per second */
	unsigned	samples_per_unit;	/* samples per unit */
	unsigned	bytes_per_unit;		/* bytes per sample unit */
	unsigned	channels;		/* # of interleaved channels */
	unsigned	encoding;		/* data encoding format */
	unsigned	data_size;		/* length of data (optional) */
} Audio_hdr;

#include <stdio.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <XPlxExt.h>
#include <XVideoIn.h>
#include <JPEG.h>
#include <movie.h>


Widget top, videoShell, jpegShell, videoWidget, jpegWidget;
void	TimerFunction(XtPointer, XtIntervalId*);
long	ElapsedTime(struct timeval*, struct timeval*);


int	 	totalNumberOfGrabs = 10;
int		outputFile;
Boolean		storeCompressedDataToDisk;
char*		fileName;
XRectangle	myRect;
jpheader	movieHeader;
int*		frameIndex;
long*		frameTimes;
int		numberBytesWritten;

 
void main(int argc, char** argv)
{
  XtAppContext	appContext;
  int		bytesWritten;
  XtIntervalId	timerID;
  
  top = XtAppInitialize(&appContext, "JPEGTest", NULL, 0,
			       &argc, argv,
			       NULL, (ArgList)NULL, 0);
  if (argc < 5)							    /* Were enough arguments supplied? */
  {								    /* No, print an error message and exit immediately */
    printf("Usage: %s <numberOfFramesToBeSaved> <width> <height> <fileName>.\n",
	   argv[0]);
    exit(-1);
  }
  frameIndex = (int*)NULL;
  totalNumberOfGrabs = (int)atoi(argv[1]);			    /* Determine how many frames to compress and save to disk */
  myRect.x = 0;
  myRect.y = 0;
  myRect.width = (unsigned short)atoi(argv[2]);			    /* Determine the dimensions of the source image */
  myRect.height = (unsigned short)atoi(argv[3]);
  fileName = argv[4];						    /* Determine file name into which to store compressed data */
  if (argc > 5)							    /* Was an option specified to skip writing data to disk? */
  {								    /* Yes, set the appropriate variable and report this */
    printf("Will not store compressed data to disk.\n");
    storeCompressedDataToDisk = False;
  }
  else								    /* No, normal operation: store data to disk */
  {
    storeCompressedDataToDisk = True;
    outputFile = open(fileName,					    /* Try to create a new file for writing data to disk */
		      (O_CREAT | O_WRONLY | O_TRUNC),
		      0644);
    if (outputFile == -1)					    /* Did the open fail? */
    {								    /* Yes, report the error and exit this program without... */
      printf("Could not open the file %s for writing.  Exiting.\n", /* ...doing anything more. */
	     fileName);
      exit(outputFile);
    }
    memset((void*)&movieHeader, 0, sizeof(jpheader));		    /* Initialize a movie header for this recording process */
    strcpy(movieHeader.magic, "j_movie");
    movieHeader.version = JPEG_VERSION_3;
    movieHeader.fps = 30;					    /* Hard-code 30 fps as a movie frame rate, just for kicks */
    movieHeader.width = myRect.width;
    movieHeader.height = myRect.height;
    movieHeader.tracks = 0;
    movieHeader.audioslice = 0;
    frameIndex = (int*)malloc(totalNumberOfGrabs * sizeof(int));    /* Allocate space to store byte offsets for each frame */
    if ((int*)NULL == frameIndex)
    {
      printf("Could not allocate space for storing byte offsets for %d frame's worth of movie.  Exiting.\n",
	     totalNumberOfGrabs);
      exit(-1);
    }
    memset(frameIndex, 0, (totalNumberOfGrabs * sizeof(int)));	    /* Initialize the frame index to zeros */
    frameTimes =						    /* Allocate space to store time taken to process each frame */
      (long*)malloc(totalNumberOfGrabs * sizeof(long));
    if ((long*)NULL == frameTimes)
    {
      printf("Could not allocate space for storing times to write each frame to disk.  Exiting.\n");
      exit(-1);
    }
    memset(frameTimes, 0, (totalNumberOfGrabs * sizeof(long)));	    /* Initialize the frame times array to zeros */
    bytesWritten = write(outputFile,				    /* Write an initial movie header to the file */
			 &movieHeader,
			 sizeof(jpheader));
    if (bytesWritten < 0)					    /* Did the attempt fail? */
    {								    /* Yes, report the error and exit immediately */
      printf("Could not write an updated movie header to the movie file.  Exiting.\n");
      close(outputFile);
      unlink(fileName);						    /* Remove the bad movie */
      exit(-1);
    }
    numberBytesWritten = sizeof(jpheader);			    /* The 1st frame begins just after the movie header */
  }
  XtVaSetValues(top,						    /* Resize the top-level widget to fit the size of the... */
		XtNwidth, (Dimension)myRect.width,		    /* ...images to be digitized and compressed. */
		XtNheight, (Dimension)myRect.height,
		NULL);
  videoWidget = XtVaCreateManagedWidget("videoIn",		    /* Create a widget for displaying video */
					xVideoInWidgetClass, 
					top,
					NULL);
  XtRealizeWidget(top);
  jpegShell = XtVaCreateManagedWidget ("jpegShell",
				       applicationShellWidgetClass, 
				       top,
				       NULL);
  jpegWidget = XtVaCreateManagedWidget("jpeg", jpegWidgetClass,	    /* Create a widget for compressing JPEG image data */
				       jpegShell,
				       XtNcompressSourceRect,
				       (XtPointer)&myRect,
				       XtNcompressDestRect,
				       (XtPointer)&myRect,
				       XtNsourceWin,
				       XtWindow(videoWidget),
				       NULL);
  XtRealizeWidget(jpegShell);
  timerID = XtAppAddTimeOut(appContext, 2500,			    /* In a little while, start doing the actual work */
			    TimerFunction, (XtPointer)NULL);
  XtAppMainLoop(appContext);
}								    /* end function main */



void TimerFunction(XtPointer clientData, XtIntervalId* id)
{
  int			bytesWritten;
  int			counter;
  int			result;
  int			grabsPerSecond;
  int			averageDiskBandwidth = 0;
  int			largestFrameSize = 0;
  int			smallestFrameSize = 0;
  int			qFactor;
  static Boolean	firstTime = True;
  Boolean		useSharedMemory;
  struct timeval	preGrab;
  struct timeval	postGrab;
  struct timeval	before;
  struct timeval	after;
  struct timeval	totalTime;
  struct timeval	timeForOneGrab;
  XPlxCImage*		jpegDataPtr;
  char			instruction;
  
  printf("Entered TimerFunction.\n\n\n");
  XtVaSetValues(jpegWidget,					    /* Tell the JPEG widget not to bother trying to draw the... */
		XtNdestWin, 0,					    /* ...images it just compressed.  This will save time and... */
		NULL);						    /* ...allow the program to write data to disk more quickly. */
  XtVaGetValues(jpegWidget,					    /* Determine if the JPEG widget is using shared memory... */
		XtNuseSharedMemory, &useSharedMemory,		    /* ...for speed purposes.  If so, point to the shared... */
		XtNjpegData, &jpegDataPtr,			    /* ...memory segment since it will be recycled. */
		XtNqFactor, &qFactor,				    /* Also get the current Q factor used by the JPEG widget */
		NULL);
  movieHeader.qfactor = qFactor;
  result = gettimeofday(&before, (struct timezone*)NULL);	    /* Determine the current time-of-day for statistics purposes */
  for (counter = 0; counter < totalNumberOfGrabs; counter++)	    /* Loop once through for each image to be saved */
  {
    frameIndex[counter] = numberBytesWritten;			    /* Mark the starting point of this frame */
/*    printf("Debugging: frameIndex[%d] is %d.\n", counter, frameIndex[counter]);*/
    XtVaSetValues(jpegWidget,					    /* Tell the JPEG widget to compress another image right now */
		  XtNgetNewImage, True,
		  NULL);
    if (useSharedMemory == False)				    /* Is the widget using shared memory for storing images? */
      XtVaGetValues(jpegWidget,					    /* No, explicitly point to the newest JPEG data (if the... */
		    XtNjpegData, &jpegDataPtr,			    /* ...widget isn't using shared memory, each new image... */
		    NULL);					    /* ...points to a different JPEG data structure). */
    if (storeCompressedDataToDisk == True)			    /* Is this test supposed to write JPEG data to disk? */
    {								    /* Yes, write one image, if possible, to disk. */
      result = gettimeofday(&preGrab, (struct timezone*)NULL);	    /* Mark time before writing information to disk */
      instruction = LOAD_JPEG;
      bytesWritten = write(outputFile, &instruction, 1);		    /* Try to write the opcode to disk */
      if (bytesWritten < 0)					    /* Did the attempt fail? */
      {								    /* Yes, report the error and return a failed status */
	printf("There was an error recording frame %d's JPEG image to disk.  Recording stopped.\n", counter);
	exit(-1);
      }
      numberBytesWritten++;					    /* Indicate that the single character was written */
      bytesWritten = write(outputFile, &(jpegDataPtr->size),	    /* Write the size in bytes of this frame's JPEG image data */
			   sizeof(int));
      if (bytesWritten < sizeof(int))				    /* Did the attempt fail? */
      {								    /* Yes, report the error and return a failed status */
	printf("There was an error recording frame %d's JPEG image size to disk.  Recording stopped.\n", counter);
	exit(-1);
      }
      numberBytesWritten += sizeof(int);			    /* Update total number of bytes written to movie file */
      bytesWritten = write(outputFile, jpegDataPtr->data,	    /* Next, write the image data itself */
			   jpegDataPtr->size);
      if (bytesWritten < 0)					    /* Did the attempt fail? */
      {								    /* Yes, report the error and return a failed status */
	printf("There was an error recording frame %d's JPEG image data to disk.  Recording stopped.\n", counter);
	exit(-1);
      }
      numberBytesWritten += jpegDataPtr->size;			    /* Note that an image's worth of data has been written */
      instruction = END_FRAME;					    /* That's the end of this frame of movie */
      bytesWritten = write(outputFile, &instruction, 1);	    /* Try to write the opcode to disk */
      if (bytesWritten < 0)					    /* Did the attempt fail? */
      {								    /* Yes, report the error and return a failed status */
	printf("There was an error recording frame %d's END_FRAME instruction to disk.  Recording stopped.\n", counter);
	exit(-1);
      }
      numberBytesWritten++;					    /* Indicate that the single character was written */
      result = gettimeofday(&postGrab, (struct timezone*)NULL);	    /* Mark time after writing information to disk */
      frameTimes[counter] = ElapsedTime(&preGrab, &postGrab);	    /* Note time taken to store information to disk */
    } /* end if storing compressed data to disk */
    if (firstTime == True)					    /* Is this the first time through the compression loop? */
    {								    /* Yes, initialize the smallest frame size so far seen to... */
      smallestFrameSize = jpegDataPtr->size;			    /* ...be the first frame compressed. */
      firstTime = False;
    }
    if (jpegDataPtr->size > largestFrameSize)			    /* Update largest and smallest compressed frames seen...  */
      largestFrameSize = jpegDataPtr->size;			    /* ...so far. */
    if (jpegDataPtr->size < smallestFrameSize)
      smallestFrameSize = jpegDataPtr->size;
  } /* end for... */
  result = gettimeofday(&after, (struct timezone*)NULL);	    /* Get the current time after all this work has been done */
  before.tv_usec += (before.tv_sec * 1000000);			    /* Convert this timeval to all microseconds */
  after.tv_usec += (after.tv_sec * 1000000);			    /* Do the same for this timeval */
  totalTime.tv_usec = after.tv_usec - before.tv_usec;		    /* Determine the total time taken for the above compressions */
  averageDiskBandwidth = (int)((float)numberBytesWritten /
			       (float)(totalTime.tv_usec / 1000000));
  timeForOneGrab.tv_usec = (totalTime.tv_usec / totalNumberOfGrabs);
  grabsPerSecond = (1000000 / timeForOneGrab.tv_usec);
  movieHeader.fps = grabsPerSecond;				    /* Set the frame rate of this movie to the speed grabbed */
  movieHeader.frames = totalNumberOfGrabs;			    /* Store the number of frames in this movie */
  movieHeader.firstValidFrame = 0;				    /* Mark the whole movie as valid for playback */
  movieHeader.lastValidFrame = totalNumberOfGrabs - 1;
  movieHeader.indexbuf = numberBytesWritten;			    /* Mark the location of the frame index information */
  bytesWritten = write(outputFile,				    /* Attempt to write the frame index info to disk */
		       frameIndex,
		       totalNumberOfGrabs * sizeof(int));
  if (bytesWritten < (totalNumberOfGrabs * sizeof(int)))	    /* Did the attempt fail? */
  {								    /* Yes, print an error message, remove the movie file, exit */
    printf("Could not write movie frame index information.  Exiting.\n");
    close(outputFile);
    unlink(fileName);						    /* Remove the bad movie */
    exit(-1);
  }
  movieHeader.bandwidth =					    /* Store average data rate for this movie, in Kilobytes */
    ((numberBytesWritten /totalNumberOfGrabs) *
     movieHeader.fps) / 1024;
  lseek(outputFile, 0L, SEEK_SET);				    /* Go back to the start of the movie to write updated... */
  bytesWritten = write(outputFile, &movieHeader, sizeof(jpheader)); /* ...movie header information. */
  if (bytesWritten < 0)						    /* Did the attempt fail? */
  {								    /* Yes, report the error and return immediately */
    printf("Could not write an updated movie header to the movie file.  Exiting.\n");
    close(outputFile);
    unlink(fileName);						    /* Remove the bad movie */
    exit(-1);
  }
  close(outputFile);						    /* Done writing JPEG data to disk */
  timeForOneGrab.tv_sec = (timeForOneGrab.tv_usec / 1000000);	    /* Convert this timeval back into seconds and microseconds */
  timeForOneGrab.tv_usec = (timeForOneGrab.tv_usec % 1000000);
  totalTime.tv_sec = (totalTime.tv_usec / 1000000);		    /* Convert this timeval back into seconds and microseconds */
  totalTime.tv_usec = (totalTime.tv_usec % 1000000);
  
  printf("Number of grabs per second:\t\t\t%d\n", grabsPerSecond);
  printf("Average compression ratio:\t\t\t%d:1\n",
	 (myRect.width * myRect.height * 3 * totalNumberOfGrabs) /
	 numberBytesWritten);
  printf("Average size of each frame:\t\t\t%d bytes.\n",
	 numberBytesWritten / totalNumberOfGrabs);
  printf("Average disk bandwidth:\t\t\t\t%d bytes per second.\n",
	 averageDiskBandwidth);
  printf("------------------------------------------------------------------------\n");
  printf("Total time taken for %d grabs:\t\t\t%d.%d seconds.\n",
	 totalNumberOfGrabs,
	 totalTime.tv_sec,
	 totalTime.tv_usec);
  printf("Average time taken for 1 grab:\t\t\t%d.%d seconds.\n",
	 timeForOneGrab.tv_sec,
	 timeForOneGrab.tv_usec);
  printf("Total number of bytes written to disk:\t\t%d bytes.\n", numberBytesWritten);
  printf("Smallest frame in bytes (after compression):\t%d bytes.\n",
	 smallestFrameSize);
  printf("Largest frame in bytes (after compression):\t%d bytes.\n",
	 largestFrameSize);
  printf("Q Factor:\t\t\t\t\t%d.\n", qFactor);
  printf("------------------------------------------------------------------------\n");
  printf("Time taken for disk I/O for each frame (only times greater than 1 msec are reported):\n");
  for (counter = 0; counter < totalNumberOfGrabs; counter++)
  {
    if (frameTimes[counter] > 0)
    {
      printf("Frame %d took %d milliseconds.\n", counter, frameTimes[counter]);
    }
  }
  XtDestroyWidget(top);
  exit(0);
  return;
} /* end function TimerFunction */



long	ElapsedTime(struct timeval*	before,
		    struct timeval*	after)
{
  return(((after->tv_sec - before->tv_sec) * 1000) +		    /* Convert the seconds part to milliseconds */
	 ((after->tv_usec - before->tv_usec) / 1000));		    /* Convert the microseconds part to milliseconds */
} /* end function ElapsedTime */
