/*
 * (c) Copyright 1991 Parallax Graphics Inc, All Rights Reserved
 *
 * Parallax Graphics, Inc.
 * 2500 Condensa Street
 * Santa Clara, California 95051
 *
 * $Log: XPlxCompress.c,v $
 * Revision 1.23  1996/01/17  19:44:21  drapeau
 * Removed a debugging message that was left in by mistake.
 *
 * Revision 1.22  1995/12/23  02:32:36  drapeau
 * Modified code to round-down to the nearest 16 pixel boundary on compress
 * widths.  Rounding down on height is unnecessary.
 *
 * Revision 1.21  1995/12/14  03:00:20  drapeau
 * Major changes to this module to accommodate the new XPlx library that
 * automatically detects the proper video extension to use.  The new
 * library attempts to use the ParallaxVideo extension if possible and
 * falls back to use the XVideo extension if need be.
 *
 * Revision 1.20  1995/06/23  19:26:12  drapeau
 * Fixed error in XPlxGetCImage(): changed round-down parameters from 8
 * to 16, according to information from Dave Arnold about the proper
 * round-down values to use when creating JPEG images that can be used
 * for software-only decoders.
 *
 * Revision 1.19  1995/02/24  03:10:11  drapeau
 * Two changes:
 *
 * 1) Added code to help this library work with multiple displays (when a
 *    process using this library is working with mulitple displays, it is
 *    not guaranteed that the Plx extension will be loaded in the same
 *    spot and therefore might occupy a different base event code for Plx
 *    protocol events.  So, this library needs to keep a mapping of
 *    Display-to-Protocol-Event-Base-Codes so the "PREAMBLE" macro will
 *    always work correctly and allow this library to communicate with
 *    the proper X extension no matter what the Display.)
 *
 * 2) Cosmetic changes to make the code more readable and editable.
 *
 * Revision 1.18  1994/10/22  02:23:30  drapeau
 * Fixed a function declaration for XPlxPutTable().
 *
 * Revision 1.17  1994/10/10  18:26:23  drapeau
 * Modified XPlxGetCImage() to conform to the JPEG baseline compression
 * specification, which states that images are compressed in 8x8 blocks.
 * This means that the width & height parameters of the XPlxGetCImage()
 * function need to be multiples of 8.  The function now enforces this
 * rule.
 *
 * Revision 1.16  1994/08/27  22:55:14  drapeau
 * ANSI-fied code, making function prototypes for better error checking.
 *
 * Revision 1.15  1994/04/22  01:33:50  drapeau
 * Made XPlxCreateCImage() more robust.  If the "data" pointer passed
 * into the function is NULL, this function will now allocate "size"
 * bytes for the caller and return the allocated structure.  If the
 * "data" pointer passed as argument is non-NULL, this function assumes
 * the caller is supplying its own memory into which to store compressed
 * image data.
 *
 * Revision 1.14  1993/11/20  01:17:38  drapeau
 * Cleaned up inclusion of header files to increase portability.
 *
 * Revision 1.13  1992/11/16  15:23:50  schen
 * Change the xPutImageReq of the SendImage to xPlxImageReq
 *
 * Revision 1.12  91/11/20  14:28:56  schen
 * *** empty log message ***
 * 
 * Revision 1.11  91/11/07  14:33:28  dave
 * use correct reply struct
 * 
 * Revision 1.10  91/11/07  14:30:26  dave
 * getimage now uses reply.size
 * 
 * Revision 1.9  91/11/06  15:12:40  dave
 * fixed planemask to have size in bytes
 * 
 * Revision 1.8  91/10/22  16:09:12  dave
 * fixed something
 * 
 * Revision 1.7  91/10/21  08:03:40  dave
 * load table now uses image request
 * 
 * Revision 1.6  91/10/19  16:15:57  dave
 * changed XPlxPutQtable to XPlxPutTable
 * 
 * Revision 1.5  91/09/17  10:31:03  dave
 * better checks for compressed image too large
 * 
 * Revision 1.4  91/08/16  14:37:02  dave
 * added XPlxPutQtable
 * 
 * Revision 1.3  91/08/03  14:09:59  eric
 * many changes
 * 
 * Revision 1.2  91/07/24  14:16:56  eric
 * added rcs header
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Id: XPlxCompress.c,v 1.23 1996/01/17 19:44:21 drapeau Exp $ Parallax Graphics Inc";
#endif

#define NEED_REPLIES
#include	<X11/Xlibint.h>
#include	<X11/Xutil.h>
#include	<plxvideo.h>

static int VideoReqCode = 0;
static int first_event, first_error;
extern Display*	plxLastOpenDpy;

/*
 * set a drawable to be of type video
 */

#define	PREAMBLE(result)	\
		LockDisplay(dpy); \
		result = XPlxFindExtension(dpy, &VideoReqCode);

#define POSTAMBLE	\
		UnlockDisplay(dpy); \
		SyncHandle();

/*
 * CreateImage
 * 
 * Allocates the memory necessary for an XPlxCImage data structure. 
 * Initializes the structure with "default" values and returns XPlxCImage. 
 * 
 */

XPlxCImage* XPlxCreateCImage(Display*		dpy,
			     char*		data,
			     int		size,
			     unsigned int	width,
			     unsigned int	height)
{
  register XPlxCImage *image;
  
  if ((image = (XPlxCImage *)
       Xcalloc(1, (unsigned) sizeof(XPlxCImage))) == NULL)
    return (XPlxCImage *) NULL;
  
  image->width = width;
  image->height = height;
  if (data != (char*)NULL)					    /* Did the programmer supply space to store image data? */
    image->data = data;						    /* Yes, use it */
  else								    /* No, allocate space for the programmer and use it */
  {
    image->data = Xcalloc(size, (unsigned)sizeof(char));	    /* Try to allocate the space. */
    if (image->data == NULL)					    /* Did the allocation fail? */
    {								    /* Yes, free space taken by the XPlxCImage structure and... */
      Xfree((char*)image);					    /* ...return a NULL structure as failure. */
      return((XPlxCImage*)NULL);
    }
  }
  image->size = size;
  return(image);
} /* end function XPlxCreateCImage */



/*
 * _DestroyImage
 * 	
 * Deallocates the memory associated with the xplxcimage data structure. 
 * this version handles the case of the image data being malloc'd
 * entirely by the library.
 */

int XPlxDestroyCImage(XPlxCImage*	ximage)
{
  if(ximage)
  {
    if(ximage->data)
      Xfree((char *)ximage->data);
    Xfree((char *)ximage);
  }
  return 1;
}



XPlxCImage* XPlxGetCImage(Display*	dpy,
			  Drawable	d,
			  GC		gc,
			  int		src_x,
			  int		src_y,
			  unsigned int	src_w,
			  unsigned int	src_h,
			  unsigned int	dst_w,
			  unsigned int	dst_h)
{
  register XExtCodes*		codes;
  xPlxGetImageReply		rep;
  register xPlxImageReq*	req;
  char*				data;
  long				nbytes;
  XPlxCImage*			image;
  
  if ((src_w < 16) || (dst_w < 16) || (src_h < 8) || (dst_h < 8))   /* Is requested image size too small any parameter? */
    return((XPlxCImage*)NULL);					    /* Yes, don't bother trying to compress */
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvGetCImage(dpy, d, gc,
			   src_x, src_y, src_w, src_h,
			   dst_w, dst_h));
  }
    
  GetReq(PlxImage, req);
  if ((src_w % 16) != 0)					    /* Assure that all width & height parameters sit on a... */
    src_w -= (src_w % 16);					    /* ...16 pixel boundary. Not sure why, but Dave Arnold... */
  if ((dst_w % 16) != 0)					    /* ...says so. */
    dst_w -= (dst_w % 16);
  req->drawable = d;						    /* Set up the standard stuff in the request */
  req->gc = gc->gid;
  req->src_x = src_x;
  req->src_y = src_y;
  req->src_w = src_w;
  req->src_h = src_h;
  req->dst_x = 0;
  req->dst_y = 0;
  req->dst_w = dst_w;
  req->dst_h = dst_h;
  req->reqType = VideoReqCode;
  req->videoReqType = X_PlxGetCImage;
  if (_XReply (dpy, (xReply *) &rep, 0, xFalse) == 0)
  {
    UnlockDisplay(dpy);
    SyncHandle();
    return (XPlxCImage *)NULL;
  }
  
  nbytes = (long)rep.length << 2;
  data = (char *) Xmalloc((unsigned) nbytes);
  if (! data)
  {
    _XEatData(dpy, (unsigned long) nbytes);
    UnlockDisplay(dpy);
    SyncHandle();
    return (XPlxCImage *) NULL;
  }
  _XReadPad (dpy, data, nbytes);
  image = XPlxCreateCImage (dpy, data, rep.size, dst_w, dst_h);
  
  POSTAMBLE
    
  return (image);
} /* end function XPlxGetCImage */



static void XPlxSendImage(register Display*		dpy,
			  register xPlxImageReq*	req,
			  register XPlxCImage*		image)
{
  long			length;
  unsigned char*	src;
  unsigned char*	dest;
  unsigned char*	shifted_src = NULL;
  
  req->leftPad = 0;
  length = image->size;
  req->length += (length + 3) >> 2;
  src = (unsigned char *)image->data;
  Data(dpy, (char *)src, length);
} /* end function XPlxSendImage */



static void PutImageRequest(register Display*		dpy,
			    Drawable			d,
			    GC				gc,
			    register XPlxCImage*	image,
			    int				src_x,
			    int				src_y,
			    unsigned int		src_w,
			    unsigned int		src_h,
			    int				dst_x,
			    int				dst_y,
			    unsigned int		dst_w,
			    unsigned int		dst_h,
			    int				flag)
{
  register XExtCodes*		codes;
  register xPlxImageReq*	req;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvPutImageRequest(dpy, d, gc, image,
			  src_x, src_y, src_w, src_h,
			  dst_x, dst_y, dst_w, dst_h,
			  flag);
    return;
  }
  
  GetReq(PlxImage, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->src_x = src_x;
  req->src_y = src_y;
  req->src_w = src_w;
  req->src_h = src_h;
  req->dst_x = dst_x;
  req->dst_y = dst_y;
  req->dst_w = dst_w;
  req->dst_h = dst_h;
  req->planeMask = image->size;
  req->format = flag;						    /* squeezed or not */
  req->reqType = VideoReqCode;
  req->videoReqType = X_PlxPutCImage;
  XPlxSendImage(dpy, req, image);
  
  POSTAMBLE
} /* end function PutImageRequest */



static void PutSubImage(register Display*		dpy,
			Drawable			d,
			GC				gc,
			register XPlxCImage*		image,
			int				src_x,
			int				src_y,
			unsigned int			src_w,
			unsigned int			src_h,
			int				dst_x,
			int				dst_y,
			unsigned int			dst_w,
			unsigned int			dst_h,
			int				flag)
{
  int				Available;
    
  if(image->size == 0)
    return;
    
#ifdef NOTDEF
  Available = ((65536 < dpy->max_request_size) ? (65536 << 2)
	       : (dpy->max_request_size << 2))
    - SIZEOF(xPutImageReq);
#endif
  Available = (dpy->max_request_size << 2) - SIZEOF(xPlxImageReq);
  
  if (image->size <= Available)
    PutImageRequest(dpy, d, gc, image, src_x, src_y, src_w, src_h, dst_x, dst_y, dst_w, dst_h, flag);
  else
    printf("XPlxPutCImage: too much data to transfer. avail = %d, size = %d\n",
	   Available, image->size);
} /* end function PutSubImage */



void		XPlxPutCImage(register Display*		dpy,
			      Drawable			d,
			      GC			gc,
			      register XPlxCImage*	image,
			      int			src_x,
			      int			src_y,
			      unsigned int		src_w,
			      unsigned int		src_h,
			      int			dst_x,
			      int			dst_y,
			      unsigned int		dst_w,
			      unsigned int		dst_h,
			      int			flag)
{
  if (image->size == 0)
    return;
  PutSubImage(dpy, d, gc, image,
	      src_x, src_y, src_w, src_h,
	      dst_x, dst_y, dst_w, dst_h,
	      flag);
  
} /* end function XPlxPutCImage */



void		XPlxPutTable(Display*	dpy,
			     Drawable	d,
			     GC		gc,
			     char*	data,
			     int	size,
			     int	dir)
{
  register XExtCodes*		codes;
  register xPlxImageReq*	req;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvPutTable(dpy, d, gc, data, size, dir);
    return;
  }
    
  GetReq(PlxImage, req);
  req->format = !dir;
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_PlxPutQTable;
  
  req->length += ((size + 3) >> 2);	
  
  Data(dpy, data, size);
  
  POSTAMBLE
}
