/*
 *   Copyright (c) 1994 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* $Header: /plx/source/Xclients/generic/RCS/XPlxUtils.c,v 0.105 1996/01/22 06:21:01 drapeau Exp $ */
/* $Log: XPlxUtils.c,v $
 * Revision 0.105  1996/01/22  06:21:01  drapeau
 * Minor enhancement to XPlxSleep(): if the argument passed in is zero,
 * exit immediately without bothering to sleep.
 *
 * Revision 0.104  1996/01/11  03:21:38  drapeau
 * Finally, one last fix to the input model to support both multiple
 * instances of a particular format (i.e., Composite, Y/C, RGB, etc.)
 * *and* multiple input "bundles" that come into the same digitizer
 * (e.g., if one of the digitizer's input wires had extra hardware to
 * pre-process an incoming analog video signal).
 *
 * Revision 0.103  1996/01/10  08:57:53  drapeau
 * Fixed minor syntax error.
 *
 * Revision 0.102  1996/01/10  08:56:49  drapeau
 * Fixed potential errors in several functions that examine the plx_IO
 * data structures.  The functions now correctly check if the data
 * structures are NULL, usually meaning that no Plx hardware was found on
 * the Display* connection being used by the calling application.
 *
 * Revision 0.101  1995/12/23  02:50:24  drapeau
 * Modified XPlxFindMatchingSignal() to use the new scheme for keeping
 * track of channel and instance information.  Instance information is
 * now stored in its own data structure, along side the channel info.  It
 * is no longer sought in the signal's "interlaced" field.
 *
 * Revision 0.100  1995/12/14  02:59:02  drapeau
 * Added the function XPlxFindExtension().  This function is the core of
 * the new XPlx library.  The function determines which video extension
 * to use, trying first to use the ParallaxVideo extension and falling
 * back to use XVideo if necessary.
 *
 * Revision 0.99  1995/10/26  20:11:09  drapeau
 * Fixed an error in XPlxSleep(): it now pays attention to time values
 * greater than one second.
 *
 * Revision 0.98  1995/10/17  20:06:44  lovvik
 * Fixed a problem in XPlxIsRTVFile().
 *
 * Revision 0.97  1995/10/17  17:44:02  lovvik
 * Changed Boolean to Bool.
 *
 * Revision 0.96  1995/10/17  17:28:27  lovvik
 * Added an include statement for <stdlib.h>.  This is used for
 * the file utility functions.
 *
 * Revision 0.95  1995/10/17  01:52:33  lovvik
 * Added three file utility functions, XPlxIsJPEGFile(), XPlxIsRTVFile(), and
 * XPlxIsJFIFFile(), which return true if the specified file is a valid jpeg
 * movie file, rtv movie file, or jfif image file.
 *
 * Revision 0.94  1995/09/20  02:09:10  drapeau
 * Added two common utility functions, XPlxGetVideoVisual() and
 * XPlxSleep(), to obtain a Visual commonly used for displaying video and
 * for putting the application to sleep for some number of microseconds.
 *
 * Revision 0.93  1995/05/23  06:50:41  drapeau
 * Added code to PlxFindDpyMap() to initialize the
 * Display-to-major-opcode mapping array before traversing it the first
 * time.
 *
 * Revision 0.92  1995/05/19  02:02:05  drapeau
 * Moved the PlxDeleteIOList() functions into this module since they are
 * generic functions for this library.
 *
 * Revision 0.91  95/02/24  03:18:09  03:18:09  drapeau ()
 * Added code to help this library work with multiple displays (when a
 * process using this library is working with mulitple displays, it is
 * not guaranteed that the Plx extension will be loaded in the same spot
 * and therefore might occupy a different base event code for Plx
 * protocol events.  So, this library needs to keep a mapping of
 * Display-to-Protocol-Event-Base-Codes so the "PREAMBLE" macro will
 * always work correctly and allow this library to communicate with the
 * proper X extension no matter what the Display.)
 * 
 * Revision 0.90  1994/07/20  22:24:47  drapeau
 * First working prototype, tested first on the HP platform.
 * */
static char xPlxUtilsRcsId[] = "$Header: /plx/source/Xclients/generic/RCS/XPlxUtils.c,v 0.105 1996/01/22 06:21:01 drapeau Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xlibint.h>
#include <plxvideo.h>
#include <movie.h>

#define PlxMaxOpenDpys	16

typedef struct _PlxDisplayMap
{
  Display*	dpy;
  int		videoReqCode;
} PlxDisplayMap;

static PlxDisplayMap	plxDpyMap[PlxMaxOpenDpys];
Display*		plxLastOpenDpy = (Display*)NULL;



Bool XPlxFindMatchingSignal(plx_IO*	plxIO,
			    int		channel,
			    int		standard,
			    int		format,
			    int		instance)
{
  plx_format_list*	formatList = (plx_format_list*) NULL;
  plx_IO_list*		ioList;
  plx_signal_list*	signalList = (plx_signal_list*) NULL;
  
  if ((plx_IO*)NULL == plxIO)					    /* If no hardware structure was available, return... */
    return(False);						    /* ...failure immediately. */
  switch (channel)						    /* Which input or output list should be used? */
  {
   case PLX_IO_OFF:						    /* Neither, return success immediately */
    return(True);
    break;
   case PLX_OUTPUT_0:						    /* Output, point to the first output on the list */
    ioList = plxIO->outputs;
    break;
   default:							    /* Point to the firts input on the list by default */
    ioList = plxIO->inputs;
    break;
  }
  while(ioList != NULL)						    /* Search until the input or output list is exhausted,... */
  {								    /* ...if necessary. */
    for (signalList = ioList->signal_list;			    /* Start with the signal list info */
	 signalList != NULL;
	 signalList = signalList->next_signal)
    {
      for (formatList = ioList->format_list;			    /* Also look at the signal format list */
	   formatList != NULL;
	   formatList = formatList->next_format)
      {
	if ((channel == ioList->channel) &&			    /* Now that all info is present, is there a matching signal? */
	    (standard == signalList->signal->standard) &&
	    (format == formatList->format) &&
	    (instance == formatList->instance))
	{		
	  return(True);						    /* Yes, return success immediately */
	}
      }	/* end for (formatList...) */
    } /* end for (signalList...) */
    ioList = ioList->next_IO;					    /* Go on to the next input or output on the list */
  }
  return(False);
} /* end function XPlxFindMatchingSignal */



Bool PlxFindDpyMap(Display*	dpy,				    /* This function tries to match the Display ptr passed in... */
		   int*		videoReqCodePtr)		    /* with one already registered for this process; if... */
{								    /* ...found, a struct with the appropriate X event base... */
  int	i;							    /* ...is returned; else a NULL struct is returned. */
  static Bool	firstTime = True;				    /* This helps apps that communicate with multiple displays. */
  
  if (True == firstTime)					    /* Initialize the PlxDpyMap first time this function is run */
  {
    for (i = 0; i < PlxMaxOpenDpys; i++)			    /* Look at each item in the static array of PlxDpyMap's */
    {
      plxDpyMap[i].dpy = (Display*)NULL;
      plxDpyMap[i].videoReqCode = 0;
    }
    firstTime = False;
  }
  for (i = 0; i < PlxMaxOpenDpys; i++)
  {
    if (dpy == plxDpyMap[i].dpy)				    /* Desired entry in the map table was found */
    {
      *videoReqCodePtr = plxDpyMap[i].videoReqCode;		    /* Update the extension protocol request base passed in */
      return(True);						    /* Also, return success code */
    }
    if (plxDpyMap[i].dpy == (Display*)NULL)			    /* Entry wasn't found; this should be 1st empty entry */
      break;							    /* Done looking, leave the for loop */
  } /* end for */
  *videoReqCodePtr = 0;						    /* Reset protocol request base passed in (this must be... */
  return(False);						    /* ...a new Display for this process) and return no success. */
} /* end function PlxFindDpyMap */



Bool PlxAddDisplayToMap(Display*	dpy,
			int		videoReqCode)
{
  int	i;
  
  for (i = 0; i < PlxMaxOpenDpys; i++)
  {
    if ((dpy == plxDpyMap[i].dpy) ||				    /* Place to insert new mapping or replace the existing... */
	(plxDpyMap[i].dpy == (Display*)NULL))			    /* ...one */
    {
      plxDpyMap[i].dpy = dpy;					    /* Make a new Display-to-Protocol-Request-Base mapping */
      plxDpyMap[i].videoReqCode = videoReqCode;
      return(True);						    /* Return successful status */
    }
  } /* end for */
  fprintf(stderr, "Plx Extension: Max number of Display connections (%d) exceeded.\n",
	  PlxMaxOpenDpys);
  return(False);						    /* Return unsuccessful status */
} /* end function PlxAddDisplayToMap */



void PlxDeletePlxIOList(plx_IO_list*	ioList)			    /* Frees memory previously allocated by XPlxQueryConfig() */
{
  plx_signal_list*	signalList = (plx_signal_list*) NULL;
  plx_format_list*	formatList = (plx_format_list*) NULL;
  plx_type_list*	typeList = (plx_type_list*) NULL;
  plx_signal_list*	nextSignal = (plx_signal_list*) NULL;
  plx_format_list*	nextFormat = (plx_format_list*) NULL;
  plx_type_list*	nextType = (plx_type_list*) NULL;
  
/*
  printf("PlxDeletePlxIOList(): Channel is %d, Instance is %d.\n",
  ioList->channel, ioList->instance);
*/
  if ((plx_IO_list*)NULL == ioList)				    /* If an empty input/output list was passed in, don't... */
    return;							    /* ...bother doing any needless work. */
  signalList = ioList->signal_list;				    /* Point to the top of the list of signals */
  while (signalList != NULL)					    /* Traverse the list until the end is reached */
  {
    nextSignal = signalList->next_signal;			    /* Temp variable points to the next signal on the list */
/*
    printf("About to free Signal standard: %d\tWidth:%d\tHeight:%d\tInterlaced:%d\n",
	   signalList->signal->standard,
	   signalList->signal->w,
	   signalList->signal->h,
	   signalList->signal->interlaced);
*/
    free(signalList->signal);					    /* Free space that holds signal information */
    free(signalList);						    /* Free the pointer to this signal structure */
    signalList = nextSignal;					    /* Point to the next signal on the list */
  }
  formatList = ioList->format_list;				    /* Point to the top of the list of signal formats */
  while (formatList != NULL)					    /* Traverse the list until the end is reached */
  {
    nextFormat = formatList->next_format;			    /* Temp variable points to the next format on the list */
    free(formatList);						    /* Free the pointer to this format structure */
    formatList = nextFormat;					    /* Point to the next format on the list */
  }
  typeList = ioList->type_list;					    /* Point to the top of the list of signal types */
  while (typeList != NULL)					    /* Traverse the list until the end is reached */
  {
    nextType = typeList->next_type;				    /* Temp variable points to the next type on the list */
    free(typeList);						    /* Free the pointer to this type structure */
    typeList = nextType;					    /* Point to the next type on the list */
  }
  return;
} /* end function PlxDeletePlxIOList */



void PlxDeletePlxHardwareList(plx_IO* ioList)
{
  plx_IO_list*	inputs;
  plx_IO_list*	outputs;
  plx_IO_list*	nextInput = (plx_IO_list*)NULL;
  plx_IO_list*	nextOutput = (plx_IO_list*)NULL;
  
/*
  printf("Entered PlxDeletePlxHardwareList.\n");
*/
  if ((plx_IO*)NULL == ioList)					    /* If an empty input/output list was passed in, don't... */
    return;							    /* ...bother doing any needless work. */
  inputs = ioList->inputs;					    /* Look at the inputs for this description of the hardware */
  while (inputs != NULL)					    /* Traverse the list until the end is reached */
  {
    nextInput = inputs->next_IO;				    /* Temp variable points to the next input on the list */
/*
    printf("Channel: %d, Instance: %d\n",
    inputs->channel, inputs->instance);
*/
    PlxDeletePlxIOList(inputs);
    free(inputs);						    /* Free the pointer to this inputs structure */
    inputs = nextInput;						    /* Move to the next input structure on the list */
  }
  outputs = ioList->outputs;					    /* Now handle the outputs for this hardware description */
  while (outputs != NULL)					    /* Traverse the list until the end is reached */
  {
    nextOutput = outputs->next_IO;				    /* Temp variable points to the next output on the list */
/*
    printf("Channel: %d\n", outputs->channel);
*/
    PlxDeletePlxIOList(outputs);
    free(outputs);						    /* Free the pointer to this outputs structure */
    outputs = nextOutput;					    /* Move to the next output structure on the list */
  }
/*
  printf("Exiting PlxDeletePlxHardwareList.\n\n");
*/
  return;
} /* end function PlxDeletePlxHardwareList */



XVisualInfo* XPlxGetVideoVisual(int		screenNumber,
				Display*	display)
{
  Visual*		defaultVisual;
  int			i;
  long			visualInfoMask;
  int			visualsMatched;
  XVisualInfo		visualHints;
  static XVisualInfo*	visualList = (XVisualInfo*)NULL;
  XVisualInfo*		theVisual;
  
  if (visualList != (XVisualInfo*)NULL)				    /* Was data saved by a previous call to this function? */
    XFree((void*)visualList);					    /* Yes, release the data since it is no longer needed */
  visualHints.class = TrueColor;				    /* Set up hints to determine the type of visual needed... */
  visualHints.screen = screenNumber;				    /* ...to create a window that can display video */
  visualInfoMask = (VisualClassMask | VisualScreenMask);	    /* Initially, try to find a TrueColor class visual */
  visualList = XGetVisualInfo(display,				    /* Try to find a visual that matches what is needed for... */
			      visualInfoMask,			    /* ... a window that will display video or TrueColor... */
			      &visualHints,			    /* ...graphics. */
			      &visualsMatched);			    /* This function may return zero visuals matched */
  for (theVisual = visualList, i = 0;				    /* Look at all the visuals returned by XGetVisualInfo... */
       i < visualsMatched;					    /* ...to find the best match.  The idea match would be... */
       i++, theVisual++)					    /* ...a 24-bit depth Visual. */
  {
    if (theVisual->depth == 24)					    /* Was an ideal match found? */
      return(theVisual);					    /* Yes, return this Visual */
  }
  for (theVisual = visualList, i = 0;				    /* If this statement is reached, no 24-bit TrueColor... */
       i < visualsMatched;					    /* ...Visual was found; in this case, fallback and try... */
       i++, theVisual++)					    /* ...to find the next best thing: any TrueColor visual... */
  {								    /* ...of depth >= 24 bits (probably a 32-bit is what will... */
    if (theVisual->depth >= 24)					    /* ...be found). */
      return(theVisual);					    /* Was a good fallback Visual found?  If so, return it. */
  }
  defaultVisual = DefaultVisual(display, screenNumber);		    /* No good Visual was found, so return the default... */
  visualHints.visualid = defaultVisual->visualid;		    /* ...Visual for this screen. */
  visualInfoMask = (VisualIDMask);
  visualList = XGetVisualInfo(display,				    /* Try to find a visual that matches what is needed for... */
			      visualInfoMask,			    /* ... a window that will display video or TrueColor... */
			      &visualHints,			    /* ...graphics. */
			      &visualsMatched);
  if (visualsMatched == 0)					    /* Did the search for a TrueColor visual fail? */
    return((XVisualInfo*)NULL);					    /* Yes, return a NULL structure to represent failure */
  return(visualList);						    /* If this statement was reached, it should be returning... */
								    /* ...the default visual for this screen. */
}								    /* end function XPlxGetVideoVisual */



void XPlxSleep(unsigned int microseconds)
{
  int			descriptorTableSize = 0;
  int			result;
  struct timeval	waitTime;				    /* Used to tell select() how long to wait until it returns */

  if (0 == microseconds)					    /* Don't bother trying to sleep if a zero time was... */
    return;							    /* ...passed in for some reason. */
  descriptorTableSize = sysconf(_SC_OPEN_MAX);			    /* Get size of descriptor table for later use by select() */
  waitTime.tv_sec = microseconds / 1000000;			    /* Set up a timeval to wait for the time passed in as... */
  waitTime.tv_usec = microseconds % 1000000;			    /* ...argument. */
  result = select(descriptorTableSize,				    /* Use select to sleep by not asking to be informed of... */
		  0, 0, 0,					    /* ...input on any file descriptors, and filling in the... */
		  &waitTime);					    /* ...timeval for the wait period. */
  return;
} /* end function XPlxSleep */



Bool XPlxIsJPEGFile(char*	filename)			    /*  Returns True if this is a valid JPEG movie file. */
{
  FILE *sourceFile;
  jpheader movieHeader;
  int numberOfRecords;
  
  sourceFile = fopen(filename, "rb");
  if(!sourceFile)
  {
    return(False);
  }
  numberOfRecords = fread(&movieHeader,				    /* Load the header of the source file. */
			  sizeof(jpheader), 1, sourceFile); 
  fclose(sourceFile);
  if(numberOfRecords == 1 &&					    /* Check to see if this is a valid movie file. */
     strncmp(movieHeader.magic, "j_movie", 7) == 0)
  {
    return(True);
  }
  else
  {
    return(False);
  }
} /* end function XPlxIsJPEGFile */



Bool XPlxIsRTVFile(char *filename)				    /*  Returns True if this is a valid rtv movie file. */
{
  FILE *sourceFile;
  int fileType;
  int index;
  char test[100];
  
  sourceFile = fopen(filename, "rb");
  if(!sourceFile)
  {
    return(False);
  }
  
  fgets(test, 7, sourceFile);					    /* Look at the first 6 bytes for the ID string. */
  fclose(sourceFile);
  if(strcmp(test, "newrtv") == 0)
  {
    return(True);
  }
  else
  {
    return(False);
  }
}



Bool XPlxIsJFIFFile(char *filename)				    /* This function returns True if this is a valid JFIF file. */
{
  FILE *sourceFile;
  int fileType;
  int index;
  char test[100];
  
  sourceFile = fopen(filename, "rb");
  if(!sourceFile)
  {
    return(False);
  }
  
  for(index = 0; index < 6; index++)				    /* Advance to the ID string. */
  {
    fileType = getc(sourceFile);
  }
  fgets(test, 5, sourceFile);					    /* Load the ID string. */
  fclose(sourceFile);
  if(strcmp(test, "JFIF") == 0)
  {
    return(True);
  }
  else
  {
    return(False);
  }
}


/* 
 * XPlxFindExtension -- See if any Parallax-compatible video extension is present
 *
 * This function determines whether a video X protocol extension is
 * available to this connection to the X server.  Currently the only
 * acceptable X extensions are the Parallax Graphics "ParallaxVideo" X
 * extension or the MIT "XVideo" extension.
 *
 * The function first searches the local Display structure's private
 * extension data to see if the XPlx (Parallax Graphics) extension is
 * present; if so, the function immediately returns a pointer to the
 * relevant information about this structure and as a side effect sets
 * the "majorOpcodePtr" variable to the major opcode of the extension as
 * installed (the "majorOpcodePtr" variable is used throughout the XPlx
 * native library).  If the XPlx extension is not already registered,
 * the function checks if the Xv extension is registered; if so, the
 * function returns a pointer to the Xv-specific information and sets
 * the majorOpcodePtr to zero, indicating that the XPlx extension should
 * not be used.
 *
 * If Xv is not found already installed, the function next tries to
 * call XInitExtension on the ParallaxVideo extension; if this fails,
 * the function tries the same approach with the Xv extension.
 *
 * If no Parallax-compatible video extension is found, the function
 * returns a NULL pointer to the caller.
 */

XExtCodes* XPlxFindExtension(Display*	display,
			     int*	majorOpcodePtr)
{
  XExtCodes*	extensionCodes;
  _XExtension*	extensionData;
  Bool		instanceFound = False;
  int		result;
  
  for (extensionData = display->ext_procs;			    /* Look at this Display's extension data for the native... */
       extensionData != (_XExtension*)NULL;			    /* ...XPlx (i.e., "ParallaxVideo") extension. */
       extensionData = extensionData->next)
  {
    result = strcmp(extensionData->name, VIDEONAME);		    /* Look for the Parallax video extension */
    if (0 == result)						    /* Was the native XVideo extension found? */
    {								    /* Yes, set the variable for the XPlx library to the... */
      *majorOpcodePtr = extensionData->codes.major_opcode;	    /* ...major opcode for this connection's extension. */
/*
      printf("Debugging:XPlxFindExtension found ParallaxVideo 1st try, major_opcode is %d.\n",
	     *majorOpcodePtr);
*/
      return(&(extensionData->codes));				    /* Return pointer to the appropriate data structure */
    }
    if ((_XExtension*)NULL == extensionData->next)		    /* The end of the list was reached */
      break;
  } /* end first for loop */
  for (extensionData = display->ext_procs;			    /* First pass failed, so look for the Xv extension */
       extensionData != (_XExtension*)NULL;
       extensionData = extensionData->next)
  {
    result = strcmp(extensionData->name, "XVideo");		    /* Look for the MIT video extension (i.e., "XVideo") */
    if (0 == result)						    /* Was the Xv extension found? */
    {
      *majorOpcodePtr = 0;					    /* Setting to zero means to use Xv for requests */
/*
      printf("Xv found on 1st try at major opcode %d.\n",
	     extensionData->codes.major_opcode);
*/
      return(&(extensionData->codes));				    /* Yes, return pointer to the appropriate data structure */
    }
    if ((_XExtension*)NULL == extensionData->next)		    /* The end of the list was reached */
      break;
  } /* end second for loop */
/*
  printf("Nothing in Display cache; Trying to XInitExtension on ParallaxVideo.\n");
*/
  extensionCodes = XInitExtension(display, VIDEONAME);		    /* No video extension was yet registered; try to register... */
  if (extensionCodes != (XExtCodes*)NULL)			    /* ...the ParallaxVideo extension.  Did it succeed? */
  {								    /* Yes, set the majorOpcodePtr appropriately and... */
/*
    printf("Successfully initialized the ParallaxVideo extension, major_opcode is %d.\n",
	   extensionCodes->major_opcode);
*/
    *majorOpcodePtr = extensionCodes->major_opcode;		    /* ...return the relevant extension information. */
    return(extensionCodes);
  }
/*
  printf("Nothing in Display cache; Trying to XInitExtension on XVideo.\n");
*/
  extensionCodes = XInitExtension(display, "XVideo");		    /* The ParallaxVideo extension isn't available; try Xv */
  if (extensionCodes != (XExtCodes*)NULL)			    /* Did the attempt to find Xv succeed? */
  {								    /* Yes, set the majorOpcodePtr to zero to indicate not to... */
/*
    printf("Successfully initialized the Xv extension, major_opcode is %d.\n",
	   extensionCodes->major_opcode);
*/
    *majorOpcodePtr = 0;					    /* ...use the native XPlx extension, then return the... */
    return(extensionCodes);					    /* ...relevant extension information. */
  }
  *majorOpcodePtr = 0;						    /* Indicate not to use the native XPlx extension */
  return((XExtCodes*)NULL);					    /* If this stmt is reached, no video extension was found */
} /* end function XPlxFindExtension */
