/*
 * (c) Copyright 1991 Parallax Graphics Inc, All Rights Reserved
 *
 * Parallax Graphics, Inc.
 * 2500 Condensa Street
 * Santa Clara, California 95051
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 * $Log: XPlxVideo.c,v $
 * Revision 1.57  1996/01/11  03:16:04  drapeau
 * Finally, one last fix to the input model to support both multiple
 * instances of a particular format (i.e., Composite, Y/C, RGB, etc.)
 * *and* multiple input "bundles" that come into the same digitizer
 * (e.g., if one of the digitizer's input wires had extra hardware to
 * pre-process an incoming analog video signal).
 *
 * Plus, fixed some coding errors that may have an effect in the future
 * if not fixed.
 *
 * Revision 1.56  1996/01/10  08:27:23  drapeau
 * Fixed errors in plxadd{format,type,signal}(): the functions did not
 * account for the possibility of duplicates being added to the plx_IO*
 * data structures so these functions never checked incoming arguments
 * against those data already added to their respective lists.  The new
 * functions now perform this check, which is crucial for overlay-based
 * servers that may provide more than one instance of the same type of
 * input.
 *
 * Revision 1.55  1995/12/28  01:32:38  lovvik
 * Fixed a problem in which the plx_IO structure returned from
 * XPlxQueryConfig was incorrect.  The fix was in plxaddformat().
 * The part of the code that allocates a new plx_IO_list for
 * every channel/instance pair was only being executed if
 * both the channel AND the instance were different.
 *
 * Revision 1.54  1995/12/23  02:20:02  drapeau
 * Added code to strip instance information from its new location, in the
 * same 32-bit word as the channel information.  This code should work
 * with all servers, current and future.
 *
 * Revision 1.53  1995/12/14  03:00:32  drapeau
 * Major changes to this module to accommodate the new XPlx library that
 * automatically detects the proper video extension to use.  The new
 * library attempts to use the ParallaxVideo extension if possible and
 * falls back to use the XVideo extension if need be.
 *
 * Revision 1.52  1995/10/17  19:13:02  drapeau
 * Minor changes to XPlxQueryConfig() and initialization code to
 * accomodate the new plx_IO data structure definition.
 *
 * Revision 1.51  1995/10/12  18:31:15  drapeau
 * * Added limited support for RTV functions and a generic "Control"
 *   request.
 * * Reformatted the code for easier editing and readability.
 * * Removed extraneous printf messages in XPlxQueryConfig().
 *
 * Revision 1.50  1995/09/28  05:00:29  drapeau
 * Modified XPlxQueryConfig() to simulate new information coming from the
 * window server that describes additional capabilities of the hardware.
 *
 * Also, cosmetic changes to make code easier to read and edit.
 *
 * Revision 1.49  1995/06/24  00:18:09  drapeau
 * Added the function XPlxXPutImage() to replace XPutImage() for the
 * overlay product on the HP (on this platform, it simply forwards the
 * request to XPutImage() as usual).
 *
 * Revision 1.48  1995/05/25  23:36:07  drapeau
 * Added new XPlx api functions:
 *
 * * XPlxSelectHardwareInstance()
 * * XPlxGetHardwareInstances()
 * * XPlxGetZoomFactor()
 * * XPlxSetZoomFactor()
 *
 * These were first implemented on the HP and IBM platforms, both of
 * which support multiple Parallax boards per X screen and hardware zoom,
 * but neither of these features are yet supported on the Sun XVideo
 * platform so the functions return default values that a Sun application
 * might expect (zoom of one, one hardware instance, don't allow the
 * program to change zoom level or select a different board).
 *
 * Revision 1.47  1995/05/19  02:03:03  drapeau
 * Removed DeletePlxIOList() functions from this module and moved them to
 * the more generic XPlxUtils.c module where they belong.
 *
 * Revision 1.46  95/04/07  01:03:19  01:03:19  drapeau ()
 * Fixed an error with XPlxQueryConfig(): the client side of this library
 * contains a global variable pointing to information about Plx hardware
 * that is returned from calling XPlxQueryConfig().  However, in no place
 * was this information ever freed, so if a process makes multiple calls
 * to XPlxQueryConfig(), the same information is added to the plx_io data
 * structures multiple times.  This can have a side effect on programs
 * that scan through the list of possible inputs by traversing the plx_io
 * data structure, since the calling programs will see duplicate signal
 * records.  The XVideoIn widget is an example of such a piece of code.
 * This can, under certain circumstances, greatly increase the time taken
 * to detect sync on a signal being selected, and unneeded duplicate
 * checks are done.
 * 
 * The fix is to free the data structures taken by previous calls to
 * XPlxQueryConfig().  Each time the function is called, it calls the
 * function DeletePlxHardwareList() to free the previous invocation's
 * results.  The net effect is that at most one copy of the hardware
 * information is ever stored in the plx_io data structure.
 * 
 * DeletePlxHardwareList() is a new function that traverses the plx_io
 * data structure created by XPlxQueryConfig().
 * 
 * Revision 1.45  1995/02/24  03:03:15  drapeau
 * Two changes:
 *
 * 1) Added code to help this library work with multiple displays (when a
 *    process using this library is working with mulitple displays, it is
 *    not guaranteed that the Plx extension will be loaded in the same
 *    spot and therefore might occupy a different base event code for Plx
 *    protocol events.  So, this library needs to keep a mapping of
 *    Display-to-Protocol-Event-Base-Codes so the "PREAMBLE" macro will
 *    always work correctly and allow this library to communicate with
 *    the proper X extension no matter what the Display.)
 *
 * 2) Cosmetic changes to make the code more readable and editable.
 *
 * Revision 1.44  1994/09/14  03:25:13  drapeau
 * Added diagnostic message to XPlxVideoInputSelect() for debugging
 * purposes.
 *
 * Revision 1.43  1994/08/27  22:51:54  drapeau
 * ANSI-fied functions, making function prototypes for them.
 *
 * Revision 1.42  1994/08/11  22:15:08  drapeau
 * Made some temporary modifications to key functions having to do with
 * input.  The modifications will be revoked when the new input model
 * (including the concept of an "instance" of input) has been implemented
 * in the server.  The modifications allow the client side to emulate the
 * old model.
 *
 * Revision 1.41  1993/11/20  01:18:19  drapeau
 * Cleaned up inclusion of header files to increase portability.
 *
 * Revision 1.4  1992/08/19  22:53:25  eric
 * make sure we send that null!
 *
 * Revision 1.3  91/09/17  10:31:59  dave
 * added string operations
 * 
 * Revision 1.2  91/07/24  14:17:13  eric
 * added rcs header
 * 
 *
 */
#ifndef lint
static char *rcsid = "@(#)$Id: XPlxVideo.c,v 1.57 1996/01/11 03:16:04 drapeau Exp $ Parallax Graphics Inc";
static char *sid_video_c = "@(#)XPlxVideo.c	1.8 5/10/89 Parallax Graphics Inc";
#endif
/*
 * XPlxVideo.c
 *
 */

#define	NEED_REPLIES
#include	<X11/Xlibint.h>
#include	<X11/Xutil.h>
#include	<plxvideo.h>

#ifdef X11R2
#define	Data16	PackData
#endif

static int VideoReqCode = 0;
static int first_event, first_error;
extern Display*	plxLastOpenDpy;

/*
 * set a drawable to be of type video
 */

#define	PREAMBLE(result)	\
LockDisplay(dpy); \
if ((GC)NULL != gc)  \
{\
   FlushGC(dpy, gc); \
 } \
result = XPlxFindExtension(dpy, &VideoReqCode);
     
#define POSTAMBLE	\
		       UnlockDisplay(dpy); \
			 SyncHandle();
     

/* variables and structurs added for rev 4.0 */
     
plx_IO plx_io =
{
  0,								    /* overlay bit */
  0,								    /* zoom bit */
  0,								    /* numJPEGCompressors bits */
  0,								    /* numMPEGDecoders bits */
  0,								    /* numH320Compressors bits */
  0,								    /* reserved bits */
  0,								    /* hardware */
  (plx_IO_list*)NULL,						    /* inputs */
  (plx_IO_list*)NULL						    /* outputs */
};



/* new functions for rev 4.0 */

void plxaddformat(register PlxVideoFormat* src)
{
  unsigned short		channel;
  unsigned short		format;
  unsigned short		instance;
  int				listEmpty = 1;			    /* Indicates whether formats list is empty; initially yes */
  unsigned int			tempBundle;
  register plx_format_list*	f0;
  register plx_IO_list*		iol;
  register plx_format_list*	lastFormat;
  
  channel = src->channel & 0x0000FFFF;				    /* Mask off the upper word of channel info to get rid... */
								    /* ...of the bundle info. */
  tempBundle = src->channel >> 16;				    /* Extract the bundle info from the 32-bit channel field */
  format = src->format & 0x0000FFFF;				    /* Mask off the upper word of format info to get rid... */
								    /* ...of the instance info. */
  instance = src->format >> 16;					    /* Extract the bundle info from the 32-bit channel field */
  iol = (src->type) ? plx_io.outputs : plx_io.inputs;
  if(iol == NULL)
  {
    iol = (plx_IO_list *) Xmalloc(sizeof(plx_IO_list));
    if(src->type)
      plx_io.outputs = iol;
    else
      plx_io.inputs = iol;
    iol->format_list = NULL;
    iol->type_list = NULL;
    iol->signal_list = NULL;
    iol->bundle = tempBundle;					    /* Fill in the plx_IO bundle info */
    iol->channel = channel;
    iol->next_IO = NULL;
  }
  for(;								    /* Find either matching channel and bundle or end of... */
      (iol->channel != channel ||				    /* ...the input list if the matching channel/bundle... */
       iol->bundle != tempBundle) && iol->next_IO;		    /* ...pair can't be found. */
      iol = iol->next_IO);
  if ((iol->channel != channel) ||				    /* Is this a new channel/bundle pair? */
      (iol->bundle != tempBundle)) 
  {
    iol->next_IO = (plx_IO_list *) Xmalloc(sizeof(plx_IO_list));
    iol = iol->next_IO;
    iol->signal_list = NULL;
    iol->format_list = NULL;
    iol->type_list = NULL;
    iol->bundle = tempBundle;					    /* Fill in the plx_IO bundle info */
    iol->channel = channel;
    iol->next_IO = NULL;
  }
  for(f0 = iol->format_list, lastFormat = iol->format_list;	    /* Look through the list of formats currently registered... */
      f0 != (plx_format_list*)NULL;				    /* ...to check for duplicates.  Don't add any duplicates. */
      f0 = f0->next_format)
  {
    lastFormat = f0;						    /* Remember the previously examined item on the list */
    if ((f0->format == format) &&				    /* Has the requested format structure already been added? */
	(f0->instance == instance))				    /* (check if the instance matches, too) */
      return;							    /* Yes, don't bother trying to add it again */
  }
  f0 = (plx_format_list *) Xmalloc(sizeof(plx_format_list));	    /* If this stmt reached, no match was found on the list */
  if (NULL == iol->format_list)					    /* Was the list empty to begin with? */
    iol->format_list = f0;					    /* Yes, make the newly-allocated node the 1st on the list */
  else								    /* No, the list had entries, so... */
    lastFormat->next_format = f0;				    /* ...add the new entry to the end of the list. */
  f0->next_format = NULL;
  f0->format = format;
  f0->instance = instance;
} /* end function plxaddformat */



void plxaddtype(register PlxVideoType*	src)
{
  unsigned short		channel;
  unsigned short		bundle;
  register plx_type_list*	lastType;
  unsigned int			tempBundle;
  register plx_type_list*	t0;
  register plx_IO_list*		iol;
  
  channel = src->channel & 0x0000FFFF;				    /* Mask off the upper word of channel info to get rid... */
								    /* ...of the bundle info. */
  tempBundle = src->channel >> 16;				    /* Extract the bundle info from the 32-bit channel field */
  iol = (src->type) ? plx_io.outputs : plx_io.inputs;		    /* Look at input list if type is 0, output list if type is 1 */
  if(iol == NULL)
  {
    iol = (plx_IO_list *) Xmalloc(sizeof(plx_IO_list));
    if(src->type)
      plx_io.outputs = iol;
    else
      plx_io.inputs = iol;
    iol->format_list = NULL;
    iol->type_list = NULL;
    iol->signal_list = NULL;
    iol->bundle = tempBundle;				    /* Fill in the plx_IO bundle info */
    iol->channel = channel;
    iol->next_IO = NULL;
  }
  for(;								    /* Find either matching channel and bundle or end of... */
      (iol->channel != channel ||				    /* ...the input list if the matching channel/bundle... */
       iol->bundle != tempBundle) && iol->next_IO;		    /* ...pair can't be found. */
      iol = iol->next_IO);
  if ((iol->channel != channel) ||				    /* Is this a new channel/bundle pair? */
      (iol->bundle != tempBundle)) 
  {
    iol->next_IO = (plx_IO_list *) Xmalloc(sizeof(plx_IO_list));
    iol = iol->next_IO;
    iol->signal_list = NULL;
    iol->format_list = NULL;
    iol->type_list = NULL;
    iol->bundle = tempBundle;					    /* Fill in the plx_IO bundle info */
    iol->channel = channel;
    iol->next_IO = NULL;
  }
  for(t0 = iol->type_list, lastType = iol->type_list;		    /* Look through the list of types currently registered... */
      t0;							    /* ...to check for duplicates.  Don't add any duplicates. */
      t0 = t0->next_type)
  {
    lastType = t0;						    /* Remember the previously examined item on the list */
    if (t0->type == src->type)					    /* Has the requested type structure already been added? */
      return;							    /* Yes, don't bother trying to add it again */
  }
  t0 = (plx_type_list *) Xmalloc(sizeof(plx_type_list));	    /* If this stmt reached, no match was found on the list */
  if (NULL == iol->type_list)					    /* Was the list empty to begin with? */
    iol->type_list = t0;					    /* Yes, make the newly-allocated node the 1st on the list */
  else								    /* No, the list had entries, so... */
    lastType->next_type = t0;					    /* ...add the new entry to the end of the list. */
  t0->next_type = NULL;
  t0->type = src->type;
} /* end function plxaddtype */



void plxaddsig(register PlxVideoSignal*	src)
{
  unsigned short		channel;
  unsigned short		bundle;
  register plx_signal_list*	lastSignal;
  unsigned int			tempBundle;
  register plx_signal_list*	s0;
  register plx_signal*		sig;
  register plx_IO_list*		iol;
  
  channel = src->channel & 0x0000FFFF;				    /* Mask off the upper word of channel info to get rid... */
								    /* ...of the bundle info. */
  tempBundle = src->channel >> 16;				    /* Extract the bundle info from the 32-bit channel field */
  iol = (src->type) ? plx_io.outputs : plx_io.inputs;
  if(iol == NULL)
  {
    iol = (plx_IO_list *) Xmalloc(sizeof(plx_IO_list));
    if(src->type)
      plx_io.outputs = iol;
    else
      plx_io.inputs = iol;
    iol->format_list = NULL;
    iol->type_list = NULL;
    iol->signal_list = NULL;
    iol->bundle = tempBundle;				    /* Fill in the plx_IO bundle info */
    iol->channel = channel;
    iol->next_IO = NULL;
  }
  for(;								    /* Find either matching channel and bundle or end of... */
      (iol->channel != channel ||				    /* ...the input list if the matching channel/bundle... */
       iol->bundle != tempBundle) && iol->next_IO;		    /* ...pair can't be found. */
      iol = iol->next_IO);
  if ((iol->channel != channel) ||				    /* Is this a new channel/bundle pair? */
      (iol->bundle != tempBundle)) 
  {								    /* ...channel structure and fill it in. */
    iol->next_IO = (plx_IO_list *) Xmalloc(sizeof(plx_IO_list));
    iol = iol->next_IO;
    iol->signal_list = NULL;
    iol->format_list = NULL;
    iol->type_list = NULL;
    iol->bundle = tempBundle;				    /* Fill in the plx_IO bundle info */
    iol->channel = channel;
    iol->next_IO = NULL;
  }
  for(s0 = iol->signal_list, lastSignal = iol->signal_list;	    /* Look through the list of signals currently registered... */
      s0;							    /* ...to check for duplicates.  Don't add any duplicates. */
      s0 = s0->next_signal)
  {
    lastSignal = s0;						    /* Remember the previously examined item on the list */
    if ((s0->signal->standard == src->standard) &&		    /* Has the requested signal structure already been added? */
	(s0->signal->w == src->w) &&
	(s0->signal->h == src->h) &&
	(s0->signal->b == src->b) &&
	(s0->signal->r == src->r) &&
	(s0->signal->interlaced == src->interlaced) &&
	(s0->signal->sync_ok == src->sync_ok))
      return;							    /* Yes, don't bother trying to add it again */
  }
  s0 = (plx_signal_list *) Xmalloc(sizeof(plx_signal_list));	    /* If this stmt reached, no match was found on the list */
  if (NULL == iol->signal_list)					    /* Was the list empty to begin with? */
    iol->signal_list = s0;					    /* Yes, make the newly-allocated node the 1st on the list */
  else								    /* No, the list had entries, so... */
    lastSignal->next_signal = s0;				    /* ...add the new entry to the end of the list. */
  s0->signal = sig = (plx_signal *) Xmalloc(sizeof(plx_signal));
  s0->next_signal = NULL;
  sig->standard = src->standard;
  sig->w = src->w;
  sig->h = src->h;
  sig->b = src->b;
  sig->r = src->r;
  sig->interlaced = src->interlaced;				    /* The source passed in may use this field for signal... */
  sig->sync_ok = src->sync_ok;					    /* ...instance info; take the info as-is. */
} /* end function plxaddsig */



plx_IO *XPlxQueryConfig(register Display*	dpy,
			Drawable		d,
			GC			gc)
{
  register XExtCodes*		codes;
  xVideoConfigReply		rep;
  register xVideoReq*		req;
  register plx_signal*		sig;
  register plx_signal_list*	siglist;
  register plx_format_list*	formlist;
  register PlxVideoSignal*	xsig;
  register PlxVideoFormat*	xform;
  register PlxVideoType*	xtype;
  register i;
  int nbytes;
  char *tbuff;
  XPlxHardwareInfo	hardwareInfo;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvQueryConfig(dpy, d, gc));
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoQueryConfig;
  if(!_XReply(dpy, (xReply *) &rep, 0, xFalse))
  {
    UnlockDisplay(dpy);
    SyncHandle();
    return (0);
  }
  
  nbytes = (rep.nformats * sizeof(PlxVideoFormat)) +
    (rep.ntypes * sizeof(PlxVideoType)) +
      (rep.nsignals * sizeof(PlxVideoSignal));
  
  tbuff = Xmalloc(nbytes);
  
  /*
     _XRead32(dpy, tbuff, nbytes);
     */
  _XReadPad(dpy, tbuff, nbytes);
  
  hardwareInfo.allBits = rep.hardware;
  plx_io.hardware = hardwareInfo.info.hardware;
/*
  printf("\n\n!!!!!!\tTemporary Hack! in XPlxQueryConfig().\n");
*/
  plx_io.numJPEGCompressors = 1;
/*
  printf("Hack ends here.\n\n");
*/
  plx_io.overlay = hardwareInfo.info.overlay;
  plx_io.zoom = hardwareInfo.info.zoom;
  plx_io.numMPEGDecoders = hardwareInfo.info.numMPEGDecoders;
  plx_io.numH320Compressors = hardwareInfo.info.numH320Compressors;
  
  PlxDeletePlxHardwareList(&plx_io);				    /* Delete old data structures from prev. call to this fct. */
  plx_io.inputs = (plx_IO_list*)NULL;
  plx_io.outputs = (plx_IO_list*)NULL;
  xform = (PlxVideoFormat *) tbuff;
  for(i = 0; i < rep.nformats; ++i)
  {
    plxaddformat(xform);
    ++xform;
  }
  xtype = (PlxVideoType *) xform;
  for(i = 0; i < rep.ntypes; ++i)
  {
    plxaddtype(xtype);
    ++xtype;
  }
  xsig = (PlxVideoSignal *) xtype;
  for(i = 0; i < rep.nsignals; ++i)
  {
    plxaddsig(xsig);
    ++xsig;
  }
  
  Xfree(tbuff);
  
  POSTAMBLE
    
  return(&plx_io);
} /* end function XPlxQueryConfig */



plx_signal *XPlxQueryVideo(register Display*	dpy,
			   Drawable		d,
			   GC			gc)
{
  register XExtCodes*	codes;
  xVideoConfigReply	rep;
  register xVideoReq*	req;
  PlxVideoSignal*	sigin;
  plx_signal*		sigout;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvQueryVideo(dpy, d, gc));
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoQueryVideo;
  if(!_XReply(dpy, (xReply *) &rep, 0, xFalse))
  {
    UnlockDisplay(dpy);
    SyncHandle();
    return (0);
  }
  
  sigin = (PlxVideoSignal *) Xmalloc(sizeof(PlxVideoSignal));
  sigout = (plx_signal *) Xmalloc(sizeof(plx_signal));
  
  /*
     _XRead32(dpy, sigin, sizeof(PlxVideoSignal));
     */
  _XReadPad(dpy, sigin, sizeof(PlxVideoSignal));
  
  sigout->standard = sigin->standard;
  sigout->w = sigin->w;
  sigout->h = sigin->h;
  sigout->b = sigin->b;
  sigout->r = sigin->r;
  sigout->interlaced = sigin->interlaced;
  sigout->sync_ok = sigin->sync_ok;
  
  Xfree(sigin);
  
  POSTAMBLE
    
    return(sigout);
}



void XPlxVideoOutput(register Display*	dpy,
		     Drawable		d,
		     GC			gc,
		     int		vx,
		     int		vy,
		     int		x,
		     int		y,
		     int		w,
		     int		h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoOutput(dpy, d, gc, vx, vy, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoOutput;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void XPlxVideoScaleOutput(register Display*	dpy,
			  Drawable		d,
			  GC			gc,
			  int			vx,
			  int			vy,
			  int			vw,
			  int			vh,
			  int			x,
			  int			y,
			  int			w,
			  int			h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoScaleOutput(dpy, d, gc, vx, vy, vw, vh, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoScaleOutput;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.vw = vw;
  xva.vh = vh;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void XPlxVideoTag(register Display*	dpy,
		  Drawable		d,
		  GC			gc,
		  int			type)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoTag(dpy, d, gc, type);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoTag;
  req->length += (sizeof(int) / 4);
  Data16(dpy, (char *) &type, sizeof(int));
  
  POSTAMBLE
}



void XPlxVideoInputSelect(register Display*	dpy,
			  Drawable		d,
			  GC			gc,
			  int			channel,
			  int			standard,
			  int			format,
			  int			type)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoSelectArgs	sel;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoInputSelect(dpy, d, gc,
			   channel, standard, format, type);
    return;
  }
  type = PLX_RGB24;						    /* !!!!!Temporary fix until input model is changed */
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoInputSelect;
  
  sel.channel = channel;
  sel.standard = standard;
  sel.format = format;
  sel.type = type;
  /*
     printf("InputSelect(): Going to use channel %d, standard %d, format %d, instance %d.\n",
     sel.channel, sel.standard, sel.format, sel.type);
     */
  
  req->length += (sizeof(xVideoSelectArgs) / 4);
  Data16(dpy, (char *)&sel, sizeof(xVideoSelectArgs));
  
  POSTAMBLE
}



void XPlxVideoOutputSelect(register Display*	dpy,
			   Drawable		d,
			   GC			gc,
			   int			channel,
			   int			standard,
			   int			format,
			   int			type)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoSelectArgs	sel;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoOutputSelect(dpy, d, gc,
			    channel, standard, format, type);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoOutputSelect;
  
  sel.channel = channel;
  sel.standard = standard;
  sel.format = format;
  sel.type = type;
  
  req->length += (sizeof(xVideoSelectArgs) / 4);
  Data16(dpy, (char *)&sel, sizeof(xVideoSelectArgs));
  
  POSTAMBLE
}



void XPlxVideoValueLoad(register Display*	dpy,
			Drawable		d,
			GC			gc,
			int			sat,
			int			cont,
			int			hue,
			int			brit)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVidValueArgs	xvva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoValueLoad(dpy, d, gc, sat, cont, hue, brit);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoValueLoad;
  
  xvva.sat = sat;
  xvva.cont = cont;
  xvva.hue = hue;
  xvva.brit = brit;
  
  req->length += (sizeof(xVidValueArgs) / 4);
  Data16(dpy, (char *)&xvva, sizeof(xVidValueArgs));
  
  POSTAMBLE
}



void XPlxVideoValueQuery(register Display*	dpy,
			 Drawable		d,
			 GC			gc,
			 int*			sat,
			 int*			cont,
			 int*			hue,
			 int*			brit)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVidValueReply	xvva;

  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoValueQuery(dpy, d, gc, sat, cont, hue, brit);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoValueQuery;
  
  if(!_XReply(dpy, (xReply *) &xvva, 0, xFalse))
  {
    UnlockDisplay(dpy);
    SyncHandle();
    return;
  }
  *sat = xvva.sat;
  *cont = xvva.cont;
  *hue = xvva.hue;
  *brit = xvva.brit;
  
  POSTAMBLE
}



void XPlxVideoValueSave(register Display*	dpy,
			Drawable		d,
			GC			gc,
			int			nvselect)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVidValueArgs		xvva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoValueSave(dpy, d, gc, nvselect);
    return;
  }
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoValueSave;
  
  xvva.select = nvselect;
  
  req->length += (sizeof(xVidValueArgs) / 4);
  Data16(dpy, (char *)&xvva, sizeof(xVidValueArgs));
  
  POSTAMBLE
}



void XPlxVideoValueRecall(register Display*	dpy,
			  Drawable		d,
			  GC			gc,
			  int			nvselect)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVidValueArgs		xvva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoValueRecall(dpy, d, gc, nvselect);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoValueRecall;
  
  xvva.select = nvselect;
  
  req->length += (sizeof(xVidValueArgs) / 4);
  Data16(dpy, (char *)&xvva, sizeof(xVidValueArgs));
  
  POSTAMBLE
}



void XPlxVideoDirection(register Display*	dpy,
			Drawable		d,
			GC			gc,
			int			xdir,
			int			ydir)
{
  register XExtCodes*	codes;
  short			dirs[2];
  register xVideoReq*	req;
  
  dirs[0] = xdir;
  dirs[1] = ydir;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoDirection(dpy, d, gc, xdir, ydir);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoDirection;
  req->length += (sizeof(dirs) / 4);
  Data16(dpy, (char *) dirs, sizeof(dirs));
  
  POSTAMBLE
}



/* old functions start here */


XPlxVideoSet(register Display*	dpy,
	     Drawable		d,
	     GC			gc)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoSet(dpy, d, gc);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoSet;
  
  POSTAMBLE
}



void		XPlxVideoLive(register Display*		dpy,
			      Drawable			d,
			      GC			gc,
			      int			vx,
			      int			vy,
			      int			x,
			      int			y,
			      int			w,
			      int			h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoLive(dpy, d, gc, vx, vy, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoLive;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxVideoFlipLive(register Display*	dpy,
		  Drawable		d,
		  GC			gc,
		  int			vx,
		  int			vy,
		  int			x,
		  int			y,
		  int			w,
		  int			h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs	xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoFlipLive(dpy, d, gc, vx, vy, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoFlipLive;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void		XPlxVideoStill(register Display*	dpy,
			       Drawable			d,
			       GC			gc,
			       int			vx,
			       int			vy,
			       int			x,
			       int			y,
			       int			w,
			       int			h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs	xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoStill(dpy, d, gc, vx, vy, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoStill;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void		XPlxVideoScale(register Display*	dpy,
			       Drawable			d,
			       GC			gc,
			       int			vx,
			       int			vy,
			       int			vw,
			       int			vh,
			       int			x,
			       int			y,
			       int			w,
			       int			h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs	xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoScale(dpy, d, gc, vx, vy, vw, vh, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoScale;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.vw = vw;
  xva.vh = vh;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void		XPlxVideoSqueezeLive(register Display*	dpy,
				     Drawable		d,
				     GC			gc,
				     int		vx,
				     int		vy,
				     int		vw,
				     int		vh,
				     int		x,
				     int		y,
				     int		w,
				     int		h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoSqueezeLive(dpy, d, gc, vx, vy, vw, vh, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoSqueezeLive;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.vw = vw;
  xva.vh = vh;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void		XPlxVideoSqueezeStill(register Display*	dpy,
				      Drawable		d,
				      GC		gc,
				      int		vx,
				      int		vy,
				      int		vw,
				      int		vh,
				      int		x,
				      int		y,
				      int		w,
				      int		h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoSqueezeStill(dpy, d, gc, vx, vy, vw, vh, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoSqueezeStill;
  
  xva.vx = vx;
  xva.vy = vy;
  xva.vw = vw;
  xva.vh = vh;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void		XPlxVideoStop(register Display*	dpy,
			      Drawable		d,
			      GC		gc)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoStop(dpy, d, gc);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoStop;
  
  POSTAMBLE
}



XPlxXPutImage(Display*	dpy,
	      Drawable	d,
	      GC	gc,
	      XImage*	image,
	      int	srcX,
	      int	srcY,
	      int	destX,
	      int	destY,
	      unsigned int	width,
	      unsigned int	height)
{
  register XExtCodes*	codes;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvPutImage(dpy, d, gc, image,
		   srcX, srcY, destX, destY,
		   width, height);
    return;
  }
  XPutImage(dpy, d, gc, image, srcX, srcY, destX, destY,
	    width, height);
} /* end function XPlxXPutImage */



void		XPlxVideoPriority(register Display*	dpy,
				  Drawable		d,
				  GC			gc,
				  int			pri)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoPriority(dpy, d, gc, pri);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoPriority;
  
  xva.xva_arg1 = pri;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxVideoFrameMode(register Display*	dpy,
		   Drawable		d,
		   GC			gc,
		   int			onoff)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoFrameMode(dpy, d, gc, onoff);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoFrameMode;
  
  xva.xva_arg1 = onoff;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxVideoFrameStart(register Display*	dpy,
		    Drawable		d,
		    GC			gc,
		    int			type)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoFrameStart(dpy, d, gc, type);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoFrameStart;
  
  xva.xva_arg1 = type;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxVideoDitherControl(register Display*	dpy,
		       Drawable			d,
		       GC			gc,
		       int			luma,
		       int			chroma)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoDitherControl(dpy, d, gc, luma, chroma);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoDitherControl;
  
  xva.xva_arg1 = luma;
  xva.xva_arg2 = chroma;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void		XPlxVideoInput(register Display*	dpy,
			       Drawable			d,
			       GC			gc,
			       int			channel)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoInput(dpy, d, gc, channel);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoInput;
  
  xva.xva_arg1 = channel;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxZoomOn(register Display*	dpy,
	   Drawable		d,
	   GC			gc,
	   int			x,
	   int			y,
	   int			w,
	   int			h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvZoomOn(dpy, d, gc, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_zoom_on;
  
  
  xva.vx = x;
  xva.vy = y;
  xva.vw = w;
  xva.vh = h;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxZoomOff(register Display*	dpy,
	    Drawable		d,
	    GC			gc,
	    int			x,
	    int			y)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvZoomOff(dpy, d, gc, x, y);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_zoom_off;
  
  xva.xva_arg1 = x;
  xva.xva_arg2 = y;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxDebug(register Display*	dpy,
	  Drawable		d,
	  GC			gc,
	  unsigned short	val)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvDebug(dpy, d, gc, val);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_Debug;
  
  xva.xva_arg1 = val;
  
  req->length += sizeof(xVideoArgs) / 4;
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



XPlxVideoValueFactory(register Display*	dpy,
		      Drawable		d,
		      GC		gc)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoValueArgs	xvva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoValueFactory(dpy, d, gc);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoValueFactory;
  
  xvva.user0 = 0;
  xvva.user1 = 0;
  xvva.saturation = 0;
  xvva.contrast = 0;
  xvva.lgain = 0;
  xvva.cgain = 0;
  xvva.hue = 0;
  xvva.brightness = 0;
  
  req->length += sizeof(xVideoValueArgs) / 4;
  Data16(dpy, (char *)&xvva, sizeof(xVideoValueArgs));
  
  POSTAMBLE
}



XPlxVideoValueSetUser0(register Display*	dpy,
		       Drawable			d,
		       GC			gc,
		       int			sat,
		       int			cont,
		       int			lgain,
		       int			cgain,
		       int			hue,
		       int			bri)
{
  _XPlxVideoValueSet(dpy, d, gc,
		     1, 0, sat, cont, lgain, cgain, hue, bri);
}

XPlxVideoValueSetUser1(register Display*	dpy,
		       Drawable			d,
		       GC			gc,
		       int			sat,
		       int			cont,
		       int			lgain,
		       int			cgain,
		       int			hue,
		       int			bri)
{
  _XPlxVideoValueSet(dpy, d, gc,
		     0, 1, sat, cont, lgain, cgain, hue, bri);
}

XPlxVideoValueSet(register Display*	dpy,
		  Drawable		d,
		  GC			gc,
		  int			sat,
		  int			cont,
		  int			lgain,
		  int			cgain,
		  int			hue,
		  int			bri)
{
  _XPlxVideoValueSet(dpy, d, gc,
		     0, 0, sat, cont, lgain, cgain, hue, bri);
}

_XPlxVideoValueSet(register Display*	dpy,
		   Drawable		d,
		   GC			gc,
		   int			user0,
		   int			user1,
		   int			sat,
		   int			cont,
		   int			lgain,
		   int			cgain,
		   int			hue,
		   int			bri)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoValueArgs	xvva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoValueSet(dpy, d, gc, user0, user1,
			sat, cont, lgain, cgain, hue, bri);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_VideoValueSet;
  
  xvva.user0 = 0;
  xvva.user1 = 0;
  xvva.saturation = sat;
  xvva.contrast = cont;
  xvva.lgain = lgain;
  xvva.cgain = cgain;
  xvva.hue = hue;
  xvva.brightness = bri;
  
  req->length += sizeof(xVideoValueArgs) / 4;
  Data16(dpy, (char *)&xvva, sizeof(xVideoValueArgs));
  
  POSTAMBLE
}



XPlxPan(register Display*	dpy,
	Drawable		d,
	GC			gc,
	int			x,
	int			y)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvPan(dpy, d, gc, x, y);
    return;
  }
    
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_Pan;
  
  xva.x_pan_x = x;
  xva.x_pan_y = y;
  
  req->length += (sizeof(xVideoArgs) / 4);
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void		XPlxScaleCopyArea(register Display*	dpy,
				  Drawable		sd,
				  Drawable		dd,
				  GC			gc,
				  int			sx,
				  int			sy,
				  int			sw,
				  int			sh,
				  int			dx,
				  int			dy,
				  int			dw,
				  int			dh)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xCVideoArgs	xcva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvVideoScaleCopyArea(dpy, sd, dd, gc,
			     sx, sy, sw, sh, dx, dy, dw, dh);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = dd;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_ScaleCopyArea;
  
  xcva.sx = sx;
  xcva.sy = sy;
  xcva.sw = sw;
  xcva.sh = sh;
  xcva.dx = dx;
  xcva.dy = dy;
  xcva.dw = dw;
  xcva.dh = dh;
  xcva.drawable = sd;
  
  req->length += (sizeof(xCVideoArgs) / 4);
  Data16(dpy, (char *)&xcva, sizeof(xCVideoArgs));
  
  POSTAMBLE
}



XPlxImageLoad(register Display*	dpy,
	      Drawable		d,
	      GC		gc,
	      int		address,
	      int		x,
	      int		y,
	      int		w,
	      int		h)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  xVideoArgs		xva;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvImageLoad(dpy, d, gc, address, x, y, w, h);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_ImageLoad;
  
  xva.vw = address & 0xffff;
  xva.vh = (address >> 16) & 0xffff;
  xva.x = x;
  xva.y = y;
  xva.w = w;
  xva.h = h;
  
  req->length += (sizeof(xVideoArgs) / 4);
  Data16(dpy, (char *)&xva, sizeof(xVideoArgs));
  
  POSTAMBLE
}



void XPlxSetVideoParameterString(register Display*	dpy,
				 Drawable		d,
				 GC			gc,
				 char*			buf)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  int			count;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvSetVideoParameterString(dpy, d, gc, buf);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_SPString;
  
  count = strlen(buf) + 1;
  
  req->length += ((count + 3) / 4);
  Data(dpy, buf, count);
  
  POSTAMBLE
} /* end function XPlxSetVideoParameterString */



char* XPlxGetVideoParameterString(register Display*	dpy,	    /* Get parameter string */
				  Drawable		d,	    /*	from the server */
				  GC			gc)
{
  register XExtCodes*	codes;
  xVideoConfigReply	rep;
  register xVideoReq*	req;
  char*			data;
  long			nbytes;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvGetVideoParameterString(dpy, d, gc));
  }
  
  GetReq(Video, req);
  
  /*
   * first set up the standard stuff in the request
   */
  
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_GPString;
  
  if (_XReply (dpy, (xReply *) &rep, 0, xFalse) == 0) {
    UnlockDisplay(dpy);
    SyncHandle();
    return (char *)NULL;
  }
  
  nbytes = (long)rep.length << 2;
  
  data = (char *) Xmalloc((unsigned) nbytes);
  
  if (! data) {
    _XEatData(dpy, (unsigned long) nbytes);
    UnlockDisplay(dpy);
    SyncHandle();
    return (char *) NULL;
  }
  
  _XReadPad (dpy, data, nbytes);
  
  POSTAMBLE
    
    return (data);
} /* end function XPlxGetVideoParameterString */



Bool XPlxSelectHardwareInstance(Display*	dpy,
				int		instance)
{
  register XExtCodes*	codes;
  GC			gc = (GC)NULL;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvSelectHardwareInstance(dpy, instance));
  }
  
  if (0 == instance)						    /* For now, only single boards per screen are supported,... */
    return(True);						    /* ...so return True if the 1st board was requested,... */
  else								    /* ...otherwise return False. */
    return(False);
} /* end function PlxSelectHardwareInstance */



int XPlxGetHardwareInstances(Display*	dpy)
{
  register XExtCodes*	codes;
  GC			gc = (GC)NULL;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvGetHardwareInstances(dpy));
  }
  
  return(1);							    /* Multiple boards not yet implemented; return 1 board */
} /* end function XPlxGetHardwareInstances */



int XPlxGetZoomFactor(Display*	dpy,
		      Drawable	d)
{
  register XExtCodes*	codes;
  GC			gc = (GC)NULL;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvGetZoomFactor(dpy, d));
  }
  
  return(1);							    /* Hardware zoom not yet supported on XVideo on Sun */
} /* end function XPlxGetZoomFactor */



Bool XPlxSetZoomFactor(Display*	dpy,
		       Drawable	d,
		       int	zoomFactor)
{
  register XExtCodes*	codes;
  GC			gc = (GC)NULL;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    return(XPlxXvSetZoomFactor(dpy, d, zoomFactor));
  }
  
  if (1 == zoomFactor)						    /* Since hardware zoom is not yet supported on Sun... */
    return(True);						    /* ...XVideo, return success only if the default zoom... */
  else								    /* ...level is requested. */
    return(False);
} /* end function XPlxSetZoomFactor */



XPlxCommand(register Display*	dpy,
	    Drawable		d,
	    GC			gc,
	    short*		buf)
{
  register XExtCodes*	codes;
  register xVideoReq*	req;
  int			count;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvCommand(dpy, d, gc, buf);
    return;
  }
  
  GetReq(Video, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->reqType = VideoReqCode;
  req->videoReqType = X_Plx_Command;
	
  count = (*buf * 2) + 2;
  req->length += ((count + 3) / 4);
  Data16(dpy, (char *) buf, count);
  
  POSTAMBLE
}



void XPlxControl(Display*	dpy,
		 Atom		op,
		 int		dataInSize,
		 char*		dataInPtr,
		 int*		dataOutSizeReturn,
		 char**		dataOutPtrReturn)
{
  register XExtCodes*		codes;
  xControlReply			rep;
  register xPlxControlReq*	req;
  char*				data;
  long				nbytes;
  GC				gc = (GC)NULL;
  
  if (((char*)NULL == dataInPtr) ||				    /* Need the place holders to be non-NULL */
      ((char**)NULL == dataOutPtrReturn) ||
      ((int*)NULL == dataOutSizeReturn))
    return;
  
  PREAMBLE(codes);
  if (0 == VideoReqCode)					    /* Using the Xv extension? */
  {								    /* Yes, call the proper Xv function instead of the XPlx one */
    XPlxXvControl(dpy, op, dataInSize, dataInPtr,
		  dataOutSizeReturn, dataOutPtrReturn);
    return;
  }
  
  GetReq(PlxControl, req);
  
  req->reqType = VideoReqCode;					    /* Set up the standard stuff in the request */
  req->videoReqType = X_PlxControl;
  req->size = dataInSize;
  req->length += ((dataInSize + 3) / 4);
  printf("+++++Debugging:XPlxControl(): is Data() call word-aligned properly?\n");
  Data(dpy, dataInPtr, dataInSize);				    /* Send the data across, if necessary */
  
  if (_XReply (dpy, (xReply *) &rep, 0, xFalse) == 0)		    /* Receive a reply, if necessary */
  {
    UnlockDisplay(dpy);
    SyncHandle();
    return;
  }
  
  nbytes = (long)rep.length << 2;				    /* Reply length is number of words; convert to # of bytes */
  data = (char *) Xmalloc((unsigned) nbytes);			    /* Allocate enough bytes to hold reply for this request */
  if (! data)
  {
    _XEatData(dpy, (unsigned long) nbytes);
    UnlockDisplay(dpy);
    SyncHandle();
    return;
  }
  _XReadPad (dpy, data, nbytes);
  
  POSTAMBLE;
  
  *dataOutSizeReturn = rep.size;				    /* Return number of valid data bytes received in the reply */
  *dataOutPtrReturn = data;					    /* Point to data received in the reply */
  return;
} /* end function XPlxControl */
