/*
 *   Copyright (c) 1993 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* $Header: /plx/source/Xclients/generic/RCS/XPlxXvCompress.c,v 1.36 1995/12/14 03:00:02 drapeau Exp $ */
/* $Log: XPlxXvCompress.c,v $
 * Revision 1.36  1995/12/14  03:00:02  drapeau
 * Major changes to this module to accommodate the new XPlx library that
 * automatically detects the proper video extension to use.  The new
 * library attempts to use the ParallaxVideo extension if possible and
 * falls back to use the XVideo extension if need be.
 *
 * Revision 1.35  1995/06/30  18:09:26  drapeau
 * Removed a hack from the XPlxGetCImage() function: the hack had been to
 * compress an extra sixteen lines to attempt to avoid random squares of
 * bad data at the bottom right corner of a JPEG image, due to a hardware
 * error.  A software workaround has been found for the hardware error,
 * so the client-side hack is no longer necessary, at least on the IBM
 * platform.  Other platform fixes will follow.
 *
 * Revision 1.34  1995/06/24  00:18:48  drapeau
 * Added support for the function XPlxXPutImage() to replace XPutImage()
 * for the overlay product on the HP.
 *
 * Revision 1.33  1995/06/23  19:26:39  drapeau
 * Fixed error in XPlxGetCImage(): changed round-down parameters from 8
 * to 16, according to information from Dave Arnold about the proper
 * round-down values to use when creating JPEG images that can be used
 * for software-only decoders.
 *
 * Revision 1.32  1995/06/21  03:27:38  drapeau
 * Fixed minor error in diagnostic message reporting: xn XPlxPutTable(),
 * the first Port was always reported as attached even if was not.  This
 * error has been fixed.
 *
 * Revision 1.31  1995/05/30  22:58:16  drapeau
 * Minor changes, using new XPlxGetCurrentHardwareInstance() instead of
 * old Xv analog version.
 *
 * Revision 1.30  1995/05/27  00:34:52  drapeau
 * Fixed an error in the way that multiple adaptors are supported.  If a
 * computer has several adaptors but only some of them are Parallax
 * adaptors, the previous version of this module would incorrectly count
 * the non-Parallax adaptors as Parallax adaptors.  This error has been
 * fixed.
 *
 * Revision 1.29  1995/05/25  18:19:38  drapeau
 * Removed the hack to skip past the first four bytes of JPEG data when
 * creating an image, destroying an image, and getting a new compressed
 * image.  The code is now in the X server on the HP and IBM platforms.
 *
 * Revision 1.28  1995/05/25  18:12:52  drapeau
 * Fixed an error in XPlxPutTable(): another incarnation of the problem
 * of a single process trying to use two Displays.  The old version only
 * called XvQueryAdaptors() once and cached the base Port ID for that
 * adaptor, which could fail if the two Adaptors on two separate Displays
 * have different starting base Port ID's.  The new version checks if the
 * Display has changed from the last call; if so, the QueryAdaptors is
 * called again.
 *
 * Revision 1.27  1995/05/25  01:59:03  drapeau
 * Optimized the MakeXvCompressionAtoms() function so that Atoms are only
 * created once per Display.  The Atoms are now stored in the Display's
 * extension data structure.
 *
 * Revision 1.26  1995/05/23  06:52:26  drapeau
 * Major changes to the module to better support multiple displays per
 * process.  This module should now work correctly when used by a single
 * process that opens multiple Display's and uses both for
 * compression/decompression requests.
 *
 * Revision 1.25  1995/04/21  20:36:43  drapeau
 * Removed extraneous debugging message now that multiple board support
 * works.
 *
 * Revision 1.24  1995/04/18  01:11:56  drapeau
 * Added support for server improvements that export 8 Ports instead of
 * the 3 previously made public.  The new Ports allow multiple JPEG
 * windows to be displayed simultaneously; the old version of both client
 * and server allowed only 3 JPEG windows maximum.
 *
 * Revision 1.23  1995/04/12  23:59:26  drapeau
 * Added support for printing diagnostic messages if the environment
 * variable "PlxDebug" is set.  This was implemented as an aid in
 * debugging.
 *
 * Revision 1.22  1994/10/22  05:30:12  drapeau
 * Made the code more ANSI-compatible.
 *
 * Revision 1.21  1994/10/14  01:01:32  drapeau
 * Fixed XPlxPutTable(), removing some code that had previously been put
 * in to assure that compression tables were installed for whatever Port
 * was being used for playback or record.  The previous version of the
 * code installed Q tables for all Ports available from the X server;
 * this version installs the Q table only for the specifically requested
 * Port.
 *
 * Revision 1.20  1994/10/10  18:26:16  drapeau
 * Modified XPlxGetCImage() to conform to the JPEG baseline compression
 * specification, which states that images are compressed in 8x8 blocks.
 * This means that the width & height parameters of the XPlxGetCImage()
 * function need to be multiples of 8.  The function now enforces this
 * rule.
 *
 * Revision 1.19  1994/10/09  02:34:27  drapeau
 * Fixed an error in XPlxPutTable(): the function now correctly sets the
 * "xvJpegPort" variable each time the function is called.  This enables
 * the function to search for JPEG ports starting from the end of the
 * list of available ports and working toward the beginning of the list.
 * Before this fix, the search occurred once, and subsequent calls to
 * this function started with the last value of xvJpegPort which was
 * most likely not at the end of the list of available Ports.
 *
 * Revision 1.18  1994/10/03  16:18:43  drapeau
 * Several changes, mainly to support output of JPEG images but also to
 * support programs that connect with multiple displays.
 *
 * * Modified MakeXvAtoms(), allowing new Atoms to be interned both when
 *   the function is called the first time and when a different Display*
 *   is passed in as argument (different X servers might use different
 *   Atom designations; that's why this test is important).
 *
 * * Modified PutImageRequest(), expicitly checking if a Port is
 *   available for displaying compressed images.
 *
 * * Added diagnostic messages for debugging purposes.  They are
 *   commented out for the code to be released.
 *
 * Revision 1.17  1994/09/26  00:09:47  drapeau
 * Removed extraneous printf messages (error messages deemed unnecessary
 * to print since they might alarm the person using the video apps).
 *
 * Revision 1.16  1994/09/23  06:24:56  drapeau
 * Modified the order in which ports are searched to find a free JPEG
 * port.  Before, ports were searched starting from the 2nd available
 * (base_id + 1) and increasing in value.  The new scheme starts at the
 * last available port and works back toward the 2nd port (the 1st port
 * is never allowed to be used for JPEG operations).
 *
 * Revision 1.15  1994/09/03  01:54:58  drapeau
 * Fixed the order of include files to take advantage of function
 * prototypes.
 *
 * Also, typecast parameters to PutImageRequest() so that they are sent
 * over the wire correctly.
 *
 * Revision 1.14  1994/06/09  22:33:43  drapeau
 * Several changes:
 *
 * 1) Added new Atoms to detect the first time a compression request is
 *    made.
 *
 * 2) Fixed XPlxGetCImage(), handling the case when a bad compression is
 *    done and therefore there are no image bytes to retrieve
 *    (nbytes <= 0).  In this case, the function should not try to
 *    consume anything from the X socket.
 *
 * 3) Added code to XPlxPutTable() to properly request a port for
 *    compression requests.
 *
 * Revision 1.13  1994/03/29  01:34:31  drapeau
 * Major changes to re-implement the XPlx->Xv protocol translation for
 * compression requests.  Andy added support for compression in the Xv
 * library, so the corresponding XPlx requests changed.
 *
 * Revision 1.12  1994/01/20  01:17:37  drapeau
 * Added a short-term fix for a problem in displaying JPEG images.  The
 * problem seems to be that when doing an XPlxGetCImage(), part of the
 * image is not gotten.  So Bob Goodwin suggested to ask for several
 * lines more than really needed, to assure that the portion needed will
 * always be safe.  I chose an arbitrary value of 16 extra lines; 8 extra
 * lines didn't work (16 may not always work but seemed to more often
 * than 8 did).
 *
 * Revision 1.11  1993/12/15  06:09:37  drapeau
 * Slight modifications to make the code more portable, and self-reliant
 * upon the Parallax software release.  Currently, not all workstations
 * release the Xv X extension as part of their distributed software, so
 * Parallax must distribute its own source and header files.  These small
 * changes allow this software to use the Parallax version of Xv.
 *
 * Revision 1.1  1993/11/20  00:40:15  drapeau
 * Initial revision
 * */
static char xPlxXvCompressRcsid[] = "$Header: /plx/source/Xclients/generic/RCS/XPlxXvCompress.c,v 1.36 1995/12/14 03:00:02 drapeau Exp $";

#define NEED_REPLIES
#include	<Xvlibint.h>
#include	<Xvplx.h>
#include	<Xvplxproto.h>					    /* Pick up Parallax-specific protocol requests */
#include	<XPlxExt.h>
#include	"WindowToPortID.h"

static char*		PlxDebugEnvVar = NULL;			    /* Library variable indicating whether to print debug msgs. */
static Bool		plxDebuggingOn = False;			    /* Accompanying internal variable to examine for debug msgs */
static XExtCodes	staticXvCodes;
XExtCodes*		_XvCodes = &staticXvCodes;
XvPortID		xvJpegPort;

#define	PREAMBLE(stat) \
 LockDisplay(dpy); \
 if (!PlxXvInitExtension(dpy)) \
   { \
     UnlockDisplay(dpy); SyncHandle(); return; \
   }

#define POSTAMBLE \
  UnlockDisplay(dpy); \
  SyncHandle()


static Atom	qTableCompressionAtom;
static Atom	qTableDecompressionAtom;
static Atom	getCImageAtom;
static Atom	putCImageAtom;
static Atom	putFirstCImageAtom;
static Atom	xvStatusAtom;


static void MakeXvCompressionAtoms(Display* dpy)		    /* Create Atoms to be used for setting Xv attributes */
{
  XvAtomsRec*		atomInfo;
  XExtData*		extensionData;
  XExtData*		newExtensionData;
  Bool			atomInfoFound = False;
  static Bool		firstTimeCalled = True;
  
  if (firstTimeCalled == True)					    /* Only do this work once */
  {
    firstTimeCalled = False;
    PlxDebugEnvVar = getenv("PlxDebug");			    /* Determine whether debugging should be enabled */
    if ((char*)NULL != PlxDebugEnvVar)				    /* Did the process have debugging enabled? */
    {
      plxDebuggingOn = True;					    /* Yes, enable debugging for this library */
    }
    else
      plxDebuggingOn = False;					    /* No, don't bother printing debugging msgs for this library */
  }
  for (extensionData = dpy->ext_data;				    /* Look at this Display's extension data for Xv Atoms info */
       extensionData != (XExtData*)NULL;
       extensionData = extensionData->next)
  {
    if (XvPlxAtomsCode == extensionData->number)		    /* Was the Xv Atoms info found? */
    {
      atomInfoFound = True;					    /* Yes, mark a flag saying so,... */
      atomInfo = (XvAtomsRec*)extensionData->private_data;	    /* ...point to the Xv Atoms data,... */
      break;							    /* ...and quit looking. */
    }
    if ((XExtData*)NULL == extensionData->next)			    /* The end of the list was reached */
      break;
  } /* end for (extensionData traversal) */
  if (False == atomInfoFound)					    /* No Xv Atoms data were found, create them and add them... */
  {								    /* ...to the Display's extension data for future use. */
    newExtensionData = Xmalloc(sizeof(XExtData));
    newExtensionData->private_data = Xmalloc(sizeof(XvAtomsRec));
    newExtensionData->free_private = (int (*)())free;
    newExtensionData->number = XvPlxAtomsCode;
    newExtensionData->next = (struct _XExtData*)NULL;
    if (extensionData != (XExtData*)NULL)			    /* If this Display already had extension info, add the... */
    {								    /* ...Xv Atoms info to the end of the extension info list. */
      newExtensionData->next = extensionData->next;
      extensionData->next = newExtensionData;
    }
    else							    /* This Display had no extension info, so add the Plx... */
      dpy->ext_data = newExtensionData;				    /* ...info as the only extension info so far. */
    atomInfo = (XvAtomsRec*)newExtensionData->private_data;	    /* Point to the Xv Atoms extension info for filling in */
    atomInfo->tableCompression =
      XInternAtom(dpy,"XV_PLX_TABLE_COMPRESSION",False);
    atomInfo->tableDecompression =
      XInternAtom(dpy,"XV_PLX_TABLE_DECOMPRESSION",False);
    atomInfo->getCImage = XInternAtom(dpy,"XV_PLX_GETCIMAGE",False);
    atomInfo->putCImage = XInternAtom(dpy,"XV_PLX_PUTCIMAGE",False);
    atomInfo->firstPutCImage =
      XInternAtom(dpy,"XV_PLX_FIRST_PUTCIMAGE",False);
    atomInfo->status = XInternAtom(dpy, "XV_STATUS", False);
    atomInfo->hue = XInternAtom(dpy,"XV_HUE",False);
    atomInfo->saturation = XInternAtom(dpy,"XV_SATURATION",False);
    atomInfo->brightness = XInternAtom(dpy,"XV_BRIGHTNESS",False);
    atomInfo->contrast = XInternAtom(dpy,"XV_CONTRAST",False);
    atomInfo->encoding = XInternAtom(dpy, "XV_ENCODING", False);
    atomInfo->valueRecall = XInternAtom(dpy, "XV_PLX_VALUE_RECALL", False);
    atomInfo->valueSave = XInternAtom(dpy, "XV_PLX_VALUE_SAVE", False);
    atomInfo->videoParameterString = XInternAtom(dpy, "XV_PLX_VIDEO_PARAMETER_STRING", False);
    atomInfo->querySignals = XInternAtom(dpy, "XV_PLX_QUERY_SIGNALS", False);
    atomInfo->crop = XInternAtom(dpy, "XV_PLX_CROP", False);
    atomInfo->zoom = XInternAtom(dpy, "XV_PLX_ZOOM", False);
    atomInfo->xImage = XInternAtom(dpy, "XV_IMAGE_XIMAGE", False);
  } /* end if atom info wasn't found */
  qTableCompressionAtom = atomInfo->tableCompression;
  qTableDecompressionAtom = atomInfo->tableDecompression;
  getCImageAtom = atomInfo->getCImage;
  putCImageAtom = atomInfo->putCImage;
  putFirstCImageAtom = atomInfo->firstPutCImage;
  xvStatusAtom = atomInfo->status;
  return;
} /* end function MakeXvCompressionAtoms */



XPlxCImage* XPlxXvGetCImage(Display*		dpy,
			    Drawable		d,
			    GC			gc,
			    int			src_x,
			    int			src_y,
			    unsigned int	src_w,
			    unsigned int	src_h,
			    unsigned int	dst_w,
			    unsigned int	dst_h)
{
  xvPlxGetImageReply		rep;				    /* Reply header buffer. */
  register xvPlxImageReq*	req;
  char*				data;
  long				nbytes;
  XPlxCImage*			image;
  WindowToPortID*		windowToPortIDMap;
  
  if ((src_w < 8) || (src_h < 8) || (dst_w < 8) || (dst_h < 8))	    /* Is requested image size too small any parameter? */
    return((XPlxCImage*)NULL);					    /* Yes, don't bother trying to compress */
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, this Window is not on the Parallax side of the... */
    return((XPlxCImage*)NULL);					    /* ...frame buffer, so it cannot be compressed.  Return now. */
  
  if ((src_w % 16) != 0)					    /* Assure that all width & height parameters sit on a... */
    src_w -= (src_w % 16);					    /* ...16 pixel boundary. Not sure why, but Dave Arnold... */
  if ((src_h % 16) != 0)					    /* ...says this needs to be. */
    src_h -= (src_h % 16);
  if ((dst_w % 16) != 0)
    dst_w -= (dst_w % 16);
  if ((dst_h % 16) != 0)
    dst_h -= (dst_h % 16);
  
  MakeXvCompressionAtoms(dpy);					    /* Make sure Atoms are all known */
  PREAMBLE(XvBadExtension);
  XvGetReq(PlxImage, req);
  /*
   * first set up the standard stuff in the request
   */
  req->drawable = d;
  req->gc = gc->gid;
  req->src_x = src_x;
  req->src_y = src_y;
  req->src_w = src_w;
  req->src_h = src_h;
  req->dst_x = 0;
  req->dst_y = 0;
  req->dst_w = dst_w;
  req->dst_h = dst_h;
  req->xvReqType = xv_PlxImage;
  req->attribute = getCImageAtom;
  req->shmseg = 0;
  req->port = windowToPortIDMap->xvPortID;			    /* Use the port for this window for the compression */
  
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: Calling XPlxGetCImage() with: dpy=%d, Drawable=%d, Port=%d, gc=%d, srcX=%d, srcY=%d, srcW=%d, srcH=%d, dstW=%d, dstH=%d.\n",
	   dpy, d, req->port, gc, src_x, src_y, src_w, src_h, dst_w, dst_h);
  }
  
  if (_XReply (dpy, (xReply *) &rep, 0, xFalse) == 0)
  {
    UnlockDisplay(dpy);
    SyncHandle();
    return (XPlxCImage *)NULL;
  }
  
  nbytes = (long)rep.length << 2;
  data = (char *) Xmalloc((unsigned) nbytes);
  if ((! data) && (nbytes > 0))					    /* Did Xmalloc fail and is there data to transfer?  */
  {								    /* Yes, eat the data from the X protocol reply */
    _XEatData(dpy, (unsigned long) nbytes);
    UnlockDisplay(dpy);
    SyncHandle();
    return (XPlxCImage *) NULL;
  }
  if (nbytes <= 0)						    /* Was an error returned from the server? */
  {								    /* Yes, don't bother creating a CImage  */
    return((XPlxCImage*)NULL);
  }
  else								    /* No, the X request went fine, create & return a CImage */
  {
    _XReadPad (dpy, data, nbytes);
    image = XPlxCreateCImage (dpy, data, rep.size, dst_w, dst_h);
  }
  POSTAMBLE;
  return(image);
} /* end function XPlxGetCImage */



static void XPlxXvSendImage(Display*		dpy,
			    xvPlxImageReq*	req,
			    XPlxCImage*		image)
{
  long			length;
  unsigned char*	src;
  unsigned char*	dest;
  unsigned char*	shifted_src = NULL;
    
  req->leftPad = 0;
  length = image->size;
  req->length += (length + 3) >> 2;
  src = (unsigned char *)image->data;
  Data(dpy, (char *)src, length);
} /* end function XPlxXvSendImage */



void XPlxXvPutImageRequest(Display*		dpy,
			   Drawable		d,
			   GC			gc,
			   XPlxCImage*		image,
			   int			src_x,
			   int			src_y,
			   unsigned int		src_w,
			   unsigned int		src_h,
			   int			dst_x,
			   int			dst_y,
			   unsigned int		dst_w,
			   unsigned int		dst_h,
			   int			flag)
{
  int				adaptorNumber = 0;
  XvAdaptorInfo*		adaptorInfo;
  XvAdaptorInfo*		adaptorList;
  static XvPortID		firstPortID;
  int				numPlxAdaptors;
  unsigned int			numberOfAdaptors;
  static int			numberOfPorts;
  int				counter;
  int				result;
  int				portIsUsed;
  register xvPlxImageReq*	req;
  static Bool			firstTime = True;
  WindowToPortID*		windowToPortIDMap;
  
  MakeXvCompressionAtoms(dpy);					    /* Assure that the Xv Atoms are known to the X server */
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, choose a port and use it for this window */
  {
    xvJpegPort = 0;						    /* Initialize portID used for JPEG compression */
    result = XvQueryAdaptors(dpy, d,				    /* Use the Xv method of getting information about the... */
			     &numberOfAdaptors,			    /* ...hardware. */
			     &adaptorList);
    if (result != Success)					    /* Did the call fail for some reason? */
      return;							    /* Yes, return a NULL pointer immediately */
    if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
      return;							    /* No, return a NULL pointer immediately */
    numPlxAdaptors = XPlxGetHardwareInstances(dpy);		    /* Determine how many of the adaptors seen are Plx */
    adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
    adaptorNumber = XPlxGetCurrentHardwareInstance(dpy);	    /* See if a specific Adaptor is being requested */
    if (adaptorNumber < 0)					    /* If no specific Adaptor requested, use the 1st one */
      adaptorNumber = 0;
    if (adaptorNumber > (numPlxAdaptors - 1))			    /* Don't allow an illegal Adaptor number to be chosen */
      adaptorNumber = 0;
    for (counter = 0;
	 counter < numberOfAdaptors;
	 counter++, adaptorInfo++)
    {
      if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		  "PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board */
      {
	if (adaptorNumber > 0)					    /* Should this function be looking at other than the 1st... */
	{							    /* ...adaptor?  If so, note that one Parallax adaptor has... */
	  adaptorNumber--;					    /* ...been seen and go to the next iteration 'til found. */
	  continue;
	}
	else
	{
	  firstPortID = adaptorInfo->base_id;			    /* Keep track of the 1st available Port for this adaptor */
	  numberOfPorts = adaptorInfo->num_ports;		    /* Keep track of the # of Ports for this adaptor */
	  break;						    /* Done searching; use this set of PortID's */
	}
      } /* end if looking at a Parallax adaptor */
    } /* end searching all Adaptors */
    xvJpegPort = firstPortID + numberOfPorts - 1;		    /* Start looking for JPEG Ports from end of Port list */
    while (xvJpegPort > firstPortID)				    /* Examine every Port except the first one, or until the... */
    {								    /* ...first unused Port is found. */
      XvGetPortAttribute(dpy, xvJpegPort, xvStatusAtom,		    /* Determine whether the requested JPEG port is in use */
			 (int*)&portIsUsed);			    /* (this atom added by Andy Goodrich) */
      if (True == plxDebuggingOn)				    /* Should a debug message be printed? */
      {								    /* Yes, print out all the ugly details */
	if ((portIsUsed & XVS_Attached) != 0)
	{
	  printf("\t\tPutImageRequest():The port %d is already attached.\n", xvJpegPort);
	}
      }
      if (((portIsUsed & XVS_Attached) != 0) &&			    /* Is the port used? Start looking at the last available... */
	  (xvJpegPort > firstPortID))				    /* ...port and go back toward the 1st port. */
	xvJpegPort--;						    /* Yes, use next avail. port, and check again */
      else							    /* No, the port is unused; no need to check any longer */
	break;
    } /* end if still decrementing the xvJpegPort variable */
    WindowToPortIDAdd(d, xvJpegPort, &windowToPortIDList);	    /* Register this port with the requested Drawable */
  } /* end if no window-to-portID map was found */
  /* !!!!!Code before this point is a hack to help get output of JPEG playback windows working!!!!! */
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, this Window is not on the Parallax side of the... */
    return;							    /* ...frame buffer, so it can't be decompressed.  Return. */
  PREAMBLE(XvBadExtension);
  XvGetReq(PlxImage, req);
  req->drawable = d;
  req->gc = gc->gid;
  req->src_x = (short)src_x;
  req->src_y = (short)src_y;
  req->src_w = (unsigned short)src_w;
  req->src_h = (unsigned short)src_h;
  req->dst_x = (short)dst_x;
  req->dst_y = (short)dst_y;
  req->dst_w = (unsigned short)dst_w;
  req->dst_h = (unsigned short)dst_h;
  req->planeMask = image->size;
  req->format = flag;						    /* Squeezed or not */
  req->xvReqType = xv_PlxImage;
  req->port = windowToPortIDMap->xvPortID;			    /* Use the port for this window for the compression */
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: Calling PutImageRequest() with: dpy=%d, Drawable=%d, Port=%d, gc=%d, srcX=%d, srcY=%d, srcW=%d, srcH=%d, destX=%d, destY=%d, destW=%d, destH=%d.\n",
	   dpy, d, req->port, gc, src_x, src_y, src_w, src_h, dst_x, dst_y, dst_w, dst_h);
  }
  if (firstTime == True)
  {
    if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
    {								    /* Yes, print out all the ugly details */
      printf("XPlx Debug: PutImageRequest(): using First CImage Atom.\n");
    }
    req->attribute = putFirstCImageAtom;
    firstTime = False;
  }
  else
    req->attribute = putCImageAtom;
  req->shmseg = 0;
  XPlxXvSendImage(dpy, req, image);
  POSTAMBLE;
} /* end function XPlxXvPutImageRequest */



void XPlxXvPutTable(Display*	dpy,
		    Drawable	d,
		    GC		gc,
		    char*	data,
		    int		size,
		    int		dir)
{
  int				adaptorNumber = 0;
  static XvPortID		firstPortID;
  int				i;
  xvSetLongPortAttributeReq*	req;				    /* Request buffer. */
  int				counter;
  int				result;
  int				portIsUsed;
  static int			firstTime = True;
  int				numPlxAdaptors;
  unsigned int			numberOfAdaptors;
  static int			numberOfPorts;
  XvAdaptorInfo*		adaptorInfo;
  XvAdaptorInfo*		adaptorList;
  XvPortID			portID;
  WindowToPortID*		windowToPortIDMap;
  static Display*		lastDisplayUsed = (Display*)NULL;
  
  MakeXvCompressionAtoms(dpy);					    /* Make sure Atoms are all known */
  if ((firstTime == True) || (dpy != lastDisplayUsed))
  {
    lastDisplayUsed = dpy;
    xvJpegPort = 0;						    /* Initialize portID used for JPEG compression */
    result = XvQueryAdaptors(dpy, d,				    /* Use the Xv method of getting information about the... */
			     &numberOfAdaptors,			    /* ...hardware. */
			     &adaptorList);
    if (result != Success)					    /* Did the call fail for some reason? */
      return;							    /* Yes, return a NULL pointer immediately */
    if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
      return;							    /* No, return a NULL pointer immediately */
    numPlxAdaptors = XPlxGetHardwareInstances(dpy);		    /* Determine how many of the adaptors seen are Plx */
    adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
    adaptorNumber = XPlxGetCurrentHardwareInstance(dpy);	    /* See if a specific Adaptor is being requested */
    if (adaptorNumber < 0)					    /* If no specific Adaptor requested, use the 1st one */
      adaptorNumber = 0;
    if (adaptorNumber > (numPlxAdaptors - 1))			    /* Don't allow an illegal Adaptor number to be chosen */
      adaptorNumber = 0;
    for (counter = 0;
	 counter < numberOfAdaptors;
	 counter++, adaptorInfo++)
    {
      if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		  "PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board */
      {
	if (adaptorNumber > 0)					    /* Should this function be looking at other than the 1st... */
	{							    /* ...adaptor?  If so, note that one Parallax adaptor has... */
	  adaptorNumber--;					    /* ...been seen and go to the next iteration 'til found. */
	  continue;
	}
	else
	{
	  firstPortID = adaptorInfo->base_id;			    /* Keep track of the 1st available Port for this adaptor */
	  numberOfPorts = adaptorInfo->num_ports;		    /* Keep track of the # of Ports for this adaptor */
	  break;						    /* Done searching; use this set of PortID's */
	}
      } /* end if looking at a Parallax adaptor */
    } /* end searching all Adaptors */
    xvJpegPort = firstPortID + numberOfPorts - 1;		    /* Hard-code the Xv PortID used for JPEG: Andy...  */
    firstTime = False;						    /* ...Goodrich's request. */
  }
  xvJpegPort = firstPortID + numberOfPorts - 1;			    /* Hard-code PortID used for JPEG: Andy Goodrich's request */
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, choose a port and use it for this window */
  {
    while (xvJpegPort > firstPortID)
    {
      XvGetPortAttribute(dpy, xvJpegPort, xvStatusAtom,		    /* Determine whether the requested JPEG port is in use */
			 (int*)&portIsUsed);			    /* (this atom added by Andy Goodrich) */
      if (True == plxDebuggingOn)				    /* Should a debug message be printed? */
      {								    /* Yes, print out all the ugly details */
	if ((portIsUsed & XVS_Attached) != 0)
	{
	  printf("\t\tXPlx Debug: XPlxPutTable():The port %d is already attached.\n", xvJpegPort);
	}
      }
      if (((portIsUsed & XVS_Attached) != 0) &&			    /* Is the port used? Start looking at the last available... */
	  (xvJpegPort > firstPortID))				    /* ...port and go back toward the 1st port. */
	xvJpegPort--;						    /* Yes, use next avail. port, do no further error checking */
      else
	break;
    } /* end if still decrementing the xvJpegPort variable */
    WindowToPortIDAdd(d, xvJpegPort, &windowToPortIDList);	    /* Register this port with the requested Drawable */
    portID = xvJpegPort;
  }
  else
    portID = windowToPortIDMap->xvPortID;
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxPutTable() calling XvSetLongPortAttribute with: dpy=%d, Port=%d, Drawable=%d, direction is %d.\n",
	   dpy, portID, d, dir);
  }
  if (dir == 0)
    XvSetLongPortAttribute(dpy, portID, qTableCompressionAtom,
			   size, (char*)data);
  else
    XvSetLongPortAttribute(dpy, portID, qTableDecompressionAtom,
			   size, (char*)data);
  return;
} /* end function XPlxXvPutTable */



Bool PlxXvInitExtension(Display* dpy)
{
  int num;
  Bool ss;
  register XExtData *exp, *nexp;
  register XExtCodes *xec;
  
  for(exp = dpy->ext_data; exp; exp = exp->next)		    /* Look at all extension data for this Display */
  {
    if(exp->number == PLXMAGICCOOKIE)				    /* Was the Parallax extension information found? */
    {								    /* Yes, use that data for the opcodes Xv should use */
      _XvCodes = (XExtCodes*) exp->private_data;
      return True;						    /* Return a Success code */
    }
    if(exp->next == NULL)					    /* Done looking at the list */
      break;
  }
  nexp = Xmalloc(sizeof(XExtData));				    /* Parallax extension info was not found */
  if (!(nexp->private_data = (XPointer)XInitExtension(dpy, XvName))) /* Try to initialize the Xv extension; was it successful? */
  {								    /* No, free the memory allocated and return a failure code */
    Xfree(nexp);
    return False;
  }
  _XvCodes = (XExtCodes*) nexp->private_data;			    /* Tell the Xv library to use the proper opcodes */
  nexp->free_private = (int (*)())free;
  nexp->number = PLXMAGICCOOKIE;
  nexp->next = (struct _XExtData*)NULL;
  if(exp)							    /* If this Display already had extension info, add the... */
  {								    /* ...Parallax info to the end of the extension info... */
    nexp->next = exp->next;					    /* ...list. */
    exp->next = nexp;
  }
  else								    /* This Display had no extension info, so add the Plx... */
    dpy->ext_data = nexp;					    /* ...info as the only extension info so far. */
  return True;
} /* end function PlxXvInitExtension */
