/*
 *   Copyright (c) 1993 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* $Header: /plx/source/Xclients/generic/RCS/XPlxXvVideo.c,v 1.43 1996/01/11 03:16:37 drapeau Exp $ */
/* $Log: XPlxXvVideo.c,v $
 * Revision 1.43  1996/01/11  03:16:37  drapeau
 * Finally, one last fix to the input model to support both multiple
 * instances of a particular format (i.e., Composite, Y/C, RGB, etc.)
 * *and* multiple input "bundles" that come into the same digitizer
 * (e.g., if one of the digitizer's input wires had extra hardware to
 * pre-process an incoming analog video signal).
 *
 * Plus, fixed some coding errors that may have an effect in the future
 * if not fixed.
 *
 * Revision 1.42  1996/01/10  08:29:29  drapeau
 * Fixed minor discrepancy in the reporting of the signal width and
 * height.  This may have to be changed back to the old way in order to
 * maintain error compatibility with overlay servers.
 *
 * Also, hard-coded the "interlaced" field to 1 since this field was used
 * by overlay servers to indicate the instance of an input type.
 *
 * Revision 1.41  1995/12/23  02:24:09  drapeau
 * Merged common code into the XPlxVideo.c module so that this module
 * contains only that code necessary to support the Xv-based video
 * extension properly.
 *
 * Revision 1.40  1995/12/20  01:00:11  drapeau
 * Modified XPlxXvQueryConfig() so that it reports that zoom and overlay
 * abilities are available on any server using the Xv extension.
 *
 * Revision 1.39  1995/12/14  03:00:50  drapeau
 * Major changes to this module to accommodate the new XPlx library that
 * automatically detects the proper video extension to use.  The new
 * library attempts to use the ParallaxVideo extension if possible and
 * falls back to use the XVideo extension if need be.
 *
 * Revision 1.38  1995/10/25  00:27:33  drapeau
 * Added reminder message for the next round of development: this module
 * should add support for the new plx_IO data structure fields.
 *
 * Revision 1.37  1995/08/17  20:23:03  drapeau
 * Fixed include file definitions to eliminate compiler errors and
 * warnings on AIX with the use of GCC.
 *
 * Revision 1.36  1995/08/17  20:17:32  drapeau
 * Modified XPlxXPutImage() to check if this platform is the AIX server.
 * If so, continue to use the existing XPutImage() call instead of the
 * newer, Andy-supplied "XvPutImage()" function.
 *
 * Revision 1.35  1995/06/24  00:31:38  drapeau
 * Fixed minor error xn XPlxXPutImage(): reversed destination and source
 * rectangles.
 *
 * Revision 1.34  1995/06/24  00:16:47  drapeau
 * Added the function XPlxXPutImage() to replace XPutImage() for the
 * overlay product on the HP.
 *
 * Revision 1.33  1995/05/31  19:34:41  drapeau
 * Removed an extraneous printf statement.
 *
 * Revision 1.32  1995/05/30  22:57:23  drapeau
 * Implemented two new functions (XPlxSelectHardwareInstance() and
 * XPlxGetCurrentHardwareInstance()) to support multiple Parallax boards
 * per Screen.  These functions had previously been implemented in the Xv
 * client-side library as convenience functions but have been moved here
 * to avoid making too many changes to the Xv library.
 *
 * Revision 1.31  1995/05/27  00:35:44  drapeau
 * Fixed an error in the way that multiple adaptors are supported.  If a
 * computer has several adaptors but only some of them are Parallax
 * adaptors, the previous version of this module would incorrectly count
 * the non-Parallax adaptors as Parallax adaptors.  This error has been
 * fixed.
 *
 * Revision 1.30  1995/05/25  02:18:34  drapeau
 * Fixed three errors:
 *
 * 1) Fixed an error in VideoChannelSelect(), in the case where a desired
 *    signal was not found or the last port checked was attached, the
 *    code used to try to forcibly use the Adaptor's base_id port.
 *    Problem was, the pointer to a current Adaptor is invalid after
 *    completing the outer for loop.  The fix is to store the base_id for
 *    the current Adaptor while still in the for loop.
 *
 * 2) Fixed an error in PlxHardwareInstances(): it never actually
 *    returned the number of hardware instances available.
 *
 * 3) Optimized MakeXvAtoms() so that Atoms are only created once per
 *    Display.  The Atoms are now stored in a new data structure that is
 *    attached to the Display's extension data field.
 *
 * Revision 1.29  1995/05/23  06:54:24  drapeau
 * Major changes to this module to better support a single process
 * opening multiple Display's and making Xv requests of both.
 *
 * Revision 1.28  1995/05/19  02:05:39  drapeau
 * Two changes:
 *
 * 1) Applied a modification that already existed for the non-Xv version
 * of this code, to XPlxQueryConfig().  The fix now deletes the old
 * plx_IO data structure before allocating a new one, thereby avoiding
 * the problem of adding duplicate signal records if a single process
 * calls XPlxQueryConfig() more than once.
 *
 * 2) Enhanced diagnostic messages for better information reporting.
 *
 * Revision 1.27  1995/05/02  20:44:38  drapeau
 * Added support for zoomed hardware when available.
 *
 * Revision 1.26  1995/04/21  20:36:51  drapeau
 * Removed extraneous debugging message now that multiple board support
 * works.
 *
 * Revision 1.25  1995/04/18  01:10:12  drapeau
 * Modified code to take advantage of server improvements, specifically,
 * the HP server now provides 8 Ports for video and JPEG operations.
 * This code uses the extra Ports mostly for JPEG operations; in the case
 * where a single process tries to use both PLX_INPUT_0 and PLX_INPUT_1,
 * the code resorts to using the old scheme of assigning the 1st input to
 * the 1st Port and the 2nd input to the 2nd Port before scanning the
 * rest of the list.
 *
 * Revision 1.24  1995/04/13  00:00:28  drapeau
 * Added support for printing diagnostic messages if the environment
 * variable "PlxDebug" is set.  This was implemented as an aid in
 * debugging.
 *
 * Revision 1.23  1994/10/03  16:19:15  drapeau
 * Added support for output of JPEG images.
 *
 * Revision 1.22  1994/09/26  00:13:42  drapeau
 * Two changes:
 *
 * 1) Begun a skeleton implementation of the Zoom feature: whenever a
 *    video application launches, this library will first reset the zoom
 *    factor to 1X.  Later versions of this library will support actual
 *    zooming.
 *
 * 2) Changed the way VideoChannelSelect() decides whether it's okay to
 *    look for a matching video signal.  Its test for whether a port is
 *    already attached to a drawable is now coded out if the channel
 *    being requested is either the output channel or the 2nd input
 *    channel.  This was done because both channels use the same Xv port
 *    (for now) and so there is a natural contention problem for apps
 *    that would use both 2nd input and output (e.g., VideoTool).  In
 *    this case, it is okay for the 2 channels to use the same port.
 *
 * Revision 1.21  1994/09/23  06:20:28  drapeau
 * Fixed an error in the way the plx_IO hardware structure is created.
 * Fewer duplicates are added to the list now.
 *
 * Also, fixed minor errors in output selection.
 *
 * Revision 1.20  1994/09/04  07:22:46  drapeau
 * Fixed function prototypes for several functions that had been
 * incorrectly declared.  Also, fixed an error in PlxVideoStill(), which
 * had been taking two extra arguments.  These two extra arguments were
 * found differing from the function prototypes, and had created an error
 * in displaying still video.  When removed, the still video
 * functionality worked fine.
 *
 * Revision 1.19  1994/08/29  20:14:58  drapeau
 * Fixed an error in XPlxVideoSqueezeStill(); the code seemed to be
 * copied straight out of XPlxVideoStill() and so forgot to set the Crop
 * atom to False.
 *
 * Also, added code to both these functions to do an extra XvPutVideo()
 * before still'ing, as per Andy's advice.
 *
 * Revision 1.18  1994/08/10  02:34:05  drapeau
 * Enhanced XPlxVideoOutputSelect() to match the functionality of the Sun
 * platform's version of this function.  Specifically, when PLX_IO_OFF is
 * selected as the output channel, output turns off (this is equivalent
 * to calling XPlxVideoStop() for a video input channel, but it's not
 * clear to me at this time whether XPlxVideoStop() also applies to
 * output channels).
 *
 * Revision 1.17  1994/08/06  01:14:24  drapeau
 * Several changes:
 *
 * * Implemented video output.
 *
 * * Removed an error in VideoChannelSelect(): a potential infinite loop
 *   could be set by changing the value of the counter variable "portID";
 *   instead, the value of the counter variable is now copied into
 *   another variable "portToUse", and that variable is modified if
 *   necessary, leaving the counter variable alone and preventing the
 *   infinite loop from occurring.
 *
 * * Added diagnostic messages used only when debugging.
 *
 * * Took into account the height of the blank interval, now that it is
 *   correctly reported.
 *
 * Revision 1.16  1994/08/01  23:15:50  drapeau
 * Added some new code to handle video output; not completely tested.
 *
 * Revision 1.15  1994/07/16  02:37:05  drapeau
 * Several major changes to the Plx-to-Xv conversion layer, mostly having
 * to do with the way inputs are handled and represented at the Plx
 * layer:
 *
 * * Removed unnecessary definition of the PlxChannelToXvEncodingID data
 *   structure (and references to it), since it is not used.
 *
 * * Removed extraneous diagnostic messages.
 *
 * * Modified XPlxQueryConfig(), removing unnecessary code to determine
 *   the available channels (i.e., digitizers).
 *
 * * Removed excess unused variable declarations from functions.
 *
 * * Changed the terminology "type" into "instance" for those functions
 *   needing to select input or output channels.  The input/output
 *   selection mechanism has been augmented to more correctly reflect the
 *   capabilities of the hardware, especially for the HP platform but
 *   also for existing XVideo-based software platforms.  The model now
 *   allows for multiple instances of a particular video signal to be
 *   selected on any given digitizer; for example, on the HP platform
 *   both digitizers have access to the same two composite input cables.
 *   The new model allows applications to select either input cable from
 *   either digitizer.  The old model did not allow this, instead
 *   assuming that the 1st composite cable was locked to the 1st
 *   digitizer, the 2nd composite cable to the 2nd digitizer.
 *
 * * "Recycled" the "interlaced" field of Plx signal data structures, now
 *   using them to store instance information as described above.  This
 *   is acceptable since the interlaced field has never been used, and
 *   will not be used for the forseeable future.
 *
 * * Removed a bad hack that forceably tied the 2nd digitizer to the 2nd
 *   available video port.  The new input model allowed this hack to be
 *   deleted.
 *
 * * Modified XPlxVideoInputSelect() to emulate previous behavior.  The
 *   "type" argument is never used anyway, so it now holds the "instance"
 *   information.  If the old "PLX_RGB24" value is passed in, the
 *   function will reset the value to the 1st instance of the signal
 *   requested.  This is the old behavior.
 *
 * Revision 1.14  1994/06/09  22:35:54  drapeau
 * Numerous changes, mostly having to do with the selection of a video
 * channel to find a valid signal, and in printing of diagnostic
 * messages.  The new code also gets valid information from sending a
 * GetLongPortAttribute request with the "query signals" Atom, thereby
 * avoiding much of the need to match information from two different data
 * structures (that returned by the query signals request and that of a
 * QueryEncodings request).
 *
 * Revision 1.13  1994/03/29  01:36:14  drapeau
 * Major changes to re-implement the XPlx->Xv protocol translation.  New
 * requests and Port Attributes were added to the Xv library, making the
 * implementation of the XPlx->Xv translation cleaner.
 *
 * Revision 1.12  1994/01/20  01:47:44  drapeau
 * Made the code more robust in handling return values from Xv queries,
 * especially in string-matching (usually replaced a call to "strcmp()"
 * with a more precise call to "strncmp()").
 *
 * Also, removed a number of diagnostic statements that were no longer
 * necessary after the code had been tested.
 *
 * Revision 1.11  1993/12/15  06:13:38  drapeau
 * Slight modifications to make the code more portable, and self-reliant
 * upon the Parallax software release.  Currently, not all workstations
 * release the Xv X extension as part of their distributed software, so
 * Parallax must distribute its own source and header files.  These small
 * changes allow this software to use the Parallax version of Xv.
 *
 * Revision 1.1  1993/11/20  00:40:15  drapeau
 * Initial revision
 * */
static char xPlxXvVideoRcsid[] = "$Header: /plx/source/Xclients/generic/RCS/XPlxXvVideo.c,v 1.43 1996/01/11 03:16:37 drapeau Exp $";

#include	<Xvlibint.h>
#include	<Xvplx.h>
#include	"WindowToPortID.h"
#include	<plxvideo.h>
#include	<XPlxExt.h>
#include	<sys/time.h>
#include	<unistd.h>
#include	<stdlib.h>

extern plx_IO	plx_io;


#define InvertedStringID	0
#define NonInvertedStringID	1
#define BWStringID		2
#define ColorStringID		3
#define GenlockStringID		4
#define FreerunStringID		5

#define MaxVideoSignalWidth	4000				    /* Pick arbitrarily large maximum possible signal size,... */
#define MaxVideoSignalHeight	3000				    /* ...for PutVideo calls when signal w&h aren't specified. */

static char*		PlxDebugEnvVar = NULL;			    /* Library variable indicating whether to print debug msgs. */
static Bool		plxDebuggingOn = False;			    /* Accompanying internal variable to examine for debug msgs */

static XvPortID	illegalOutputPort;

static Atom	hueAtom;
static Atom	saturationAtom;
static Atom	brightnessAtom;
static Atom	contrastAtom;
static Atom	encodingIDAtom;
static Atom	valueRecallAtom;
static Atom	valueSaveAtom;
static Atom	videoParameterStringAtom;
static Atom	querySignalsAtom;
static Atom	cropAtom;
static Atom	xvStatusAtom;
static Atom	zoomAtom;
static Atom	xImageAtom;

static void MakeXvAtoms(Display* dpy)				    /* Create Atoms to be used for setting Xv attributes */
{
  XvAtomsRec*		atomInfo;
  XExtData*		extensionData;
  XExtData*		newExtensionData;
  Bool			atomInfoFound = False;
  static Bool		firstTimeCalled = True;
  
  if (firstTimeCalled == True)					    /* Only do this work once */
  {
    firstTimeCalled = False;
    PlxDebugEnvVar = getenv("PlxDebug");			    /* Determine whether debugging should be enabled */
    if ((char*)NULL != PlxDebugEnvVar)				    /* Did the process have debugging enabled? */
    {
      plxDebuggingOn = True;					    /* Yes, enable debugging for this library */
    }
    else
      plxDebuggingOn = False;					    /* No, don't bother printing debugging msgs for this library */
  }
  for (extensionData = dpy->ext_data;				    /* Look at this Display's extension data for Xv Atoms info */
       extensionData != (XExtData*)NULL;
       extensionData = extensionData->next)
  {
    if (XvPlxAtomsCode == extensionData->number)		    /* Was the Xv Atoms info found? */
    {
      atomInfoFound = True;					    /* Yes, mark a flag saying so,... */
      atomInfo = (XvAtomsRec*)extensionData->private_data;	    /* ...point to the Xv Atoms data,... */
      break;							    /* ...and quit looking. */
    }
    if ((XExtData*)NULL == extensionData->next)			    /* The end of the list was reached */
      break;
  } /* end for (extensionData traversal) */
  if (False == atomInfoFound)					    /* No Xv Atoms data were found, create them and add them... */
  {								    /* ...to the Display's extension data for future use. */
    newExtensionData = Xmalloc(sizeof(XExtData));
    newExtensionData->private_data = Xmalloc(sizeof(XvAtomsRec));
    newExtensionData->free_private = (int (*)())free;
    newExtensionData->number = XvPlxAtomsCode;
    newExtensionData->next = (struct _XExtData*)NULL;
    if (extensionData != (XExtData*)NULL)			    /* If this Display already had extension info, add the... */
    {								    /* ...Xv Atoms info to the end of the extension info list. */
      newExtensionData->next = extensionData->next;
      extensionData->next = newExtensionData;
    }
    else							    /* This Display had no extension info, so add the Plx... */
      dpy->ext_data = newExtensionData;				    /* ...info as the only extension info so far. */
    atomInfo = (XvAtomsRec*)newExtensionData->private_data;	    /* Point to the Xv Atoms extension info for filling in */
    atomInfo->tableCompression =
      XInternAtom(dpy,"XV_PLX_TABLE_COMPRESSION",False);
    atomInfo->tableDecompression =
      XInternAtom(dpy,"XV_PLX_TABLE_DECOMPRESSION",False);
    atomInfo->getCImage = XInternAtom(dpy,"XV_PLX_GETCIMAGE",False);
    atomInfo->putCImage = XInternAtom(dpy,"XV_PLX_PUTCIMAGE",False);
    atomInfo->firstPutCImage =
      XInternAtom(dpy,"XV_PLX_FIRST_PUTCIMAGE",False);
    atomInfo->status = XInternAtom(dpy, "XV_STATUS", False);
    atomInfo->hue = XInternAtom(dpy,"XV_HUE",False);
    atomInfo->saturation = XInternAtom(dpy,"XV_SATURATION",False);
    atomInfo->brightness = XInternAtom(dpy,"XV_BRIGHTNESS",False);
    atomInfo->contrast = XInternAtom(dpy,"XV_CONTRAST",False);
    atomInfo->encoding = XInternAtom(dpy, "XV_ENCODING", False);
    atomInfo->valueRecall = XInternAtom(dpy, "XV_PLX_VALUE_RECALL", False);
    atomInfo->valueSave = XInternAtom(dpy, "XV_PLX_VALUE_SAVE", False);
    atomInfo->videoParameterString = XInternAtom(dpy, "XV_PLX_VIDEO_PARAMETER_STRING", False);
    atomInfo->querySignals = XInternAtom(dpy, "XV_PLX_QUERY_SIGNALS", False);
    atomInfo->crop = XInternAtom(dpy, "XV_PLX_CROP", False);
    atomInfo->zoom = XInternAtom(dpy, "XV_PLX_ZOOM", False);
    atomInfo->xImage = XInternAtom(dpy, "XV_IMAGE_XIMAGE", False);
  } /* end if atom info wasn't found */
  hueAtom = atomInfo->hue;
  saturationAtom = atomInfo->saturation;
  brightnessAtom = atomInfo->brightness;
  contrastAtom = atomInfo->contrast;
  encodingIDAtom = atomInfo->encoding;
  valueRecallAtom = atomInfo->valueRecall;
  valueSaveAtom = atomInfo->valueSave;
  videoParameterStringAtom = atomInfo->videoParameterString;
  querySignalsAtom = atomInfo->querySignals;
  cropAtom = atomInfo->crop;
  zoomAtom = atomInfo->zoom;
  xvStatusAtom = atomInfo->status;
  xImageAtom = atomInfo->xImage;
  return;
} /* end function MakeXvAtoms */



void AddVideoFormat(Display*		dpy,
		    XvPortID		portID,
		    PlxVideoFormat*	format)
{
  int			result;
  int			counter;
  int			encodingCounter;
  unsigned int		numberOfEncodings;
  XvplxSignalRec*	signalInfo;
  PlxVideoSignal	signal;
  unsigned int		tempInstance;
  
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  result = XvGetLongPortAttribute(dpy, portID,			    /* Use the new call to get encoding information about... */
				  querySignalsAtom,		    /* ...the signal standards supported by this port. */
				  &numberOfEncodings,
				  (char**)&signalInfo);
  if (result != Success)
  {
    printf("XvPlx layer: AddVideoFormat(); XvGetLongPortAttributed failed.\n");
    return;
  }
  if (numberOfEncodings == 0)					    /* Are there any encodings for this Port? */
    return;							    /* No, return without doing anymore work */
  numberOfEncodings = numberOfEncodings / sizeof(XvplxSignalRec);   /* "numberOfEncodings" is actually the # of bytes returned */
  for (counter = 0; counter < numberOfEncodings; counter++)	    /* Examine each Encoding supported by the hardware */
  {
/*
    printf("=======Encoding %d:\t\t", counter);
    printf("Channel:\t%d\tStandard:\t%d\tFormat:\t%d\tInstance:\t%d\tEncoding:\t%08x\n",
	   signalInfo->channel,
	   signalInfo->standard,
	   signalInfo->format,
	   signalInfo->interlaced,
	   signalInfo->encoding_id);
    printf("\t\t\t\tWidth:\t%d\tHeight:\t%d\tBlank:\t%d",
	   signalInfo->width,
	   signalInfo->height,
	   signalInfo->blank);
    if (0 == signalInfo->type)
      printf("\tType:\tInput\n");
    else
      printf("\tType:\tOutput\n");
*/
    if (format->type == signalInfo->type)			    /* Only add info for the type of signal (i.e., input... */
    {								    /* ...or output) being requested. */
      format->format = signalInfo->format;
      tempInstance = signalInfo->interlaced;			    /* For overlay servers, instance info is in the... */
      tempInstance = tempInstance << 16;			    /* ...interlaced field of the singal info; put it into... */
      format->format |= tempInstance;				    /* ...the upper word of the format structure. */
      format->channel = signalInfo->channel;
      plxaddformat(format);					    /* Add this format to the list of supported formats  */
      plxaddtype((PlxVideoType*)format);			    /* Add this signal type to the list of supported types */
      signal.channel = signalInfo->channel;			    /* Set up the information for this signal, based on the... */
      signal.type = format->type;				    /* ...current format and the Encoding information just... */
      signal.standard = signalInfo->standard;			    /* ...retrieved. */
      signal.w = signalInfo->width;
      signal.h = signalInfo->height;				    /* Restore the proper height and blanking interval... */
      signal.b = signalInfo->blank;				    /* ...height. */
      signal.r = signalInfo->rate;
      signal.interlaced = 1;					    /* For now, hard-code the interlaced field to 1 since... */
      signal.sync_ok = signalInfo->sync_ok;			    /* ...interlaced is being used by server to report... */
								    /* ...instance info. */
      plxaddsig(&signal);					    /* Add this signal info to the list of supported signals */
    }
    signalInfo++;						    /* Examine the next Signal info structure on the list */
  }
  return;
} /* end function AddVideoFormat */



plx_IO*	XPlxXvQueryConfig(Display*	dpy,
			  Drawable	d,
			  GC		gc)
{
  int			adaptorNumber = 0;
  int			result;
  int			counter;
  int			numPlxAdaptors;
  unsigned int		numberOfAdaptors;
  unsigned long		numberOfEncodings;
  PlxVideoFormat	format;
  XvAdaptorInfo*	adaptorList;
  XvAdaptorInfo*	adaptorInfo;
  XPlxHardwareInfo	hardwareInfo;
  
  PlxDeletePlxHardwareList(&plx_io);				    /* Delete old data structures from prev. call to this fct. */
  plx_io.inputs = (plx_IO_list*)NULL;
  plx_io.outputs = (plx_IO_list*)NULL;
  plx_io.hardware = 0;						    /* Re-initialize the hardware type for this structure */
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  result = XvQueryAdaptors(dpy, XDefaultRootWindow(dpy),	    /* Use the Xv method of getting information about the... */
			   &numberOfAdaptors,			    /* ...hardware. */
			   &adaptorList);
  if (result != Success)					    /* Did the call fail for some reason? */
    return((plx_IO*)NULL);					    /* Yes, return a NULL pointer immediately */
  if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
    return((plx_IO*)NULL);					    /* No, return a NULL pointer immediately */
  numPlxAdaptors = XPlxGetHardwareInstances(dpy);		    /* Determine how many of the adaptors seen are Plx */
  adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
  adaptorNumber = XPlxGetCurrentHardwareInstance(dpy);		    /* See if a specific Adaptor is being requested */
  if (adaptorNumber < 0)					    /* If no specific Adaptor requested, use the 1st one */
    adaptorNumber = 0;
  if (adaptorNumber > (numPlxAdaptors - 1))			    /* Don't allow an illegal Adaptor number to be chosen */
    adaptorNumber = 0;
  for (counter = 0;						    /* Look at each Adaptor, searching for an XVideo card */
       counter < numberOfAdaptors;				    /* Note: all Adaptors must be searched, not just Plx... */
       counter++, adaptorInfo++)				    /* ...ones, since Plx adaptors might not be contiguous... */
  {								    /* ...(e.g., could be Plx, then non-Plx, then another Plx). */
    if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		"PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board */
    {
      if (adaptorNumber > 0)					    /* Should this function be looking at other than the 1st... */
      {								    /* ...adaptor?  If so, note that one Parallax adaptor has... */
	adaptorNumber--;					    /* ...been seen and go to the next iteration 'til found. */
	continue;
      }
      else							    /* Found the appropriate Parallax adaptor */
      {
	plx_io.hardware = PLX_XVIDEO_24;			    /* Fill in the appropriate structure for an XVideo board */ 
	if (((int)adaptorInfo->type & XvInputMask) != 0)	    /* Does this Adaptor support video input? */
	{
	  format.type = 0;					    /* Note that this is an input channel, not an output channel */
	  AddVideoFormat(dpy, adaptorInfo->base_id, &format);	    /* Add encoding information for this Port */
	} /* end if adaptorInfo is input... */
	if (((int)adaptorInfo->type & XvOutputMask) != 0)	    /* Does this Adaptor support video output? */
	{
	  format.channel = PLX_OUTPUT_0;
	  format.type = 1;
	  AddVideoFormat(dpy, adaptorInfo->base_id + 1, &format);   /* Add encoding information for this Port */
	  illegalOutputPort = adaptorInfo->base_id;		    /* Don't allow any output operations on the 1st Port for now */
	}
	break;							    /* Done searching; use this set of PortID's */
      }	/* end else (appropriate Parallax adaptor was found) */
    } /* end if looking at a Parallax adaptor */
    else							    /* No, this is not a Parallax board; go to the next... */
      continue;							    /* ...item on the list, searching for a Parallax board. */
  } /* end for counter... */
  plx_io.numJPEGCompressors = 1;				    /* For now, hard-code that there's a JPEG chip available */
  plx_io.overlay = 1;						    /* For now, anything with Xv has overlay available */
  plx_io.zoom = 1;						    /* Ditto for zoom */
  plx_io.numMPEGDecoders = hardwareInfo.info.numMPEGDecoders;
  plx_io.numH320Compressors = hardwareInfo.info.numH320Compressors;
  return(&plx_io);
} /* end function XPlxXvQueryConfig */



plx_signal* XPlxXvQueryVideo(Display*	dpy,
			     Drawable	d,
			     GC		gc)
{
  Bool			encodingFound = False;
  unsigned int		numberOfEncodings;
  WindowToPortID*	windowToPortIDMap;
  XvEncodingID		encodingID;
  XvplxSignalRec*	signalInfo;
  plx_signal*		theSignal;
  int			result;
  int			counter;
  int			portStatus;
  struct timeval	waitTime;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvGetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Determine which encoding in use for this XvPort */
		     encodingIDAtom, (int*)&encodingID);
  result = XvGetLongPortAttribute(dpy,				    /* Use the new call to get encoding information about... */
				  windowToPortIDMap->xvPortID,	    /* ...signal standards supported by this port. */
				  querySignalsAtom,
				  &numberOfEncodings,
				  (char**)&signalInfo);
  if (result != Success)
  {
    printf("XvPlx layer: XPlxQueryVideo(); XvGetLongPortAttributed failed.\n");
    return;
  }
  if (numberOfEncodings == 0)					    /* Are there any encodings for this Port? */
    return((plx_signal*)NULL);					    /* No, return without doing anymore work */
  numberOfEncodings = numberOfEncodings / sizeof(XvplxSignalRec);
  for (counter = 0; counter < numberOfEncodings; counter++)	    /* Examine each Encoding supported by the hardware */
  {
    if (signalInfo->encoding_id == encodingID)			    /* Was the requested XvEncodingID found? */
    {
      counter = numberOfEncodings;				    /* Yes, break out of the loop */
      encodingFound = True;					    /* Set a variable indicating success */
    }
    else							    /* No, look at the next Encoding structure */
    {
      signalInfo++;
    }
  }
  if (encodingFound == False)					    /* Was the requested Encoding found?  If not, return... */
    return((plx_signal*)NULL);					    /* ...a NULL Parallax signal pointer. */
  theSignal = (plx_signal*)malloc(sizeof(plx_signal));		    /* Allocate memory to store a pointer to a Parallax signal */
  if (theSignal == (plx_signal*)NULL)				    /* Did the allocation fail? */
    return((plx_signal*)NULL);					    /* Yes, return a NULL Parallax signal pointer */
  theSignal->standard = signalInfo->standard;			    /* Fill in the signal information for this encoding ID */
  theSignal->w = signalInfo->width;
  theSignal->h = signalInfo->height - signalInfo->blank;
  theSignal->b = 0;						    /* Forget about the blank interval for current HP release */
  theSignal->r = signalInfo->rate;
  theSignal->interlaced = signalInfo->interlaced;
  waitTime.tv_sec = 0;
  waitTime.tv_usec = 200000;					    /* Wait for 1 hundred milliseconds */
  select(sysconf(_SC_OPEN_MAX),					    /* Wait for some time while the video hardware sync's up */
	 0, 0, 0,
	 &waitTime);
  XvGetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Determine whether video is live on the requested port */
		     xvStatusAtom,				    /* (this atom added by Andy Goodrich) */
		     (int*)&portStatus);
  theSignal->sync_ok = ((portStatus & XVS_Sync) != 0) ? 1 : 0;
  return(theSignal);						    /* Return a pointer to the signal structure */
} /* end function XPlxQueryVideo */



void XPlxXvVideoScaleOutput(Display*	dpy,
			    Drawable	d,
			    GC		gc,
			    int		vx,
			    int		vy,
			    int		vw,
			    int		vh,
			    int		x,
			    int		y,
			    int		w,
			    int		h)
{
  XvPortID		portToUse;
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  if (windowToPortIDMap->xvPortID == illegalOutputPort)		    /* Was the 1st port for this Drawable unusable for output? */
    portToUse = windowToPortIDMap->xvPortID + 1;		    /* Yes, use the next available Port for output */
  else								    /* No, the 1st Port found attached to this Drawable was... */
    portToUse = windowToPortIDMap->xvPortID;			    /* ...a legitimate output port. */
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoScaleOutput() going to call XvGetVideo with: dpy=%d, Drawable=%d, Port=%d, gc=%d, x=%d, y=%d, w=%d, h=%d, vx=%d, vy=%d, vw=%d, vh=%d.\n",
	   dpy, d, portToUse, gc, x, y, w, h, vx, vy, vw, vh);
  }
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvGetVideo(dpy, portToUse, d, gc,
	     x, y, w, h,
	     vx, vy, vw, vh);
} /* end function XPlxVideoScaleOutput */



void XPlxXvVideoTag(Display*	dpy,
		    Drawable	d,
		    GC		gc,
		    int		type)
{
  return;							    /* For now, this function has no meaning for the HP product */
} /* end function XPlxVideoTag */



Bool FindMatchingVideoSignal(Display*		dpy,
			     Drawable		d,
			     XvPortID		portID,
			     int		channel,
			     int		standard,
			     int		format,
			     int		instance,
			     WindowToPortID*	windowToPortIDMap)
{
  int			result;
  int			signalCounter;
  int			numEncodingsReturned;
  unsigned int		encodingToUse;
  XvplxSignalRec*	signalInfo;
  
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  result = XvGetLongPortAttribute(dpy,				    /* Use the new call to get encoding information about... */
				  portID,			    /* ...this port. */
				  querySignalsAtom,
				  &numEncodingsReturned,
				  (char**)&signalInfo);
  if (result != Success)
  {
    printf("XvPlx layer: FindMatchingVideoSignal(); XvGetLongPortAttributed failed.\n");
    return;
  }
  numEncodingsReturned = numEncodingsReturned / sizeof(XvplxSignalRec);
  for (signalCounter = 0;					    /* Look at each Encoding returned, searching for an exact... */
       signalCounter < numEncodingsReturned;			    /* ...match. */
       signalCounter++, signalInfo++)
  {
    if ((signalInfo->channel == channel) &&			    /* Did the channel, standard, format, and instance of... */
	(signalInfo->standard == standard) &&			    /* ...the signal match? */
	(signalInfo->format == format) &&
	(signalInfo->interlaced == instance))
    {								    /* Yes, determine which channel is being requested */
      encodingToUse = signalInfo->encoding_id;			    /* Store the encoding to modify before requesting it */
      if (True == plxDebuggingOn)
      {
	printf("\tFindMatchingVideoSignal: Going to use channel %d, standard %d, format %d, instance %d, encoding %08x.\n",
	       channel, standard, format, instance, encodingToUse);
      }
      XvSetPortAttribute(dpy,					    /* Yes, tell this Port to use the requested Encoding */
			 portID,
			 encodingIDAtom,
			 encodingToUse);
      if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* Did an existing Window-to-XvPortID mapping exist? */
      {								    /* No, create one now that a channel has been selected */
	WindowToPortIDAdd(d, portID, &windowToPortIDList);
      }
      return(True);						    /* A matching signal was successfully found */
    }
  }	/* end (for signalCounter...) */
  return(False);						    /* No matching signal was found */
} /* end function FindMatchingVideoSignal */



void VideoChannelSelect(Display*	dpy,
			Drawable	d,
			GC		gc,
			int		channel,
			int		standard,
			int		format,
			int		instance)
{
  int			adaptorNumber = 0;
  int			result;
  int			counter;
  int			signalCounter;
  int			numEncodingsReturned;
  int			portStatus;
  Bool			signalMatched;
  int			numPlxAdaptors;
  unsigned int		numberOfAdaptors;
  WindowToPortID*	windowToPortIDMap;
  XvPortID		firstPort;
  XvPortID		portID;
  XvPortID		portToUse;
  XvAdaptorInfo*	adaptorList;
  XvAdaptorInfo*	adaptorInfo;
  XvplxSignalRec*	signalInfo;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap != NULL)				    /* Yes, this Window already has an XvPort associated... */
  {								    /* ...with it; use it. */
    signalMatched = FindMatchingVideoSignal(dpy, d,		    /* Try to find a video signal that matches */
					    windowToPortIDMap->xvPortID,
					    channel,
					    standard, format,
					    instance,
					    windowToPortIDMap);
    return;
  }
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  result = XvQueryAdaptors(dpy, XDefaultRootWindow(dpy),	    /* Use the Xv method of getting information about the... */
			   &numberOfAdaptors,			    /* ...hardware. */
			   &adaptorList);
  if (result != Success)					    /* Did the call fail for some reason? */
    return;							    /* Yes, return a NULL pointer immediately */
  if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
    return;							    /* No, return a NULL pointer immediately */
  numPlxAdaptors = XPlxGetHardwareInstances(dpy);		    /* Determine how many of the adaptors seen are Plx */
  adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
  adaptorNumber = XPlxGetCurrentHardwareInstance(dpy);		    /* See if a specific Adaptor is being requested */
  if (adaptorNumber < 0)					    /* If no specific Adaptor requested, use the 1st one */
    adaptorNumber = 0;
  if (adaptorNumber > (numPlxAdaptors - 1))			    /* Don't allow an illegal Adaptor number to be chosen */
    adaptorNumber = 0;
  for (counter = 0;						    /* Look at each Adaptor, searching for an XVideo card */
       counter < numberOfAdaptors;				    /* Note: all Adaptors must be searched, not just Plx... */
       counter++, adaptorInfo++)				    /* ...ones, since Plx adaptors might not be contiguous... */
  {								    /* ...(e.g., could be Plx, then non-Plx, then another Plx). */
    if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		"PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board */
    {
      if (adaptorNumber > 0)					    /* Should this function be looking at other than the 1st... */
      {								    /* ...adaptor?  If so, note that one Parallax adaptor has... */
	adaptorNumber--;					    /* ...been seen and go to the next iteration 'til found. */
	continue;
      }
    } /* end if looking at a Parallax adaptor */
    firstPort = adaptorInfo->base_id;				    /* In case this loop fails, save the 1st valid Port */
    for (portID = adaptorInfo->base_id;				    /* Look at each port supported by this adaptor */
	 portID < adaptorInfo->base_id + adaptorInfo->num_ports;
	 portID++)
    {
      if (adaptorInfo->base_id == portID)			    /* If this is the first time through the loop, and not... */
      {								    /* ...looking at the 1st input channel, then start... */
	if ((PLX_INPUT_1 == channel) ||				    /* ...searching with the 2nd port for this Adaptor; this... */
	    (PLX_OUTPUT_0 == channel))				    /* ...is to handle the case where one process tries to... */
	{							    /* ...use two Ports, but the 1st Port isn't tied until a... */
	  portToUse = portID + 1;				    /* PutVideo is called, so the 2nd search for a Port might... */
	}							    /* ...might try to use the 1st Port over again. */
	else							    /* This must be the 1st input, it's okay to start looking... */
	  portToUse = portID;					    /* ...from the 1st available port. */
      }	/* end if this is 1st time through loop */
      else							    /* This is not the 1st time through the loop; increment... */
	portToUse++;						    /* ...the port being examined. */
      if (True == plxDebuggingOn)				    /* Should a debug message be printed? */
      {								    /* Yes, print out all the ugly details */
	printf("XPlx Debug: VideoChannelSelect(): portToUse is %d.\n", portToUse);
	printf("\tChannel is %d, standard is %d, format is %d, instance is %d.\n\n",
	       channel, standard, format, instance);
      }
      XvGetPortAttribute(dpy, portToUse,			    /* Determine whether this port is busy */
			 xvStatusAtom,
			 (int*)&portStatus);
      if (True == plxDebuggingOn)				    /* Should a debug message be printed? */
      {								    /* Yes, print out all the ugly details */
	if ((portStatus & XVS_Attached) != 0)
	{
	  printf("\t\tThe port %d is already attached.\n", portToUse);
	}
      }
      if ((portStatus & XVS_Attached) != 0)			    /* Is this port busy showing video? */
	continue;						    /* Yes, look at the next port in the loop */
      signalMatched = FindMatchingVideoSignal(dpy, d, portToUse,    /* Try to find a video signal that matches */
					      channel,
					      standard, format,
					      instance,
					      windowToPortIDMap);
      if (signalMatched == True)				    /* Was a matching signal found? */
	return;							    /* Yes, return without needing to search any further */
    } /* end (for portID...) */
  } /* end (for counter...) */
  if ((portStatus & XVS_Attached) != 0)				    /* Were all the ports already taken? */
  {								    /* Yes, go ahead and steal the 1st Port on the adaptor,... */
    if ((channel == PLX_INPUT_1) ||				    /* ...unless the process is seeking the 2nd input or... */
	(channel == PLX_OUTPUT_0))				    /* ...the output channel, in which case use the 2nd Port. */
    {
      portToUse = firstPort + 1;
    }
    else
      portToUse = firstPort;
    signalMatched = FindMatchingVideoSignal(dpy, d, portToUse,	    /* Try to find a video signal that matches */
					    channel,
					    standard, format,
					    instance,
					    windowToPortIDMap);
  }
  return;							    /* If this statement is reached, no matches were found */
} /* end function VideoChannelSelect */



void XPlxXvVideoInputSelect(Display*	dpy,
			    Drawable	d,
			    GC		gc,
			    int		channel,
			    int		standard,
			    int		format,
			    int		instance)
{
  if (instance == PLX_RGB24)					    /* If using the old-style "type" argument? */
    instance = 0;						    /* Yes, interpret that to mean "use the 1st instance"... */
								    /* ...of the type of video input being requested. */
  VideoChannelSelect(dpy, d, gc,				    /* This function does the work for both input and output */
		     channel, standard, format, instance);
} /* end function XPlxVideoInputSelect */



void XPlxXvVideoOutputSelect(Display* 	dpy,
			     Drawable	d,
			     GC		gc,
			     int	channel,
			     int	standard,
			     int	format,
			     int	instance)
{
  int			adaptorNumber = 0;
  XvAdaptorInfo*	adaptorList;
  XvAdaptorInfo*	adaptorInfo;
  int			counter;
  Bool			done;
  int			numPlxAdaptors;
  unsigned int		numberOfAdaptors;
  XvPortID		portToUse;
  int			result;
  WindowToPortID*	windowToPortIDMap;
  
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  if (channel == PLX_IO_OFF)					    /* Is PLX_IO_OFF being requested? */
  {
    windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);  /* Is there a mapping b/w requested Window and an XvPort? */
    if (windowToPortIDMap != NULL)				    /* Yes, this Window already has an XvPort associated... */
    {
      XvStopVideo(dpy, windowToPortIDMap->xvPortID, d);		    /* Explicitly stop drawing video for this window */
    }
    return;							    /* Don't bother trying to select a null channel */
  }
  if (instance == PLX_RGB24)					    /* If using the old-style "type" argument? */
    instance = 0;						    /* Yes, interpret that to mean "use the 1st instance"... */
								    /* ...of the type of video input being requested. */
  if (channel == PLX_OUTPUT_JPEG)				    /* Special hack for the current HP release: want to... */
  {								    /* ...produce output from a JPEG playback stream. */
    done = False;
    while (done == False)					    /* Keep deleting all references to the Drawable passed in... */
    {								    /* ...as argument until all such refs are gone. */
      windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList); /* Look for another reference to the Drawable passed in */
      if (windowToPortIDMap == NULL)				    /* Was one found? */
	done = True;						    /* No, stop trying to delete window refs */
      else							    /* Yes, another ref was found, delete it */
      {
	XvStopVideo(dpy, windowToPortIDMap->xvPortID, d);	    /* Explicitly stop drawing video for this window */
	windowToPortIDMap = WindowToPortIDDelete(d, windowToPortIDList);
	windowToPortIDList = windowToPortIDMap;			    /* Point to the new head of the list, in case it was changed */
      }
    } /* end while (done == False) */
    result = XvQueryAdaptors(dpy, XDefaultRootWindow(dpy),	    /* Use the Xv method of getting information about the... */
			     &numberOfAdaptors,			    /* ...hardware. */
			     &adaptorList);
    if (result != Success)					    /* Did the call fail for some reason? */
      return;							    /* Yes, return a NULL pointer immediately */
    if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
      return;							    /* No, return a NULL pointer immediately */
    numPlxAdaptors = XPlxGetHardwareInstances(dpy);		    /* Determine how many of the adaptors seen are Plx */
    adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
    adaptorNumber = XPlxGetCurrentHardwareInstance(dpy);	    /* See if a specific Adaptor is being requested */
    if (adaptorNumber < 0)					    /* If no specific Adaptor requested, use the 1st one */
      adaptorNumber = 0;
    if (adaptorNumber > (numPlxAdaptors - 1))			    /* Don't allow an illegal Adaptor number to be chosen */
      adaptorNumber = 0;
    for (counter = 0;						    /* Look at each Adaptor, searching for an XVideo card */
	 counter < numberOfAdaptors;				    /* Note: all Adaptors must be searched, not just Plx... */
	 counter++, adaptorInfo++)				    /* ...ones, since Plx adaptors might not be contiguous... */
    {								    /* ...(e.g., could be Plx, then non-Plx, then another Plx). */
      if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		  "PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board */
      {
	if (adaptorNumber > 0)					    /* Should this function be looking at other than the 1st... */
	{							    /* ...adaptor?  If so, note that one Parallax adaptor has... */
	  adaptorNumber--;					    /* ...been seen and go to the next iteration 'til found. */
	  continue;
	}
	else							    /* Found the appropriate Parallax adaptor */
	{
	  portToUse = adaptorInfo->base_id;
	  break;						    /* Done searching; use this set of PortID's */
	}
      } /* end if looking at a Parallax adaptor */
      else							    /* No, this is not a Parallax board; go to the next... */
	continue;						    /* ...item on the list, searching for a Parallax board. */
    } /* end for (looking at all adaptors) */
    channel = PLX_OUTPUT_0;					    /* Reset the requested channel to ask for the output channel */
    VideoChannelSelect(dpy, d, gc,				    /* This function does the work for both input and output */
		       channel, standard, format, instance);
    windowToPortIDMap = WindowToPortIDDelete(d, windowToPortIDList);
    windowToPortIDList = windowToPortIDMap;			    /* Point to the new head of the list, in case it was changed */
    WindowToPortIDAdd(d, portToUse, &windowToPortIDList);	    /* Force JPEG playback to use the 1st port for this window */
    return;
  } /* end (if channel is meant to output JPEG playback) */
  if (channel == PLX_OUTPUT_JPEG_OFF)				    /* Is JPEG output off being requested? */
  {
    windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);  /* Is there a mapping b/w requested Window and an XvPort? */
    if (windowToPortIDMap != NULL)				    /* Yes, this Window already has an XvPort associated... */
    {
      if (windowToPortIDMap->xvPortID == illegalOutputPort)	    /* Was the 1st port for this Drawable unusable for output? */
	portToUse = windowToPortIDMap->xvPortID + 1;		    /* Yes, use the next available Port for output */
      else							    /* No, the 1st Port found attached to this Drawable was... */
	portToUse = windowToPortIDMap->xvPortID;		    /* ...a legitimate output port. */
      XvStopVideo(dpy, portToUse, d);				    /* Explicitly stop drawing video for this window */
    } /* end if (windowTo... != NULL) */
    done = False;
    while (done == False)					    /* Keep deleting all references to the Drawable passed in... */
    {								    /* ...as argument until all such refs are gone. */
      windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList); /* Look for another reference to the Drawable passed in */
      if (windowToPortIDMap == NULL)				    /* Was one found? */
	done = True;						    /* No, stop trying to delete window refs */
      else							    /* Yes, another ref was found, delete it */
      {
	XvStopVideo(dpy, windowToPortIDMap->xvPortID, d);	    /* Explicitly stop drawing video for this window */
	windowToPortIDMap = WindowToPortIDDelete(d, windowToPortIDList);
	windowToPortIDList = windowToPortIDMap;			    /* Point to the new head of the list, in case it was changed */
      }
    } /* end while (done == False) */
    return;							    /* Don't bother trying to select a null channel */
  }
  VideoChannelSelect(dpy, d, gc,				    /* This function does the work for both input and output */
		     channel, standard, format, instance);
  return;
} /* end function XPlxVideoOutputSelect */



void XPlxXvVideoValueLoad(Display*	dpy,
			  Drawable	d,
			  GC		gc,
			  int		sat,
			  int		cont,
			  int		hue,
			  int		brit)
{
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Set the Saturation attribute */
		     saturationAtom,
		     (2000 * sat / 256) - 1000);		    /* Convert from XPlx range to Xv range of valid values */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Set the Contrast attribute */
		     contrastAtom,
		     (2000 * cont / 256) - 1000);		    /* Convert from XPlx range to Xv range of valid values */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Set the Hue attribute */
		     hueAtom,
		     (2000 * hue / 256) - 1000);		    /* Convert from XPlx range to Xv range of valid values */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Set the Brightness attribute */
		     brightnessAtom,
		     (2000 * brit / 256) - 1000);		    /* Convert from XPlx range to Xv range of valid values */
} /* end function XPlxVideoValueLoad */



void XPlxXvVideoValueQuery(Display*	dpy,
			   Drawable	d,
			   GC		gc,
			   int*		sat,
			   int*		cont,
			   int*		hue,
			   int*		brit)
{
  int			retrievedValue;
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvGetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Get the Saturation attribute */
		     saturationAtom,
		     &retrievedValue);
  *sat = (256 * (retrievedValue + 1000)) / 2000;		    /* Convert from Xv range to XPlx range of valid values */
  XvGetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Get the Contrast attribute */
		     contrastAtom,
		     &retrievedValue);
  *cont = (256 * (retrievedValue + 1000)) / 2000;		    /* Convert from Xv range to XPlx range of valid values */
  XvGetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Get the Hue attribute */
		     hueAtom,
		     &retrievedValue);
  *hue = (256 * (retrievedValue + 1000)) / 2000;		    /* Convert from Xv range to XPlx range of valid values */
  XvGetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Get the Brightness attribute */
		     brightnessAtom,
		     &retrievedValue);
  *brit = (256 * (retrievedValue + 1000)) / 2000;		    /* Convert from Xv range to XPlx range of valid values */
} /* end function XPlxVideoValueQuery */



void XPlxXvVideoValueSave(Display*	dpy,
			  Drawable	d,
			  GC		gc,
			  int		nvselect)
{
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Call the corresponding Xv function with the new Atom... */
		     valueSaveAtom, nvselect);			    /* ...defined for writing to NVRAM. */
} /* end function XPlxVideoValueSave */
	


void XPlxXvVideoValueRecall(Display*	dpy,
			    Drawable	d,
			    GC		gc,
			    int		nvselect)
{
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Call the corresponding Xv function with the new Atom... */
		     valueRecallAtom, nvselect);		    /* ...defined for reading from NVRAM. */
} /* end function XPlxVideoValueRecall */



void XPlxXvVideoLive(Display*	dpy,
		     Drawable	d,
		     GC		gc,
		     int	vx,
		     int	vy,
		     int	x,
		     int	y,
		     int	w,
		     int	h)
{
  plx_signal*		tempSignalInfo;
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  tempSignalInfo = XPlxQueryVideo(dpy, d, gc);			    /* Get information about the signal currently in use */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Tell this Port to crop video to the Drawable instead... */
		     cropAtom, True);				    /* ...of squeezing it to fit. */
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoLive() going to call XvPutVideo with: dpy=%d, Drawable=%d, Port=%d, gc=%d, vx=%d, vy=%d, w=%d, h=%d, x=%d, y=%d, w=%d, h=%d.\n",
	   dpy, d, windowToPortIDMap->xvPortID, gc,
	   vx, vy, tempSignalInfo->w, tempSignalInfo->h,
	   x, y, w, h);
  }
  XvPutVideo(dpy, windowToPortIDMap->xvPortID, d, gc,		    /* Turn on live, cropped video for this window */
	     vx, vy, tempSignalInfo->w, tempSignalInfo->h,
	     x, y, w, h);
  return;
} /* end function XPlxXvVideoLive */



void XPlxXvVideoStill(Display*	dpy,
		      Drawable	d,
		      GC	gc,
		      int	vx,
		      int	vy,
		      int	x,
		      int	y,
		      int	w,
		      int	h)
{
  plx_signal*		tempSignalInfo;
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  tempSignalInfo = XPlxQueryVideo(dpy, d, gc);			    /* Get information about the signal currently in use */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Tell this Port to crop video to the Drawable instead... */
		     cropAtom, True);				    /* ...of squeezing it to fit. */
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoStill() going to call XvPutVideo with: dpy=%d, Drawable=%d, Port=%d, gc=%d, vx=%d, vy=%d, w=%d, h=%d, x=%d, y=%d, w=%d, h=%d.\n",
	   dpy, d, windowToPortIDMap->xvPortID, gc,
	   vx, vy, w, h,
	   x, y, w, h);
  }
  XvPutVideo(dpy, windowToPortIDMap->xvPortID, d, gc,		    /* Make sure live video is running in the window first  */
	     vx, vy, w, h,
	     x, y, w, h);
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoStill() *now* going to call XvStopVideo with: dpy=%d, Drawable=%d, Port=%d.\n",
	   dpy, windowToPortIDMap->xvPortID, d);
  }
  XvStopVideo(dpy, windowToPortIDMap->xvPortID, d);		    /* Andy says to use this call to still video */
  return;
} /* end function XPlxVideoStill */



void XPlxXvVideoSqueezeLive(Display*	dpy,
			    Drawable	d,
			    GC		gc,
			    int		vx,
			    int		vy,
			    int		vw,
			    int		vh,
			    int		x,
			    int		y,
			    int		w,
			    int		h)
{
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Tell this Port to squeeze video to fit the Drawable... */
		     cropAtom, False);				    /* ... instead of cropping it to the window. */
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoSqueezeLive() going to call XvPutVideo with: dpy=%d, Drawable=%d, Port=%d, gc=%d, vx=%d, vy=%d, w=%d, h=%d, x=%d, y=%d, w=%d, h=%d.\n",
	   dpy, d, windowToPortIDMap->xvPortID, gc,
	   vx, vy, vw, vh,
	   x, y, w, h);
  }
  XvPutVideo(dpy, windowToPortIDMap->xvPortID, d, gc,		    /* Perform the corresponding Xv call */
	     vx, vy, vw, vh,
	     x, y, w, h);
} /* end function XPlxXvVideoSqueezeLive */



void XPlxXvVideoSqueezeStill(Display*	dpy,
			     Drawable	d,
			     GC		gc,
			     int	vx,
			     int	vy,
			     int	vw,
			     int	vh,
			     int	x,
			     int	y,
			     int	w,
			     int	h)
{
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,		    /* Tell this Port to crop video to the Drawable instead... */
		     cropAtom, False);				    /* ...of squeezing it to fit. */
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoSqueezeStill() going to call XvPutVideo with: dpy=%d, Drawable=%d, Port=%d, gc=%d, vx=%d, vy=%d, w=%d, h=%d, x=%d, y=%d, w=%d, h=%d.\n",
	   dpy, d, windowToPortIDMap->xvPortID, gc,
	   vx, vy, vw, vh,
	   x, y, w, h);
  }
  XvPutVideo(dpy, windowToPortIDMap->xvPortID, d, gc,		    /* Make sure live video is running in the window first  */
	     vx, vy, vw, vh,
	     x, y, w, h);
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoSqueezeStill() *now* going to call XvStopVideo with: dpy=%d, Drawable=%d, Port=%d.\n",
	   dpy, d, windowToPortIDMap->xvPortID);
  }
  XvStopVideo(dpy, windowToPortIDMap->xvPortID, d);		    /* Andy says to use this call to still video */
  return;
} /* end function XPlxXvVideoSqueezeStill */



void XPlxXvVideoStop(Display*	dpy,
		     Drawable	d,
		     GC		gc)
{
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  if (True == plxDebuggingOn)					    /* Should a debug message be printed? */
  {								    /* Yes, print out all the ugly details */
    printf("XPlx Debug: XPlxVideoStop() going to call XvStopVideo with: dpy=%d, Drawable=%d, Port=%d.\n",
	   dpy, windowToPortIDMap->xvPortID, d);
  }
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvStopVideo(dpy, windowToPortIDMap->xvPortID, d);		    /* Perform the corresponding Xv call */
} /* end function XPlxXvVideoStop */



void XPlxXvSetVideoParameterString(Display*	dpy,		    /* Send a parameter string to the server */
				   Drawable	d,
				   GC		gc,
				   char*	buf)
{
  int			parameterStringID;
  WindowToPortID*	windowToPortIDMap;
  
  if (buf == (char*)NULL)					    /* Don't allow a NULL string to be sent */
    return;
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  XvSetLongPortAttribute(dpy,					    /* Make the corresponding Xv function call */
			 windowToPortIDMap->xvPortID,
			 videoParameterStringAtom,
			 strlen(buf) + 1,
			 buf);
} /* end function XPlxSetVideoParameterString */



Bool XPlxXvSelectHardwareInstance(Display*	dpy,
				  int		instance)
{
  XExtData*	extensionData;
  XExtData*	newExtensionData;
  Bool		instanceFound = False;
  
  for (extensionData = dpy->ext_data;				    /* Look at this Display's extension data for Xv Atoms info */
       extensionData != (XExtData*)NULL;
       extensionData = extensionData->next)
  {
    if (XPlxHardwareInstanceCode == extensionData->number)	    /* Was the Parallax hardware instance code found? */
    {
      extensionData->private_data = (char*)instance;		    /* Store the new instance number in the Display's area */
      return(True);
    }
    if ((XExtData*)NULL == extensionData->next)			    /* The end of the list was reached */
      break;
  } /* end for (extensionData traversal) */
  if (False == instanceFound)					    /* No instance number was found; add it to the... */
  {								    /* ...Display's extension data for future use. */
    newExtensionData = Xmalloc(sizeof(XExtData));
    newExtensionData->private_data = Xmalloc(sizeof(int));
    newExtensionData->private_data = (char*)instance;		    /* Store the new instance number in the Display's area */
    newExtensionData->free_private = (int (*)())free;
    newExtensionData->number = XPlxHardwareInstanceCode;
    newExtensionData->next = (struct _XExtData*)NULL;
    if (extensionData != (XExtData*)NULL)			    /* If this Display already had extension info, add the... */
    {								    /* ...instance info to the end of the extension info list. */
      newExtensionData->next = extensionData->next;
      extensionData->next = newExtensionData;
    }
    else							    /* This Display had no extension info, so add the Plx... */
      dpy->ext_data = newExtensionData;				    /* ...info as the only extension info so far. */
    return(True);						    /* New info was successfully added to the Display; outtahere */
  }
  else								    /* This stmt should never be reached, but if it is, then... */
    return(False);						    /* ...return an error condition. */
} /* end function XPlxXvSelectHardwareInstance */



int XPlxXvGetHardwareInstances(Display*	dpy)
{
  int			result;
  int			counter;
  unsigned int		numberOfAdaptors;
  int			numPlxBoards = 0;
  XvAdaptorInfo*	adaptorList;
  XvAdaptorInfo*	adaptorInfo;
  
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  result = XvQueryAdaptors(dpy, XDefaultRootWindow(dpy),	    /* Use the Xv method of getting information about the... */
			   &numberOfAdaptors,			    /* ...hardware. */
			   &adaptorList);
  if (result != Success)					    /* Did the call fail for some reason? */
    return(0);							    /* Yes, return no Adaptors immediately */
  if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
    return(0);							    /* No, return no Adaptors immediately */
  adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
  for (counter = 0;						    /* Look at each Adaptor, searching for an XVideo card */
       counter < numberOfAdaptors;
       counter++, adaptorInfo++)
  {
    if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		"PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board and increment the... */
      numPlxBoards++;						    /* ...total number of Plx boards seen. */
    else							    /* No, this is not a Parallax board; go to the next... */
      continue;							    /* ...item on the list, searching for a Parallax board. */
  } /* end looking at all adaptors */
  return(numPlxBoards);						    /* Return number of Parallax boards seen */
} /* end function XPlxXvGetHardwareInstances */



int XPlxGetCurrentHardwareInstance(Display*	dpy)
{
  XExtData*	extensionData;
  XExtData*	newExtensionData;
  Bool		instanceFound = False;
  
  for (extensionData = dpy->ext_data;				    /* Look at this Display's extension data for Xv Atoms info */
       extensionData != (XExtData*)NULL;
       extensionData = extensionData->next)
  {
    if (XPlxHardwareInstanceCode == extensionData->number)	    /* Was the Parallax hardware instance code found? */
    {
      return((int)extensionData->private_data);			    /* Yes, return the hardware instance stored for this Display */
    }
    if ((XExtData*)NULL == extensionData->next)			    /* The end of the list was reached */
      break;
  } /* end for (extensionData traversal) */
  return(0);							    /* Hardware instance not found; return default (1st board) */
} /* end function XPlxGetCurrentHardwareInstance */



int XPlxXvGetZoomFactor(Display*	dpy,
			Drawable	d)
{
  int			adaptorNumber = 0;
  int			result;
  int			counter;
  int			numPlxAdaptors;
  unsigned int		numberOfAdaptors;
  WindowToPortID*	windowToPortIDMap;
  XvAdaptorInfo*	adaptorList;
  XvAdaptorInfo*	adaptorInfo;
  int			zoomFactor;
  
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap != NULL)				    /* Yes, this Window already has an XvPort associated... */
  {								    /* ...with it; use it. */
    printf("GetZoomFactor(): Get the current zoom factor.\n");
    XvGetPortAttribute(dpy, windowToPortIDMap->xvPortID,	    /* Determine whether this port is busy */
		       zoomAtom,
		       (int*)&zoomFactor);
    return(zoomFactor);
  }
  result = XvQueryAdaptors(dpy, XDefaultRootWindow(dpy),	    /* Use the Xv method of getting information about the... */
			   &numberOfAdaptors,			    /* ...hardware. */
			   &adaptorList);
  if (result != Success)					    /* Did the call fail for some reason? */
    return(1);							    /* Yes, return default zoom factor of 1:1 immediately */
  if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
    return(1);							    /* No, return default zoom factor of 1:1 immediately */
  numPlxAdaptors = XPlxGetHardwareInstances(dpy);		    /* Determine how many of the adaptors seen are Plx */
  adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
  adaptorNumber = XPlxGetCurrentHardwareInstance(dpy);		    /* See if a specific Adaptor is being requested */
  if (adaptorNumber < 0)					    /* If no specific Adaptor requested, use the 1st one */
    adaptorNumber = 0;
  if (adaptorNumber > (numPlxAdaptors - 1))			    /* Don't allow an illegal Adaptor number to be chosen */
    adaptorNumber = 0;
  for (counter = 0;						    /* Look at each Adaptor, searching for an XVideo card */
       counter < numberOfAdaptors;
       counter++, adaptorInfo++)
  {
    if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		"PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board */
    {
      if (adaptorNumber > 0)					    /* Should this function be looking at other than the 1st... */
      {								    /* ...adaptor?  If so, note that one Parallax adaptor has... */
	adaptorNumber--;					    /* ...been seen and go to the next iteration 'til found. */
	continue;
      }
      if (True == plxDebuggingOn)				    /* Should a debug message be printed? */
      {								    /* Yes, print out all the ugly details */
	printf("XPlx Debug: XPlxGetZoomFactor(): Get zoom factor from port %d.\n",
	       adaptorInfo->base_id);
      }
      XvGetPortAttribute(dpy, adaptorInfo->base_id,		    /* Determine whether this Adaptor is zoomed, and at what... */
			 zoomAtom,				    /* ...factor. */
			 (int*)&zoomFactor);
      return(zoomFactor);
    } /* end if looking at a Parallax adaptor */
  } /* end (for counter...) */
  return(1);							    /* If this statement is reached, return default zoom level */
} /* end function XPlxXvGetZoomFactor */



Bool XPlxXvSetZoomFactor(Display*	dpy,
			 Drawable	d,
			 int		zoomFactor)
{
  int			adaptorNumber = 0;
  int			result;
  int			counter;
  int			numPlxAdaptors;
  unsigned int		numberOfAdaptors;
  WindowToPortID*	windowToPortIDMap;
  XvAdaptorInfo*	adaptorList;
  XvAdaptorInfo*	adaptorInfo;
  
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap != NULL)				    /* Yes, this Window already has an XvPort associated... */
  {								    /* ...with it; use it. */
    result = XvSetPortAttribute(dpy, windowToPortIDMap->xvPortID,   /* Set the zoom factor for this Adaptor */
				zoomAtom,
				zoomFactor);
    return(result);
  }
  result = XvQueryAdaptors(dpy, XDefaultRootWindow(dpy),	    /* Use the Xv method of getting information about the... */
			   &numberOfAdaptors,			    /* ...hardware. */
			   &adaptorList);
  if (result != Success)					    /* Did the call fail for some reason? */
    return(False);						    /* Yes, return fail-to-set-zoom result */
  if (numberOfAdaptors == 0)					    /* Is there any hardware connected to this display? */
    return(False);						    /* No, return fail-to-set-zoom result */
  numPlxAdaptors = XPlxGetHardwareInstances(dpy);		    /* Determine how many of the adaptors seen are Plx */
  adaptorInfo = adaptorList;					    /* Point to the 1st Adaptor info structure on the list */
  adaptorNumber = XPlxGetCurrentHardwareInstance(dpy);		    /* See if a specific Adaptor is being requested */
  if (adaptorNumber < 0)					    /* If no specific Adaptor requested, use the 1st one */
    adaptorNumber = 0;
  if (adaptorNumber > (numPlxAdaptors - 1))			    /* Don't allow an illegal Adaptor number to be chosen */
    adaptorNumber = 0;
  for (counter = 0;						    /* Look at each Adaptor, searching for an XVideo card */
       counter < numberOfAdaptors;
       counter++, adaptorInfo++, adaptorNumber--)
  {
    if (strncmp(adaptorInfo->name,				    /* Is this Adaptor a Parallax board of some sort? */
		"PGItv2", strlen("PGItv2")) == 0)		    /* Yes, assume it's an XVideo board */
    {
      if (adaptorNumber > 0)					    /* Should this function be looking at other than the 1st... */
      {								    /* ...adaptor?  If so, note that one Parallax adaptor has... */
	adaptorNumber--;					    /* ...been seen and go to the next iteration 'til found. */
	continue;
      }
      if (True == plxDebuggingOn)				    /* Should a debug message be printed? */
      {								    /* Yes, print out all the ugly details */
	printf("XPlx Debug: XPlxSetZoomFactor(): Set zoom factor to  %d.\n", zoomFactor);
      }
      result = XvSetPortAttribute(dpy, adaptorInfo->base_id,	    /* Set zoom factor for this Adaptor */
				  zoomAtom,
				  zoomFactor);
      return(result);
    } /* end if looking at a Parallax adaptor */
  } /* end (for counter...) */
  return(False);						    /* If this statement is reached, return failure code */
} /* end function XPlxXvSetZoomFactor */



XPlxXvPutImage(Display*		dpy,
	       Drawable		d,
	       GC		gc,
	       XImage*		image,
	       int		srcX,
	       int		srcY,
	       int		destX,
	       int		destY,
	       unsigned int	width,
	       unsigned int	height)
{
#ifdef ibm
  
  XPutImage(dpy, d, gc, image, srcX, srcY, destX, destY,
	    width, height);
#else  /* Use the new XvPutImage function written by Andy Goodrich */
  
  WindowToPortID*	windowToPortIDMap;
  
  windowToPortIDMap = WindowToPortIDFind(d, windowToPortIDList);    /* Is there a mapping b/w requested Window and an XvPort? */
  if (windowToPortIDMap == (WindowToPortID*)NULL)		    /* No, return immediately */
    return;
  MakeXvAtoms(dpy);						    /* Assure that the Xv Atoms are known to the X server */
  if (width > image->width)					    /* Assure that the image parameters don't exceed... */
    width = image->width;					    /* ...the width or height of the image data. */
  if (height > image->height)
    height = image->height;
  XvPutImage(dpy, windowToPortIDMap->xvPortID, d, gc,		    /* Call a special function to show the image on the... */
	     xImageAtom, 0, image,				    /* ...display; this was written primarily for the HP... */
	     destX, destY, width, height,			    /* ...platform, which has problems with putting video... */
	     srcX, srcY, image->width, image->height);		    /* ...data with an XPutImage() call. */
  
#endif /* if IBM platform */
  
} /* end function XPlxXvPutImage */



/* --------------------------------------------------------------------------- */
/* Past this point, none of the functions are implemented.  They are obsolete. */
/* --------------------------------------------------------------------------- */

void XPlxXvVideoPriority(register Display*	dpy,
			 Drawable		d,
			 GC			gc,
			 int			pri)
{
  printf("XPlxVideoPriority(): Not implemented.\n");
  return;
}

XPlxXvVideoFrameMode(dpy, d, gc, onoff)
     register Display *dpy;
     Drawable d;
     GC gc;
{
  printf("XPlxVideoFrameMode(): Not implemented.\n");
  return;
}

XPlxXvVideoFrameStart(dpy, d, gc, type)
     register Display *dpy;
     Drawable d;
     GC gc;
{
  printf("XPlxVideoFrameStart(): Not implemented.\n");
  return;
}

XPlxXvVideoDitherControl(dpy, d, gc, luma, chroma)
     register Display *dpy;
     Drawable d;
     GC gc;
{
  printf("XPlxVideoDitherControl(): Not implemented.\n");
  return;
}

void		XPlxXvVideoInput(register Display*	dpy,
				 Drawable		d,
				 GC			gc,
				 int			channel)
{
  printf("XPlxVideoInput(): Not implemented.\n");
  return;
}

XPlxXvZoomOn(dpy, d, gc, x, y, w, h)
register Display *dpy;
Drawable d;
GC gc;
int x, y, w, h;
{
  printf("XPlxVideoZoomOn(): Not implemented.\n");
  return;
}


XPlxXvZoomOff(dpy, d, gc, x, y)
register Display *dpy;
Drawable d;
GC gc;
{
  printf("XPlxVideoZoomOff(): Not implemented.\n");
  return;
}

XPlxXvDebug(dpy, d, gc, val)
     register Display *dpy;
     Drawable d;
     GC gc;
     unsigned short val;
{
  printf("XPlxVideoDebug(): Not implemented.\n");
  return;
}

XPlxXvVideoValueFactory(dpy, d, gc)
     register Display *dpy;
     Drawable d;
     GC gc;
{
  printf("XPlxVideoValueFactory(): Not implemented.\n");
  return;
}

XPlxXvVideoValueSet(dpy, d, gc, user0, user1, sat, cont, lgain, cgain, hue, bri)
     register Display *dpy;
     Drawable d;
     GC gc;
{
  printf("XPlxVideoValueSet(): Not implemented.\n");
  return;
}


XPlxXvPan(dpy, d, gc, x, y)
     register Display *dpy;
     Drawable d;
     GC gc;
     int x, y;
{
  printf("XPlxVideoPan(): Not implemented.\n");
  return;
}



char* XPlxXvGetVideoParameterString(dpy, d, gc) /* get parameter string */
     register Display *dpy;					    /*	from the server */
     Drawable d;
     GC gc;
{
  printf("XPlxGetVideoParameterString(): Not implemented.\n");
  return((char*)NULL);
}

XPlxXvCommand(dpy, d, gc, buf)
     register Display *dpy;
     Drawable d;
     GC gc;
     short *buf;
{
  printf("XPlxVideoCommand(): Not implemented.\n");
  return;
}


XPlxXvVideoFlipLive(dpy, d, gc, vx, vy, x, y, w, h)
register Display *dpy;
Drawable d;
GC gc;
{
  printf("XPlxVideoFlipLive(): Not implemented.\n");
  return;
}



void		XPlxXvVideoScale(register Display*	dpy,
				 Drawable		d,
				 GC			gc,
				 int			vx,
				 int			vy,
				 int			vw,
				 int			vh,
				 int			x,
				 int			y,
				 int			w,
				 int			h)
{
  printf("XPlxVideoScale(): Not implemented.\n");
  return;
}

XPlxXvImageLoad(dpy, d, gc, address, x, y, w, h)
     register Display *dpy;
     Drawable d;
     GC gc;
     int x, y, w, h;
{
  printf("XPlxVideoImageLoad(): Not implemented.\n");
  return;
}


XPlxXvVideoSet(dpy, d, gc)
register Display *dpy;
Drawable d;
GC gc;
{
  printf("XPlxVideoSet(): Not implemented.\n");
  return;
}


void XPlxXvVideoDirection(dpy, d, gc, xdir, ydir)
register Display *dpy;
Drawable d;
GC gc;
int xdir, ydir;
{
  printf("XPlxVideoDirection(): Not implemented.\n");
  return;
}


void XPlxXvVideoOutput(dpy, d, gc, vx, vy, x, y, w, h)
register Display *dpy;
Drawable d;
GC gc;
{
  printf("XPlxVideoOutput(): Not implemented.\n");
  return;
}

void		XPlxXvVideoScaleCopyArea(register Display*	dpy,
					 Drawable		sd,
					 Drawable		dd,
					 GC			gc,
					 int			sx,
					 int			sy,
					 int			sw,
					 int			sh,
					 int			dx,
					 int			dy,
					 int			dw,
					 int			dh)
{
  printf("XPlxXvVideoScaleCopyArea(): Not implemented.\n");
  return;
}

void XPlxXvControl(Display*	dpy,
		   Atom		op,
		   int		dataInSize,
		   char*	dataInPtr,
		   int*		dataOutSizeReturn,
		   char**	dataOutPtrReturn)
{
  printf("XPlxControl(): Not implemented.\n");
  return;
}
