/*
 * (c) Copyright 1991 Parallax Graphics Inc, All Rights Reserved
 *
 * Parallax Graphics, Inc.
 * 2500 Condensa Street
 * Santa Clara, California 95051
 *
 * $Log: panasonic.c,v $
 * Revision 1.4  1994/01/20  01:32:44  drapeau
 * Removed extraneous conditional compilation directives, since including
 * "string.h" always works no matter what OS platform.
 *
 * Revision 1.3  1993/11/19  04:32:13  drapeau
 * Modified the code to add support for SVR4.
 *
 * Revision 1.2  91/08/19  13:09:40  eric
 * added rcs header
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Id: panasonic.c,v 1.4 1994/01/20 01:32:44 drapeau Exp $ Parallax Graphics Inc";
#endif

/*
 *	panasonic.c - low level routines for driving the Panasonic Video Players
 *
 *	Written by John Malleo-Roach
 *
 */

/* INCLUDE FILES */

#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/termios.h>
#include	<sys/filio.h>
#include	<fcntl.h>
#include	<string.h>
#include	<errno.h>
#include	"alert.h"

/* CONSTANTS */

#define BUFLEN			32

#define ModelNumber		"TQ-3031F"

/* Define the instructions for communicating to a TQ-3031F */

#define Ack				'\006'
#define Nak				'\025'
#define Stx				'\002'

#define Stop			"\002SP\003\015\012"
#define Reject			"\002EJ\003\015\012"
#define Start			"\002LD\003\015\012"

#define Still			"\002PA\003\015\012"
#define Continue		"\002RS\003\015\012"

#define Play			"\002PF\003\015\012"
#define PlayStop		"\002PF.%05d:\003\015\012"
#define Reverse			"\002PR\003\015\012"
#define FastForward		"\002FF\003\015\012"
#define FastReverse		"\002FR\003\015\012"
#define SlowForward		"\002LF\003\015\012"
#define SlowReverse		"\002LR\003\015\012"
#define ScanForward		"\002CF\003\015\012"
#define ScanReverse		"\002CR\003\015\012"
#define StepForward		"\002TF\003\015\012"
#define StepReverse		"\002TR\003\015\012"

#define Search			"\002SR%05d:\003\015\012"

#define AudioOn			"\002AS\003\015\012"
#define AudioOff		"\002AR\003\015\012"
#define VideoOn			"\002VS\003\015\012"
#define VideoOff		"\002VR\003\015\012"
#define Ch1On			"\002A132767:\003\015\012"
#define Ch1Off			"\002A1\003\015\012"
#define Ch2On			"\002A232767:\003\015\012"
#define Ch2Off			"\002A2\003\015\012"
#define IndexOn			"\002DS\003\015\012"
#define IndexOff		"\002DR\003\015\012"

#define Clear			"\002AC\003\015\012"
#define FrameRequest	"\002NO\003\015\012"
#define StatusRequest	"\002PS\003\015\012"
#define DiscRequest		"\002PS1:\003\015\012"
#define ModelRequest	""

/* Various other constants for the video player */

#define AUDIO_OFF			401
#define LEFT_ONLY			402
#define RIGHT_ONLY			403
#define STEREO				404

#define VIDEO_OFF			410
#define VIDEO_ON			411

/* Player status codes, these are the same in all device driver files */

#define PLAYER_STATUS_UNKNOWN	0
#define PLAYER_STATUS_STOP		1
#define PLAYER_STATUS_STILL		2
#define PLAYER_STATUS_PLAY		3
#define PLAYER_STATUS_EMPTY		4

/* GLOBALS */

extern int PlayerId;			/* File descriptor for the Pansonic player	*/

static char buffer[BUFLEN];		/* Buffer for messages from the player		*/
static char message[256];		/* Scratch area for messages				*/

/* FUNCTIONS */

/* Send a command to the panasonic player */

static int Send(command,count)
char *command;
int count;
{
	int i;

/* Flush any exising characters in the queue */
	
	if (ioctl(PlayerId,TCFLSH,2) < 0) {
#ifdef DEBUG
		perror("Couldn't flush I/O queues!");
#endif
		return(-1);
	}
#ifdef DEBUG
	fprintf(stderr,"Sending command : [ ");
	for (i=0; i<strlen(command); i++) fprintf(stderr,"%3o ",command[i]);
	fprintf(stderr,"]\n");
#endif
	if (write(PlayerId,command,strlen(command)) < 0) {
		SystemError(errno,"Write to device failed!");
		return (-1);
	}
	while (count) {
		i = read(PlayerId,buffer,1);
		if (i == 0) {
#ifdef DEBUG
			fprintf(stderr,"No reply on count %d on previous command\n",count);
#endif
			continue;
		}
		else if (buffer[0] == Ack) {
#ifdef DEBUG
			fprintf(stderr,"Received ACK(%d) for previous command\n",count);
#endif
		}
		else if (buffer[0] == Nak) {
#ifdef DEBUG
			fprintf(stderr,"Received NAK(%d) for previous command\n",count);
#endif
			return(-1);
		}
		else {
#ifdef DEBUG
			fprintf(stderr,"Received (%d) byte for previous command\n",buffer[0]);
#endif
		}
		count = count - i;
	}
	return(NULL);
}

/* Read the 5 byte status message returned by the player */

static int ReadPlayerStatus()
{
	int i;
	char c[2];

/* Read the 2 byte status byte 1 byte at a time. */

	i  = read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,&(buffer[0]),1);
	i += read(PlayerId,&(buffer[1]),1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	if (i != 9) {
		ShowAlert("Reading player status word failed!",NULL);
		return(-1);
	}
	buffer[2] = 0;

	return(NULL);
}

/* Initialize the Panasonic player to a known status */

int InitializePanasonic(device,speed)
char *device;
int speed;
{
	char command[16];
	int retcode;
    struct termios ttbuf;

/* Open up port for communicating with the player */

	if ((PlayerId = open(device, O_RDWR | O_NDELAY)) < 0) {
		SystemError(errno,"Couldn't open Panasonic player device!");
		return(-1);
	}

/* Set the default speed to 9600 baud and RAW character mode */
	
    (void) ioctl(PlayerId, TCGETS, &ttbuf);

    ttbuf.c_iflag		= (long) 0;
    ttbuf.c_oflag		= (long) 0;
    ttbuf.c_lflag		= (long) 0;

	switch (speed) {
	case 0:
	    ttbuf.c_cflag	= (long) B9600 | CS8 | CLOCAL | CREAD | (B9600 << IBSHIFT) ;
		break;
	case 1:
	    ttbuf.c_cflag	= (long) B1200 | CS8 | CLOCAL | CREAD | (B1200 << IBSHIFT) ;
		break;
	case 2:
	    ttbuf.c_cflag	= (long) B2400 | CS8 | CLOCAL | CREAD | (B2400 << IBSHIFT) ;
		break;
	case 3:
	    ttbuf.c_cflag	= (long) B4800 | CS8 | CLOCAL | CREAD | (B4800 << IBSHIFT) ;
		break;
	case 4:
	    ttbuf.c_cflag	= (long) B9600 | CS8 | CLOCAL | CREAD | (B9600 << IBSHIFT) ;
		break;
	}

    ttbuf.c_cc[VMIN]	= 0;
    ttbuf.c_cc[VTIME]	= 10;

	if ((retcode = ioctl(PlayerId, TCSETSW, &ttbuf)) < 0) {
		SystemError(errno,"Couldn't initialize Panasonic player!");
		close(PlayerId);
		return(-1);
	}

/* Flush any garbage out of the I/O queues */
	
	if ((retcode = ioctl(PlayerId, TCFLSH, 2)) < 0) {
		SystemError(errno,"Couldn't flush I/O queues!");
		close(PlayerId);
		return(-1);
	}

/* See if we recognize this player: Not possible on this model

	if (Send(ModelRequest,1) < 0) {
		close(PlayerId);
		return(-1);
	}
	retcode = read(PlayerId,buffer,BUFLEN);
	if (strcmp(buffer,ModelNumber) != 0) {
		sprintf(message,"Unsupported model number %s",buffer);
		ShowAlert(message,NULL);
		close(PlayerId);
		return(-1);
	}
*/

/* Reset any previous errors */

	if (Send(Clear,7) < 0) {
		ShowAlert("Failed to reset Panasonic internal status",NULL);
		close(PlayerId);
		return(-1);
	}

/* Find out player status and get user to load media */

	if (Send(StatusRequest,1) < 0) {
		sprintf(message,"Communication error with Panasonic player");
		ShowAlert(message,NULL);
		close(PlayerId);
		return(-1);
	}
	if (ReadPlayerStatus() < 0) {
		sprintf(message,"Communication error with Panasonic player");
		ShowAlert(message,NULL);
		close(PlayerId);
		return(-1);
	}

/* Process on current status */

	while (strcmp(buffer,"EJ") == NULL) {
		ShowNote("Select OKAY after loading video media into Player",NULL);
		if (Send(StatusRequest,1) < 0) {
			sprintf(message,"Communication error with Panasonic player");
			ShowAlert(message,NULL);
			close(PlayerId);
			return(-1);
		}
 		if (ReadPlayerStatus() < 0) {
			sprintf(message,"Communication error with Panasonic player");
			ShowAlert(message,NULL);
			close(PlayerId);
			return(-1);
		}
	}
	if (strcmp(buffer,"PA") != NULL) {
		if (Send(Still,7) < 0) {
			ShowAlert("Panasonic player initialization failed",NULL);
			close(PlayerId);
			return(-1);
		}
	}
 
	sprintf(command,Search,0);
      
	/* Send out a search frame request */
      
	if (Send(command,7) < 0) {
		ShowAlert("Panasonic player initialization failed!",NULL,NULL);
		close(PlayerId);
		return(-1);
	}
	  
	return(NULL);
}

/* Command : Return simple player status code */

int PanasonicStatus()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}
	if (Send(StatusRequest,1) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (ReadPlayerStatus() < 0) {
		sprintf(message,"Communication error with Panasonic player");
		ShowAlert(message,NULL);
		close(PlayerId);
		return(-1);
	}
	return(PLAYER_STATUS_UNKNOWN);
}

/* Command : Return current frame number */

int PanasonicFrameNumber()
{
	char c[2];
	int number,i;

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}
	if (Send(FrameRequest,1) < 0) {
		ShowAlert("Player frame number request failed!",NULL);
		return(-1);
	}
	i  = read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,&(buffer[0]),1);
	i += read(PlayerId,&(buffer[1]),1);
	i += read(PlayerId,&(buffer[2]),1);
	i += read(PlayerId,&(buffer[3]),1);
	i += read(PlayerId,&(buffer[4]),1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	i += read(PlayerId,c,1);
	if (i != 13) {
		ShowAlert("Reading player frame number failed!",NULL);
		return(-1);
	}
	buffer[5] = '\0';
	number = atoi(buffer);

	return(number);
}

/* Set Audio parameters */

int SetPanasonicAudio(mode)
int mode;
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}
	switch (mode) {
	case AUDIO_OFF:
		if (Send(AudioOff,7) < 0) {
			ShowAlert("Unable to turn audio track off!",NULL);
			return(-1);
		}
		break;
	case LEFT_ONLY:
		if (Send(AudioOff,7) < 0) {
			ShowAlert("Unable to turn audio track on!",NULL);
			return(-1);
		}
		if (Send(Ch1Off,7) < 0) {
			ShowAlert("Unable to turn right audio track off!",NULL);
			return(-1);
		}
		if (Send(Ch2On,7) < 0) {
			ShowAlert("Unable to turn left audio track on!",NULL);
			return(-1);
		}
		break;
	case RIGHT_ONLY:
		if (Send(AudioOff,7) < 0) {
			ShowAlert("Unable to turn audio track on!",NULL);
			return(-1);
		}
		if (Send(Ch1On,7) < 0) {
			ShowAlert("Unable to turn right audio track on!",NULL);
			return(-1);
		}
		if (Send(Ch2Off,7) < 0) {
			ShowAlert("Unable to turn left audio track off!",NULL);
			return(-1);
		}
		break;
	case STEREO:
		if (Send(AudioOff,7) < 0) {
			ShowAlert("Unable to turn audio track on!",NULL);
			return(-1);
		}
		if (Send(Ch1On,7) < 0) {
			ShowAlert("Unable to turn right audio track on!",NULL);
			return(-1);
		}
		if (Send(Ch2On,7) < 0) {
			ShowAlert("Unable to turn left audio track on!",NULL);
			return(-1);
		}
		break;
	default:
		sprintf(message,"Unknown mode requested - %d",mode);
		ShowAlert(message,NULL);
		return(-1);
	}
	return(NULL);
}

/* This is a noop for now, as the player doesn't seem to support this option */

int SetPanasonicVideo(mode)
int mode;
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}
	switch (mode) {
	case VIDEO_OFF:
		if (Send(VideoOff,7) < 0) {
			ShowAlert("Unable to turn video off!",NULL);
			return(-1);
		}
	case VIDEO_ON:
		if (Send(VideoOn,7) < 0) {
			ShowAlert("Unable to turn video on!",NULL);
			return(-1);
		}
		break;
	default:
		sprintf(message,"Unknown mode requested - %d",mode);
		ShowAlert(message,NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Seek and display a single frame */

int ShowPanasonicFrame(frame)
int frame;
{
	int i;
	char command[16];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameter */

	if ((frame < 0) || (frame > 54000)) {
		sprintf(message,"Invalid frame address requested : %d",frame);
		ShowAlert(message,NULL);
		return(-1);
	}

	sprintf(command,Search,frame);

/* Send out a search frame request */

	if (Send(command,7) < 0) {
		ShowAlert("Search command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : Seek and display a complete video segment */

int ShowPanasonicSegment(start,end,cardname)
int start,end;
char *cardname;
{
	int i;
	char command[16];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameters */

	if ((start < 0) || (start > 54000)) {
		sprintf(message,"Invalid starting frame address requested : %d",start);
		ShowAlert(message,NULL);
		return(-1);
	}

	if ((end < 0) || (end > 54000)) {
		sprintf(message,"Invalid ending frame address requested : %d",end);
		ShowAlert(message,NULL);
		return(-1);
	}

	if (start >= end) {
		sprintf(message,"Starting frame (%d) not less than end (%d)",start,end);
		ShowAlert(message,NULL);
		return(-1);
	}

	sprintf(command,Search,start);

/* Send out a search frame request */

	if (Send(command,7) < 0) {
		ShowAlert("Search command failed!",NULL);
		return(-1);
	}

	sprintf(command,PlayStop,end);

/* Start play with stop point */

	if (Send(command,1) < 0) {
		ShowAlert("Play with stop command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : play forward */

int PanasonicPlay()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Wait here until playing has started */

	if (Send(Play,7) < 0) {
		ShowAlert("Play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : play reverse */

int PanasonicReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Wait here until playing has started */

	if (Send(Reverse,7) < 0) {
		ShowAlert("Play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : Continue an interrupted segment */

int PanasonicContinue()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Wait here until playing has started again */

	if (Send(Continue,7) < 0) {
		ShowAlert("Continue command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : pause the player */

int PanasonicStill()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Wait here for return confirmation */

	if (Send(Still,7) < 0) {
		ShowAlert("Still command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : Stop all playback */

int PanasonicStop()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Wait here for return confirmation */

	if (Send(Stop,7) < 0) {
		ShowAlert("Stop command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : move 1 frame forward */

int PanasonicStepForward()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Take one step forward */

	if (Send(StepForward,7) < 0) {
		ShowAlert("Step forward play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Move 1 frame backwards */

int PanasonicStepReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Take one step back */

	if (Send(StepReverse,7) < 0) {
		ShowAlert("Step reverse-play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : move slowly forward */

int PanasonicSlowForward()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Take slow steps forward */

	if (Send(SlowForward,7) < 0) {
		ShowAlert("Slow forward-play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : move slowly backwards */

int PanasonicSlowReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Take slow steps back */

	if (Send(SlowReverse,7) < 0) {
		ShowAlert("Variable reverse-play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : move fast forward */

int PanasonicFastForward()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Take fast steps forward */

	if (Send(FastForward,7) < 0) {
		ShowAlert("Fast forward-play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : move fast backwards */

int PanasonicFastReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Take fast steps back */

	if (Send(FastReverse,7) < 0) {
		ShowAlert("Variable reverse-play command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* Command : Provide visual frame indicator */

int PanasonicIndex(status)
int status;
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}

/* Turn on/off frame location display */

	if (status) {
		if (Send(IndexOn,7) < 0) {
			ShowAlert("Player frame display command failed!",NULL);
			return(-1);
		}
	}
	else {
		if (Send(IndexOff,7) < 0) {
			ShowAlert("Player frame display command failed!",NULL);
			return(-1);
		}
	}

	return(NULL);
}

/* Eject the video source */

int PanasonicEject()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Panasonic Player not initialized",NULL);
		return(-1);
	}
	if (Send(Reject,7) < 0) {
		ShowAlert("Player ejection command failed!",NULL);
		return(-1);
	}

	return(NULL);
}

/* end of panasonic.c */
