/*
 * (c) Copyright 1991 Parallax Graphics Inc, All Rights Reserved
 *
 * Parallax Graphics, Inc.
 * 2500 Condensa Street
 * Santa Clara, California 95051
 *
 * $Log: pioneer.c,v $
 * Revision 1.14  1994/01/20  01:33:04  drapeau
 * Removed extraneous conditional compilation directives, since including
 * "string.h" always works no matter what OS platform.
 *
 * Revision 1.13  1993/11/19  04:32:13  drapeau
 * Modified the code to add support for SVR4.
 *
 * Revision 1.2  91/08/19  13:09:30  eric
 * added rcs header
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Id: pioneer.c,v 1.14 1994/01/20 01:33:04 drapeau Exp $ Parallax Graphics Inc";
#endif

/*
 *	pioneer.c - low level routines for driving Pioneer players
 *
 *	Written by John Malleo-Roach
 *
 */

/* INCLUDE FILES */

#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/termios.h>
#include	<sys/filio.h>
#include	<fcntl.h>
#include	<string.h>
#include	<errno.h>
#include	"alert.h"

/* CONSTANTS */

#define BUFLEN			32

/* Define the instructions for communicating to a Pioneer player */

#define Ready			'R'
#define Error			'E'

#define DoorOpen		"OP\r"
#define Reject			"RJ\r"
#define Start			"SA\r"
#define Play			"PL\r"
#define Pause			"PA\r"
#define Still			"ST\r"
#define StepForward		"SF\r"
#define StepReverse		"SR\r"
#define MultiForward	"MF\r"
#define MultiReverse	"MR\r"
#define Speed			"SP\r"
#define Search			"SE\r"
#define StopMarker		"SM\r"
#define Frame			"FR\r"
#define Time			"TM\r"
#define Chapter			"CH\r"
#define AudioControl	"AD\r"
#define VideoControl	"VD\r"
#define DisplayControl	"DS\r"
#define Clear			"CL\r"
#define FrameRequest	"?F\r"
#define ModeRequest		"?P\r"
#define DiscRequest		"?D\r"
#define ModelRequest	"?X\r"
#define ClearScreen		"CS\r"
#define Print			"PR\r"

/* Various other constants for the Pioneer player */

#define AUDIO_OFF			401
#define LEFT_ONLY			402
#define RIGHT_ONLY			403
#define STEREO				404

#define VIDEO_OFF			410
#define VIDEO_ON			411

/* Player status codes, these are the same in all device driver files */

#define PLAYER_STATUS_UNKNOWN	0
#define PLAYER_STATUS_STOP		1
#define PLAYER_STATUS_STILL		2
#define PLAYER_STATUS_PLAY		3
#define PLAYER_STATUS_EMPTY		4

/* GLOBALS */

extern int PlayerId;			/* File descriptor for the Pioneer player	*/
static char buffer[BUFLEN];		/* Buffer for messages from the player		*/
static char message[256];		/* Scratch area for messages				*/

/* FUNCTIONS */

/* Send a command to the pioneer player and collect response if any */

static int Send(command,count,sleeptime)
char *command;
int count,sleeptime;
{
	int i;

/* Flush any exising characters in the queue */
	
	if (ioctl(PlayerId,TCFLSH,2) < 0) {
#ifdef DEBUG
		perror("Couldn't flush I/O queues!");
#endif
		return(-1);
	}
#ifdef DEBUG
	fprintf(stderr,"Sending command : %s\n",command);
#endif
	if (write(PlayerId,command,strlen(command)) < 0) {
		SystemError(errno,"Write to device failed!");
		return (-1);
	}
	if (sleeptime) sleep(sleeptime);
	while (count) {
		i = read(PlayerId,buffer,BUFLEN);
		if (i == 0) {
#ifdef DEBUG
			 fprintf(stderr,"Received 0 bytes in read in Send\n");
#endif
		}
		else if (buffer[0] == Ready) {
			buffer[1] = NULL;
#ifdef DEBUG
			fprintf(stderr,"Received READY on count %d\n",count);
#endif
			break;
		}
		else if (buffer[0] == Error) {
			buffer[3] = NULL;
			sprintf(message,"Player returned Error on previous command : %s",buffer);
			ShowAlert(message,NULL);
			return(-1);
		}
		else {
			buffer[i-1] = NULL;
			break;
		}
		count--;
	}
	return(NULL);
}

/* Initialize the pioneer player to a known state */

int InitializePioneer(device,speed)
char *device;
int speed;
{
	int i,retcode;
    struct termios ttbuf;

/* Open up port for communicating with the player */

	if ((PlayerId = open(device, O_RDWR | O_NDELAY)) < 0) {
		SystemError(errno,"Couldn't open Pioneer player device!");
		return(-1);
	}

/* Set the default speed to 4800 baud and RAW character mode */
	
    (void) ioctl(PlayerId, TCGETS, &ttbuf);

    ttbuf.c_iflag		= (long) 0;
    ttbuf.c_oflag		= (long) 0;
    ttbuf.c_lflag		= (long) 0;

	switch (speed) {
	case 0:
	    ttbuf.c_cflag = (long) B4800 | CS8 | CLOCAL | CREAD | (B4800 << IBSHIFT) ;
		break;
	case 1:
	    ttbuf.c_cflag = (long) B1200 | CS8 | CLOCAL | CREAD | (B1200 << IBSHIFT) ;
		break;
	case 2:
	    ttbuf.c_cflag = (long) B2400 | CS8 | CLOCAL | CREAD | (B2400 << IBSHIFT) ;
		break;
	case 3:
	    ttbuf.c_cflag = (long) B4800 | CS8 | CLOCAL | CREAD | (B4800 << IBSHIFT) ;
		break;
	case 4:
	    ttbuf.c_cflag = (long) B9600 | CS8 | CLOCAL | CREAD | (B9600 << IBSHIFT) ;
		break;
	}

    ttbuf.c_cc[VMIN]	= 0;
    ttbuf.c_cc[VTIME]	= 10;

	if ((retcode = ioctl(PlayerId, TCSETSW, &ttbuf)) < 0) {
		SystemError(errno,"Couldn't initialize device!");
		close(PlayerId);
		return(-1);
	}

/* Flush any garbage out of the I/O queues */
	
	if ((retcode = ioctl(PlayerId, TCFLSH, 2)) < 0) {
		SystemError(errno,"Couldn't flush I/O queues!");
		close(PlayerId);
		return(-1);
	}

/* See if we recognize this player */

	if (Send(ModelRequest,1,1) < 0) {
		close(PlayerId);
		return(-1);
	}
/* I have turned off model number verification
	if (strncmp(buffer,ModelNumber1,strlen(ModelNumber1)) != 0) {
		if (strncmp(buffer,ModelNumber2,strlen(ModelNumber2)) != 0) {
			if (strncmp(buffer,ModelNumber3,strlen(ModelNumber3)) != 0) {
				if (strncmp(buffer,ModelNumber4,strlen(ModelNumber4)) != 0) {
					sprintf(message,"Unsupported model number %s",buffer);
					ShowAlert(message,NULL);
					close(PlayerId);
					return(-1);
				}
			}
		}
	}
*/
/* Find out player status and get user to load video media */

	if (Send(DiscRequest,1,1) < 0) {
		close(PlayerId);
		return(-1);
	}
	if (buffer[0] == '1') {
		if (buffer[1] != 'X') {
/* Block until reject returns R when disc has spun down */
			if (Send(Reject,5,0) < 0) {
				close(PlayerId);
				return(-1);
			}
		}
/* Block until reject returns R when disc is out */
		if (Send(Reject,5,0) < 0) {
			close(PlayerId);
			return(-1);
		}
	}
	else {
		if (Send(ModeRequest,1,1) < 0) {
			return(-1);
		}
		if (buffer[2] == '1') {
/* Block until reject returns R when disc is out */
			if (Send(Reject,5,0) < 0) {
				close(PlayerId);
				return(-1);
			}
		}
	}

	ShowNote("Select OKAY after loading video media into Player",NULL);

/* Wait for player to spin up and return ready status */

	if (Send(Start,20,0) < 0) {
		close(PlayerId);
		return(-1);
	}
	
	return(NULL);
}

/* Return a simple status code for the player */

int PioneerStatus()
{
	int status;

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

/* Get play mode status */

	if (Send(ModeRequest,1,1) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(PLAYER_STATUS_UNKNOWN);
	}
	switch (buffer[2]) {
	case '0':
		status = PLAYER_STATUS_EMPTY;
		break;
	case '1':
		status = PLAYER_STATUS_STOP;
		break;
	case '4':
		status = PLAYER_STATUS_PLAY;
		break;
	case '5':
		status = PLAYER_STATUS_STILL;
		break;
	default:
		status = PLAYER_STATUS_UNKNOWN;
		break;
	}
	return(status);
}

/* Return current frame number from player */

int PioneerFrameNumber()
{
	int number;

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (Send(FrameRequest,1,1) < 0) {
		ShowAlert("Player frame number request failed!",NULL);
		return(-1);
	}
	buffer[5] = '\0';
	number = atoi(buffer);

	return(number);
}

/* Set Audio parameters */

int SetPioneerAudio(mode)
int mode;
{
	char command[8];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}
	switch (mode) {
	case AUDIO_OFF:
		sprintf(command,"0%s",AudioControl);
		break;
	case LEFT_ONLY:
		sprintf(command,"1%s",AudioControl);
		break;
	case RIGHT_ONLY:
		sprintf(command,"2%s",AudioControl);
		break;
	case STEREO:
		sprintf(command,"3%s",AudioControl);
		break;
	default:
		sprintf(message,"Unknown mode requested for audio - %d",mode);
		ShowAlert(message,NULL);
		return(-1);
	}

	if (Send(command,5,0) < 0) {
		ShowAlert("Audio parameters not changed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Set whether the video is visible or not */

int SetPioneerVideo(mode)
int mode;
{
	char command[8];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}
	switch (mode) {
	case VIDEO_OFF:
		sprintf(command,"0%s",VideoControl);
		break;
	case VIDEO_ON:
		sprintf(command,"1%s",VideoControl);
		break;
	default:
		sprintf(message,"Unknown mode requested for video - %d",mode);
		ShowAlert(message,NULL);
		return(-1);
	}

	if (Send(command,5,0) < 0) {
		ShowAlert("Video parameters not changed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Seek and show a single frame */

int ShowPioneerFrame(frame)
int frame;
{
	char command[16];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameter */

	if ((frame < 0) || (frame > 54000)) {
		sprintf(message,"Invalid Frame address - %d",frame);
		ShowAlert(message,NULL);
		return(-1);
	}

/* Send out a search frame request */

	sprintf(command,"%s%05d%s",Frame,frame,Search);

	if (Send(command,5,0) < 0) {
		ShowAlert("Frame selection failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Seek and show a single frame by elapsed time */

static int ShowPioneerTime(minutes,seconds)
int minutes,seconds;
{
	char number[8];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameters */

	if ((minutes < 0) || (minutes > 60)) {
		ShowAlert("Invalid time (minutes) - %d",minutes,NULL);
		return(-1);
	}

	if ((seconds < 0) || (seconds > 60)) {
		ShowAlert("Invalid time (seconds) - %d",seconds,NULL);
		return(-1);
	}

/* Send out a search for time request */

	write(PlayerId,Time,strlen(Time));
	sprintf(number,"%2d%2d",minutes,seconds);
	write(PlayerId,number,strlen(number));
	write(PlayerId,Search,strlen(Search));

/* Wait here until sucessful */

	read(PlayerId,buffer,BUFLEN);
	return(NULL);
}

/* Seek and show a complete video segment */

int ShowPioneerSegment(start,end)
int start,end;
{
	char command[16];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameters */

	if ((start < 0) || (start > 54000)) {
		sprintf(message,"Invalid starting frame address requested : %d",start);
		ShowAlert(message,NULL);
		return(-1);
	}

	if ((end < 0) || (end > 54000)) {
		sprintf(message,"Invalid ending frame address requested : %d",end);
		ShowAlert(message,NULL);
		return(-1);
	}

	if (start >= end) {
		sprintf(message,"Starting frame (%d) not less than end (%d)",start,end);
		ShowAlert(message,NULL);
		return(-1);
	}

/* Send out a search frame request */

	sprintf(command,"%s%05d%s",Frame,start,Search);

	if (Send(command,5,0) < 0) {
		ShowAlert("Frame selection failed!",NULL);
		return(-1);
	}

/* Start play with stop point */
	
	sprintf(command,"%s%05d%s",Frame,end,Play);

	if (Send(command,0,0) < 0) {
		ShowAlert("Play activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Seek and show a complete video segment by times */

static int ShowPioneerTimeSegment(smin,ssec,emin,esec)
int smin,ssec,emin,esec;
{
	char number[8];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameters */

	if ((smin < 0) || (smin > 60)) {
		ShowAlert("Invalid time (start minutes) - %d",smin,NULL);
		return(-1);
	}

	if ((ssec < 0) || (ssec > 60)) {
		ShowAlert("Invalid time (start seconds) - %d",ssec,NULL);
		return(-1);
	}

	if ((emin < 0) || (emin > 60)) {
		ShowAlert("Invalid time (end minutes) - %d",emin,NULL);
		return(-1);
	}

	if ((esec < 0) || (esec > 60)) {
		ShowAlert("Invalid time (end seconds) - %d",esec,NULL);
		return(-1);
	}

/* Send out a search frame request */

	write(PlayerId,Time,strlen(Time));
	sprintf(number,"%2d%2d",smin,ssec);
	write(PlayerId,number,strlen(number));
	write(PlayerId,Search,strlen(Search));

/* Wait here until sucessful */

	read(PlayerId,buffer,BUFLEN);

/* Start play with stop point */
	
	write(PlayerId,Time,strlen(Time));
	sprintf(number,"%2d%2d",emin,esec);
	write(PlayerId,number,strlen(number));
	write(PlayerId,Play,strlen(Play));

/* Wait here until done

	read(PlayerId,buffer,BUFLEN);
*/
	return(NULL);
}

/* Command : play forward */

int PioneerPlay()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (Send(Play,5,0) < 0) {
		ShowAlert("Play activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : play reverse */

int PioneerReverse()
{
	char command[12];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	sprintf(command,"60%s",Speed);

	if (Send(command,5,0) < 0) {
		ShowAlert("Reverse activation failed!",NULL);
		return(-1);
	}

	if (Send(MultiReverse,5,0) < 0) {
		ShowAlert("Reverse activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Continue an interrupted segment */

int PioneerContinue()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (Send(Play,5,0) < 0) {
		ShowAlert("Play continuation activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : pause during playback */

int PioneerStill()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (Send(Still,5,0) < 0) {
		ShowAlert("Freeze frame activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Stop playback */

int PioneerStop()
{
	int status;

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	status = PioneerStatus();

	if (status != PLAYER_STATUS_STOP) {
		if (Send(Reject,5,0) < 0) {
			ShowAlert("Stop player activation failed!",NULL);
			return(-1);
		}
	}
	return(NULL);
}

/* Command : move 1 frame forward */

int PioneerStepForward()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (Send(StepForward,5,0) < 0) {
		ShowAlert("Step forward activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move 1 frame backward */

int PioneerStepReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (Send(StepReverse,5,0) < 0) {
		ShowAlert("Step reverse activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move slow forward */

int PioneerSlowForward()
{
	char command[12];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	sprintf(command,"5%s",Speed);

	if (Send(command,5,0) < 0) {
		ShowAlert("Scan forward activation failed!",NULL);
		return(-1);
	}

	if (Send(MultiForward,5,0) < 0) {
		ShowAlert("Scan forward activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move slow reverse */

int PioneerSlowReverse()
{
	char command[12];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	sprintf(command,"5%s",Speed);

	if (Send(command,5,0) < 0) {
		ShowAlert("Scan reverse activation failed!",NULL);
		return(-1);
	}

	if (Send(MultiReverse,5,0) < 0) {
		ShowAlert("Scan reverse activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move fast forward */

int PioneerFastForward()
{
	char command[12];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	sprintf(command,"200%s",Speed);

	if (Send(command,5,0) < 0) {
		ShowAlert("Scan forward activation failed!",NULL);
		return(-1);
	}

	if (Send(MultiForward,5,0) < 0) {
		ShowAlert("Scan forward activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move fast reverse */

int PioneerFastReverse()
{
	char command[12];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	sprintf(command,"200%s",Speed);

	if (Send(command,5,0) < 0) {
		ShowAlert("Scan reverse activation failed!",NULL);
		return(-1);
	}

	if (Send(MultiReverse,5,0) < 0) {
		ShowAlert("Scan reverse activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : display visual frame location */

int PioneerIndex(status)
int status;
{
	char command[8];

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (status) sprintf(command,"1%s",DisplayControl);
	else sprintf(command,"0%s",DisplayControl);

	if (Send(command,5,0) < 0) {
		ShowAlert("Player status display activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Ejects the platter from the player */

int PioneerEject()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("Video Player not initialized",NULL);
		return(-1);
	}

	if (Send(Reject,5,1) < 0) {
		ShowAlert("Player ejection activation failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* end of pioneer.c */
