/*
 * (c) Copyright 1991-1993 Parallax Graphics Inc, All Rights Reserved
 *
 * Parallax Graphics, Inc.
 * 2500 Condensa Street
 * Santa Clara, California 95051
 *
 */
/*
 *
 *	rasterfile.c - Support Solaris 1.X raster files
 *
 *	Set of utilities to recreate the use of pr_load
 *	and pr_save.
 *
 *	Modified by John Malleo-Roach
 *
 */
static char rasterfileRcsId[] = "$Id: rasterfile.c,v 1.2 1995/12/22 23:56:03 drapeau Exp $";

/* INCLUDE FILES */

#include <stdio.h>
#include "rasterfile.h"
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/* CONSTANTS */

/* Number of bytes in an XImage */

#define	bytes_in_ximage(x)	((x)->bytes_per_line*(x)->height)

#define	ESCAPE	128

/* EXTERNAL FUNCTIONS */

extern Display *VAADisplay;
extern int      VAAScreen;

/* FUNCTIONS */

/* Get Sun rasterfile header from stream. */

static int ras_load_header(file, rh)
FILE	       		*file;
struct rasterfile	*rh;
{
    int	rc;

    if (rh == 0) return(-1);
    rc = fread((char *)rh, 1, sizeof(struct rasterfile), file);
    if ((rc == sizeof(struct rasterfile)) && (rh->ras_magic == RAS_MAGIC))
		return(0);
    else return(-1);
}

/* Get the image's colormap, if there is one */

static int ras_load_colormap(file, rh, map_length, red, green, blue)
FILE				*file;
struct rasterfile 	*rh;
int					*map_length;
u_char				*red, *green, *blue;
{
    if (!file || !rh || !red || !green || !blue || !map_length) return(-1);
 
/* Read colormap data, if any */

    switch (rh->ras_maptype) {
    case RMT_NONE :
    case RMT_RAW :
		if (rh->ras_maplength != 0) {
		    fseek(file, rh->ras_maplength, 1);
		}
		*map_length = 0; return(0); break;
    case RMT_EQUAL_RGB :

/* Allocate space for the colormap data */

		*map_length = rh->ras_maplength / 3;
		if (fread(red, 1, *map_length, file) != *map_length) return(-1);
		if (fread(green, 1, *map_length, file) != *map_length) return(-1);
		if (fread(blue, 1, *map_length, file) != *map_length) return(-1);
		return(0); break;
    default :
		return(-1);
    }
}

/*
 * Load the image into an X11 image data structure
 * Assumes that the header and colormap have
 * already been allowed.
 */

static XImage *ras_load_image(file, rh)
FILE				*file;
struct rasterfile	*rh;
{
    XImage			*r_ximage;
    register u_char	*tmp_ptr;
    u_char	        tmp_buf[4];
	XVisualInfo     info;
    register int	i, j, ras_bytes, bytes_left;

    if (!rh) return(0);

/* Create the XImage for the data */

	XMatchVisualInfo(VAADisplay,VAAScreen, 24, TrueColor,(XVisualInfo *) &info);
/* FIXUP: This probably changes in Solaris 2.3 */
	if (strncmp(XServerVendor(VAADisplay),"X11/NeWS",8) == 0) {
		r_ximage = XCreateImage(VAADisplay,info.visual,24,ZPixmap,0,
						 0,rh->ras_width,rh->ras_height,32,0);
	}
	else {
		r_ximage = XCreateImage(VAADisplay,info.visual,32,ZPixmap,0,
						 0,rh->ras_width,rh->ras_height,32,0);
	}
    if (!r_ximage) return(0);
	i = bytes_in_ximage(r_ximage);
	r_ximage->data = (char *)malloc(i);
	memset(r_ximage->data, 0, i);		    

/* Get the image data out of the file */

    switch (rh->ras_type) {
    case RT_OLD :
    case RT_STANDARD :
		if (rh->ras_depth == 24) {
/*
 * Assume that data is in the order BGR, but 24-bits
 * deep. Needs expanding to 32-bits.
 */
		    ras_bytes = rh->ras_width * rh->ras_depth;
		    ras_bytes = (ras_bytes/16 + (ras_bytes%16 ? 1 : 0))*2;
		    bytes_left = ras_bytes - (3 * rh->ras_width);
		    tmp_ptr = (u_char *)(r_ximage->data + 1);
		    for (i=0; i<rh->ras_height; i++) {
				for (j=0; j<rh->ras_width; j++) {
				    if (fread(tmp_ptr, 1, 3, file) != 3)
					return(NULL);
				    tmp_ptr += 4;
				}

/* Skip over the padding pixels */

				for (j=0; j<bytes_left; j++)
				    fgetc(file);	
		    }
		}
		else {

/* Standard image just read the whole thing */

		    if ((fread(r_ximage->data, 1, bytes_in_ximage(r_ximage), file)) != 
			    bytes_in_ximage(r_ximage))
			return(NULL);
		}
		break;
    case RT_BYTE_ENCODED :

/* Image is byte encoded */

		if (ras_read_rle(file, r_ximage->data, rh->ras_length) == -1) return(0);
		break;
    }
    return(r_ximage);
}

/*
 * Decode a run length encoded image, Images are encoded as follows,
 * runs must be of size three or greater, runs start with the value 
 * 128 (ESCAPE) followed by a count byte and pixel value byte. If the
 * count = 0, then it is a single 128 pixel, if the count is 1 followed
 * by 128 then it is two 128 pixels in a row. Anything else is a byte
 * count pair. Single pixels not associated with an escape are just
 * treated as that pixel value. See the Pixrects manual for more
 * details.
 *
 * Examples 
 *	<ESC><100><25>	- a run of 100 pixels of value 25
 *	<ESC><0>	- A single pixel of value 128
 *	<ESC><1><ESC>	- Two pixels of value 128
 *	<100><25><23>	- Indiviual pixels of different values
 */

int ras_read_rle(file, im_data, bytes_in_image)
FILE	*file;
u_char	*im_data;
{
    register u_char	*buf;
    u_char		pix_val, tmp_count;
    int			count, times;

    buf = im_data;
    while (bytes_in_image--) {
		if (fread(&pix_val, 1, 1, file) != 1) {
		    ShowAlert("Read on image data failed",NULL);
		    return(-1);
		}
		switch (pix_val) {
		case ESCAPE :

/* Get the count */

		    if (fread(&tmp_count, 1, 1, file) != 1) {
				ShowAlert("Read on image data failed",NULL);
				return(-1);
		    }
		    if (tmp_count == 0) {
/* 
 * If the value of count is 0 this is really
 * a 128 pixel
 */
				*(buf++) = ESCAPE;
				bytes_in_image--;
				break;
		    }

/* Get the pixel value */

		    if (fread(&pix_val, 1, 1, file) != 1) {
				ShowAlert("Read on image data failed",NULL);
				return(-1);
		    }
		    if ((tmp_count == 1) && (pix_val == ESCAPE)) {
/*
 * If we have ESC 1 ESC its really 2 128 pixels
 * in a row.
 */
				*buf = *(buf+1) = ESCAPE;
				bytes_in_image -=2;
				break;
		    }

/* Increment by one */

		    count = tmp_count + 1;
		    while (count--) {
/*
 * Its a run of pixels of 3 or more expand
 */
				*(buf++) = pix_val;
		    }
	 	    bytes_in_image -= 2;
		    break;
		default :

/* Default condition is just the pixel value */

		    *(buf++) = pix_val;
		    break;
		}
    }
    return(0);
}

/*
 * Dump the specified ximage and colour map to a file
 */

pr_dump_ximage(file, ximage, map_length, r, g, b)
FILE	*file;
XImage	*ximage;
int	     map_length;
u_char	*r, *g, *b;
{
    struct rasterfile	rh;
    int					rc;
    u_char	       		*data_ptr, *out_data;
    register int       	i, ras_bytes, bytes_to_copy;

    if (!ximage) return(-1);

    rh.ras_magic     = RAS_MAGIC;
    rh.ras_width     = ximage->width;
    rh.ras_height    = ximage->height;
/* This needs more work here
    rh.ras_depth     = ximage->depth;
*/
    rh.ras_depth     = 32;;
    rh.ras_type      = RT_STANDARD;
	if (map_length == 0) {
		rh.ras_maptype   = RMT_NONE;
		rh.ras_maplength = 0;
    }
	else {
		rh.ras_maptype   = RMT_EQUAL_RGB;
		rh.ras_maplength = 3*map_length;
    }
	    
    /* 
     * Dump everything to the file 
     * 1. Calculate the amount of image data to be dumped
     * 2. The rasterfile header
     * 3. the colour map if there is one
     * 4. the image data
     */

    i = ximage->width * rh.ras_depth;
    ras_bytes = (i/16 + (i%16 ? 1 : 0))*2;
    if ((out_data = (u_char *) malloc(ras_bytes)) == NULL) return(-1);
    memset(out_data, 0, ras_bytes);

    rh.ras_length = ras_bytes * ximage->height;
    rc = fwrite(&rh, sizeof(struct rasterfile), 1, file);
    if(rh.ras_maptype == RMT_EQUAL_RGB) {
		rc = fwrite(r, sizeof(u_char), map_length, file);
		rc = fwrite(g, sizeof(u_char), map_length, file);
		rc = fwrite(b, sizeof(u_char), map_length, file);
    }

    /* Write the image data line by line ensures a good raster file */

    data_ptr = (u_char *) ximage->data;
    
    if (rh.ras_depth == 1)
		 bytes_to_copy = ximage->width/8 + (ximage->width%8 ? 1 : 0);
    else bytes_to_copy = ximage->width * rh.ras_depth/8;
    
    for (i=0; i<ximage->height; i++) {
		memcpy(out_data, data_ptr, bytes_to_copy);
		rc = fwrite(out_data, sizeof(u_char), ras_bytes, file);
		memset(out_data, 0, ras_bytes);
		data_ptr += ximage->bytes_per_line;
    }
    free(out_data);
    return(0);
}

XImage *pr_load_ximage(filename,r,g,b,mapsize)
char *filename;
u_char *r,*g,*b;
int *mapsize;
{
	XImage             *image;
    register int	    i;
    FILE       		   *file; 
    struct rasterfile	rh;

    /* 
     * Load the XImage with the raster file data
     */

    if ((file = fopen(filename,"r")) == NULL) {
		ShowAlert("Could not open file",NULL);
		return(NULL);
    }

    if (ras_load_header(file, &rh) == -1) {
		ShowAlert("Error reading rasterfile header",NULL);
		fclose(file); return(NULL);
    }	

	if ((rh.ras_maptype != RMT_NONE) &&  (rh.ras_maplength != 0)) {
	    if(ras_load_colormap(file, &rh, mapsize, r, g, b) == -1) {
			fclose(file); return(NULL);
	    }
	}
	else {
	    *mapsize = 0;
	}
    if ((image = ras_load_image(file, &rh)) == (XImage *)NULL) {
	    fclose(file); return(NULL);
	}
    fclose(file);
	return(image);
}

/* end of raster.c */
