/*
 * (c) Copyright 1991 Parallax Graphics Inc, All Rights Reserved
 *
 * Parallax Graphics, Inc.
 * 2500 Condensa Street
 * Santa Clara, California 95051
 *
 * $Log: visca.c,v $
 * Revision 1.5  1994/01/20  01:33:34  drapeau
 * Removed extraneous conditional compilation directives, since including
 * "string.h" always works no matter what OS platform.
 *
 * Revision 1.4  1993/11/19  04:32:13  drapeau
 * Modified the code to add support for SVR4.
 *
 * Revision 1.2  91/08/19  13:09:45  eric
 * added rcs header
 * 
 *
 */

#ifndef lint
static char *rcsid = "@(#)$Id: visca.c,v 1.5 1994/01/20 01:33:34 drapeau Exp $ Parallax Graphics Inc";
#endif

/*
 *	visca.c - low level routines for the Sony VISCA Protocol
 *
 *  Note that this implemention only handles one device for now
 *
 *	Written by John Malleo-Roach
 *
 */

/* INCLUDE FILES */

#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/termios.h>
#include	<sys/filio.h>
#include	<fcntl.h>
#include	<string.h>
#include	<errno.h>
#include	"alert.h"

/* CONSTANTS */

#define BUFLEN			32

/* Define the instructions for communicating with Sony VISCA */

#define Completion		'\121'
#define Error			'\140'
#define Ack				'\101'

#define StopStatus      '\000'
#define EjectStatus     '\030'
#define StillStatus     '\040'
#define PlayStatus      '\050'

#define Clear			"\210\001\000\001\377"
#define Address         "\210\060\001\377"
#define StatusRequest	"\201\011\002\001\377"
#define FrameRequest    "\201\011\002\003\377"

#define Stop			"\201\001\002\001\000\377"
#define Eject			"\201\001\002\001\030\377"
#define Still			"\201\001\002\001\040\377"

#define Play			"\201\001\002\001\050\377"
#define Reverse			"\201\001\002\001\070\377"
#define FastForward		"\201\001\002\001\010\377"
#define FastReverse		"\201\001\002\001\020\377"
#define SlowForward		"\201\001\002\001\046\377"
#define SlowReverse		"\201\001\002\001\066\377"
#define StepForward		"\201\001\002\002\002\377"
#define StepReverse		"\201\001\002\002\003\377"

#define SearchPlay		"\201\001\002\003\041\000\000\000\000\050\377"
#define SearchStill		"\201\001\002\003\041\000\000\000\000\040\377"
#define StopAt			"\201\003\002\041\000\000\000\000\001\000\377"

#define IndexOn			"\201\001\002\025\001\377"
#define IndexOff		"\201\001\002\025\000\377"

#define VideoOn			"\201\001\003\041\001\001\000\000\000\000\377"
#define VideoOff		"\201\001\003\041\001\001\017\017\000\000\377"
#define AudioOn			"\201\001\003\041\001\010\000\000\000\000\377"
#define AudioOff		"\201\001\003\041\001\010\017\017\000\000\377"

/* Various other constants for the video player */

#define AUDIO_OFF			401
#define LEFT_ONLY			402
#define RIGHT_ONLY			403
#define STEREO				404

#define VIDEO_OFF			410
#define VIDEO_ON			411

/* Player status codes, these are the same in all device driver files */

#define PLAYER_STATUS_UNKNOWN	0
#define PLAYER_STATUS_STOP		1
#define PLAYER_STATUS_STILL		2
#define PLAYER_STATUS_PLAY		3
#define PLAYER_STATUS_EMPTY		4

/* GLOBALS */

extern int PlayerId;		/* File descriptor for the VISCA video player  */

static char buffer[BUFLEN];	/* Buffer for messages from the player		  */
static char message[256];	/* Scratch space for error messages			  */

/* FUNCTIONS */

/* convert decimal to bcd hex */

static int bcd(x)
int x;
{
	int i,j;

	i = x / 10;
	j = x % 10;
	x = i * 16;
	x = x + j;
	return(x);
}

/* convert bcd hex to decimal */

static int unbcd(x)
int x;
{
	int i,j;

	i = x / 16;
	j = x % 16;
	x = i * 10;
	x = x + j;
	return(x);
}

/* Compute length of command */

static int comlen(s)
unsigned char *s;
{
	int i;

#ifdef DEBUG
	fprintf(stderr,"Sending command : ");
#endif
	for (i=1; *s != 255; s++,i++) {
#ifdef DEBUG
		fprintf(stderr,"0x%02x ",*s);
#endif
	}
#ifdef DEBUG
	fprintf(stderr,"0x%02x\n",*s);
#endif
	return(i);
}

/* Send a byte to the VISCA player */

static int Send(command,count,timeout)
char *command;
int   count;
int   timeout;
{
	int i,j;

/* Flush any exising characters in the queue */
	
	if (ioctl(PlayerId,TCFLSH,2) < 0) {
#ifdef DEBUG
		perror("Couldn't flush I/O queues!");
#endif
		return(-1);
	}
	if (write(PlayerId,command,comlen(command)) < 0) {
		SystemError(errno,"Write to device failed!");
		return (-1);
	}
	for (j=0; j<count; j++) buffer[j] = 0; j = 0;
	while (count) {
		i = read(PlayerId,&(buffer[j]),count);
		if (i < 0) {
			SystemError(errno,"Read failed");
			return(-1);
		}
#ifdef DEBUG
		if (i != count) fprintf(stderr,"Received %d bytes instead of %d\n",i,count);
#endif
		if (buffer[j+1] == Ack) {
#ifdef DEBUG
			fprintf(stderr,"Received ACK(%d) on previous command\n",count);
#endif
		}
		else if (buffer[j+1] == Completion) {
#ifdef DEBUG
			fprintf(stderr,"Received COMP(%d) on previous command\n",count);
#endif
		}
		else if (buffer[j+1] == Error) {
			switch (buffer[j+2]) {
			case '\001':
				ShowAlert("VISCA: Message length error",NULL);
				break;
			case '\002':
				ShowAlert("VISCA: Syntax error",NULL);
				break;
			case '\003':
				ShowAlert("VISCA: Command buffer full",NULL);
				break;
			case '\004':
				ShowAlert("VISCA: Command cancelled",NULL);
				break;
			case '\005':
				ShowAlert("VISCA: No socket (to be cancelled)",NULL);
				break;
			case '\100':
				ShowAlert("VISCA: Power off",NULL);
				break;
			case '\101':
				ShowAlert("VISCA: Command failed",NULL);
				break;
			case '\102':
				ShowAlert("VISCA: Search error",NULL);
				break;
			case '\103':
				ShowAlert("VISCA: Condition error",NULL);
				break;
			case '\106':
				ShowAlert("VISCA: Counter type error",NULL);
				break;
			case '\107':
				ShowAlert("VISCA: Tuner error",NULL);
				break;
			case '\110':
				ShowAlert("VISCA: Emergency stop error",NULL);
				break;
			case '\111':
				ShowAlert("VISCA: Media is unmounted",NULL);
				break;
			case '\112':
				ShowAlert("VISCA: Register error",NULL);
				break;
			case '\113':
				ShowAlert("VISCA: Register mode setting error",NULL);
				break;
			}
			count = i;
		}
		if (i) {
			count -= i;
			j     += i;
		}
		if ((--timeout == 0) && count) {
			ShowAlert("Timeout error on last request with VISCA device",NULL);
			count = 0;
		}
	}
#ifdef DEBUG
	fprintf(stderr,"Command reply   : ");
	for (i=0; i<j; i++) {
		fprintf(stderr,"0x%02x ",(unsigned char) buffer[i]);
	}
	fprintf(stderr,"\n");
#endif
	return(j);
}

/* Initialize the VISCA player to a known status */

int InitializeVISCA(device,speed)
char *device;
int speed;
{
	int retcode;
    struct termios ttbuf;

/* Open up port for communicating with the video player */

	if ((PlayerId = open(device, O_RDWR | O_NDELAY)) < 0) {
		SystemError(errno,"Couldn't open VISCA video player device!");
		return(-1);
	}

/* Set the default speed to 9600 baud and RAW character mode */
	
    (void) ioctl(PlayerId, TCGETS, &ttbuf);

    ttbuf.c_iflag		= (long) 0;
    ttbuf.c_oflag		= (long) 0;
    ttbuf.c_lflag		= (long) 0;

	switch (speed) {
	case 0:
	    ttbuf.c_cflag	= (long) B9600 | CS8 | CLOCAL | CREAD | (B9600 << IBSHIFT) ;
		break;
	case 1:
	    ttbuf.c_cflag	= (long) B1200 | CS8 | CLOCAL | CREAD | (B1200 << IBSHIFT) ;
		break;
	case 2:
	    ttbuf.c_cflag	= (long) B2400 | CS8 | CLOCAL | CREAD | (B2400 << IBSHIFT) ;
		break;
	case 3:
	    ttbuf.c_cflag	= (long) B4800 | CS8 | CLOCAL | CREAD | (B4800 << IBSHIFT) ;
		break;
	case 4:
	    ttbuf.c_cflag	= (long) B9600 | CS8 | CLOCAL | CREAD | (B9600 << IBSHIFT) ;
		break;
	}

    ttbuf.c_cc[VMIN]	= 0;
    ttbuf.c_cc[VTIME]	= 10;

	if ((retcode = ioctl(PlayerId, TCSETSW, &ttbuf)) < 0) {
		SystemError(errno,"Couldn't initialize device!");
		close(PlayerId);
		return(-1);
	}

/* Flush any garbage out of the I/O queues */
	
	if ((retcode = ioctl(PlayerId, TCFLSH, 2)) < 0) {
		SystemError(errno,"Couldn't flush I/O queues!");
		close(PlayerId);
		return(-1);
	}

/* Reset the VISCA network */

	if (Send(Clear,5,5) < 0) {
		ShowAlert("Failed to reset VISCA network",NULL);
		close(PlayerId);
		return(-1);
	}
	if (Send(Address,4,5) < 0) {
		ShowAlert("Failed to assign VISCA addresses",NULL);
		close(PlayerId);
		return(-1);
	}

/* Find out player status and get user to load video media */

	if (Send(StatusRequest,4,5) < 0) {
		close(PlayerId);
		return(-1);
	}

/* Process on current status */

	while (buffer[2]  == EjectStatus) {
		ShowNote("Select OKAY after loading video media into Player",NULL);
		if (Send(StatusRequest,4,5) < 0) {
			close(PlayerId);
			return(-1);
		}
	}
	return(NULL);
}

/* Command : Return simple player status code */

int VISCAStatus()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}
	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	switch (buffer[2]) {
	case StopStatus:
		return(PLAYER_STATUS_STOP);
		break;
	case EjectStatus:
		return(PLAYER_STATUS_EMPTY);
		break;
	case StillStatus:
		return(PLAYER_STATUS_STILL);
		break;
	case PlayStatus:
		return(PLAYER_STATUS_PLAY);
		break;
	default:
		return(PLAYER_STATUS_UNKNOWN);
		break;
	}
}

/* Command : Return current frame number */

int VISCAFrameNumber()
{
	int number,i;

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!");
		return(-1);
	}
	if (Send(FrameRequest,13,5) < 0) {
		ShowAlert("Player frame number request failed!",NULL);
		return(-1);
	}
	number  = unbcd((int) buffer[6]);
	number += unbcd((int) buffer[5]) * 30;
	number += unbcd((int) buffer[4]) * 1800;
	number += unbcd((int) buffer[3]) * 108000;
	return(number);
}

/* Set Audio parameters */

int SetVISCAAudio(mode)
int mode;
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}
	switch (mode) {
	case AUDIO_OFF:
		if (Send(AudioOff,6,5) < 0) {
			ShowAlert("Player audio fade out request failed!",NULL);
			return(-1);
		}
		break;
	case STEREO:
		if (Send(AudioOn,6,5) < 0) {
			ShowAlert("Player audio fade in request failed!",NULL);
			return(-1);
		}
		break;
	case LEFT_ONLY:
	case RIGHT_ONLY:
		ShowNote("This option not supported for this unit",NULL);
		break;
	default:
		sprintf(message,"Unknown mode requested - %d",mode);
		ShowAlert(message,NULL);
		return(-1);
	}
	return(NULL);
}

/* This is a noop for now, as the player doesn't seem to support this option */

int SetVISCAVideo(mode)
int mode;
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}
	switch (mode) {
	case VIDEO_OFF:
		if (Send(VideoOff,6,5) < 0) {
			ShowAlert("Player video fade out request failed!",NULL);
			return(-1);
		}
		break;
	case VIDEO_ON:
		if (Send(VideoOn,6,5) < 0) {
			ShowAlert("Player video fade in request failed!",NULL);
			return(-1);
		}
		break;
	default:
		sprintf(message,"Unknown mode requested - %d",mode);
		ShowAlert(message,NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Seek and display a single frame */

int ShowVISCAFrame(frame)
int frame;
{
	int i;

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameter */

	if ((frame < 0) || (frame > 216000)) {
		sprintf(message,"Invalid Frame address requested : %d",frame);
		ShowAlert(message,NULL);
		return(-1);
	}

/* Convert from frame number to time code */

	memcpy(buffer,SearchStill,11);
	i = frame % 30;
	buffer[8] = (unsigned char) bcd(i);
	frame = frame / 30;
	i = frame % 60;
	buffer[7] = (unsigned char) bcd(i);
	frame = frame / 60;
	i = frame % 60;
	buffer[6] = (unsigned char) bcd(i);
	frame = frame / 60;
	buffer[5] = (unsigned char) bcd(frame);

/* Send out a search frame request */

	if (Send(buffer,6,300) < 0) {
		ShowAlert("Search command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Seek and display a complete video segment */

int ShowVISCASegment(start,end,repeat)
int start,end,repeat;
{
	int i;

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* Check for valid parameters */

	if ((start < 0) || (start > 216000)) {
		sprintf(message,"Invalid starting frame address requested - %d",start);
		ShowAlert(message);
		return(-1);
	}

	if ((end < 0) || (end > 216000)) {
		sprintf(message,"Invalid ending frame address requested - %d",end);
		ShowAlert(message);
		return(-1);
	}

	if (start >= end) {
		sprintf(message,"Starting frame (%d) not less than end (%d)",start,end);
		ShowAlert(message,NULL);
		return(-1);
	}

/* Convert from frame number to time code */

	memcpy(buffer,SearchPlay,11);
	i = start % 30;
	buffer[8] = (unsigned char) bcd(i);
	start = start / 30;
	i = start % 60;
	buffer[7] = (unsigned char) bcd(i);
	start = start / 60;
	i = start % 60;
	buffer[6] = (unsigned char) bcd(i);
	start = start / 60;
	buffer[5] = (unsigned char) bcd(start);

/* Send out a search frame request */

	if (Send(buffer,6,300) < 0) {
		ShowAlert("Search command failed!",NULL);
		return(-1);
	}

/* Convert from frame number to time code */

	memcpy(buffer,StopAt,11);
	i = end % 30;
	buffer[7] = (unsigned char) bcd(i);
	end = end / 30;
	i = end % 60;
	buffer[6] = (unsigned char) bcd(i);
	end = end / 60;
	i = end % 60;
	buffer[5] = (unsigned char) bcd(i);
	end = end / 60;
	buffer[4] = (unsigned char) bcd(end);

/* Send out a stop at request (don't wait for completion) */

	if (Send(buffer,3,5) < 0) {
		ShowAlert("Stop at command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : play forward */

int VISCAPlay()
{
/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Wait here until playing has started */

	if (Send(Play,6,5) < 0) {
		ShowAlert("Play command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : play reverse */

int VISCAReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Wait here until playing has started */

	if (Send(Reverse,6,5) < 0) {
		ShowAlert("Play command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Continue an interrupted segment */

int VISCAContinue()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Wait here until playing has started again */

	if (Send(Play,6,5) < 0) {
		ShowAlert("Continue command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : pause the player */

int VISCAStill()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}
	if (buffer[2] == StopStatus) {
		ShowAlert("Player is stopped!",NULL);
		return(-1);
	}

/* Wait here for return confirmation */

	if (Send(Still,6,5) < 0) {
		ShowAlert("Still command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Stop all playback */

int VISCAStop()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == StopStatus) return(NULL);
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Wait here for return confirmation */

	if (Send(Stop,6,5) < 0) {
		ShowAlert("Stop command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move 1 frame forward */

int VISCAStepForward()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] != StillStatus) {
		ShowAlert("Player is not in pause mode!",NULL);
		return(-1);
	}

/* Take one step forward */

	if (Send(StepForward,6,5) < 0) {
		ShowAlert("Forward one frame command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Move 1 frame backwards */

int VISCAStepReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] != StillStatus) {
		ShowAlert("Player is not in pause mode!",NULL);
		return(-1);
	}

/* Take one step back */

	if (Send(StepReverse,6,5) < 0) {
		ShowAlert("Reverse one frame command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move slowly forward */

int VISCASlowForward()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Take slow steps forward */

	if (Send(SlowForward,6,5) < 0) {
		ShowAlert("Slow forward-play command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move slowly backwards */

int VISCASlowReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Take slow steps back */

	if (Send(SlowReverse,6,5) < 0) {
		ShowAlert("Slow reverse-play command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move fast forward */

int VISCAFastForward()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Take fast steps forward */

	if (Send(FastForward,6,5) < 0) {
		ShowAlert("Fast forward-play command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : move fast backwards */

int VISCAFastReverse()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Take fast steps back */

	if (Send(FastReverse,6,5) < 0) {
		ShowAlert("Fast reverse-play command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* Command : Provide visual frame indicator */

int VISCAIndex(status)
int status;
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) {
		ShowAlert("Player is empty!",NULL);
		return(-1);
	}

/* Turn on/off frame location display */

	if (status) {
		if (Send(IndexOn,6,5) < 0) {
			ShowAlert("Player status display command failed!",NULL);
			return(-1);
		}
	}
	else {
		if (Send(IndexOff,6,5) < 0) {
			ShowAlert("Player status display command failed!",NULL);
			return(-1);
		}
	}
	return(NULL);
}

/* Eject the video source */

int VISCAEject()
{

/* Make sure the player is initialized correctly */

	if (PlayerId < 0) {
		ShowAlert("VISCA Player not initialized",NULL);
		return(-1);
	}

/* See if we are in a reasonable mode */

	if (Send(StatusRequest,4,5) < 0) {
		ShowAlert("Player status request failed!",NULL);
		return(-1);
	}
	if (buffer[2] == EjectStatus) return(NULL);
	if (buffer[2] != StopStatus) {
		if (Send(Stop,6,5) < 0) {
			ShowAlert("Player STOP request failed!",NULL);
			return(-1);
		}
	}
	if (Send(Eject,6,5) < 0) {
		ShowAlert("Player ejection command failed!",NULL);
		return(-1);
	}
	return(NULL);
}

/* end of visca.c */
