/*
 *   Copyright (c) 1993 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* $Header: /plx/source/Xclients/generic/2.17/src/lib/xt/RCS/JPEG.c,v 1.64 1996/02/01 03:14:33 drapeau Exp $ */
/* $Log: JPEG.c,v $
 * Revision 1.64  1996/02/01  03:14:33  drapeau
 * Fixed errors in the way that compress and display rectangles were
 * calculating their widths and heights.  The widget was not honoring the
 * JPEG rule that suggests image sizes should be a multiple of 8 (and in
 * fact, according to Dave Arnold, should be a multiple of 16 for some
 * reason).
 *
 * The widget is now more robust about checking widths and heights for
 * rectangles, as well as checking for illegal widths and heights (no
 * width or height should be less than 8, and the "displaySourceRect"
 * should not consider a larger area than the "compressDestRect", since
 * the compressDestRect represents the final, actual width and height of
 * a JPEG image.
 *
 * Revision 1.63  1996/01/10  09:31:19  drapeau
 * Fixed lots of minor compiler warnings to sanity-check the module.
 *
 * Revision 1.62  1996/01/06  01:31:43  drapeau
 * Improved the portability of the widget in that it checks more cleanly
 * whether the X server being used is an overlay server.  This makes the
 * JPEG automatically detect whether it should track ConfigureNotify
 * events so it can properly redraw JPEG images after being moved.
 *
 * Revision 1.61  1995/12/19  00:12:19  drapeau
 * Fixed a minor error found by Paul: the widget installs its own X error
 * handler to determine whether it's possible to use shared memory.
 * However, the widget had not taken into account the possibility that
 * the application programmer might already have overridden the default
 * error handler.  The new version of the widget now correctly saves the
 * previous error handler and restores that error handler when done.
 *
 * Revision 1.60  1995/12/14  02:55:46  drapeau
 * Minor fixes to check for NULL JPEG images.
 *
 * Revision 1.59  1995/10/26  07:33:43  drapeau
 * Minor modification: default qFactor is now 50 to reflect the
 * difference in the way the Parallax hardware library interprets Qtable
 * data.  An old-style server interprets a Q=100 table the same as a
 * new-style server interprets a Q=50 table.
 *
 * Revision 1.58  1995/10/25  00:29:49  drapeau
 * Fixed an error in the way Q tables were being used on the
 * decompression side: the new, requested Q table either wasn't being put
 * into the decompressGC, or (in the IBM case) the Q table was being put
 * before a Port was attached to the Drawable in question.  The function
 * that creates the decompress Q table when necessary now takes an extra
 * argument that indicates whether to forcibly put the Q table.  The
 * calling function has enough information to pass in the correct value
 * for this new argument.
 *
 * Revision 1.57  1995/10/17  19:09:57  drapeau
 * Two changes:
 *
 * Fixed CreateJPEGWidgetQTablesIfNecessary() so that it does not call
 * XPlxPutTable() each time it's called; the new table is put only when a
 * new GC is created.
 *
 * Minor change in hardwareParams data structure due to change in the
 * "plx_io" data structure to accommodate new information about hardware
 * capabilities.
 *
 * Revision 1.56  1995/10/05  19:40:58  drapeau
 * Fixed minor error in Initialize() method: the widget should now use
 * the Parallax shared memory extension, not the MIT shared memory
 * extension (although in some cases this will turn out to be the same
 * extension).
 *
 * Revision 1.55  1995/09/28  03:39:06  drapeau
 * Modified the widget to use special Parallax versions of X11 shared
 * memory extension functions.  In order for shared memory functions to
 * work correctly on the Solaris 2.X product. the server side created a
 * "PLX-SHM" extension which is similar but less functional than the
 * MIT-SHM extension (specifically, the XShm{Get,Put}Image() functions
 * are not implemented, and possibly other functions).  On the client
 * side, this created a name space conflict:  applications that used both
 * XPlxShm{Get,Put}CImage() and XShm{Get,Put}Image() functions were not
 * able to specify the extension to which to send each request.
 *
 * The solution involved cloning certain functions from the
 * XShm... extension and making them part of the XPlx library; these
 * calls connect with the PLX-SHM extension.  The JPEG widget now uses
 * these calls, which have a prefix of "XPlxShm..." instead of "XShm...".
 *
 * Revision 1.54  1995/09/21  01:42:42  drapeau
 * Added code to take advantage of enhanced server reporting
 * capabilities: the XPlxQueryConfig() function new reports more specific
 * information about the capabilities of the hardware, including whether
 * JPEG hardware is available or not.
 *
 * Revision 1.53  1995/09/19  20:51:46  drapeau
 * Replaced GetTrueColorVisual() with more portable, new function
 * XPlxGetVideoVisual() that should work cleanly across platforms.
 *
 * Also, removed extraneous function prototypes.
 *
 * Revision 1.52  1995/09/15  01:48:26  drapeau
 * Fixed an error in the way the jfifOutputFile resource was implemented.
 * This fix is similar to that for the jfifInputFile resource: in the
 * SetValues() method, the widget only used to check if the
 * jfifOutputFile *pointer* was changed between the current copy of the
 * widget and the proposed copy.  However, if an application programmer
 * uses the same string space but writes different data into that area
 * with the intent of writing a new JFIF file, the widget was not able to
 * catch this case.  The fix is to check every time the SetValues()
 * method is called: check not only if the pointer has changed but if the
 * current and new strings differ (i.e., a strcmp).
 *
 * Also, cleaned up the code, especially the SetValues() method, moving
 * the long checks for jfifInputFile and jfifOutputFile resource changes
 * into their own functions.  The intent was to enhance the readability
 * of the code.
 *
 * Revision 1.51  1995/08/17  20:24:54  drapeau
 * Modified SetValues() method: now whenever the widget is told to point
 * to a new JPEG data segment, the widget informs itself that it has new
 * data to display.
 *
 * Revision 1.50  1995/08/15  06:29:43  drapeau
 * Three fixes:
 *
 * 1) Added support for the IBM AIX overlay server, allowing the widget
 *    to track ConfigureNotify events as happens with the HP overlay
 *    server.  This allows window moves to cause the widget to redraw
 *    itself.
 *
 * 2) Fixed an error in CreateJPEGWidgetQTablesIfNecessary(), which did
 *    not always put the decompression Q table.  The error exhibited
 *    itself in the MovieTool application: when changing image quality,
 *    the compression Q table was being put but not the decompression Q
 *    table, resulting in images that displayed incorrectly.
 *
 * 3) Fixed annoying compiler warnings, mostly as a result of calls to
 *    malloc().
 *
 * Revision 1.49  1995/07/12  17:55:39  drapeau
 * Modified SetValues() method: when a new qFactor is requested for the
 * widget, make sure that the compression GC gets the new Qtable
 * installed immediately.
 *
 * Revision 1.48  1995/06/21  03:36:55  drapeau
 * Moved the creation of the first decompressGC from the Realize() method
 * to the Redisplay() method, via addition of a new function
 * "CreateJPEGWidgetQTablesIfNecessary()".  This is a workaround for a
 * limitation of the Xv-based X servers (e.g., HP overlay, IBM overlay),
 * which does not keep track of private GC data (such as Q tables) in the
 * server; instead, the client side is burdened with keeping track of
 * which GC's and Drawables have Q table data with them.  Compounding
 * this problem, there is no explicit way to "attach" an Xv Port to a
 * Drawable, so functions that check whether a Port is already in use
 * will sometimes get information contrary to what one would expect.  In
 * this case, a Q table might already have been put into a Port, but the
 * Port is not rendered attached so subsequent calls to XPlxPutTable()
 * are allowed to put new Q table data to the same Port.
 *
 * By delaying the creation of the decompressGC (and thus an
 * XPlxPutTable() call) until just before the first XPlxPutCImage() is
 * done, the chance of a different Q table being put into a Port before
 * the decompression is greatly reduced, although strictly speaking a
 * race condition is still possible, while unlikely.
 *
 * Revision 1.47  1995/06/20  21:32:35  drapeau
 * Removed more code that has to do with Q tables: evidently it wasn't
 * enough to quit zig-zagging the Q tables before saving a JFIF file,
 * because the 8-bit values were also shifted right and left by one bit
 * on loading and saving.  This one-bit correction (i.e., power of two
 * multiplication/division) is now not necessary, and has been removed.
 *
 * Revision 1.46  1995/05/27  07:42:11  drapeau
 * Fixed a semantic error in the way JFIF images are loaded into the JPEG
 * widget.  The previous method of opening JFIF images did not easily
 * allow programmers to use the same string pointer to specify different
 * JFIF file names (e.g., using "char* fileName" to successively contain
 * the filename "image1", "image2", "image3", etc.) and have the JPEG
 * widget understand that the filename changed even though the pointer
 * pointed to the same area of memory).  The widget now checks not only
 * of the jfifInputFile pointer has changed but if the string is
 * different from the last time the widget was told to open a JFIF file.
 *
 * This change should be applied to the jfifOutputFile resource as well,
 * but has not been done for this version of the widget.
 *
 * Revision 1.45  1995/05/25  02:28:34  drapeau
 * Removed code that manually zigzag's the Q table being used when saving
 * JFIF files; also removed code that un-zigzag's the Q table when
 * opening JFIF files.
 *
 * Revision 1.44  1995/05/02  01:32:09  drapeau
 * Added support for hardware zoom: at Initialize() time the widget asks
 * the connection for the current zoom factor in use.  Currently that
 * zoom factor is used only in calculating whether JPEG compression
 * requests are greater than that allowed by the C-Cube chip (1024 * the
 * zoom factor).
 *
 * Revision 1.43  1995/03/03  22:14:00  drapeau
 * Fixed an error in trying to allocate shared memory images: if a
 * process tries to create more shared memory segments than the system
 * allows, XPlxGetShmInfo() returns NULL (indicating that no shared
 * memory segment was allocated).  However, the JPEG widget was not
 * checking for this condition.  The widget now does so, and falls back
 * to non-shared-memory operation.
 *
 * Revision 1.42  1994/10/23  00:46:02  drapeau
 * Fixed incompatibilities between function prototypes and calls to those
 * functions, for better ANSI compliance.
 *
 * Revision 1.41  1994/10/14  21:14:29  drapeau
 * Removed event handlers on the widget during the Destroy method, if
 * they are installed.
 *
 * Revision 1.40  1994/10/14  02:00:22  drapeau
 * Two major changes:
 *
 * 1) Fixed an error in SetValues() that had neglected to install new Q
 *    tables when the programmer requests a new destWin resource.
 *
 * 2) Added HP-specific code to handle window move events.  On an overlay
 *    product, the window server may not necessarily move the window's
 *    pixel data when the window is moved.  In this case, the widget's
 *    window needs to register interest in window move events.  This
 *    version of the widget checks if it is connecting with an HP server;
 *    if so, the widget assumes it's working with an overlay board and so
 *    registers an Xt handler for window move events.
 *
 * Revision 1.39  1994/09/03  07:10:51  drapeau
 * Three changes:
 *
 * 1) Fixed an error in the way JFIF files were loaded.  In the
 *    Initialize() method, if a JFIF file was loaded it might overwrite
 *    the existing pointer to a JPEG data segment (if shared memory was
 *    being used).  This would result in a memory leak.
 *
 * 2) If the jfifInputFile resource is set but the load was unsuccessful,
 *    the jfifInputFile String is set back to NULL, so the application
 *    programmer can tell that the JPEG widget has no JFIF input file in
 *    use.
 *
 * 3) Added an old hack for writing to the JPEG window; the hack is to
 *    tag the window for graphics then later to re-tag it for video
 *    override.  This is done to support older servers that don't
 *    correctly implement window id's.
 *
 * Revision 1.38  1994/09/02  06:52:34  drapeau
 * Two major changes:
 *
 * 1) Fixed errors in the way the JPEG widget changed its display
 *    parameters after reading in a JFIF file.  It now correctly resets
 *    the default DisplaySourceRect rectangle to fit the size of the JFIF
 *    file (before it would retain its default value of 640x480; if the
 *    JFIF image were smaller than this, the result would be stride
 *    errors when displaying the JFIF image).
 *
 * 2) Added a new resource, "drawMethod", allowing application
 *    programmers to draw on top of JPEG windows.  This resource is
 *    intended for applications that have some need to draw
 *    graphics-over-JPEG, for example, annotating a JPEG image, adding a
 *    caption to a JPEG image, or placing a logo at the bottom right of a
 *    JPEG movie (a la CNN, NBC, etc.).
 *
 * Revision 1.37  1994/09/01  19:12:58  drapeau
 * Fixed diagnostic printing; only do so if the diagnostics resource is
 * set to True.
 *
 * Revision 1.36  1994/08/30  00:09:26  drapeau
 * Added extra diagnostic message to help debug non-shared memory
 * decompression.
 *
 * Revision 1.35  1994/08/06  00:42:54  drapeau
 * Added diagnostic messages to functions for reading JFIF files.
 *
 * Revision 1.34  1994/07/23  05:27:55  drapeau
 * Several changes:
 *
 * * Cleaned up the way shared memory is used by the widget.  The widget
 *   did not allow the application programmer to forcibly disable the use
 *   of shared memory -- if the widget detected that shared memory usage
 *   was possible, the widget automatically used it.  This has been
 *   changed; the application programmer may now set the
 *   "useSharedMemory" resource to False at initialization time; if this
 *   is done, the widget will not use shared memory for storing JPEG
 *   data.
 *
 * * Added diagnostic messages to aid in debugging of the widget.
 *
 * Revision 1.33  1994/06/14  00:41:12  drapeau
 * Removed a typographical error in the Resize method, allowing the JPEG
 * widget to resize its default compression rectangle to the size of the
 * widget's window.  Before this error was fixed, the compression
 * rectangle would always resize to 480 lines tall.
 *
 * Revision 1.32  1994/06/09  23:50:55  drapeau
 * Minor changes to assure that compression happens correctly when the
 * JPEG widget's window is resized.
 *
 * Revision 1.31  1994/05/10  00:17:16  drapeau
 * Added code to implement a diagnostics resource for printing widget
 * diagnostics.
 *
 * Revision 1.30  1994/04/30  00:48:37  drapeau
 * Several changes:
 * * Removed unused variables.
 * * Cleaned up code to create a TrueColor window for the display part of
 *   the widget.  It should now correctly create a TrueColor window in
 *   all cases possible.
 * * Moved the code to implement getting a TrueColor visual to another
 * * module, since other modules share this common functionality and it
 * * needs to be portable across platforms, too.
 *
 * Revision 1.29  1994/04/28  05:28:42  drapeau
 * Removed the hack workaround for errors in the OpenWindows server.
 * The only affected server should now be OpenWindows 3.X under SunOS 4.X.
 *
 * Revision 1.28  1994/04/21  23:57:22  drapeau
 * Fixed an error in the SetValues() function.  When the widget is asked
 * both to get and put a new image, the order in which these requests are
 * fulfilled matters.  A get request must be done before a put, so that
 * the get can produce compressed image data for the put to display.
 * This version of the code takes care of the getNewImage before
 * servicing the putNewImage request.
 *
 * Revision 1.27  1994/03/29  01:43:03  drapeau
 * Inserted a function "JPEGWidgetDebug()" that does nothing except to
 * serve as a placeholder for setting breakpoints in debuggers that don't
 * let the programmer specify a specific C module to examine.
 *
 * Also, cleaned up the way JPEG data are allocated to account for errors
 * in the HP product's hardware.  The HP hardware, when asked to compress
 * an image, returns an extra four bytes at the beginning of the image.
 * These four bytes are unusable but to preserve the transparency of the
 * libraries the four bytes are hidden in allocation and free requests.
 *
 * Revision 1.26  1994/01/12  17:38:19  drapeau
 * Added code to allow the JPEG widget to automatically determine whether
 * it's okay to use shared memory for storage of JPEG image data.
 * Also, removed the extra two calls to XPlxShmGetCImage(), since they
 * are no longer part of the API (I had previously added the last two
 * arguments to that function but later found they were unnecessary).
 * Also, added simple diagnostics to aid in debugging, especially on the HP.
 *
 * Revision 1.24  1993/12/16  01:44:11  drapeau
 * Added code to detect whether it's okay to use shared memory for JPEG
 * images.  Also, minor changes to eliminate compiler warnings and
 * diagnostic messages.
 *
 * Revision 1.23  1993/11/19  17:51:11  drapeau
 * Enabled the use of shared memory again.  Also, fixed an error when
 * telling the widget not to compress the contents of a window anymore
 * (this is done by setting the "sourceWin" resource to Window 0; it is a
 * special code telling the JPEG widget not to compress anything anymore).
 * The error was that even when pointing to Window 0, the widget would try
 * to install a QTable into that non-existant window.
 *
 * Revision 1.22  1993/11/19  04:44:37  drapeau
 * Temporarily disabled the use of shared memory until server problems have
 * been fixed for OpenWindows on Solaris 2.X.
 *
 * Revision 1.21  1993/11/04  21:54:04  drapeau
 * Several changes to clean up the code and work around errors in the
 * OpenWindows server:
 * * An error occurred in the default value of the 'getNewImage' resource;
 *   it should be set to True to enable the widget to display correctly the
 *   first time it is told to Redisplay itself.
 * * Extra error checking occurs when trying to attach a shared memory segment
 *   to the widget.
 * * The Redisplay() method checked whether the compressSourceRect and
 *   compressDestRect were larger than the dimensions the CCube chip could
 *   handle (1K x 1K), but the check occurred in the wrong place.  This was
 *   a simple fix, moving the code a little earlier.
 * * A workaround for an OpenWindows display error was placed in the Redisplay()
 *   method.  The workaround tags the display window for graphics after the
 *   window was damaged, but subsequently tags the window for overlay.  This
 *   prevents the display from "flashing black".  This error does not exist in
 *   the X11R5 server, and eventually this workaround should be deleted.
 * * A semantic error occurred in waiting for shared memory completion events;
 *   the wrong event code was passed to the XShmGetEventBase() function.
 *
 * Revision 1.20  93/09/09  17:40:08  drapeau
 * Changed permissions for this file; no code changes made.
 * 
 * Revision 1.19  93/08/30  19:34:54  drapeau
 * Added code to the SetValues() method to better handle changes in
 * the qFactor resource.  Before, changes in the qFactor resource were being
 * ignored; this was an error.  They are no longer being ignored, although
 * visible results are not immediately forced on the widget; the results
 * will not be noticeable until the next Redisplay() is called.
 * 
 * Revision 1.18  93/08/25  23:24:34  drapeau
 * Fixed an error in redisplaying images after loading them via the
 * jfifInputFile resource.
 * 
 * Revision 1.17  93/08/25  17:08:02  drapeau
 * Modified the Redisplay() method to accommodate a limitation of the CCube
 * JPEG chip: the chip is limited to compressing an area of 1024x1024 pixels,
 * so the Redisplay() method now checks if the compressSourceRect and
 * compressDestRect rectangles exceed these boundaries.  If so, the method
 * fixes the rectangles to stay within the limits of these boundaries.
 * 
 * Revision 1.16  93/08/16  15:48:28  drapeau
 * One significant change to the SetValues() method: when the putNewImage
 * resource is set, the widget now correctly sets a widget variable
 * informing the widget that there is new, valid data to display.
 * Other changes are cosmetic only.
 * 
 * Revision 1.15  93/08/16  09:49:29  drapeau
 * Removed the static variables for default compression and display rectangles,
 * moving those data structures into the widget's data structure itself.  This
 * was done to avoid possible problems when creating multiple widgets, and to
 * clarify the encapsulation in the JPEG widget (if this were a true object,
 * then all widget-specific data would be part of the widget's class structure,
 * nothing would be static, global variables).
 * Also, cleaned up the way the Redisplay() method is done, checking for extra
 * boundary conditions.
 * 
 * Revision 1.14  93/08/10  18:18:50  drapeau
 * Cleaned up the way shared memory was used.  Also, made the widget more
 * efficient in general, mostly in the way the Redisplay() method was being
 * called and how the SetValues() method was deciding when and how to redraw
 * the widget.
 * 
 * Revision 1.13  93/07/16  18:00:52  drapeau
 * Cleaned up code some, added debugging messages to test performance and
 * configurability of the widget.  Also, added code to resize the rectangles
 * used for compression and decompression based on the parameters in the
 * jpegData resource.  For example, when this widget is made to point to a new
 * jpegData image, it might be a different size than a previous image.  This
 * widget now acts on possible changes in width and height of the image.
 * Also, problems with using shared memory have been eliminated.  Before, this
 * widget could only be used in synchronous mode.  New code that checks for
 * a "ShmCompletionCode" event prevents overwriting the shared memory buffer
 * before it is ready to overwrite.  No great performance loss seems to result,
 * however, this has not been strictly tested for performance.
 * 
 * Revision 1.12  93/07/16  10:35:26  drapeau
 * Major changes to this widget, the majority of which have to do with the
 * addition of the ability to use shared memory.  The widget will now
 * automatically use shared memory if it can, to increase performance.
 * Also, a new resource was added: putNewImage.  This was put into place
 * largely because of shared memory: when using shared memory, it makes no
 * sense to destroy a shared memory segment and create a new one every time
 * a new image is being compressed and/or decompressed; it is preferable to
 * create the shared memory segment only at initialization time and use that
 * segment over and over.  However, there was no way to tell the widget that
 * new data have been written to the shared memory segment.  The new resource,
 * "putNewImage", does exactly that: it signals to the widget that there are
 * new data in the shared memory segment to be displayed.
 * 
 * Revision 1.11  93/06/11  16:50:07  drapeau
 * Minor change to the Resize() method: if the display rectangle being used
 * is the default display rectangle (as opposed to a programmer-supplied
 * rectangle), change the width & height fields to reflect the new size
 * of the widget.  In this way, the JPEG image will always fill the widget's
 * window and will squeeze to the correct size when the widget is made
 * smaller.
 * 
 * Revision 1.1  93/06/10  17:06:25  drapeau
 * Initial revision
 *
 */
static char jpegRcsid[] = "$Header: /plx/source/Xclients/generic/2.17/src/lib/xt/RCS/JPEG.c,v 1.64 1996/02/01 03:14:33 drapeau Exp $";

#include <X11/Xatom.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "JPEGP.h"
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#define offset(field) XtOffsetOf(JPEGRec, jpeg.field)

static XtResource resources[] =					    /* {name, class, type, size, offset, ... */
{								    /* ...default_type, default_addr}, */
{
  XtNsourceWin,
  XtCSourceWin,
  XtRWindow,
  sizeof(Window),
  offset(sourceWin),
  XtRImmediate,
  (XtPointer) 0
  },
{
  XtNcompressSourceRect,
  XtCCompressSourceRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(compressSourceRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNcompressDestRect,
  XtCCompressDestRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(compressDestRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNjpegData,
  XtCJPEGData,
  XtRPointer,
  sizeof(XPlxCImage*),
  offset(jpegData),
  XtRImmediate,
  (XtPointer) 0
  },
{
  XtNdestWin,
  XtCDestWin,
  XtRWindow,
  sizeof(Window),
  offset(destWin),
  XtRImmediate,
  (XtPointer) 0
  },
{
  XtNdisplaySourceRect,
  XtCDisplaySourceRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(displaySourceRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNdisplayDestRect,
  XtCDisplayDestRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(displayDestRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNsqueezeDest,
  XtCSqueezeDest,
  XtRBoolean,
  sizeof(Boolean),
  offset(squeezeDest),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNqFactor,
  XtCQFactor,
  XtRInt,
  sizeof(int),
  offset(qFactor),
  XtRImmediate,
  (XtPointer) 50
  },
{
  XtNgetNewImage,
  XtCGetNewImage,
  XtRBoolean,
  sizeof(Boolean),
  offset(getNewImage),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNputNewImage,
  XtCPutNewImage,
  XtRBoolean,
  sizeof(Boolean),
  offset(putNewImage),
  XtRImmediate,
  (XtPointer) False
  },
{
  XtNjfifInputFile,
  XtCJFIFInputFile,
  XtRString,
  sizeof(String),
  offset(jfifInputFile),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNjfifOutputFile,
  XtCJFIFOutputFile,
  XtRString,
  sizeof(String),
  offset(jfifOutputFile),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNuseSharedMemory,
  XtCUseSharedMemory,
  XtRBoolean,
  sizeof(Boolean),
  offset(useSharedMemory),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNdiagnostics,
  XtCDiagnostics,
  XtRBoolean,
  sizeof(Boolean),
  offset(diagnostics),
  XtRImmediate,
  (XtPointer) False
  },
{
  XtNdrawMethod,
  XtCDrawMethod,
  XtRPointer,
  sizeof(XtPointer),
  offset(drawMethod),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNzoomFactor,
  XtCZoomFactor,
  XtRInt,
  sizeof(int),
  offset(zoomFactor),
  XtRImmediate,
  (XtPointer) 1
  },
{
  XtNjpegAvailable,
  XtCJPEGAvailable,
  XtRBoolean,
  sizeof(Boolean),
  offset(jpegAvailable),
  XtRImmediate,
  (XtPointer) True
  },
};

#undef offset

static int	shmMajorOpcode;
static Bool	xShmAttachErrorOccurred;
static plx_IO*	hardwareParams;					    /* Holds all relevant info about the board being used. */

static void JPEGAction(/* Widget, XEvent*, String*, Cardinal* */);

static XtActionsRec actions[] =
{
  /* {name, procedure}, */
    {"jpeg",	JPEGAction},
};

static char translations[] =
"<Key>:		jpeg()	\n\
";

/* Declaration of methods */

static void 		Initialize(Widget, Widget, ArgList, Cardinal*);
static void		Realize(Widget, XtValueMask*, XSetWindowAttributes*);
static void 		Redisplay(Widget, XEvent*, Region);
static void 		Destroy(Widget);
static void 		Resize(Widget);
static Boolean		SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void		CreateJPEGDisplayWindow(JPEGWidget,
						XtValueMask*,
						XSetWindowAttributes*);
static void		PrintWidgetDiagnostics(JPEGWidget, char*);
static XPlxCImage*	jpeg_load_jfif(JPEGWidget, char*);
static int		jpeg_save_jfif(JPEGWidget, XPlxCImage*, char*, unsigned char*);
int			JPEGWidgetShmErrorHandler(Display* display,
						  XErrorEvent* errorEvent);
static void		RegisterJPEGWidgetEventHandler(JPEGWidget);
static void		JPEGWidgetEventHandler(Widget, XtPointer, XEvent*, Boolean*);
static void		CreateJPEGWidgetQTablesIfNecessary(JPEGWidget, Boolean);
static Boolean		JPEGWidgetCheckJFIFInputFile(JPEGWidget, JPEGWidget);
static void		JPEGWidgetCheckJFIFOutputFile(JPEGWidget, JPEGWidget);
static void		AdjustCompressRectangles(JPEGWidget);
static void		AdjustDisplayRectangles(JPEGWidget);


JPEGClassRec jpegClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"JPEG",
    /* widget_size		*/	sizeof(JPEGRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* jpeg fields */
    /* empty			*/	0
  }
};

WidgetClass jpegWidgetClass = (WidgetClass)&jpegClassRec;


static void JPEGWidgetDebug(void)				    /* Used to make it easy to tell stupid debuggers what... */
{								    /* ...file to switch to for debugging. */
  return;
}



static void Initialize(Widget		treq,
		       Widget		tnew,
		       ArgList		args,
		       Cardinal*	numArgs)
{
  Display*			display = XtDisplay(tnew);
  XPlxCImage*			jfifImage;
  XPlxHardwareInfo		hardwareInfo;			    /* Info about hardware capabilities from XPlxQueryConfig() */
  int				majorVersion;
  int				minorVersion;
  JPEGWidget			newWidget = (JPEGWidget) tnew;
  Bool				okayToUseSharedMemory;
  XErrorHandler			oldErrorHandler;
  Window			rootWindow;
  int				screen;
  static XShmSegmentInfo*	sharedMemInfo = (XShmSegmentInfo*)NULL;
  Bool				sharedPixmaps;
  int				shmFirstEvent;
  int				shmFirstError;
  Window			sourceWindow;
  Status			status;
  XWindowAttributes		windowAttributes;
  
  JPEGWidgetDebug();						    /* Make it easy to set breakpoints in stupid debuggers */
  screen = XDefaultScreen(display);
  rootWindow = RootWindow(display, screen);
  if (newWidget->jpeg.sourceWin == 0)				    /* If the widget doesn't specify a valid sourceWin, then... */
    sourceWindow = rootWindow;					    /* ...point to the Root Window for later calls. */
  else
    sourceWindow = newWidget->jpeg.sourceWin;
  newWidget->jpeg.defaultCompressSourceRect.x = 0;
  newWidget->jpeg.defaultCompressSourceRect.y = 0;
  newWidget->jpeg.defaultCompressSourceRect.width = 640;
  newWidget->jpeg.defaultCompressSourceRect.height = 480;
  
  newWidget->jpeg.defaultCompressDestRect.x = 0;
  newWidget->jpeg.defaultCompressDestRect.y = 0;
  newWidget->jpeg.defaultCompressDestRect.width = 640;
  newWidget->jpeg.defaultCompressDestRect.height = 480;
  
  newWidget->jpeg.defaultDisplaySourceRect.x = 0;
  newWidget->jpeg.defaultDisplaySourceRect.y = 0;
  newWidget->jpeg.defaultDisplaySourceRect.width = 640;
  newWidget->jpeg.defaultDisplaySourceRect.height = 480;
  
  newWidget->jpeg.defaultDisplayDestRect.x = 0;
  newWidget->jpeg.defaultDisplayDestRect.y = 0;
  newWidget->jpeg.defaultDisplayDestRect.width = 640;
  newWidget->jpeg.defaultDisplayDestRect.height = 480;
  
  if (newWidget->jpeg.compressSourceRect == (XRectangle*)NULL)
    newWidget->jpeg.compressSourceRect =
      &(newWidget->jpeg.defaultCompressSourceRect);
  
  if (newWidget->jpeg.compressDestRect == (XRectangle*)NULL)
    newWidget->jpeg.compressDestRect =
      &(newWidget->jpeg.defaultCompressDestRect);
  
  if (newWidget->jpeg.displaySourceRect == (XRectangle*)NULL)
    newWidget->jpeg.displaySourceRect =
      &(newWidget->jpeg.defaultDisplaySourceRect);
  
  if (newWidget->jpeg.displayDestRect == (XRectangle*)NULL)
    newWidget->jpeg.displayDestRect =
      &(newWidget->jpeg.defaultDisplayDestRect);
  
  if (newWidget->jpeg.destWin == (Window)0)			    /* Was an already existing window chosen for displaying... */
  {								    /* ...JPEG images?  If not, try to replicate the source... */
    status = XGetWindowAttributes(display,			    /* ...window's information about window depth, etc. */
				  sourceWindow,
				  &windowAttributes);
    if (status != 0)						    /* Did the inquiry succeed? */
    {								    /* Yes, use the source window's values for creation of... */
      newWidget->core.depth = windowAttributes.depth;		    /* ...the destination window. */
      newWidget->core.colormap = windowAttributes.colormap;
      newWidget->jpeg.visualClass = windowAttributes.visual->class;
    }
    else							    /* No, inquiry failed.  Report the error and set default... */
    {								    /* ...values. */
      XtWarning("JPEG Widget: Could not get information about the source window.\n");
      newWidget->core.depth = 24;
      newWidget->core.colormap = (Colormap)NULL;
      newWidget->jpeg.visualClass = TrueColor;
    }
  }
  newWidget->jpeg.compressGC = (GC)NULL;
  newWidget->jpeg.decompressGC = (GC)NULL;
  newWidget->jpeg.qTableSize = MakeQTables(newWidget->jpeg.qFactor, /* Initialize the Q table for this JPEG window */
					   &(newWidget->jpeg.qTable));
  newWidget->core.width = newWidget->jpeg.displayDestRect->width;
  newWidget->core.height = newWidget->jpeg.displayDestRect->height;
  newWidget->jpeg.shmExtensionExists =				    /* Determine the major protocol opcode for the shared... */
    XQueryExtension(display, PlxSharedMemoryExtensionName,	    /* ...memory extension. */
		    &(newWidget->jpeg.shmMajorOpcode),
		    &shmFirstEvent, &shmFirstError);
  if (newWidget->jpeg.shmExtensionExists == False)		    /* Was the shared memory extension found? */
    newWidget->jpeg.useSharedMemory = False;			    /* No, set the appropriate variable for this widget */
  else								    /* Yes, the shared memory extension was found */
  {
    shmMajorOpcode = newWidget->jpeg.shmMajorOpcode;		    /* Initialize variables specifically set for shared memory */
    xShmAttachErrorOccurred = False;
    okayToUseSharedMemory =					    /* Determine whether it's okay to use shared memory for... */
      XPlxShmQueryVersion(display, &majorVersion, &minorVersion,    /* ...storing and displaying JPEG images. */
			  &sharedPixmaps);
    
  }
  if ((okayToUseSharedMemory == True) &&			    /* Is it okay to use shared memory, and did the...*/
      (newWidget->jpeg.useSharedMemory == True))		    /* ...application programmer permit it? */
  {								    /* Yes, create the necessary resources */
    sharedMemInfo = XPlxGetShmInfo();				    /* Create a shared memory info segment for this widget's... */
								    /* ...JPEG data. */
    if (sharedMemInfo != (XShmSegmentInfo*)NULL)		    /* Was a shared memory segment successfully allocated? */
    {								    /* Yes, try attaching it with the X server */
      newWidget->jpeg.jpegData =				    /* Create a shared memory image for storing JPEG images */
	XPlxShmCreateCImage(display,
			    sharedMemInfo->shmaddr,		    /* This is where JPEG data will be stored */
			    sharedMemInfo,
			    0,					    /* There is no image yet, so size is zero */
			    newWidget->jpeg.compressSourceRect->width,
			    newWidget->jpeg.compressSourceRect->height);
      oldErrorHandler = XSetErrorHandler(JPEGWidgetShmErrorHandler); /* Intercept XShmAttach errors */
      XSync(display, False);					    /* Install the error handler now, before doing anymore work */
      status  = XPlxShmAttach(display, sharedMemInfo);		    /* Put the shared memory buffer into use */
      XSync(display, False);					    /* Make sure this request gets processed before continuing */
      XSetErrorHandler(oldErrorHandler);			    /* Reinstate the previously-installed error handler */
      XSync(display, False);					    /* Remove the error handler now, before doing anymore work */
      if ((status == 0) || (xShmAttachErrorOccurred == True))	    /* Did the shared memory scheme fail for some reason? */
      {								    /* Yes, back off and use non shared memory for JPEG images */
	XPlxFreeShmInfo(newWidget->jpeg.jpegData->info);
	newWidget->jpeg.useSharedMemory = False;
	newWidget->jpeg.jpegData = NULL;
      }
    } /* end if shared memory segment successfully attached */
    else							    /* No, a shared memory segment was not successfully... */
    {								    /* ...allocated; explicitly disable shared memory usage. */
      newWidget->jpeg.useSharedMemory = False;
    }
  } /* end if okay to use shared memory */
  else								    /* No, it's not okay to use shared memory for whatever... */
  {								    /* ...reason; explicitly turn it off. */
    newWidget->jpeg.useSharedMemory = False;
  }
  PrintWidgetDiagnostics(newWidget, "Initialize");
  newWidget->jpeg.dataToDisplay = False;			    /* At this point there is no valid JPEG image loaded */
  newWidget->jpeg.lastJFIFInputFileName = (char*)NULL;		    /* Initialize instance variables indicating that no JFIF... */
  newWidget->jpeg.lastJFIFOutputFileName = (char*)NULL;		    /* ...input or output files were yet specified. */
  if (newWidget->jpeg.jfifInputFile != (String)NULL)		    /* Was a JFIF file specified from which to load an image */
  {								    /* Yes, attempt to load the file into the jpegData resource */
    jfifImage = jpeg_load_jfif(newWidget,
			       newWidget->jpeg.jfifInputFile);
    if (jfifImage != (XPlxCImage*)NULL)				    /* Did the image load succeed? */
    {
      newWidget->jpeg.jpegData = jfifImage;			    /* Yes, point to the new JPEG image data */
      newWidget->jpeg.getNewImage = False;			    /* Yes, do not grab new data; the jfifInputFile has the data */
      newWidget->jpeg.dataToDisplay = True;
      newWidget->jpeg.defaultDisplaySourceRect.width =		    /* Update default display variables to reflect the size.. */
	newWidget->jpeg.jpegData->width;			    /* ...of the JPEG image just passed in to the widget. */
      newWidget->jpeg.defaultDisplaySourceRect.height =
	newWidget->jpeg.jpegData->height;
    }
    else							    /* No, the JFIF image load failed; reset the resource... */
    {								    /* ...containing the name of the JFIF file. */
      newWidget->jpeg.jfifInputFile = (String)NULL;
    }
  }
  if (newWidget->jpeg.jfifOutputFile != (String)NULL)		    /* Was a JFIF file specified to which to save an image */
  {								    /* Yes, attempt to save any current JPEG data into the... */
    if (newWidget->jpeg.jpegData != (XPlxCImage*)NULL)		    /* ...file specified by the "jfifOutputFile" resource */
      jpeg_save_jfif(newWidget,
		     newWidget->jpeg.jpegData,
		     newWidget->jpeg.jfifOutputFile,
		     newWidget->jpeg.qTable);
  }
  newWidget->jpeg.zoomFactor = XPlxGetZoomFactor(display,	    /* Get the default zoom ratio for this connection */
						 rootWindow);
  hardwareParams =						    /* Determine the type of hardware being used and get... */
    XPlxQueryConfig(display, rootWindow,			    /* ...other relevant information about it. */
		    XDefaultGC(display, XDefaultScreen(display)));
  if ((plx_IO*)NULL == hardwareParams)				    /* Was any Plx hardware found on this X connection? */
  {								    /* No, disable JPEG capabilities and report the problem... */
    newWidget->jpeg.jpegAvailable = False;			    /* ...if diagnostics call for it. */
    if (newWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
    {
      printf("JPEG Widget: Initialize(): Warning: no Parallax hardware was found.\n");
    }
    return;
  }
  hardwareInfo.allBits = hardwareParams->hardware;		    /* Point to struct describing this board's capabilities */
  if (0 == hardwareParams->numJPEGCompressors)			    /* Does this board have any JPEG compressors? */
  {
    newWidget->jpeg.jpegAvailable = False;			    /* No, tell the widget that it can't compress/decompress */
    if (newWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
    {
      printf("JPEG Widget: Initialize(): Warning: no JPEG hardware is available.\n");
    }
  }
  return;
}								    /* end function Initialize */


static void Realize(Widget			w,
		    XtValueMask*		valueMask,
		    XSetWindowAttributes*	attributes)
{
  Display*		display = XtDisplay(w);
  JPEGWidget		newWidget = (JPEGWidget) w;
  
  CreateJPEGDisplayWindow(newWidget, valueMask, attributes);	    /* Create a default window for this widget */
  if (newWidget->jpeg.sourceWin != (Window)0)			    /* Was a valid source window specified? */
  {								    /* Yes, create a GC needed for storing a Q table for... */
    newWidget->jpeg.compressGC = XCreateGC(display,		    /* ...compressing the image in the source window */
					   newWidget->jpeg.sourceWin,
					   0, 0);
    XPlxPutTable(display,					    /* Install the GC for the compression (source) window */
		 newWidget->jpeg.sourceWin,
		 newWidget->jpeg.compressGC,
		 (char*)newWidget->jpeg.qTable,
		 newWidget->jpeg.qTableSize,
		 0);
  }
  RegisterJPEGWidgetEventHandler(newWidget);			    /* Look for window move events if on an overlay product */
  return;
}								    /* end function Realize */


static void Redisplay(Widget	w,
		      XEvent*	event,
		      Region	region)
{
  JPEGWidget		theWidget = (JPEGWidget) w;
  static int		resetTag = True;
  int			squeezeDest = 1;
  Bool			result = False;
  XEvent		unusedEvent;
  
  if (!XtIsRealized(w))						    /* Don't bother trying to draw if the widget is not... */
    return;							    /* ...or has not been realized */
  if (False == theWidget->jpeg.jpegAvailable)			    /* Don't bother compressing or decompressing if there's... */
    return;							    /* ...no JPEG hardware available for this connection. */
  if (theWidget->jpeg.getNewImage == True)			    /* Is there any JPEG data to compress?  If not, make some... */
  {								    /* ...by compressing whatever image is in the source window */
    if (theWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
    {
      printf("JPEG Widget: Redisplay(): getNewImage is True, sourceWin is %d.\n",
	     theWidget->jpeg.sourceWin);
    }
    theWidget->jpeg.getNewImage = False;			    /* Reset this resource, since a new image will be obtained */
    if (theWidget->jpeg.sourceWin == (Window)0)			    /* Is the source window a valid window ID? */
      return;							    /* No, don't bother trying to draw the image */
    theWidget->jpeg.dataToDisplay = True;			    /* Tell this method that there is valid JPEG data to display */
    
    AdjustCompressRectangles(theWidget);			    /* Assure that legal compression parameters are used */
    
    if (theWidget->jpeg.useSharedMemory == True)		    /* Is this widget using shared memory? */
    {
      theWidget->jpeg.jpegData->width =
	theWidget->jpeg.compressDestRect->width;
      theWidget->jpeg.jpegData->height =
	theWidget->jpeg.compressDestRect->height;
      if (theWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
      {
	printf("JPEG Widget: Redisplay(): shmCompress: %d, %d, %d, %d, %d, %d, %d, %d.\n",
	       XtDisplay(theWidget),				    /* Yes, compress the contents of the source window and... */
	       theWidget->jpeg.sourceWin,			    /* ...store the compressed data in the appropriate field... */
	       theWidget->jpeg.compressGC,			    /* ... for this widget. */
	       theWidget->jpeg.jpegData,
	       theWidget->jpeg.compressSourceRect->x,
	       theWidget->jpeg.compressSourceRect->y,
	       (unsigned int)theWidget->jpeg.compressSourceRect->width,
	       (unsigned int)theWidget->jpeg.compressSourceRect->height);
      }
      XPlxShmGetCImage(XtDisplay(theWidget),			    /* Yes, compress the contents of the source window and... */
		       theWidget->jpeg.sourceWin,		    /* ...store the compressed data in the appropriate field... */
		       theWidget->jpeg.compressGC,		    /* ... for this widget. */
		       theWidget->jpeg.jpegData,
		       theWidget->jpeg.compressSourceRect->x,
		       theWidget->jpeg.compressSourceRect->y,
		       (unsigned int)theWidget->jpeg.compressSourceRect->width,
		       (unsigned int)theWidget->jpeg.compressSourceRect->height);
    }
    else							    /* No, the widget isn't using shared memory */
    {
      if (theWidget->jpeg.jpegData != (XPlxCImage*)NULL)	    /* Does the widget have any existing JPEG data? */
	XPlxDestroyCImage(theWidget->jpeg.jpegData);		    /* Yes, free old image & data since it's no longer needed */
      if (theWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
      {
	printf("JPEG Widget: Redisplay(): Non-shared Compress: %d, %d, %d, %d, %d, %d, %d, %d, %d.\n",
	       XtDisplay(theWidget),				    /* Yes, compress the contents of the source window and... */
	       theWidget->jpeg.sourceWin,			    /* ...store the compressed data in the appropriate field... */
	       theWidget->jpeg.compressGC,			    /* ... for this widget. */
	       theWidget->jpeg.compressSourceRect->x,
	       theWidget->jpeg.compressSourceRect->y,
	       (unsigned int)theWidget->jpeg.compressSourceRect->width,
	       (unsigned int)theWidget->jpeg.compressSourceRect->height,
	       theWidget->jpeg.compressDestRect->width,
	       theWidget->jpeg.compressDestRect->height);
      }
      theWidget->jpeg.jpegData =				    /* Compress the contents of the source window and store... */
	XPlxGetCImage(XtDisplay(theWidget),			    /* ...the compressed data in the appropriate field for... */
		      theWidget->jpeg.sourceWin,		    /* ...this widget. */
		      theWidget->jpeg.compressGC,
		      theWidget->jpeg.compressSourceRect->x,
		      theWidget->jpeg.compressSourceRect->y,
		      theWidget->jpeg.compressSourceRect->width,
		      theWidget->jpeg.compressSourceRect->height,
		      theWidget->jpeg.compressDestRect->width,
		      theWidget->jpeg.compressDestRect->height);
    } /* end else (widget is not using shared memory) */
    
    AdjustDisplayRectangles(theWidget);				    /* Assure that legal de-compression parameters are used */
    
  } /* end (if theWidget->jpeg.getNewImage == True) */
  squeezeDest = (theWidget->jpeg.squeezeDest == True) ? 1 : 0;
  if ((theWidget->jpeg.destWin == (Window)0) ||			    /* Did the programmer explicitly disable rendering of... */
      (theWidget->jpeg.dataToDisplay == False))			    /* ...JPEG images, or is there no data to display? */
    return;							    /* No reason to try to display images, return immediately */
  CreateJPEGWidgetQTablesIfNecessary(theWidget, False);		    /* Only create Q tables if new decompress GC was created */
  if (event != (XEvent*)NULL)					    /* Did the toolkit call this method?  (As opposed to the... */
  {								    /* ...widget calling this itself) */
    XPlxVideoTag(XtDisplay(theWidget),				    /* Yes, tag the window for graphics again, to avoid... */
		 theWidget->jpeg.destWin,			    /* ...flicker (this is a hack that should be removed when... */
		 theWidget->jpeg.decompressGC,			    /* ...the OpenWindows server is fixed; this is not... */
		 PLX_VIDEO);					    /* ...necessary for the X11R5 server). */
    resetTag = True;						    /* Tell the widget to set the tag to OVR next time it... */
  } /* ...draws, but only if called by the widget itself. */
  if ((event == (XEvent*)NULL) && (resetTag == True))		    /* Did the widget call this method itself?  If so, check... */
  {								    /* ...if the tag needs to be set to overlay; this is true... */
    resetTag = False;						    /* ...if the last draw used the GRAPHICS tag. */
    XPlxVideoTag(XtDisplay(theWidget),				    /* Yes, tag the window for graphics again, to avoid... */
		 theWidget->jpeg.destWin,			    /* ...flicker (this is a hack that should be removed when... */
		 theWidget->jpeg.decompressGC,			    /* ...the OpenWindows server is fixed; this is not... */
		 PLX_VIDEO_OVR);				    /* ...necessary for the X11R5 server). */
  }
  if (theWidget->jpeg.useSharedMemory == True)			    /* Is this widget using shared memory? */
  {
    if (theWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
    {
      printf("JPEG Widget: Redisplay(): ShmPutCImage: %d %d %d %d %d %d %d %d %d %d %d %d %d %d.\n",
	     XtDisplay(theWidget),				    /* Yes, decompress and display the JPEG image data into... */
	     theWidget->jpeg.destWin,				    /* ...the destination window. */
	     theWidget->jpeg.decompressGC,			    /* Remember to set the send_event flag to "True", to... */
	     theWidget->jpeg.jpegData,				    /* ...ensure that the X server sends an event back to... */
	     theWidget->jpeg.displaySourceRect->x,		    /* ...this process when the shared memory transfer has... */
	     theWidget->jpeg.displaySourceRect->y,		    /* ...completed. */
	     theWidget->jpeg.displaySourceRect->width,
	     theWidget->jpeg.displaySourceRect->height,
	     theWidget->jpeg.displayDestRect->x,
	     theWidget->jpeg.displayDestRect->y,
	     theWidget->jpeg.displayDestRect->width,
	     theWidget->jpeg.displayDestRect->height,
	     squeezeDest, True);
    }
    XPlxShmPutCImage(XtDisplay(theWidget),			    /* Yes, decompress and display the JPEG image data into... */
		     theWidget->jpeg.destWin,			    /* ...the destination window. */
		     theWidget->jpeg.decompressGC,		    /* Remember to set the send_event flag to "True", to... */
		     theWidget->jpeg.jpegData,			    /* ...ensure that the X server sends an event back to... */
		     theWidget->jpeg.displaySourceRect->x,	    /* ...this process when the shared memory transfer has... */
		     theWidget->jpeg.displaySourceRect->y,	    /* ...completed. */
		     theWidget->jpeg.displaySourceRect->width,
		     theWidget->jpeg.displaySourceRect->height,
		     theWidget->jpeg.displayDestRect->x,
		     theWidget->jpeg.displayDestRect->y,
		     theWidget->jpeg.displayDestRect->width,
		     theWidget->jpeg.displayDestRect->height,
		     squeezeDest, True);
    while (result == False)					    /* Wait until the shared memory transfer has completed */
    {
      result = XCheckTypedWindowEvent(XtDisplay(theWidget),	    /* Check if the last shared memory JPEG put image has... */
				      theWidget->jpeg.destWin,	    /* ...completed. */
				      XPlxShmGetEventBase(XtDisplay(theWidget)) + ShmCompletion,
				      &unusedEvent);
    }
  }
  else
  {
    if (theWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
    {
      printf("JPEG Widget: Redisplay(): non-shared decompress: %d %d %d %d %d %d %d %d %d %d %d %d %d.\n",
	     XtDisplay(theWidget),
	     theWidget->jpeg.destWin,
	     theWidget->jpeg.decompressGC,
	     theWidget->jpeg.jpegData,
	     theWidget->jpeg.displaySourceRect->x,
	     theWidget->jpeg.displaySourceRect->y,
	     theWidget->jpeg.displaySourceRect->width,
	     theWidget->jpeg.displaySourceRect->height,
	     theWidget->jpeg.displayDestRect->x,
	     theWidget->jpeg.displayDestRect->y,
	     theWidget->jpeg.displayDestRect->width,
	     theWidget->jpeg.displayDestRect->height,
	     squeezeDest);
    }
    XPlxPutCImage(XtDisplay(theWidget),				    /* No, the widget isn't using shared memory. */
		  theWidget->jpeg.destWin,
		  theWidget->jpeg.decompressGC,
		  theWidget->jpeg.jpegData,
		  theWidget->jpeg.displaySourceRect->x,
		  theWidget->jpeg.displaySourceRect->y,
		  theWidget->jpeg.displaySourceRect->width,
		  theWidget->jpeg.displaySourceRect->height,
		  theWidget->jpeg.displayDestRect->x,
		  theWidget->jpeg.displayDestRect->y,
		  theWidget->jpeg.displayDestRect->width,
		  theWidget->jpeg.displayDestRect->height,
		  squeezeDest);
  }
  if (theWidget->jpeg.drawMethod != (XtExposeProc)NULL)		    /* Did the app pgmr specify a custom drawing method? */
  {								    /* Yes, call the pgmr's drawing method to draw over JPEG */
    (*theWidget->jpeg.drawMethod)(w, event, region);		    /* Perform the method call */
  }
  return;
} /* end function Redisplay */


static void Destroy(Widget w)
{
  JPEGWidget	theWidget = (JPEGWidget) w;
  
  if (theWidget->jpeg.compressGC != (GC)0)			    /* Release any resources that might have been allocated... */
    XFreeGC(XtDisplay(w), theWidget->jpeg.compressGC);		    /* ...to this widget. */
  if (theWidget->jpeg.decompressGC != (GC)0)
    XFreeGC(XtDisplay(w), theWidget->jpeg.decompressGC);
  if (theWidget->jpeg.jpegData != (XPlxCImage*)NULL)		    /* Is there a JPEG image to destroy? */
  {								    /* Yes, determine what kind and take appropriate action */
    if (theWidget->jpeg.useSharedMemory == True)		    /* Is this widget using shared memory? */
    {
      XPlxShmDetach(XtDisplay(theWidget),			    /* Yes, tell the X server that this shared memory segment... */
		    theWidget->jpeg.jpegData->info);		    /* ...is no longer needed. */
      XPlxFreeShmInfo(theWidget->jpeg.jpegData->info);		    /* Free the shared memory segment no longer needed */
    }
    else
      XPlxDestroyCImage(theWidget->jpeg.jpegData);
  }
  if (theWidget->jpeg.eventHandlerInstalled == True)		    /* If this widget is running on an overlay product then... */
  {								    /* ...it has an event handler looking for window move... */
    XtRemoveEventHandler(theWidget->jpeg.eventHandlerWidget,	    /* ...events.  Remove it before destroying the widget. */
			 StructureNotifyMask,
			 False,
			 JPEGWidgetEventHandler,
			 (XtPointer)theWidget);
  }
  return;
}								    /* end function Destroy */



static void Resize(Widget w)
{
  JPEGWidget	theWidget = (JPEGWidget) w;
  
  if (theWidget->jpeg.displayDestRect ==			    /* Is the widget using the default display rectangle? */
      &(theWidget->jpeg.defaultDisplayDestRect))
  {								    /* Yes, make the display track the size of the JPEG... */
    theWidget->jpeg.displayDestRect->width =			    /* ...display window. */
      (unsigned short) theWidget->core.width;
    theWidget->jpeg.displayDestRect->height =
      (unsigned short) theWidget->core.height;
    theWidget->jpeg.defaultCompressSourceRect.width =
      (unsigned short) theWidget->core.width;
    theWidget->jpeg.defaultCompressSourceRect.height =
      (unsigned short) theWidget->core.height;
    theWidget->jpeg.defaultCompressDestRect.width = 
      (unsigned short) theWidget->core.width;
    theWidget->jpeg.defaultCompressDestRect.height =
      (unsigned short) theWidget->core.height;
  }
  return;
}								    /* end function Resize */


static Boolean SetValues(Widget		current,
			 Widget		request,
			 Widget		new,
			 ArgList	args,
			 Cardinal*	numArgs)
{
  JPEGWidget	currentWidget = (JPEGWidget) current;
  JPEGWidget	newWidget = (JPEGWidget) new;
  Boolean	redisplayNeeded = False;
  Boolean	createNewGCs = False;
  Boolean	jfifResult = False;
  int		squeezeDest = 1;
  
  if (newWidget->jpeg.destWin !=				    /* Did the application programmer request to draw images... */
      currentWidget->jpeg.destWin)				    /* ...in a different window? */
  {								    /* Yes, do some error checking */
    if ((newWidget->jpeg.destWin ==				    /* Check for the unlikely case that this widget is... */
	 currentWidget->core.window) &&				    /* ...requested to draw into its window when previously... */
	(currentWidget->core.depth == 0))			    /* ...it was created as an InputOnly window. */
    {
      XtWarning("JPEG Widget: the display window \"destWin\" was \
      created as an InputOnly window and cannot be used for displaying\n\
      JPEG images.");
      newWidget->jpeg.destWin = currentWidget->jpeg.destWin;
    }
    else							    /* The new destination window should be legal; install... */
    {								    /* ...the appropriate resources. */
      createNewGCs = True;					    /* This variable is examined later in the function; if... */
      redisplayNeeded = True;					    /* ...set, appropriate resources will be created. */
    }
  }
  if (newWidget->jpeg.jpegAvailable !=				    /* Did the application programmer illegally try to change... */
      currentWidget->jpeg.jpegAvailable)			    /* ...the widget's notion of whether JPEG is available? */
  {								    /* Yes, report the error and restore the widget to health */
    XtWarning("JPEG Widget: jpegAvailable is a read-only resource. Restoring original value.");
    newWidget->jpeg.jpegAvailable =
      currentWidget->jpeg.jpegAvailable;
  }
  if (newWidget->jpeg.useSharedMemory !=			    /* Did the application programmer illegally try to make... */
      currentWidget->jpeg.useSharedMemory)			    /* ...the widget use shared memory or turn it off? */
  {								    /* Yes, report the error and restore the widget to health */
    XtWarning("JPEG Widget: useSharedMemory is a read-only resource. Restoring original value.");
    newWidget->jpeg.useSharedMemory =
      currentWidget->jpeg.useSharedMemory;
  }
  if (newWidget->jpeg.sourceWin != currentWidget->jpeg.sourceWin)   /* Is the widget to compress data from a new window? */
  {								    /* Yes, continue */
    if (newWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
    {
      printf("JPEG Widget: SetValues(): sourceWin changed to %d.\n",
	     newWidget->jpeg.sourceWin);
    }
    createNewGCs = True;					    /* This variable is examined later in the function; if... */
    newWidget->jpeg.getNewImage = True;				    /* ...set, appropriate resources will be created. */
  }								    /* End (if sourceWin's differ) */
  if (createNewGCs == True)					    /* Was the sourceWin or destWin changed? */
  {								    /* Yes, create new resources for (de)compressing images */
    if (newWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
    {
      printf("JPEG Widget: SetValues(): createNewGC's is True.\n");
    }
    if (newWidget->jpeg.compressGC != (GC)NULL)			    /* Does a compress GC exist for the source window? */
    {
      XFreeGC(XtDisplay(newWidget), newWidget->jpeg.compressGC);    /* Yes, destroy it and create a new one */
      newWidget->jpeg.compressGC = (GC)NULL;
    }
    if (newWidget->jpeg.sourceWin != 0)				    /* Was a valid window requested?  A value of "0" means to... */
    {								    /* ...disable compression. */
      newWidget->jpeg.compressGC = XCreateGC(XtDisplay(newWidget),  /* Create a new GC for the new window to be used as an... */
					     newWidget->jpeg.sourceWin, /* ...image source. */
					     0, 0);
      if (newWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
      {
	printf("JPEG Widget: SetValues(): About to XPlxPutTable the compressGC %d.\n",
	       newWidget->jpeg.compressGC);
      }
      XPlxPutTable(XtDisplay(newWidget),			    /* Make sure the compression table is installed in the... */
		   newWidget->jpeg.sourceWin,			    /* ...new window to be used for compressing image data. */
		   newWidget->jpeg.compressGC,
		   (char*)newWidget->jpeg.qTable,
		   newWidget->jpeg.qTableSize,
		   0);
      if (newWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
      {
	printf("JPEG Widget: SetValues(): Successfully PutTable'd the compressGC %d.\n",
	       newWidget->jpeg.compressGC);
      }
    } /* end (if sourceWin differed) */
    if ((newWidget->jpeg.destWin != currentWidget->jpeg.destWin) && /* Did the programmer request to draw in a different... */
	(newWidget->jpeg.destWin != (Window)0))			    /* ...window, and is the window valid? */
    {
      if (newWidget->jpeg.decompressGC != (GC)NULL)		    /* Does a decompress GC exist for the dest window? */
      {
	XFreeGC(XtDisplay(newWidget),				    /* Yes, destroy it and create a new one */
		newWidget->jpeg.decompressGC);
	newWidget->jpeg.decompressGC = (GC)NULL;
      }
      newWidget->jpeg.decompressGC =				    /* Create a new GC for the new window to be used as an... */
	XCreateGC(XtDisplay(newWidget),				    /* ...image destination. */
		  newWidget->jpeg.destWin,
		  0, 0);
      XPlxPutTable(XtDisplay(newWidget),			    /* Yes, install the Q table in the new decompression... */
		   newWidget->jpeg.destWin,			    /* ...(destination) window's GC */
		   newWidget->jpeg.decompressGC,
		   (char*)newWidget->jpeg.qTable,
		   newWidget->jpeg.qTableSize,
		   1);
    } /* end (if destWin differed) */
  } /* end (if createNewGCs == True) */
  if (newWidget->jpeg.compressSourceRect !=			    /* Was a new value requested?  And, is it valid? */
      currentWidget->jpeg.compressSourceRect)
  {
    if (newWidget->jpeg.compressSourceRect != (XRectangle*)NULL)
      redisplayNeeded = True;
    else
    {
      XtWarning("JPEG Widget: compressSourceRect must point to an XRectangle.\n\
		Using currently-stored value.");
      newWidget->jpeg.compressSourceRect = currentWidget->jpeg.compressSourceRect;
    }
  }
  if (newWidget->jpeg.compressDestRect !=			    /* Was a new value requested?  And, is it valid? */
      currentWidget->jpeg.compressDestRect)
  {
    if (newWidget->jpeg.compressDestRect != (XRectangle*)NULL)
      redisplayNeeded = True;
    else
    {
      XtWarning("JPEG Widget: compressDestRect must point to an XRectangle.\n\
		Using currently-stored value.");
      newWidget->jpeg.compressDestRect = currentWidget->jpeg.compressDestRect;
    }
  }
  if (newWidget->jpeg.displaySourceRect !=			    /* Was a new value requested?  And, is it valid? */
      currentWidget->jpeg.displaySourceRect)
  {
    if (newWidget->jpeg.displaySourceRect != (XRectangle*)NULL)
      redisplayNeeded = True;
    else
    {
      XtWarning("JPEG Widget: displaySourceRect must point to an XRectangle.\n\
		Using currently-stored value.");
      newWidget->jpeg.displaySourceRect = currentWidget->jpeg.displaySourceRect;
    }
  }
  if (newWidget->jpeg.displayDestRect !=			    /* Was a new value requested?  And, is it valid? */
      currentWidget->jpeg.displayDestRect)
  {
    if (newWidget->jpeg.displayDestRect != (XRectangle*)NULL)
      redisplayNeeded = True;
    else
    {
      XtWarning("JPEG Widget: displayDestRect must point to an XRectangle.\n\
		Using currently-stored value.");
      newWidget->jpeg.displayDestRect = currentWidget->jpeg.displayDestRect;
    }
  }
  if (newWidget->jpeg.jpegData != currentWidget->jpeg.jpegData)	    /* Is the widget being requested to point to new JPEG data? */
  {								    /* Yes; is the new JPEG data non-NULL? */
    if (newWidget->jpeg.jpegData != (XPlxCImage*)NULL)		    /* Yes, indicate that this widget should display the new... */
    {								    /* ...image in the destination window. */
      if (newWidget->jpeg.useSharedMemory == True)		    /* Is this widget using shared memory?  If so, then it... */
      {								    /* ...is illegal to point to a different jpegData structure. */
	XtWarning("JPEG Widget: This widget is using shared memory; it is illegal\n\
		to point to another jpegData structure.");
	newWidget->jpeg.jpegData = currentWidget->jpeg.jpegData;
      }
      redisplayNeeded = True;					    /* Regardless, this widget needs to be redrawn */
      newWidget->jpeg.putNewImage = True;			    /* Tell the widget to display the new image now */
      newWidget->jpeg.defaultDisplaySourceRect.width =		    /* Update default display variables to reflect the size.. */
	newWidget->jpeg.jpegData->width;			    /* ...of the JPEG image just passed in to the widget. */
      newWidget->jpeg.defaultDisplaySourceRect.height =
	newWidget->jpeg.jpegData->height;
      newWidget->jpeg.defaultDisplayDestRect.width =
	newWidget->jpeg.jpegData->width;
      newWidget->jpeg.defaultDisplayDestRect.height =
	newWidget->jpeg.jpegData->height;
    } /* end (if jpegData != (XPlxCImage*)NULL) */
    else							    /* No, an error occurred (this widget should not be made... */
    {								    /* ...to point to a NULL image); report and correct for... */
      XtWarning("JPEG Widget: Tried to point to a NULL JPEG image.\n\
	Using currently-stored image data.");			    /* ...the error. */
      newWidget->jpeg.jpegData = currentWidget->jpeg.jpegData;
    }
  }
  jfifResult = JPEGWidgetCheckJFIFInputFile(newWidget,		    /* Determine if a new JFIF file needs to be read from disk */
					    currentWidget);
  if (True == jfifResult)					    /* Was a JFIF file read in from disk? */
    redisplayNeeded = True;					    /* Yes, indicate that the widget needs refreshing */
  JPEGWidgetCheckJFIFOutputFile(newWidget, currentWidget);	    /* Save a JFIF file if appropriate */
  if ((newWidget->jpeg.jfifOutputFile !=			    /* Did the programmer request the JPEG data to be saved... */
      currentWidget->jpeg.jfifOutputFile) &&			    /* ...as a JFIF file, *and*, does the widget currently... */
      (newWidget->jpeg.jpegData != (XPlxCImage*)NULL))		    /* ...have JPEG data to save? */
  {
    if (newWidget->jpeg.jfifOutputFile == (String)NULL)		    /* Yes, is the filename valid? */
    {
      XtWarning("JPEG Widget: JFIF file name was NULL.  No save will be done.");
      newWidget->jpeg.jfifOutputFile = currentWidget->jpeg.jfifOutputFile;
    }
    else							    /* Yes, the filename is valid, try to save current JPEG... */
    {								    /* ...data as a JFIF file. */
      jpeg_save_jfif(newWidget,
		     newWidget->jpeg.jpegData,
		     newWidget->jpeg.jfifOutputFile,
		     newWidget->jpeg.qTable);
    }
  }
  if (newWidget->jpeg.squeezeDest !=				    /* Was a change in the crop-or-squeeze flag made? */
      currentWidget->jpeg.squeezeDest)
    redisplayNeeded = True;					    /* Yes, indicate that this widget needs redrawing */
  if (newWidget->jpeg.qFactor != currentWidget->jpeg.qFactor)	    /* Was the compression quality changed? */
  {
    if ((newWidget->jpeg.qFactor < 25) ||			    /* Yes, assure that the requested quality falls within... */
	(newWidget->jpeg.qFactor > 1000))			    /* ...accepted ranges? */
    {
      XtWarning("JPEG Widget: qFactor must be between 25 and 1000.\n\
	Restoring original values.");
      newWidget->jpeg.qFactor = currentWidget->jpeg.qFactor;
    }
    else							    /* Requested Q factor fell within accepted range */
    {
      if (newWidget->jpeg.diagnostics == True)			    /* Was the appropriate resource set to print diagnostics... */
      {
	printf("JPEG Widget: SetValues(): qFactor changed to %d.\n",
	       newWidget->jpeg.qFactor);
      }
      newWidget->jpeg.qTableSize =				    /* Initialize the Q table for this JPEG widget */
	MakeQTables(newWidget->jpeg.qFactor,
		    &(newWidget->jpeg.qTable));
      if (newWidget->jpeg.sourceWin != (Window)0)		    /* Was a valid source window specified? */
      {
	XPlxPutTable(XtDisplay(newWidget),			    /* Yes, install the GC for the compression (source) window */
		     newWidget->jpeg.sourceWin,
		     newWidget->jpeg.compressGC,
		     (char*)newWidget->jpeg.qTable,
		     newWidget->jpeg.qTableSize,
		     0);
      }
      if (newWidget->jpeg.destWin != (Window)0)			    /* Was a valid destination window specified? */
      {
	CreateJPEGWidgetQTablesIfNecessary(newWidget, True);	    /* Yes, install the Q table in the dest window's GC */
      }
    }
  }
  if (newWidget->jpeg.getNewImage == True)			    /* Did the programmer request a fresh compressed image? */
  {
    Redisplay((Widget)newWidget, (XEvent*)NULL, (Region)NULL);	    /* Force a redisplay of widget, bypassing normal mechanisms */
    newWidget->jpeg.getNewImage = False;
  }
  if ((newWidget->jpeg.putNewImage !=				    /* Did the programmer place new JPEG data into this widget? */
       currentWidget->jpeg.putNewImage) &&
      (newWidget->jpeg.putNewImage == True))
  {
    if ((False == newWidget->jpeg.jpegAvailable) ||		    /* Is the board for this server missing JPEG ability, or... */
	((XPlxCImage*)NULL == newWidget->jpeg.jpegData))	    /* ...is there no JPEG data after all? */
      return(False);						    /* Yes, no need to redisplay anything */
    newWidget->jpeg.putNewImage = False;			    /* Reset the resource to prevent an endless loop of redraws */
    newWidget->jpeg.defaultDisplaySourceRect.width =		    /* Update default display variables to reflect the size.. */
      newWidget->jpeg.jpegData->width;				    /* ...of the JPEG image just created. */
    newWidget->jpeg.defaultDisplaySourceRect.height =
      newWidget->jpeg.jpegData->height;
    newWidget->jpeg.defaultDisplayDestRect.width =
      newWidget->jpeg.jpegData->width;
    newWidget->jpeg.defaultDisplayDestRect.height =
      newWidget->jpeg.jpegData->height;
    newWidget->jpeg.dataToDisplay = True;			    /* Tell the widget that there is valid data to display */
    Redisplay((Widget)newWidget, (XEvent*)NULL, (Region)NULL);	    /* Force a redisplay of widget, bypassing normal mechanisms */
    if (newWidget->jpeg.squeezeDest == True)
      squeezeDest = 1;
    else
      squeezeDest = 0;
  }
  return(redisplayNeeded);
}								    /* end function SetValues */



static void CreateJPEGDisplayWindow(JPEGWidget			widget,
				    XtValueMask*		valueMask,
				    XSetWindowAttributes*	attributes)
{
  int			screenNumber;
  Display*		display = XtDisplay(widget);
  XVisualInfo*		theVisual;
  
  widget->core.width = widget->jpeg.displayDestRect->width;
  widget->core.height = widget->jpeg.displayDestRect->height;
  screenNumber = XScreenNumberOfScreen(XtScreen(widget));
  theVisual = XPlxGetVideoVisual(screenNumber, display);	    /* Try to get the best TrueColor visual for this widget */
  if (widget->jpeg.destWin == (Window)0)			    /* Was a destination window specified for display of JPEG... */
  {								    /* ...images? */
    widget->core.colormap = XCreateColormap(display,		    /* Create a default colormap to install for this window. */
					    XtWindow(XtParent(widget)), /* See O'Reilly Xlib Reference Manual entry for... */
					    theVisual->visual,	    /* ..."XCreateWindow()" to learn why a default colormap... */
					    AllocNone);		    /* ...needs to be created at all. */
    attributes->colormap = widget->core.colormap;		    /* Set necessary attributes for a 24-bit video window.  ... */
    attributes->background_pixmap = None;			    /* ...Absent from this set of attributes are the... */
    attributes->background_pixel = BlackPixel(display, screenNumber); /* ..."border_pixmap" and "border_pixel" attributes; I... */
    attributes->bit_gravity = CenterGravity;			    /* ...don't know why they can't seem to be set here.  ... */
    attributes->win_gravity = NorthWestGravity;			    /* ...Using the values recommended by the O'Reilly book... */
    attributes->backing_store = NotUseful;			    /* ...(Volume 1) of "None" and "BlackPixel(etc.)"... */
    attributes->backing_pixel = BlackPixel(display, screenNumber);  /* ...causes a "BadMatch" protocol error at run-time.  ... */
    attributes->save_under = False;				    /* ...However, these initial values seem to be sufficient. */
    *valueMask |= (CWColormap | CWBackPixmap | CWBackPixel |
		   CWBitGravity | CWWinGravity |
		   CWBackingStore | CWBackingPixel |
		   CWSaveUnder);
    widget->core.depth = theVisual->depth;			    /* Reset the widget's depth to the TrueColor depth */
    XtCreateWindow((Widget)widget, InputOutput, theVisual->visual,  /* Yes, create window into which video will be displayed */
		   *valueMask, attributes);
    widget->jpeg.destWin = widget->core.window;			    /* Point to the newly-created display window */
  }
  else								    /* No, create a dummy window that won't be used for... */
  {								    /* ...displaying JPEG images */
    widget->core.border_width = 0;				    /* InputOnly windows must have a border width of zero. */
    widget->core.depth = 0;
    *valueMask = (XtValueMask)0;
    XtCreateWindow((Widget)widget, InputOnly, theVisual->visual,
		   *valueMask, attributes);
  }
  
  return;
}								    /* end function CreateJPEGDisplayWindow */


static void JPEGAction(Widget	w,
		       XEvent*	event,
		       String*	params,
		       Cardinal*	numParams)
{
  return;							    /* Right now this function does nothing */
}								    /* end function JPEGAction */



static void PrintWidgetDiagnostics(JPEGWidget	widget,
				   char*	calledFromWhere)
{
  if (widget->jpeg.diagnostics == False)			    /* Was the appropriate resource set to print diagnostics... */
    return;							    /* ...about this widget?  If not, return right now. */
  printf("JPEG Widget: Called from function %s.\n",
	 calledFromWhere);
  if (widget->jpeg.useSharedMemory == True)
    printf("+++++Using Shared Memory.\n");
  else
    printf("=====*Not* using shared memory.\n");
  return;
}								    /* end function PrintWidgetDiagnostics */


/*-------------------------- cut here --------------------------*/
/*
 *
 *	The rest of this file contains routines for loading
 *	and saving JFIF images
 *
 *	Adapted from code written by John Malleo-Roach
 *
 */


static void SystemError(int, char*);
static void ShowAlert(JPEGWidget, char*);

/* For undoing a zigzaged load/store of qtable */

static int zigzag[64] = {
	0, 1, 8,16, 9, 2, 3,10,
	17,24,32,25,18,11, 4, 5,
	12,19,26,33,40,48,41,34,
	27,20,13, 6, 7,14,21,28,
	35,42,49,56,57,50,43,36,
	29,22,15,23,30,37,44,51,
	58,59,52,45,38,31,39,46,
	53,60,61,54,47,55,62,63
};

static u_char jfif_header[] = {
/* SOI    2 bytes */
	0xff, 0xd8,
/* APP0  18 bytes */
	            0xff, 0xe0, 0x00, 0x10, 0x4a, 0x46,
	0x49, 0x46, 0x00, 0x01,	0x01, 0x00, 0x00, 0x01,
	0x00, 0x01,	0x00, 0x00,
/* SOF   19 bytes */
	                        0xff, 0xc0, 0x00, 0x11,
	0x08, 0x00, 0x00, 0x00, 0x00, 0x03,	0x01, 0x21,
	0x00, 0x02, 0x11, 0x01,	0x03, 0x11, 0x01,
/* DQT  134 bytes */
	                                          0xff,
	0xdb, 0x00, 0x84, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,	0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00,
/* DHT 420 bytes */
	                              0xff, 0xc4, 0x01,
	0xa2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01, 0x01,
	0x01, 0x01,	0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00,	0x00, 0x01, 0x02, 0x03, 0x04, 0x05,
	0x06, 0x07,	0x08, 0x09, 0x0a, 0x0b,	0x10, 0x00,
	0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05,
	0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7d, 0x01,
	0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12, 0x21,
	0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22,
	0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
	0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
	0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
	0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
	0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
	0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
	0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
	0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
	0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
	0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
	0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
	0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
	0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
	0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
	0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
	0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
	0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
	0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
	0xf9, 0xfa,	0x01, 0x00, 0x03, 0x01, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04,
	0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x11,
	0x00, 0x02, 0x01, 0x02, 0x04, 0x04, 0x03, 0x04,
	0x07, 0x05, 0x04, 0x04, 0x00, 0x01, 0x02, 0x77,
	0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
	0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
	0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
	0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
	0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
	0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
	0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
	0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
	0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
	0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
	0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
	0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
	0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
	0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
	0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
	0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
	0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
	0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
	0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
	0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
	0xf9, 0xfa,
/* SOS  14 bytes */
	            0xff, 0xda, 0x00, 0x0c,	0x03, 0x01,
	0x00, 0x02, 0x11, 0x03, 0x11, 0x00, 0x3f, 0x00
};

/* FUNCTIONS */

static int jpeg_save_jfif(JPEGWidget		widget,		    /* Save one frame of video as a jfif file */
			  XPlxCImage*		image,
			  char*			file,
			  unsigned char*	table)
{
  int      fd,i;
  
  if ((XPlxCImage*)NULL == image)				    /* Was a NULL JPEG image passed in as argument? */
    return(-1);							    /* Yes, don't bother trying to save anything */
  if ((fd = open(file,O_CREAT|O_WRONLY|O_TRUNC,0644)) == -1) {
    SystemError(errno,"Couldn't open jfif file for output");
    return(-1);
  }
  jfif_header[25] = (image->height  >> 8) & 0xff;		    /* Update with width and height of compressed image */
  jfif_header[26] = (image->height)  & 0xff;
  jfif_header[27] = (image->width >> 8) & 0xff;
  jfif_header[28] = (image->width) & 0xff;
  
  for (i=0; i<64; i++)						    /* Load the XVideo quanization tables specified by the user */
  {
    jfif_header[i+ 44] = (table[i + 5]);
    jfif_header[i+109] = (table[i + 70]);
  }
  if (write(fd,jfif_header,sizeof(jfif_header)) < 0)		    /* Put out the jfif header info */
  {
    SystemError(errno,"JFIF header write failed!");
    close(fd);
    unlink(file);
    return(-1);
  }
  if (write(fd,image->data,image->size) < 0)			    /* Save the compressed image data */
  {
    SystemError(errno,"JFIF scan data write failed!");
    close(fd);
    unlink(file);
    return(-1);
  }
  close(fd);							    /* End of JFIF file */
  if ((char*)NULL == widget->jpeg.lastJFIFOutputFileName)	    /* Was space for a filename allocated yet? */
    widget->jpeg.lastJFIFOutputFileName = (char*)malloc(MAXPATHLEN); /* No, try to allocate space for a new file name */
  if ((char*)NULL == widget->jpeg.lastJFIFOutputFileName)	    /* Did the allocation fail? */
  {								    /* Yes, report the error if necessary */
    ShowAlert(widget, "XPlxCreateCImage failed: Unknown error");
    if ((char*)NULL == widget->jpeg.lastJFIFOutputFileName)	    /* Free space taken by the JFIF input filename, if necessary */
    {
      free (widget->jpeg.lastJFIFOutputFileName);
      widget->jpeg.lastJFIFOutputFileName = (char*)NULL;
    }
    return(-1);
  }
  strcpy(widget->jpeg.lastJFIFOutputFileName, file);
  return(0);
}								    /* end function jpeg_save_jfif */



static int ParseJFIF(JPEGWidget	widget,				    /* Parse the jfif data into a form the c3lib can handle */
		     u_char*	buffer,
		     int	len,
		     int*	width,
		     int*	height)
{
  int done,count;
  u_char c,*p,*buf;
  int haveq,haveh;
  int i,j,k,l,m,n,rsti;
  u_char qtab[128];
  u_char htab[1024];
  
  buf = buffer; done = count = haveh = haveq = 0;
  while (!done)
  {
    c = *buf++; count++;
    if (c != 0xff) return(-1);
    c = *buf++; count++;
    switch (c)
    {
     case 0xd8:							    /* SOI */
      break;
     case 0xfe:							    /* COM */
      p = buf;
      i  = ((int) *p++) << 8; 
      i += ((int) *p++); 
      count += i;
      buf += i; break;
     case 0xdd:							    /* RST */
      p = buf;
      i  = ((int) *p++) << 8; 
      i += ((int) *p++); 
      rsti  = ((int) *p++) << 8; 
      rsti += ((int) *p++); 
      count += i;
      buf += i; break;
     case 0xe0:							    /* APP0 */
      p = buf;
      i  = ((int) *p++) << 8; 
      i += ((int) *p++); 
      count += i;
      if (*p++ != 'J') return(-2);
      if (*p++ != 'F') return(-2);
      if (*p++ != 'I') return(-2);
      if (*p++ != 'F') return(-2);
      buf += i; break;
     case 0xdb:							    /* DQT */
      p = buf;
      i  = ((int) *p++) << 8; 
      i += ((int) *p++); 
      count += i;
      j  = i - 2;
      while (j)
      {
	haveq++;
	l = (int) *p++;
	if (l > 1) return(-3);
	k = (l << 6) + l;
	qtab[k++] = l;
	for (l=0; l<64; l++)
	{
	  qtab[l + k] = (*p++);
	}
	j -= 65;
      }
      buf += i; break;
     case 0xc0:							    /* SOF */
      p = buf;
      i  = ((int) *p++) << 8; 
      i += ((int) *p++); 
      count += i;
      if (*p++ != 8) return(-4);
      *height  = ((int) *p++) << 8;
      *height += ((int) *p++);
      *width   = ((int) *p++) << 8;
      *width  += ((int) *p++);
      if ((j = *p++) != 3) return(-5);
      for (k=0; k<j; k++)
      {
	l = *p++; n = *p++; p++;
	m = (n >> 4) & 0xf;
	n = n & 0xf; 
	switch (l)
	{
	 case 1:
	  if ((m != 2) || (n != 1)) return(-5);
	  break;
	 case 2:
	 case 3:
	  if ((m != 1) || (n != 1)) return(-5);
	  break;
	 default:
	  return(0);
	}
      }
      buf += i; break;
     case 0xc4:							    /* DHT */
      p = buf;
      i  = ((int) *p++) << 8; 
      i += ((int) *p++); 
      count += i;
      j = i - 2;
      while (j)
      {
	htab[haveh++] = *p++;
	j--; l = 0;
	for (k=0; k<16; k++)
	{
	  htab[haveh] = *p++;
	  l += htab[haveh++];
	}
	j -= 16;
	for (k=0; k<l; k++) htab[haveh++] = *p++;
	j -= l;
      }
      buf += i; break;
     case 0xda:							    /* SOS */
      p = buf;
      i  = ((int) *p++) << 8; 
      i += ((int) *p++); 
      count += i;
      if ((j = *p++) != 3) return(-5);
      for (k=0; k<j; k++)
      {
	l = *p++; n = *p++;
	m = (n >> 4) & 0xf;
	n = n & 0xf; 
	switch (l)
	{
	 case 1:
	  if ((m != 0) || (n != 0)) return(-5);
	  break;
	 case 2:
	 case 3:
	  if ((m != 1) || (n != 1)) return(-5);
	  break;
	 default:
	  return(0);
	}
      }
      done = 1; buf += i; break;
     default:
      return(-1);
    }
  }
  if (!haveh || !haveq)						    /* Make sure the file defines qtables and htables */
    return(-6);
  i = 0;							    /* Reassemble the data in the right form for the C3lib */
  buffer[i++] = 0xff;
  buffer[i++] = 0xdb;
  j = haveq * 65 + 2;
  buffer[i++] = (j >> 8) & 0xff;
  buffer[i++] = j & 0xff;
  j -= 2;
  for (k=0; k<j; k++) buffer[i++] = qtab[k];			    /* restart interval added */
  buffer[i++] = 0xff;
  buffer[i++] = 0xdd;
  j = 4;
  buffer[i++] = (j >> 8) & 0xff;
  buffer[i++] = j & 0xff;
  buffer[i++] = (rsti >> 8) & 0xff;
  buffer[i++] = rsti & 0xff;					    /* end of restart interval */
  buffer[i++] = 0xff;
  buffer[i++] = 0xc4;
  j = haveh + 2;
  buffer[i++] = (j >> 8) & 0xff;
  buffer[i++] = j & 0xff;
  j -= 2;
  for (k=0; k<j; k++) buffer[i++] = htab[k];
  for (j=count; j<len; j++) buffer[i++] = buffer[j];
  return(i);
}								    /* end function ParseJFIF */



static XPlxCImage* jpeg_load_jfif(JPEGWidget	widget,		    /* Load one frame of video from a JFIF file */
				  char*		file)
{
  XPlxCImage*	i;
  char*		data;
  int        	fd,size;
  int        	width,height;
  
  if ((fd = open(file,O_RDONLY)) == -1)
  {
    SystemError(errno,"Couldn't open JFIF file for input");
    return((XPlxCImage *) NULL);
  }
  size  = lseek(fd,0,SEEK_END);					    /* Figure out the size of the JFIF data */
  lseek(fd,0,SEEK_SET);
  if (widget->jpeg.useSharedMemory == False)			    /* Is the widget using shared memory? */
    data = (char*)malloc(size);					    /* No, allocate space necessary to store a JPEG image */
  else								    /* Yes, the widget is using shared memory; point to the... */
    data = widget->jpeg.jpegData->data;				    /* ...shared memory already allocated. */
  if (read(fd,data,size) < 0)					    /* Get the compressed frame out */
  {
    SystemError(errno,"JFIF file data read failed!");
    if (widget->jpeg.useSharedMemory == False)
      free(data);
    close(fd);
    return((XPlxCImage *) NULL);
  }
  close(fd);
  width = height = 0;
  size = ParseJFIF(widget, (u_char*)data, size, &width, &height);
  if (size < 0)
  {
    switch (size)
    {
     case -1:
      ShowAlert(widget, "Invalid marker sequence in file!");
      break;
     case -2:
      ShowAlert(widget, "Unrecognized APP0 marker in file!");
      break;
     case -3:
      ShowAlert(widget, "Too many qtables defined in file!");
      break;
     case -4:
      ShowAlert(widget, "Only 8 bit precision supported!");
      break;
     case -5:
      ShowAlert(widget, "Only YCbCr 422 format supported!");
      break;
     case -6:
      ShowAlert(widget, "JFIF file is missing DHT/DQT sequence!");
      break;
    }
    if (widget->jpeg.useSharedMemory == False)
      free(data);
    return((XPlxCImage *) NULL);
  }
  if (!width || !height)
  {
    ShowAlert(widget, "File is not a valid JFIF file");
    if (widget->jpeg.useSharedMemory == False)
      free(data);
    return((XPlxCImage *) NULL);
  }
  if (widget->jpeg.useSharedMemory == False)			    /* Is this widget using shared memory? */
    i = (XPlxCImage *) XPlxCreateCImage(XtDisplay(widget),	    /* No, allocate space for a JPEG image */
					NULL,size,width,height);
  else								    /* Yes, the widget is using shared memory; no need to... */
  {								    /* ...allocate space, just point to existing shared memory. */
    i = widget->jpeg.jpegData;
    widget->jpeg.jpegData->width = width;			    /* Update the shared memory image's parameters */
    widget->jpeg.jpegData->height = height;
    widget->jpeg.jpegData->size = size;
  }
  if ((char*)NULL == widget->jpeg.lastJFIFInputFileName)	    /* Was space for a filename allocated yet? */
    widget->jpeg.lastJFIFInputFileName = (char*)malloc(MAXPATHLEN); /* No, try to allocate space for a new file name */
  if ((i == NULL) ||						    /* If for some reason the image is NULL, or if the space... */
      ((char*)NULL == widget->jpeg.lastJFIFInputFileName))	    /* ...for the image's name cannot be allocated, return... */
  {								    /* ...an error. */
    ShowAlert(widget, "XPlxCreateCImage failed: Unknown error");
    if (widget->jpeg.useSharedMemory == False)
      free(data);
    if ((char*)NULL == widget->jpeg.lastJFIFInputFileName)	    /* Free space taken by the JFIF input filename, if necessary */
    {
      free (widget->jpeg.lastJFIFInputFileName);
      widget->jpeg.lastJFIFInputFileName = (char*)NULL;
    }
    return((XPlxCImage *) NULL);
  }
  if (widget->jpeg.useSharedMemory == False)			    /* Is the widget using shared memory?  If not, copy the... */
  {								    /* ...image data into the widget and free the temporary... */
    memcpy(i->data, data, size);				    /* ...data allocated for this function to read from disk. */
    free(data);
  }
  strcpy(widget->jpeg.lastJFIFInputFileName, file);		    /* Update the last input file name seen */
  return(i);							    /* Return a nice pointer to the newly-created JPEG image */
}								    /* end function jpeg_load_jfif */



static void SystemError(int theError, char* errorMessage)
{
  return;							    /* For now, SystemError will do nothing */
}								    /* end function SystemError */



static void ShowAlert(JPEGWidget	widget,
		      char*		theAlert)
{
  if (widget->jpeg.diagnostics == True)				    /* Should the widget print an error message?  */
    printf("JPEG Widget Warning: %s.\n", theAlert);		    /* Yes, print an error message before returning */
  return;
}								    /* end function ShowAlert */



int JPEGWidgetShmErrorHandler(Display*		display,
			      XErrorEvent*	errorEvent)
{
  if ((errorEvent->request_code == shmMajorOpcode) &&		    /* Was it the shared memory attach request that failed? */
      (errorEvent->minor_code == X_ShmAttach))
  {
    xShmAttachErrorOccurred = True;
    return(0);							    /* Yes, return an error code */
  }
  else
    return(1);							    /* No, something else failed, return success */
} /* end function JPEGShmWidgetErrorHandler */



static void  RegisterJPEGWidgetEventHandler(JPEGWidget newWidget)   /* Look for window move events if on an overlay product */
{
  Widget	childWidget;
  Widget	parentWidget;
  
  if ((plx_IO*)NULL == hardwareParams)				    /* If no Plx hardware was found, don't bother... */
    return;							    /* ...registering an event handler. */
  if (0 == hardwareParams->overlay)				    /* Is this server an overlay product? */
    return;							    /* No, no need to track ConfigureNotify events */
  childWidget = (Widget)newWidget;
  parentWidget = XtParent((Widget)newWidget);
  while ((XtIsShell(parentWidget) == False) &&			    /* Keep looking up the widget tree until a Shell widget... */
	 (parentWidget != (Widget)NULL))			    /* ... is found or the top of the widget tree is reached;... */
  {								    /* ...a Shell widget will surely receive... */
    childWidget = parentWidget;					    /* ...ConfigureNotify events. */
    parentWidget = XtParent(childWidget);
  }
  XtAddEventHandler(parentWidget,				    /* ...Seek StructureNotify events, and redraw when received. */
		    StructureNotifyMask,
		    False,					    /* Don't bother looking for non-maskable events */
		    JPEGWidgetEventHandler,
		    (XtPointer)newWidget);
  newWidget->jpeg.eventHandlerInstalled = True;
  newWidget->jpeg.eventHandlerWidget = parentWidget;
  return;
} /* end function RegisterJPEGWidgetEventHandler */



static void JPEGWidgetEventHandler(Widget	widget,
				   XtPointer	clientData,
				   XEvent*	theEvent,
				   Boolean*	dispatchOtherEventHandlers)
{
  if (theEvent->type != ConfigureNotify)			    /* Don't bother doing anything if the window doesn't move */
    return;
  Redisplay((Widget)clientData, (XEvent*)NULL, (Region)NULL);	    /* Force a redisplay of widget, bypassing normal mechanisms */
  return;
} /* end function JPEGWidgetEventHandler */



static void CreateJPEGWidgetQTablesIfNecessary(JPEGWidget	theWidget,
					       Boolean		absolutely)
{
  Display*		display = XtDisplay(theWidget);
  
  if ((GC)NULL == theWidget->jpeg.decompressGC)			    /* Does a decompress GC exist for the dest window? */
  {
    theWidget->jpeg.decompressGC =				    /* No, create a GC needed for storing a Q table for... */
      XCreateGC(display,					    /* ...decompressing the image in the destination window. */
		theWidget->jpeg.destWin,
		0, 0);
    absolutely = True;						    /* A new GC was created, so absolutely put a new Q table */
  }
  if (True == absolutely)					    /* Should a new Q table be put? */
  {								    /* Yes, absolutely */
    XPlxPutTable(display,					    /* Install the Q table in the decompression (destination)... */
		 theWidget->jpeg.destWin,			    /* ... window's GC */
		 theWidget->jpeg.decompressGC,
		 (char*)theWidget->jpeg.qTable,
		 theWidget->jpeg.qTableSize,
		 1);
  }
  return;
} /* end function CreateJPEGWidgetQTablesIfNecessary */



static Boolean JPEGWidgetCheckJFIFInputFile(JPEGWidget	newWidget,
					    JPEGWidget	currentWidget)
{
  XPlxCImage*	jfifImage;
  int		stringCompareResult = 0;
  Boolean	stringsDiffer = False;
  
  if ((newWidget->jpeg.jfifInputFile != NULL) &&		    /* Does the current widget point to valid strings? */
      (newWidget->jpeg.lastJFIFInputFileName != NULL))
  {								    /* Yes, determine if they represent different filenames */
    stringCompareResult = strcmp(newWidget->jpeg.jfifInputFile,
				 newWidget->jpeg.lastJFIFInputFileName);
    stringsDiffer = (stringCompareResult != 0) ? True : False;
  }
  else								    /* No, at least one of them is a NULL pointer */
    stringsDiffer = False;
  if ((newWidget->jpeg.jfifInputFile !=				    /* Did the programmer request a new JFIF file be read in? */
       currentWidget->jpeg.jfifInputFile) ||
      (True == stringsDiffer))
  {
    if (newWidget->jpeg.jfifInputFile == (String)NULL)		    /* Yes, is the filename valid? */
    {
      XtWarning("JPEG Widget: JFIF file name was NULL.  Reverting to previous value.");
      newWidget->jpeg.jfifInputFile = currentWidget->jpeg.jfifInputFile;
      return(False);						    /* Not necessary to redisplay */
    }
    jfifImage = jpeg_load_jfif(newWidget,
			       newWidget->jpeg.jfifInputFile);
    if (jfifImage == (XPlxCImage*)NULL)				    /* Did the read fail? */
    {								    /* Yes, report the error and revert to previous values */
      XtWarning("JPEG Widget: jfifInputFile was not a valid JFIF file.");
      newWidget->jpeg.jpegData = currentWidget->jpeg.jpegData;
      newWidget->jpeg.jfifInputFile = (String)NULL;		    /* Reset resource showing name of JFIF file not loaded */
      return(False);						    /* Not necessary to redisplay */
    }
    if ((currentWidget->jpeg.jpegData != (XPlxCImage*)NULL) &&
	(currentWidget->jpeg.useSharedMemory == False))
      XPlxDestroyCImage(currentWidget->jpeg.jpegData);
    newWidget->jpeg.jpegData = jfifImage;			    /* Point to the new JPEG image data */
    newWidget->jpeg.putNewImage = True;				    /* Tell the widget to display the new image now */
    newWidget->jpeg.getNewImage = False;			    /* Don't grab new data to compress; new data was just...  */
    newWidget->jpeg.defaultDisplaySourceRect.width =		    /* Update default display variables to reflect the size.. */
      newWidget->jpeg.jpegData->width;				    /* ...of the JPEG image just passed in to the widget. */
    newWidget->jpeg.defaultDisplaySourceRect.height =
      newWidget->jpeg.jpegData->height;
    return(True);						    /* Tell the widget to redraw itself with the new data */
  } /* end if new & current widgets differ or strings differ */
  return(False);						    /* If this stmt reached, no reason to refresh the widget */
} /* end function JPEGWidgetCheckJFIFInputFile */



static void JPEGWidgetCheckJFIFOutputFile(JPEGWidget	newWidget,
					  JPEGWidget	currentWidget)
{
  int		stringCompareResult = 0;
  Boolean	stringsDiffer = False;
  
  if ((XPlxCImage*)NULL == newWidget->jpeg.jpegData)		    /* Don't bother doing anything if there's no JPEG data... */
    return;							    /* ...to save. */
  if ((newWidget->jpeg.jfifOutputFile != NULL) &&		    /* Does the current widget point to valid strings? */
      (newWidget->jpeg.lastJFIFOutputFileName != NULL))
  {								    /* Yes, determine if they represent different filenames */
    stringCompareResult = strcmp(newWidget->jpeg.jfifOutputFile,
				 newWidget->jpeg.lastJFIFOutputFileName);
    stringsDiffer = (stringCompareResult != 0) ? True : False;
  }
  else								    /* No, at least one of them is a NULL pointer */
    stringsDiffer = False;
  if ((newWidget->jpeg.jfifOutputFile !=			    /* Did the programmer request a new JFIF file be saved? */
       currentWidget->jpeg.jfifOutputFile) ||
      (True == stringsDiffer))
  {
    if (newWidget->jpeg.jfifOutputFile == (String)NULL)		    /* Yes, is the filename valid? */
    {
      XtWarning("JPEG Widget: JFIF file name was NULL.  Reverting to previous value.");
      newWidget->jpeg.jfifOutputFile = currentWidget->jpeg.jfifOutputFile;
      return;							    /* Nothing else to do; get outtahere */
    }
    jpeg_save_jfif(newWidget,					    /* Attempt to save the current JPEG data as a JFIF file */
		   newWidget->jpeg.jpegData,
		   newWidget->jpeg.jfifOutputFile,
		   newWidget->jpeg.qTable);
  } /* end if a new output filename was requested */
  return;
} /* end function JPEGWidgetCheckJFIFOutputFile */



static void AdjustCompressRectangles(JPEGWidget theWidget)	    /* Assure that both the compressSource and compressDest... */
{								    /* ...rectangles obey C-Cube chip and JPEG size limits. */
  theWidget->jpeg.compressDestRect->width =			    /* Assure that the image to be compressed doesn't exceed... */
    (theWidget->jpeg.compressDestRect->width >			    /* ...the dimensions of the CCube chip (1K x 1K). */
     (1024 * theWidget->jpeg.zoomFactor) ?
     (1024 * theWidget->jpeg.zoomFactor) :
     theWidget->jpeg.compressDestRect->width);
  theWidget->jpeg.compressDestRect->height =
    (theWidget->jpeg.compressDestRect->height >
     (1024 * theWidget->jpeg.zoomFactor) ?
     (1024 * theWidget->jpeg.zoomFactor) :
     theWidget->jpeg.compressDestRect->height);
  theWidget->jpeg.compressSourceRect->width =
    (theWidget->jpeg.compressSourceRect->width >
     (1024 * theWidget->jpeg.zoomFactor) ?
     (1024 * theWidget->jpeg.zoomFactor) :
     theWidget->jpeg.compressSourceRect->width);
  theWidget->jpeg.compressSourceRect->height =
    (theWidget->jpeg.compressSourceRect->height >
     (1024 * theWidget->jpeg.zoomFactor) ?
     (1024 * theWidget->jpeg.zoomFactor) :
     theWidget->jpeg.compressSourceRect->height);
  
  if ((theWidget->jpeg.compressSourceRect->width % 16) != 0)	    /* Assure that all width & height parameters sit on a... */
    theWidget->jpeg.compressSourceRect->width -=		    /* ...16 pixel boundary. Not sure why, but Dave Arnold... */
      (theWidget->jpeg.compressSourceRect->width % 16);		    /* ...says so. */
  if ((theWidget->jpeg.compressDestRect->width % 16) != 0)
    theWidget->jpeg.compressDestRect->width -=
      (theWidget->jpeg.compressDestRect->width % 16);
  
  if ((theWidget->jpeg.compressSourceRect->height % 16) != 0)
    theWidget->jpeg.compressSourceRect->height -=
      (theWidget->jpeg.compressSourceRect->height % 16);
  if ((theWidget->jpeg.compressDestRect->height % 16) != 0)
    theWidget->jpeg.compressDestRect->height -=
      (theWidget->jpeg.compressDestRect->height % 16);
  
  if (theWidget->jpeg.compressSourceRect->width < 16)		    /* Assure that at least a minimally-sized JPEG image is... */
    theWidget->jpeg.compressSourceRect->width = 16;		    /* ...created. */
  if (theWidget->jpeg.compressSourceRect->height < 16)
    theWidget->jpeg.compressSourceRect->height = 16;
  
  if (theWidget->jpeg.compressDestRect->width < 16)
    theWidget->jpeg.compressDestRect->width = 16;
  if (theWidget->jpeg.compressDestRect->height < 16)
    theWidget->jpeg.compressDestRect->height = 16;
  
  return;
} /* end function AdjustCompressRectangles */



static void AdjustDisplayRectangles(JPEGWidget w)		    /* Assure that both the displaySource and displayDest... */
{								    /* ...rectangles obey C-Cube chip and JPEG size limits. */
  w->jpeg.defaultDisplaySourceRect.width =			    /* Update default display variables to reflect the size.. */
    w->jpeg.jpegData->width;					    /* ...of the JPEG image just created. */
  w->jpeg.defaultDisplaySourceRect.height =
    w->jpeg.jpegData->height;
  w->jpeg.defaultDisplayDestRect.width =
    w->jpeg.jpegData->width;
  w->jpeg.defaultDisplayDestRect.height =
    w->jpeg.jpegData->height;
  
  if (w->jpeg.displaySourceRect->width >			    /* Is the width of JPEG image to be considered larger... */
      w->jpeg.compressDestRect->width)				    /* ...than the image itself? */
    w->jpeg.displaySourceRect->width =				    /* Yes, don't allow this error to occur; adjust the width... */
      w->jpeg.compressDestRect->width;				    /* ...being considered to fit the actual image width.  */
  
  if (w->jpeg.displaySourceRect->height >			    /* Is the height of JPEG image to be considered larger... */
      w->jpeg.compressDestRect->height)				    /* ...than the image itself? */
    w->jpeg.displaySourceRect->height =				    /* Yes, don't allow this error to occur; adjust height... */
      w->jpeg.compressDestRect->height;				    /* ...being considered to fit the actual image height.  */
  
  if ((w->jpeg.displaySourceRect->width % 16) != 0)		    /* Assure that all width & height parameters sit on a... */
    w->jpeg.displaySourceRect->width -=				    /* ...16 pixel boundary. Not sure why, but Dave Arnold... */
      (w->jpeg.displaySourceRect->width % 16);			    /* ...says so. */
  if ((w->jpeg.displayDestRect->width % 16) != 0)
    w->jpeg.displayDestRect->width -=
      (w->jpeg.displayDestRect->width % 16);
  
  if ((w->jpeg.displaySourceRect->height % 16) != 0)
    w->jpeg.displaySourceRect->height -=
      (w->jpeg.displaySourceRect->height % 16);
  if ((w->jpeg.displayDestRect->height % 16) != 0)
    w->jpeg.displayDestRect->height -=
      (w->jpeg.displayDestRect->height % 16);
  
  if (w->jpeg.displaySourceRect->width < 16)			    /* Assure that at least a minimally-sized JPEG image is... */
    w->jpeg.displaySourceRect->width = 16;			    /* ...created. */
  if (w->jpeg.displaySourceRect->height < 16)
    w->jpeg.displaySourceRect->height = 16;
  
  if (w->jpeg.displayDestRect->width < 16)
    w->jpeg.displayDestRect->width = 16;
  if (w->jpeg.displayDestRect->height < 16)
    w->jpeg.displayDestRect->height = 16;
  
  return;
} /* end function AdjustDisplayRectangles */
