/*
 *   Copyright (c) 1993 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* $Header: /plx/source/Xclients/generic/RCS/XVideoIn.c,v 0.65 1996/01/25 04:03:52 drapeau Exp $ */
/* $Log: XVideoIn.c,v $
 * Revision 0.65  1996/01/25  04:03:52  drapeau
 * Fixed a minor error in DestroyVideoStill(): the function was checking
 * the wrong widget variable to determine whether an image needed to be
 * destroyed.
 *
 * Revision 0.64  1996/01/22  06:20:15  drapeau
 * Added a new resource, "stillDelay", to allow control over the time
 * spent waiting between the time that the video digitizer is stopped and
 * the time a still image Pixmap or XImage is grabbed.
 *
 * Revision 0.63  1996/01/20  00:28:39  drapeau
 * Cleaned up code to destroy still images to handle both cases (i.e.,
 * when the still image is either a Pixmap or an XImage).
 *
 * Revision 0.62  1996/01/19  04:04:44  drapeau
 * Added support for both XImages and their faster Pixmap counterparts
 * for still images.  The widget detects which to use and uses that.
 *
 * Revision 0.61  1996/01/11  03:18:51  drapeau
 * Finally, one last fix to the input model to support both multiple
 * instances of a particular format (i.e., Composite, Y/C, RGB, etc.)
 * *and* multiple input "bundles" that come into the same digitizer
 * (e.g., if one of the digitizer's input wires had extra hardware to
 * pre-process an incoming analog video signal).
 *
 * Revision 0.60  1996/01/10  09:30:39  drapeau
 * Fixed lots of minor compiler warnings to sanity-check the module.
 *
 * Revision 0.59  1996/01/10  09:16:38  drapeau
 * Two changes:
 *
 * 1) Added a new resource, "stillImage", that gives a copy of the
 * widget's XImage to the application programmer if the programmer wishes
 * a copy of the widget's still image.
 *
 * 2) The widget now correctly checks if hardwareParams and associated
 * info is non-NULL.  If NULL, that means there is no Plx hardware
 * available on the Display* being used for this connection.
 *
 * Revision 0.58  1996/01/06  01:32:20  drapeau
 * Improved the portability of the widget in that it checks more cleanly
 * whether the X server being used is an overlay server.  This makes the
 * widget automatically detect whether it should track ConfigureNotify
 * events so it can properly redraw still images after being moved.
 *
 * Revision 0.57  1996/01/05  21:22:13  drapeau
 * Removed extraneous debugging message in CreateVideoPixmap().
 *
 * Revision 0.56  1995/12/23  02:43:48  drapeau
 * Removed extraneous, debugging diagnostic messages.
 *
 * Revision 0.55  1995/12/23  02:17:37  drapeau
 * Enhanced the widget to deal with multiple instances of similar inputs
 * better.
 *
 * Also, removed the XClearWindow() call from the Redisplay() method so
 * as to eliminate flashing when the widget's window is redrawn.
 *
 * Revision 0.54  1995/12/19  00:10:55  drapeau
 * Major modifications to the autoSearch feature of the widget, to
 * enhance its usefulness.  The new version of this feature re-defines
 * the autoSearch resource as an "int" rather than a "Boolean" and allows
 * the application programmer to do auto-searching in three ways:
 *
 * 1) Set to "True": search for an available input source on *all*
 *    possible inputs (i.e., digitizers)
 *
 * 2) Set to "AutoSearchInput0": search for an available input source on
 *    the 1st digitizer only (PLX_INPUT_0)
 *
 * 3) Set to "AutoSearchInput1": search for an available input source on
 *    the 2nd digitizer only (PLX_INPUT_1)
 *
 * To disable the autoSearch feature completely, set the resource to
 * "False".
 *
 * Revision 0.53  1995/11/13  20:33:49  drapeau
 * Exported the widget's GC so that programmers doing graphics-over-video
 * operations can re-tag the GC.  This exporting of the GC is dangerous,
 * but perhaps a necessary evil to enable GOV.
 *
 * Revision 0.52  1995/10/27  20:47:09  drapeau
 * Put a short sleep between the time that video is stilled and the time
 * an XImage is grabbed from that still video, to give the video
 * digitizer time to finish digitizing the frame.
 *
 * Revision 0.51  1995/09/28  21:55:51  drapeau
 * Enhanced diagnostic messages used in debugging the widget and Xlib
 * layer.
 *
 * Revision 0.50  1995/09/19  21:12:00  drapeau
 * Replaced calls to GetTrueColorVisual() with a new, more portable
 * version now in the XPlx library: XPlxGetVideoVisual().
 *
 * Revision 0.49  1995/08/17  20:26:02  drapeau
 * Modified the widget so that it checks for ConfigureNotify events if
 * running on either the HP overlay server or the IBM overlay server.
 *
 * Revision 0.48  1995/06/24  00:21:27  drapeau
 * Modified the Redisplay() method by replacing the call XPutImage() with
 * a new function XPlxXPutImage(), which is used for displaying still
 * images.  This was put in as support for the HP overlay product, which
 * ran into some problems when overriding XPutImage() in some cases.
 *
 * Revision 0.47  1995/05/25  18:24:08  drapeau
 * Changed the behavior of the widget, especially when auto search is
 * off.  If the "autoSearch" resource is False, then this widget will
 * always refresh video even if a video sync signal is not detected.
 * This should address concerns of those who want video to be attempted
 * when, say, an input cable isn't connected, so that when the input
 * cable *is* connected the signal can be seen immediately.
 *
 * A side effect of this is that video may be incorrectly digitized: for
 * example, if the widget is set to display PAL video but an NTSC source
 * is plugged in, the widget's window size will be incorrect and the
 * color will be wrong when the input cable is actually plugged in.  This
 * can be corrected by re-selecting the signal type.
 *
 * Revision 0.46  1995/05/19  02:00:28  drapeau
 * Removed GC_info code (since it's not necessary on the HP or IBM
 * platform).
 *
 * Also, added an extra wait just after selecting a video source to give
 * it time to settle in, before querying whether the signal has proper
 * sync.
 *
 * Revision 0.45  1995/04/12  23:00:50  drapeau
 * Added code to keep track of GC-Window pairs.  This code is
 * library-wide and is to help assure that no single process installs
 * more than one GC per Window.  This was done to work around the
 * operation of the Solaris 2.4 server with regard to window id's (i.e.,
 * tags).
 *
 * Revision 0.44  1995/04/07  01:34:32  drapeau
 * Modified the hack that correctly detects the signal type.  In this
 * scheme, the first vote is taken immediately.  If a False (i.e., no
 * sync) result is returned, the widget doesn't bother to vote the other
 * two times and instead immediately moves on to the next source.  If a
 * True is returned, the widget pauses for 500 milliseconds, then queries
 * again.  If a False is returned this 2nd time, again the widget kicks
 * out and moves to the next source.  If a True was returned, however,
 * the widget pauses yet once more (again for 500 milliseconds), then
 * queries a third time.
 *
 * In this scheme, a maximum of three votes are taken.  A True result is
 * obtained only if all three votes are True.
 *
 * Revision 0.43  1995/04/06  22:55:19  drapeau
 * Added yet another hack to the widget.  Actually, replaced one hack
 * with another.  This hack still addresses the problem of "false
 * positives", i.e., when asking the server whether an NTSC signal is
 * present when in fact a PAL signal is actually there, sometimes the
 * server will return True when it should return False.
 *
 * The latest hack calls XPlxVideoInputSelect() once, then calls
 * XPlxQueryVideo() three times, pausing for about half a second between
 * calls to give the digitizer chip time to adjust to the new signal type
 * being requested by ...InputSelect().  Even this time may not be
 * enough, which is why the ...QueryVideo() is called three times: in
 * order for the signal to be considered valid, ...QueryVideo() must
 * return True for its sync_ok field all three times it is called,
 * otherwise the widget moves on to the next signal if possible.
 *
 * This delay should really be placed in the X server or hardware library
 * so that all XPlx applications have the benefit of waiting for the
 * digitizer chip to detect sync correctly.  In addition, it appears
 * possible that the chip itself can report back what type of signal it
 * is detecting, thereby alleviating the need for the widget to select
 * each of NTSC, PAL, and SECAM types.  In other words, it should be
 * possible for XPlxVideoInputSelect() to specify only the digitizer and
 * wire (i.e., channel and format) and let the hardware detect which
 * signal standard (i.e., NTSC, PAL, SECAM) is present.  This would save
 * a lot of time on searching for a valid input signal type.
 *
 * Another change made in this revision: additional diagnostic messages
 * relating to the voting scheme.
 *
 * Revision 0.42  1995/04/05  16:49:15  drapeau
 * Replaced last revision's hack with a simpler hack.  The hack was put
 * in to solve the problem of falsely detecting a valid sync signal
 * (e.g., a PAL signal is plugged into the input but when asking if a
 * valid NTSC signal is present, a True result is returned).  The simpler
 * hack is to select the desired video source, then query if sync is
 * present (using XPlxQueryVideo()), thenrepeating this series twice more
 * in order to get three total tries.  The best 2 out of 3 results is
 * taken.
 *
 * The question remains: is it enough to select the video signal once
 * (using XPlxVideoInputSelect()) and do the XPlxQueryVideo() three
 * times, or does the XPlxVideoInputSelect() also have to be performed
 * three times?
 *
 * Revision 0.41  1995/04/05  02:23:59  drapeau
 * Added some hacks to the code to do alternate auto-searching for valid
 * video sources.  This was done to correct the problem in which
 * VideoTool and other Xt-based video apps would incorrectly detect NTSC
 * sync when in fact a PAL signal was present.
 *
 * Revision 0.40  1995/03/09  23:50:17  drapeau
 * Minor fix: initialized the widget instance variable
 * "eventHandlerInstalled" in case an event handler is not installed for
 * this widget.  On Sun platforms to this date, no event handler needs to
 * be installed (it's normally installed on the HP platform to detect
 * when the widget's window has been moved).
 *
 * Revision 0.39  1995/02/24  03:38:14  drapeau
 * Fixed an error in the way XPlxSetVideoParameterString() was being
 * used.  For GENLOCK and FREERUN it should be an "output_config" instead
 * of an "input_config" string.  This version fixes the problem.
 *
 * Revision 0.38  1994/10/26  01:51:35  drapeau
 * Added a new resource, "drawMethod", to the widget, allowing
 * application programmers to draw graphics over the video without the
 * graphics being overwritten by video.
 *
 * Revision 0.37  1994/10/22  02:11:00  drapeau
 * Used a portable version of the "time.h" include file.
 *
 * Revision 0.36  1994/10/14  21:14:39  drapeau
 * Removed event handlers on the widget during the Destroy method, if
 * they are installed.
 *
 * Revision 0.35  1994/10/14  02:01:20  drapeau
 * Added HP-specific code to handle window move events.  On an overlay
 * product, the window server may not necessarily move the window's pixel
 * data when the window is moved.  In this case, the widget's window
 * needs to register interest in window move events.  This version of the
 * widget checks if it is connecting with an HP server; if so, the widget
 * assumes it's working with an overlay board and so registers an Xt
 * handler for window move events.
 *
 * Revision 0.34  1994/09/23  06:43:28  drapeau
 * Minor changes to enhance diagnostic reporting for debugging purposes.
 *
 * Revision 0.33  1994/09/14  03:22:05  drapeau
 * Added a sleep (via the select() system call) between the time inputs
 * are selected and the time the hardware is queried to see if the new
 * input has a valid sync on it.
 *
 * Revision 0.32  1994/09/04  06:38:22  drapeau
 * Minor change to improve performance when getting a new still frame of
 * video.  In SetValues() when the "newFrame" resource is set, the method
 * immediately calls the widget's Redisplay() method instead of waiting
 * for the toolkit to call the widget's Redisplay() method.
 *
 * Revision 0.31  1994/09/03  02:05:50  drapeau
 * Several error fixes:
 *
 * * Modified Redisplay(); before it did not paint its window if there
 *   was no sync signal found.  The improvement now considers that even
 *   if there is no sync signal, the widget might be in still mode and
 *   there is a still image to display.  In this case, it's irrelevant
 *   whether a sync signal is present.
 *
 * * SetValues(), after trying to set a new input signal: if it fails,
 *   the original input signal is reset, but had not done a QueryVideo()
 *   to reset the widget's sync status resource.  The code now does so.
 *
 * * FindBestVideoSourceMatch() now eliminates the call to InputSelect()
 *   if no signal is found.  The calling function will reset the input
 *   selection instead.
 *
 * Revision 0.30  1994/08/29  22:17:54  drapeau
 * Changed the default for the autoSearch resource for the VideoIn
 * widget; autoSearch is now enabled by default.
 *
 * Revision 0.29  1994/08/27  06:12:33  drapeau
 * Added a new resource, autoSearch, that allows the widget to
 * automatically search for a valid video input on all channels if the
 * programmer's first choice has no valid video.  By default, this is set
 * to False.
 *
 * Revision 0.28  1994/08/27  05:14:52  drapeau
 * Explicitly stop video input before destroying the widget and its
 * window.
 *
 * Revision 0.27  1994/08/24  02:08:29  drapeau
 * Fixed two minor memory leaks, now properly freeing data after calling
 * XPlxQueryVideo().
 *
 * Revision 0.26  1994/08/06  01:15:15  drapeau
 * Cosmetic change: moved PrintPlx* functions earlier in this module so
 * the Initialize() function could see them.
 *
 * Revision 0.25  1994/07/20  21:56:06  drapeau
 * Major changes to clean up the way input selection and management is
 * done.  These changes were prompted by the PowerVideo/700 but the new
 * input model will apply to all future products.  Mostly, the addition
 * of the "instance" field is important.  The instance field applies when
 * trying to distinguish two similar hardware inputs (e.g., when two
 * composite cables are available on a single digitizer).  This allows
 * software running on VIO products to select between the "normal"
 * composite input and the genlock input (which also serves as a separate
 * composite input).
 *
 * Revision 0.24  1994/06/11  00:34:11  drapeau
 * Minor change to eliminate ANSI compiler warnings.  Declared function
 * prototype for XPlxQueryVideo().
 *
 * Revision 0.23  1994/06/09  23:52:47  drapeau
 * Two major changes:
 *
 * 1) Changed implementation of handling still video from using Pixmaps
 *    to using XImages, in order to support the HP platform.
 *
 * 2) Added diagnostic messages to aid in debugging, especially for the
 *    HP platform.
 *
 * Revision 0.22  1994/05/10  02:13:12  drapeau
 * Fixed minor error in changing widget sizes after signal format was
 * changed; this fix includes changing the size of the default source
 * rectangle.
 *
 * Revision 0.21  1994/05/10  00:14:07  drapeau
 * Added code to implement a diagnostics resource for printing widget
 * diagnostics.
 *
 * Revision 0.20  1994/04/30  00:43:16  drapeau
 * Moved the code to implement getting a TrueColor visual to another
 * module, since other modules share this common functionality and it
 * needs to be portable across platforms, too.
 *
 * Revision 0.19  1994/04/28  05:31:20  drapeau
 * Major changes to the widget, mostly having to do with support for PAL
 * and SECAM formats.  The widget now automatically checks all possible
 * input combinations for a valid video signal, if the default signal
 * configuration does not have video sync on it.
 *
 * Revision 0.18  1994/01/20  01:54:12  drapeau
 * Fixed an error in the way a TrueColor Visual was being sought; the
 * search was not correctly searching through the list of Visuals.
 *
 * Also initialized the startingY resource to a reasonable value
 * (startingY has to do with taking the blank interval into account when
 * showing video).
 *
 * Revision 0.17  1993/12/20  22:03:39  drapeau
 * Fixed an error in SetValues(): when a new channel was requested,
 * SetValues() failed to correctly change the channel.
 *
 * Revision 0.16  1993/11/17  18:45:46  drapeau
 * Modified code to take advantage of the new resource "videoWindow"; this
 * resource points to a window to be used by this widget when using XPlx
 * extension calls.  It does not necessarily have to point to the widget's
 * window but can indicate any valid X window to allow the XPlx extension calls
 * to be used.
 *
 * Revision 0.15  93/09/02  12:12:06  drapeau
 * Modified the SetColorValues() function.  The function now checks if any
 * color values (brightness, contrast, saturation, hue) were actually
 * changed to determine whether it's necessary to call XPlxVideoValueLoad().
 * This change was made in response to an error found by Dave Plain of VI Corp
 * in Europe.  He found that calling the widget's SetValues() method before its
 * Realize() method was called caused the widget to dump core.  This is because
 * SetValues() always calls SetColorValues(), never bothering to check whether
 * it's actually necessary to do so.
 * The problem with calling SetColorValues() before the Realize() method is
 * called is that there is no X window for the widget until Realize() is
 * called (it is the job of the Realize() method to create an X window for the
 * widget, according to Xt specs).  Therefore, no XPlx... calls can be made
 * until the widget has a window to work with.
 * This solution is probably a short-term solution; a longer-term solution might
 * be to retain the temporary window created by the Initialize() method until
 * Realize() creates the widget's "real" X window; at that point, the temporary
 * window could be destroyed but in the meantime all other widget methods would
 * have an X window to use for XPlx... calls.
 * 
 * Revision 0.14  93/08/10  17:51:38  drapeau
 * Several changes:
 * * Added function "CreateVideoPixmap()" to hold still images correctly.
 *    This was done to implement correct behavior for the widget when told to
 *   display a still frame of video: when video is set to still, then each time
 *   the Redisplay() method is called this widget should show the last image
 *   that was captured, instead of grabbing a new still frame.
 * * Fixed errors in the default source and display rectangle code.  It turned
 *   out not to be enough to make the variables "defaultSourceRect" and
 *   "defaultDisplayRect" statics (evidently, an application that creates
 *   several VideoIn widgets will share a single copy of these variables).  The
 *   fix was to make these XRectangle's widget fields that are invisible to the
 *   application programmer.
 * * Cosmetic changes were made, to remove extraneous diagnostic messages.
 * 
 * Revision 0.13  93/06/29  15:00:23  drapeau
 * One important change, several cosmetic changes.
 * The important change was to make the definition of the default display
 * and source rectangles "static", so that only the functions inside this C
 * module can modify them.  If they are not declared static, it is possible
 * that other C modules using the same variable name (e.g., the VideoIn widget)
 * might modify these by mistake.
 * The cosmetic changes were simply to remove extraneous printf statements.
 * 
 * Revision 0.12  93/06/11  16:36:27  drapeau
 * Simplified the method in which resizing is done, especially when trying
 * to conform to 4:3 aspect ratio.  Now, the displayRect is not taken into
 * account when determining resizing; the display rectangle of the widget
 * should now track the size of the widget.  In other words, the widget
 * should now always fill its window with video; there will be no black
 * edges that contain no video.
 * 
 * Revision 0.11  93/06/10  17:09:23  drapeau
 * Updated the copyright message; no other changes made.
 * 
 * Revision 0.10  93/06/10  16:55:02  drapeau
 * First working version.  This version allows the programmer to specify a
 * display rectangle as well as a source rectangle.  It is not clear whether
 * the programmer should be allowed this much flexibility, though: it might be
 * better to implement a policy whereby the size of the widget always tracks
 * the size of the display rectangle; in other words, instead of letting the
 * programmer specify a displayRect, the programmer would simply modify the
 * Core part's width and height fields.  In this case, the video would always
 * occupy the whole window.  This is probably fine, since the window will not
 * be used to display anything else.
 *  */

static char xVideoInRcsid[] = "$Header: /plx/source/Xclients/generic/RCS/XVideoIn.c,v 0.65 1996/01/25 04:03:52 drapeau Exp $";

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "XVideoInP.h"
#include <unistd.h>
#include <sys/time.h>
#include <stdio.h>
#include <sys/types.h>

#define offset(field) XtOffsetOf(XVideoInRec, xVideoIn.field)

static XtResource resources[] =
{				/* {name, class, type, size, offset, default_type, default_addr}, */
{
  XtNbrightness,
  XtCBrightness,
  XtRInt,
  sizeof(int),
  offset(brightness),
  XtRImmediate,
  (XtPointer) DefaultBrightness
  },
{
  XtNcontrast,
  XtCContrast,
  XtRInt,
  sizeof(int),
  offset(contrast),
  XtRImmediate,
  (XtPointer) DefaultContrast
  },
{
  XtNhue,
  XtCHue,
  XtRInt,
  sizeof(int),
  offset(hue),
  XtRImmediate,
  (XtPointer) DefaultHue
  },
{
  XtNsaturation,
  XtCSaturation,
  XtRInt,
  sizeof(int),
  offset(saturation),
  XtRImmediate,
  (XtPointer) DefaultSaturation
  },
{
  XtNcropped,
  XtCCropped,
  XtRBoolean,
  sizeof(Boolean),
  offset(cropped),
  XtRImmediate,
  (XtPointer) False
  },
{
  XtNnormalAspectRatio,
  XtCNormalAspectRatio,
  XtRBoolean,
  sizeof(Boolean),
  offset(normalAspectRatio),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNsourceRect,
  XtCSourceRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(sourceRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNdisplayRect,
  XtCDisplayRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(displayRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNlive,
  XtCLive,
  XtRBoolean,
  sizeof(Boolean),
  offset(live),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNnewFrame,
  XtCNewFrame,
  XtRBoolean,
  sizeof(Boolean),
  offset(newFrame),
  XtRImmediate,
  (XtPointer) False
  },
{
  XtNsaveColorSettings,
  XtCSaveColorSettings,
  XtRInt,
  sizeof(int),
  offset(saveColorSettings),
  XtRImmediate,
  (XtPointer) FACTORY
  },
{
  XtNrestoreColorSettings,
  XtCRestoreColorSettings,
  XtRInt,
  sizeof(int),
  offset(restoreColorSettings),
  XtRImmediate,
  (XtPointer) FACTORY
  },
{
  XtNsignalStandard,
  XtCSignalStandard,
  XtRInt,
  sizeof(int),
  offset(signalStandard),
  XtRImmediate,
  (XtPointer) PLX_NTSC
  },
{
  XtNsignalFormat,
  XtCSignalFormat,
  XtRInt,
  sizeof(int),
  offset(signalFormat),
  XtRImmediate,
  (XtPointer) PLX_COMP
  },
{
  XtNhardwareType,
  XtCHardwareType,
  XtRInt,
  sizeof(int),
  offset(hardwareType),
  XtRImmediate,
  (XtPointer) PLX_XVIDEO_24
  },
{
  XtNinterlaced,
  XtCInterlaced,
  XtRBoolean,
  sizeof(Boolean),
  offset(interlaced),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNsync,
  XtCSync,
  XtRBoolean,
  sizeof(Boolean),
  offset(sync),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNsignalWidth,
  XtCSignalWidth,
  XtRDimension,
  sizeof(Dimension),
  offset(signalWidth),
  XtRImmediate,
  (XtPointer) 640
  },
{
  XtNsignalHeight,
  XtCSignalHeight,
  XtRDimension,
  sizeof(Dimension),
  offset(signalHeight),
  XtRImmediate,
  (XtPointer) 480
  },
{
  XtNblankIntervalHeight,
  XtCBlankIntervalHeight,
  XtRDimension,
  sizeof(Dimension),
  offset(blankIntervalHeight),
  XtRImmediate,
  (XtPointer) 45
  },
{
  XtNshowBlankInterval,
  XtCShowBlankInterval,
  XtRBoolean,
  sizeof(Boolean),
  offset(showBlankInterval),
  XtRImmediate,
  (XtPointer) False
  },
{
  XtNchannel,
  XtCChannel,
  XtRInt,
  sizeof(int),
  offset(channel),
  XtRImmediate,
  (XtPointer) PLX_INPUT_0
  },
{
  XtNvideoParameterString,
  XtCVideoParameterString,
  XtRString,
  sizeof(String),
  offset(videoParameterString),
  XtRImmediate,
  (XtPointer) "COLOR"
  },
{
  XtNresizeParent,
  XtCResizeParent,
  XtRBoolean,
  sizeof(Boolean),
  offset(resizeParent),
  XtRImmediate,
  (XtPointer) False
  },
{
  XtNdiagnostics,
  XtCDiagnostics,
  XtRBoolean,
  sizeof(Boolean),
  offset(diagnostics),
  XtRImmediate,
  (XtPointer) False
  },
{
  XtNhardwareParams,
  XtCHardwareParams,
  XtRPointer,
  sizeof(plx_IO*),
  offset(hardwareParams),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNinstance,
  XtCInstance,
  XtRInt,
  sizeof(int),
  offset(instance),
  XtRImmediate,
  (XtPointer) 0
  },
{
  XtNautoSearch,
  XtCAutoSearch,
  XtRInt,
  sizeof(int),
  offset(autoSearch),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNdrawMethod,
  XtCDrawMethod,
  XtRPointer,
  sizeof(XtPointer),
  offset(drawMethod),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNgc,
  XtCGC,
  XtRPointer,
  sizeof(XtPointer),
  offset(gc),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNstillImage,
  XtCStillImage,
  XtRPointer,
  sizeof(XtPointer),
  offset(stillImage),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNstillDelay,
  XtCStillDelay,
  XtRInt,
  sizeof(int),
  offset(stillDelay),
  XtRImmediate,
  (XtPointer) 100
  },
};

#undef offset

static void		XVideoInAction(/* Widget, XEvent*, String*, Cardinal* */);

static XtActionsRec actions[] =
{
{"xVideoIn",	XVideoInAction},				    /* {name, procedure}, */
};

static char translations[] =
  "<Key>:		xVideoIn()	\n";

/* Declaration of methods */

static void 		Initialize(Widget, Widget, ArgList, Cardinal*);
static void		Realize(Widget, XtValueMask*, XSetWindowAttributes*);
static void 		Redisplay(Widget, XEvent*, Region);
static void	 	Destroy(Widget);
static void 		Resize(Widget);
static Boolean		SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void		FindBestVideoSourceMatch(XVideoInWidget, plx_IO_list*);
static void		AdjustSourceWidthAndHeight(XVideoInWidget);
static void		AdjustDisplayWidthAndHeight(XVideoInWidget);
static void		AdjustCoreWidthAndHeight(XVideoInWidget, Dimension*, Dimension*);
static void		SetColorValues(XVideoInWidget, XVideoInWidget);
static void		PrintWidgetDiagnostics(XVideoInWidget, char*);
static void		PreventIllegalWidgetChanges(XVideoInWidget, XVideoInWidget);
static void		SaveOrRestoreColorSettings(XVideoInWidget, XVideoInWidget);
static XtGeometryResult QueryGeometry(Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static Boolean		CheckForValidVideoParameterString(String);
static void		CreateVideoPixmap(XVideoInWidget widget);
static void		RegisterXVideoInWidgetEventHandler(XVideoInWidget);
static void		XVideoInWidgetEventHandler(Widget, XtPointer, XEvent*, Boolean*);
static void		GetVideoPixmap(XVideoInWidget, XRectangle*);
static void		GetVideoImage(XVideoInWidget, XRectangle*);
static void		ShowStillImage(XVideoInWidget, XRectangle*);
static void		DestroyVideoStill(XVideoInWidget);


XVideoInClassRec xVideoInClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"XVideoIn",
    /* widget_size		*/	sizeof(XVideoInRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	TRUE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	QueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* xVideoIn fields */
    /* empty			*/	0
  }
};

WidgetClass xVideoInWidgetClass = (WidgetClass)&xVideoInClassRec;



void PrintPlxInputList(plx_IO_list*	inputs)
{
  plx_signal_list*	signalList = (plx_signal_list*) NULL;
  plx_format_list*	formatList = (plx_format_list*) NULL;
  
  printf("PrintInputList(): Channel is %d, Bundle is %d.\n",
	 inputs->channel,
	 inputs->bundle);
  for (signalList = inputs->signal_list;
       signalList != NULL;
       signalList = signalList->next_signal)
  {
    printf("Signal standard: ");
    switch (signalList->signal->standard)
    {
     case PLX_NTSC:
      printf("NTSC (%d)\t", signalList->signal->standard);
      break;
     case PLX_PAL:
      printf("PAL (%d)\t", signalList->signal->standard);
      break;
     case PLX_SECAM:
      printf("SECAM (%d)\t", signalList->signal->standard);
      break;
    } /* end switch on the signal standard */
    printf("Width:%d\tHeight:%d\tInterlaced:%d\n",
	   signalList->signal->w,
	   signalList->signal->h,
	   signalList->signal->interlaced);
  } /* end for looking at each signal on the list */
  for (formatList = inputs->format_list;
       formatList != NULL;
       formatList = formatList->next_format)
  {
    printf("Signal format: ");
    switch (formatList->format)
    {
     case PLX_COMP:
      printf("Composite (%d)", formatList->format);
      break;
     case PLX_YC:
      printf("YC (%d)", formatList->format);
      break;
     case PLX_YUV:
      printf("YUV (%d)", formatList->format);
      break;
     case PLX_RGB:
      printf("RGB (%d)", formatList->format);
      break;
    } /* end switch format */
    printf(" instance %d\n", formatList->instance);
  } /* end looking at each format */
  return;
} /* end function PrintPlxInputList */



static void PrintPlxHardwareList(plx_IO* ioList)
{
  plx_IO_list*	inputs;
  
  printf("Entered PrintPlxHardwareList.\n");
  if ((plx_IO*)NULL == ioList)
  {
    printf("The hardware list is empty; no Plx hardware was found.\n");
    return;
  }
  for (inputs = ioList->inputs;
       inputs != NULL;
       inputs = inputs->next_IO)
  {
    PrintPlxInputList(inputs);
  }
  printf("Exiting PrintPlxHardwareList.\n\n");
  return;
} /* end function PrintPlxHardwareList */



static void XVideoInWidgetDebug(void)				    /* Used to make it easy to tell stupid debuggers what... */
{								    /* ...file to switch to for debugging. */
  return;
}



static void Initialize(Widget		treq,
		       Widget		tnew,
		       ArgList		args,
		       Cardinal*	numArgs)
{
  int			brightness, contrast, hue, saturation;
  int			screen;
  int			screenNumber;
  GC			tempGC;
  Display*		display = XtDisplay(tnew);
  XRectangle*		theRect = (XRectangle*)NULL;
  XVisualInfo*		visualInfo;
  XVideoInWidget	newWidget = (XVideoInWidget) tnew;
  plx_IO_list*		inputList = (plx_IO_list*) NULL;
  
  XVideoInWidgetDebug();
  screenNumber = XScreenNumberOfScreen(XtScreen(tnew));
  visualInfo = XPlxGetVideoVisual(screenNumber, display);	    /* Try to find the best TrueColor visual possible */
  if (visualInfo == (XVisualInfo*)NULL)				    /* Did the search for a TrueColor visual fail? */
    return;							    /* Yes, return without doing anything else */
  else								    /* The search succeeded; set this widget's depth to that... */
    newWidget->core.depth = visualInfo->depth;			    /* ...of the best TrueColor visual found. */
  screen = XDefaultScreen(display);
  newWidget->xVideoIn.eventHandlerInstalled = False;		    /* By default, no event handler need be installed */
  newWidget->xVideoIn.videoWindow =
    XCreateSimpleWindow(display,				    /* Create a window; this is necessary for subsequent... */
			RootWindow(display, screen),		    /* ...calls to XPlx extensions, which take a Drawable... */
			0, 0,					    /* ...as argument.  This window will be released when... */
			1, 1,					    /* ...this function exits, since it is not needed for... */
			1,					    /* ...anything outside the scope of this function. */
			BlackPixel(display, screen),
			BlackPixel(display, screen));
  tempGC = XDefaultGC(display,					    /* Create a temporary GC; also necessary for subsequent... */
		      XDefaultScreen(display));			    /* ...calls to XPlx extensions.  This will be freed... */
								    /* ...at the end of this function, when no longer needed. */
  newWidget->xVideoIn.gc = tempGC;				    /* Make the widget's GC point to this temporary GC for now */
  newWidget->xVideoIn.startingY = 0;				    /* Initialize this variable to a reasonable default value */
  newWidget->xVideoIn.stillImagePixmap = (XImage*)NULL;		    /* Tell the widget that there's no valid Pixmap yet */
  newWidget->xVideoIn.pixmapExists = False;			    /* No still video Pixmap has been created for this widget */
  newWidget->xVideoIn.hardwareParams =				    /* Determine the type of hardware being used and get... */
    XPlxQueryConfig(display, newWidget->xVideoIn.videoWindow, tempGC); /* ...other relevant information about it. */
/*
  PrintPlxHardwareList(newWidget->xVideoIn.hardwareParams);
*/
  if ((plx_IO*)NULL == newWidget->xVideoIn.hardwareParams)	    /* If hardwareParams is NULL, no hardware was found, so... */
    inputList = (plx_IO_list*)NULL;				    /* ...indicate that no inputs were found. */
  else								    /* Inputs were found */
  {
    for (inputList = newWidget->xVideoIn.hardwareParams->inputs;    /* Search through all available inputs for this hardware,... */
	 inputList != (plx_IO_list*) NULL;			    /* ...looking for the input channel being requested. */
	 inputList = inputList->next_IO)
    {
      if (inputList->channel == newWidget->xVideoIn.channel)	    /* Was the requested channel found? */
      {								    /* Yes, determine of this channel can supply requested... */
	FindBestVideoSourceMatch(newWidget, inputList);		    /* ...video format */
	break;							    /* No need to search any longer */
      }
    }
  } /* end else (inputs were found) */
  if (inputList == (plx_IO_list*) NULL)				    /* Was the requested channel found? */
  {
    newWidget->xVideoIn.channel = PLX_IO_OFF;			    /* No, reset input channel resource */
    newWidget->xVideoIn.sync = False;				    /* Also, indicate there's no sync for requested channel */
    newWidget->xVideoIn.interlaced = True;
    newWidget->xVideoIn.signalWidth = 640;
    newWidget->xVideoIn.signalHeight = 525 - 43;		    /* NTSC height minus NTSC blanking interval height */
    newWidget->xVideoIn.blankIntervalHeight = 43;
    newWidget->xVideoIn.signalFormat = PLX_NTSC;
  }
  newWidget->xVideoIn.defaultSourceRect.x = 0;			    /* Set default size of rectangle to digitize from the... */
  newWidget->xVideoIn.defaultSourceRect.y = 0;			    /* ...incoming video signal */
  newWidget->xVideoIn.defaultSourceRect.width =
    newWidget->xVideoIn.signalWidth;
  if (newWidget->xVideoIn.showBlankInterval == True)		    /* Take into account the blanking interval when setting... */
    newWidget->xVideoIn.defaultSourceRect.height =		    /* ...the default height of the default source rectangle. */
      newWidget->xVideoIn.signalHeight +
	newWidget->xVideoIn.blankIntervalHeight;
  else
    newWidget->xVideoIn.defaultSourceRect.height =
      newWidget->xVideoIn.signalHeight;
  if (newWidget->xVideoIn.sourceRect != (XRectangle*)NULL)	    /* Did the application programmer specify a sourceRect... */
  {								    /* ...from which to digitize an incoming video signal? */
    if (newWidget->xVideoIn.sourceRect->x <= 0)			    /* Yes, assure that the values specified are valid */
      newWidget->xVideoIn.sourceRect->x = 0;
    if (newWidget->xVideoIn.sourceRect->y <= 0)
      newWidget->xVideoIn.sourceRect->y = 0;
    if ((newWidget->xVideoIn.sourceRect->width <= 0) ||		    /* Make sure that width is valid: not less than zero, and... */
	(newWidget->xVideoIn.sourceRect->width >		    /* ...not greater than the height of the incoming video... */
	 newWidget->xVideoIn.signalWidth))			    /* ...signal */
      newWidget->xVideoIn.sourceRect->width =
	newWidget->xVideoIn.signalWidth;
    if ((newWidget->xVideoIn.sourceRect->height <= 0) ||	    /* Make sure that height is valid: not less than zero,... */
	(newWidget->xVideoIn.sourceRect->height >		    /* ...and not greater than the height of the incoming... */
	 newWidget->xVideoIn.defaultSourceRect.height))		    /* ...video signal (the defaultSourceRect.height field... */
      newWidget->xVideoIn.sourceRect->height =			    /* ...at this point has been correctly calculated to... */
	newWidget->xVideoIn.defaultSourceRect.height;		    /* ...account for blank interval height, so use it). */
    theRect = newWidget->xVideoIn.sourceRect;
  }
  else
    theRect = &(newWidget->xVideoIn.defaultSourceRect);
  newWidget->xVideoIn.defaultDisplayRect.x = 0;			    /* Set default size of rectangle used to display the... */
  newWidget->xVideoIn.defaultDisplayRect.y = 0;			    /* ...incoming video signal */
  newWidget->xVideoIn.defaultDisplayRect.width = theRect->width;
  newWidget->xVideoIn.defaultDisplayRect.height = theRect->height;
  if (newWidget->xVideoIn.displayRect != (XRectangle*)NULL)	    /* Did the application programmer specify a displayRect... */
  {								    /* ...for displaying an incoming video signal? */
    if (newWidget->xVideoIn.displayRect->x <= 0)		    /* Yes, assure that the values specified are valid */
      newWidget->xVideoIn.displayRect->x = 0;
    if (newWidget->xVideoIn.displayRect->y <= 0)
      newWidget->xVideoIn.displayRect->y = 0;
    if ((newWidget->xVideoIn.displayRect->width <= 0) ||	    /* Make sure that width is valid: not less than zero, and... */
	(newWidget->xVideoIn.displayRect->width >		    /* ...not greater than the width of the rectangle... */
	 theRect->width))					    /* ...digitizing the incoming video signal. */
      newWidget->xVideoIn.displayRect->width = theRect->width;
    if ((newWidget->xVideoIn.displayRect->height <= 0) ||	    /* Make sure that height is valid: not less than zero,... */
	(newWidget->xVideoIn.displayRect->height >		    /* ...and not greater than the height of the rectangle... */
	 theRect->height))					    /* ...digitizing the incoming video signal. */
      newWidget->xVideoIn.displayRect->height = theRect->height;
  }
  PrintWidgetDiagnostics(newWidget, "Initialize, before setting core width & depth.");
  AdjustSourceWidthAndHeight(newWidget);			    /* Make sure that sourceWidth&Height fields are valid */
  if (newWidget->xVideoIn.sourceRect != (XRectangle*)NULL)
  {
    newWidget->core.width =					    /* Calculate default width and height appropriate for... */
      (Dimension)newWidget->xVideoIn.sourceRect->width;		    /* ...showing video for the current format. */
    newWidget->core.height =
      (Dimension)newWidget->xVideoIn.sourceRect->height;
  }
  else
  {
    newWidget->core.width = (Dimension)newWidget->xVideoIn.defaultSourceRect.width;
    newWidget->core.height = (Dimension)newWidget->xVideoIn.defaultSourceRect.height;
  }
  XPlxVideoValueQuery(display, newWidget->xVideoIn.videoWindow,	    /* Get current video configuration values and store them... */
		      tempGC,					    /* ...for possible later use in this function. */
		      &saturation, &contrast, &hue, &brightness);
  
  if (CheckForValidVideoParameterString(newWidget->xVideoIn.videoParameterString) == False)
  {
    XtWarning("XVideoIn Widget: videoParameterString was not an acceptable value.");
    newWidget->xVideoIn.videoParameterString = "COLOR";
  }
  
  if ((newWidget->xVideoIn.brightness < 0) ||			    /* Assure that initial values are within accepted parameters */
      (newWidget->xVideoIn.brightness > 255))
  {
    XtWarning("XVideoIn Widget: brightness must be between \
      	      0 and 255.  Using currently-stored NVRAM value.");
    newWidget->xVideoIn.brightness = brightness;
  }
  
  if ((newWidget->xVideoIn.contrast < 0) ||
      (newWidget->xVideoIn.contrast > 255))
  {
    XtWarning("XVideoIn Widget: contrast must be between \
      	      0 and 255.  Using currently-stored NVRAM value.");
    newWidget->xVideoIn.contrast = contrast;
  }
  
  if ((newWidget->xVideoIn.hue < 0) ||
      (newWidget->xVideoIn.hue > 255))
  {
    XtWarning("XVideoIn Widget: hue must be between \
      	      0 and 255.  Using currently-stored NVRAM value.");
    newWidget->xVideoIn.hue = hue;
  }
  
  if ((newWidget->xVideoIn.saturation < 0) ||
      (newWidget->xVideoIn.saturation > 255))
  {
    XtWarning("XVideoIn Widget: saturation must be between \
      	      0 and 255.  Using currently-stored NVRAM value.");
    newWidget->xVideoIn.saturation = saturation;
  }
  return;
}								    /* end function Initialize */



static void Realize(Widget			w,
		    XtValueMask*		valueMask,
		    XSetWindowAttributes*	attributes)
{
  int			screenNumber;
  char			tempString[256];
  Display*		display = XtDisplay(w);
  XVisualInfo*		theVisual;
  XVideoInWidget	newWidget = (XVideoInWidget) w;
  Colormap		colormap;
  
  XDestroyWindow(display, newWidget->xVideoIn.videoWindow);	    /* Free up temporary window previously created */
  screenNumber = XScreenNumberOfScreen(XtScreen(newWidget));
  theVisual = XPlxGetVideoVisual(screenNumber, display);	    /* Try to get the best TrueColor visual for this widget */
  if (theVisual == (XVisualInfo*)NULL)				    /* Did the search for an appropriate visual fail? */
  {								    /* Yes, report the failure and exit the program immediately */
    XtWarning("XVideoIn Widget: Could not find the TrueColor \
		visual being sought.");
    exit(-1);
  }
  colormap = XCreateColormap(display,				    /* Create a default colormap to install for this window. */
			     XtWindow(XtParent(newWidget)),	    /* See O'Reilly Xlib Reference Manual entry for... */
			     theVisual->visual,			    /* ..."XCreateWindow()" to learn why a default colormap... */
			     AllocNone);			    /* ...needs to be created at all. */
  
  attributes->colormap = colormap;				    /* Set necessary attributes for a 24-bit video window.  ... */
  attributes->background_pixmap = None;				    /* ...Absent from this set of attributes are the... */
  attributes->background_pixel = BlackPixel(display, screenNumber); /* ..."border_pixmap" and "border_pixel" attributes; I... */
  attributes->bit_gravity = ForgetGravity;			    /* ...don't know why they can't seem to be set here.  ... */
  attributes->win_gravity = NorthWestGravity;			    /* ...Using the values recommended by the O'Reilly book... */
  attributes->backing_store = NotUseful;			    /* ...(Volume 1) of "None" and "BlackPixel(etc.)"... */
  attributes->backing_pixel = BlackPixel(display, screenNumber);    /* ...causes a "BadMatch" protocol error at run-time.  ... */
  attributes->save_under = False;				    /* ...However, these initial values seem to be sufficient. */
  *valueMask |= (CWColormap | CWBackPixmap | CWBackPixel |
		 CWBitGravity |CWWinGravity |
		 CWBackingStore | CWBackingPixel |
		 CWSaveUnder);
  
  XtCreateWindow((Widget)newWidget, InputOutput, theVisual->visual, /* Create the window into which video will be displayed */
		 *valueMask, attributes);
  newWidget->xVideoIn.videoWindow = XtWindow(newWidget);	    /* Point to the permanent window for this widget */
  newWidget->xVideoIn.gc =					    /* Create a graphics context just for this widget; later... */
    XCreateGC(display,						    /* ...this GC will be tagged to indicate that this window... */
	      newWidget->xVideoIn.videoWindow,			    /* ...is a video window. */
	      0, 0);
  XPlxVideoTag(display,						    /* Indicate that this window is a video window */
	       XtWindow(newWidget),
	       newWidget->xVideoIn.gc,
	       PLX_VIDEO);
  XPlxVideoInputSelect(display, XtWindow(newWidget),		    /* Configure the video input hardware to show the... */
		       newWidget->xVideoIn.gc,			    /* ...requested video signal */
		       newWidget->xVideoIn.channel,
		       newWidget->xVideoIn.signalStandard,
		       newWidget->xVideoIn.signalFormat,
		       newWidget->xVideoIn.instance);
  XPlxVideoValueLoad(display, XtWindow(newWidget),		    /* Configure this video window with the correct color values */
		     newWidget->xVideoIn.gc,
		     newWidget->xVideoIn.saturation,
		     newWidget->xVideoIn.contrast,
		     newWidget->xVideoIn.hue,
		     newWidget->xVideoIn.brightness);
  if ((strcmp(newWidget->xVideoIn.videoParameterString,		    /* Was the requested video parameter an output function? */
	      "GENLOCK") == 0) ||
      (strcmp(newWidget->xVideoIn.videoParameterString,
	      "FREERUN") == 0))
  {								    /* Yes, format the correct string */
    sprintf(tempString, "output_config %s ",
	    newWidget->xVideoIn.videoParameterString);
  }
  else								    /* No, the requested video parameter was an input function */
  {
  sprintf(tempString, "input_config %s ",
	  newWidget->xVideoIn.videoParameterString);
  }
  XPlxSetVideoParameterString(display, XtWindow(newWidget),
			      newWidget->xVideoIn.gc,
			      tempString);
  PrintWidgetDiagnostics(newWidget, "Realize");
  RegisterXVideoInWidgetEventHandler(newWidget);		    /* Look for window move events if on an overlay product */
  return;
}								    /* end function Realize */



static void Redisplay(Widget	w,
		      XEvent*	event,
		      Region	region)
{
  XVideoInWidget	theWidget = (XVideoInWidget) w;
  XRectangle*		sourceRect;
  XRectangle		displayRect;
  
  if (!XtIsRealized(w))						    /* Don't bother trying to draw if the widget is not... */
    return;							    /* ...or has not been realized */
  if ((theWidget->xVideoIn.live == True) &&			    /* Was live video being requested but no sync was present? */
      (theWidget->xVideoIn.sync == False) &&
      (theWidget->xVideoIn.autoSearch != False))		    /* If autoSearch is off, always try to show live video */
    return;							    /* If this is so, don't bother trying to continue */
  if (theWidget->xVideoIn.displayRect != (XRectangle*)NULL)	    /* Did the application programmer choose to explicitly... */
  {								    /* ...describe a display rectangle for showing video? */
    displayRect.x = theWidget->xVideoIn.displayRect->x;		    /* Yes, use the application programmer's values for... */
    displayRect.y = theWidget->xVideoIn.displayRect->y;		    /* display of the video signal in the X window, instead... */
    displayRect.width = theWidget->xVideoIn.displayRect->width;	    /* ...of the default display values. */
    displayRect.height = theWidget->xVideoIn.displayRect->height;
  }
  else								    /* No, the application programmer did not explicitly... */
  {								    /* ...choose display coordinates; use widget-specific... */
    displayRect.x = 0;						    /* ...defaults instead. */
    displayRect.y = 0;
    displayRect.width = (unsigned short)theWidget->core.width;
    displayRect.height = (unsigned short)theWidget->core.height;
  }
  if (theWidget->xVideoIn.sourceRect != (XRectangle*)NULL)	    /* Did the application programmer choose to explicitly... */
  {								    /* ...describe a source rectangle for digitizing video? */
    sourceRect = theWidget->xVideoIn.sourceRect;		    /* Yes, use the application programmer's values for... */
  }								    /* ...digitization of the video signal from the source. */
  else								    /* No, the application programmer did not explicitly... */
  {								    /* ...choose source coordinates; use widget-specific... */
    sourceRect = &theWidget->xVideoIn.defaultSourceRect;
  }
  if (displayRect.width > sourceRect->width)			    /* The size of the display rectangle may not be larger... */
    displayRect.width = sourceRect->width;			    /* ...than the size of the source rectangle in either... */
  if (displayRect.height > sourceRect->height)			    /* ...dimension.  Enforce these constraints. */
    displayRect.height = sourceRect->height;
/*  XClearWindow(XtDisplay(theWidget), XtWindow(theWidget));*/	    /* Make sure there is no video residue after resizing */
  if (theWidget->xVideoIn.live == True)				    /* Should the video be live? */
  {								    /* Yes */
    if (theWidget->xVideoIn.cropped == True)
      XPlxVideoLive(XtDisplay(theWidget),			    /* Yes, display live, cropped video in the widget's window */
		    XtWindow(theWidget),
		    theWidget->xVideoIn.gc,
		    (int)sourceRect->x,
		    (int)theWidget->xVideoIn.startingY,
		    (int)displayRect.x,
		    (int)displayRect.y,
		    (int)displayRect.width,
		    (int)displayRect.height);
    else
      XPlxVideoSqueezeLive(XtDisplay(theWidget),		    /* No, show scaled, live video in the widget's window */
			   XtWindow(theWidget),
			   theWidget->xVideoIn.gc,
			   (int)sourceRect->x,
			   (int)theWidget->xVideoIn.startingY,
			   (int)sourceRect->width,
			   (int)sourceRect->height,
			   (int)displayRect.x,
			   (int)displayRect.y,
			   (int)displayRect.width,
			   (int)displayRect.height);
  } /* end if live == True */
  else								    /* No, video is still; determine whether to create a new... */
  {								    /* ...Pixmap or not, then display whatever is appropriate. */
    if (theWidget->xVideoIn.pixmapExists == True)		    /* Is there a still image to be used for display? */
      ShowStillImage(theWidget, &displayRect);			    /* Yes, display it by whatever means are appropriate */
    else							    /* No, there is no image to use; still the video then... */
    {								    /* ...create a new still image for next time */
      if (theWidget->xVideoIn.cropped == True)			    /* Should the video be cropped?  Yes */
	XPlxVideoStill(XtDisplay(theWidget),			    /* Display still, cropped video in the widget's window */
		       XtWindow(theWidget),
		       theWidget->xVideoIn.gc,
		       (int)sourceRect->x,
		       (int)theWidget->xVideoIn.startingY,
		       (int)displayRect.x,
		       (int)displayRect.y,
		       (int)displayRect.width,
		       (int)displayRect.height);
      else							    /* No, the video should be squeezed, not cropped */
	XPlxVideoSqueezeStill(XtDisplay(theWidget),
			      XtWindow(theWidget),
			      theWidget->xVideoIn.gc,
			      (int)sourceRect->x,
			      (int)theWidget->xVideoIn.startingY,
			      (int)sourceRect->width,
			      (int)sourceRect->height,
			      (int)displayRect.x,
			      (int)displayRect.y,
			      (int)displayRect.width,
			      (int)displayRect.height);
      XPlxSleep(theWidget->xVideoIn.stillDelay * 1000);		    /* Sleep for a while, possibly, before trying to grab... */
      CreateVideoPixmap(theWidget);				    /* ...the image from the digitized video. */
    } /* end else (no Pixmap has been created) */
  } /* end else (video is still, not live) */
  if (theWidget->xVideoIn.drawMethod != (XtExposeProc)NULL)	    /* Did the app pgmr specify a custom drawing method? */
  {								    /* Yes, call the pgmr's drawing method to draw over video */
    (*theWidget->xVideoIn.drawMethod)(w, event, region);	    /* Perform the method call */
  }
  return;
}								    /* end function Redisplay */


static void Destroy(Widget w)
{
  XVideoInWidget	theWidget = (XVideoInWidget) w;
  
  XPlxVideoStop(XtDisplay(theWidget),				    /* Just to make sure, stop video right now before... */ 
		theWidget->xVideoIn.videoWindow,		    /* ...destroying the window and widget. */
		theWidget->xVideoIn.gc);
  XFreeGC(XtDisplay(theWidget), theWidget->xVideoIn.gc);	    /* Destroy the GC created for this video widget */
  DestroyVideoStill(theWidget);					    /* Free resources for any still image that might be in use */
  if (theWidget->xVideoIn.eventHandlerInstalled == True)	    /* If this widget is running on an overlay product then... */
  {								    /* ...it has an event handler looking for window move... */
    XtRemoveEventHandler(theWidget->xVideoIn.eventHandlerWidget,    /* ...events.  Remove it before destroying the widget. */
			 StructureNotifyMask,
			 False,
			 XVideoInWidgetEventHandler,
			 (XtPointer)theWidget);
  }
  return;
}								    /* end function Destroy */


static void Resize(Widget w)
{
  XVideoInWidget	theWidget = (XVideoInWidget) w;
  Dimension		desiredWidth, desiredHeight;
  Dimension		compromiseWidth, compromiseHeight;
  XRectangle*		displayRect = (XRectangle*)NULL;
  XtGeometryResult	result;
  XWindowAttributes	attributes;
  
  if (theWidget->xVideoIn.displayRect != (XRectangle*)NULL)	    /* Did the programmer specify a rectangle in which to... */
    displayRect = theWidget->xVideoIn.displayRect;		    /* ...display video?  If so, point to it; if not, use the... */
  else								    /* ...default display rectangle. */
  {
    displayRect = &(theWidget->xVideoIn.defaultDisplayRect);
    displayRect->width = theWidget->core.width;			    /* If the default displayRect is being used, make the... */
    displayRect->height = theWidget->core.height;		    /* ...displayRect's size track the size of the widget. */
  }
  AdjustSourceWidthAndHeight(theWidget);			    /* Adjust to account for incoming video signal size */
  AdjustDisplayWidthAndHeight(theWidget);			    /* Make sure that display parameters are also acceptable */
  AdjustCoreWidthAndHeight(theWidget, &desiredWidth, &desiredHeight);
  if ((desiredWidth != theWidget->core.width) ||		    /* Was width or height of this widget modified after... */
      (desiredHeight != theWidget->core.height))		    /* ...conforming to size constraints? */
  {
    result = XtMakeResizeRequest((Widget)theWidget,		    /* Now that the parent has been resized, try to resize... */
				 desiredWidth,			    /* ...this widget. */
				 desiredHeight,
				 &compromiseWidth,
				 &compromiseHeight);
    if (result == XtGeometryAlmost)				    /* The requested size was not quite acceptable; a... */
    {								    /* ...compromise was suggested; accept this compromise. */
      result = XtMakeResizeRequest((Widget)theWidget,		    /* Ask for the compromise size */
				   compromiseWidth,
				   compromiseHeight,
				   NULL, NULL);
    }
    if (theWidget->xVideoIn.resizeParent == True)		    /* Did the app. programmer request the parent widget be... */
    {								    /* ...resized whenever this widget is? */
      XGetWindowAttributes(XtDisplay(theWidget),		    /* Determine the border_width of this widget's parent */
			   XtWindow(XtParent((Widget)theWidget)),
			   &attributes);
      XtResizeWidget(XtParent((Widget)theWidget),		    /* Break the rules by trying to explicitly resize this... */
		     desiredWidth, desiredHeight,		    /* ...widget's parent. */
		     attributes.border_width);
    }								    /* end (if ... resizeParent == True...) */
  }								    /* end (if widthOrHeightChanged == True...) */
  return;
}								    /* end function Resize */



static Boolean SetValues(Widget		current,
			 Widget		request,
			 Widget		new,
			 ArgList	args,
			 Cardinal*	numArgs)
{
  XVideoInWidget	currentWidget = (XVideoInWidget) current;
  XVideoInWidget	newWidget = (XVideoInWidget) new;
  plx_signal*		tempSignal;
  char			tempString[256];
  Boolean		redisplayNeeded = False;
  plx_IO_list*		inputList = (plx_IO_list*) NULL;
  
  PreventIllegalWidgetChanges(currentWidget, newWidget);	    /* Don't let read-only fields be modified by the application */
  
  if ((newWidget->xVideoIn.channel !=				    /* Was a new input requested? */
       currentWidget->xVideoIn.channel) ||
      (newWidget->xVideoIn.signalStandard !=			    /* Was a new signal standard requested? */
       currentWidget->xVideoIn.signalStandard) ||
      (newWidget->xVideoIn.instance !=				    /* Was a new "instance" (i.e., input cable) requested? */
       currentWidget->xVideoIn.instance) ||
      (newWidget->xVideoIn.signalFormat !=			    /* Was a new signal format requested? */
       currentWidget->xVideoIn.signalFormat))
  {								    /* Yes, see if the request can be satisfied */
    if (newWidget->xVideoIn.channel == PLX_IO_OFF)		    /* Should this widget turn off video? */
    {
      XPlxVideoStop(XtDisplay(newWidget),			    /* Yes, forcibly disable drawing video */
		    newWidget->xVideoIn.videoWindow,
		    newWidget->xVideoIn.gc);
      newWidget->xVideoIn.sync = False;				    /* There's no video to draw on channel "PLX_IO_OFF" */
    }
    else							    /* No, a valid video input channel was chosen, try it out */
    {
/*
   PrintPlxHardwareList(newWidget->xVideoIn.hardwareParams);
*/
      if ((plx_IO*)NULL == newWidget->xVideoIn.hardwareParams)	    /* If hardwareParams is NULL, no hardware was found, so... */
	inputList = (plx_IO_list*)NULL;				    /* ...indicate that no inputs were found. */
      else							    /* Inputs were found */
      {
	for (inputList = newWidget->xVideoIn.hardwareParams->inputs; /* Search all available inputs for this hardware,... */
	     inputList != (plx_IO_list*) NULL;			    /* ...looking for the input channel being requested. */
	     inputList = inputList->next_IO)
	{
	  if (inputList->channel == newWidget->xVideoIn.channel)    /* Was the requested channel found? */
	  {							    /* Yes, determine if this channel can supply requested... */
	    FindBestVideoSourceMatch(newWidget, inputList);	    /* ...video format */
	    if (newWidget->xVideoIn.sync == True)		    /* Was video found on the new requested channel? */
	      redisplayNeeded = True;				    /* Yes, tell the widget a redraw is needed */
	    break;						    /* No need to search any longer */
	  }
	}
      }	/* end else (inputs were found) */
      if ((inputList == (plx_IO_list*) NULL) ||			    /* Did the search for the requested channel fail, OR... */
	  ((newWidget->xVideoIn.sync == False) &&		    /* ...was there no active video on the new channel when... */
	   (newWidget->xVideoIn.autoSearch != False)))		    /* ...the widget is searching for the 1st good signal?  */
      {								    /* That's right: either the requested channel wasn't... */
	newWidget->xVideoIn.channel =				    /* ...found or it *was* found but no active video on it. */
	  currentWidget->xVideoIn.channel;			    /* Revert the widget to its original settings: channel,... */
	newWidget->xVideoIn.signalStandard =			    /* ...signal standard,... */
	  currentWidget->xVideoIn.signalStandard;
	newWidget->xVideoIn.signalFormat =			    /* ...and signal format. */
	  currentWidget->xVideoIn.signalFormat;
	newWidget->xVideoIn.instance =
	  currentWidget->xVideoIn.instance;
	XPlxVideoInputSelect(XtDisplay(newWidget),		    /* Yes, try to re-configure the video input hardware to... */
			     newWidget->xVideoIn.videoWindow,	    /* ...show the newly-requested video signal. */
			     newWidget->xVideoIn.gc,
			     newWidget->xVideoIn.channel,
			     newWidget->xVideoIn.signalStandard,
			     newWidget->xVideoIn.signalFormat,
			     newWidget->xVideoIn.instance);
	tempSignal = XPlxQueryVideo(XtDisplay(newWidget),	    /* Re-check if sync was found on the original video input */
				    newWidget->xVideoIn.videoWindow,
				    newWidget->xVideoIn.gc);
	newWidget->xVideoIn.sync = tempSignal->sync_ok;		    /* Update the status of video sync for this widget */
	free(tempSignal);					    /* Free space no longer needed by this application */
      } /* end (if inputList is NULL or sync is False) */
    } /* end (videoIn.channel was not PLX_IO_OFF) */
  } /* end (if channel, standard, or format where changed) */
  if (newWidget->xVideoIn.sourceRect !=				    /* Did the source rectangle change? */
      currentWidget->xVideoIn.sourceRect)
    redisplayNeeded = True;					    /* Yes, redraw the widget with video */
  if (newWidget->xVideoIn.displayRect !=			    /* Did the destination rectangle change? */
      currentWidget->xVideoIn.displayRect)
    redisplayNeeded = True;					    /* Yes, redraw the widget with video */
  AdjustSourceWidthAndHeight(newWidget);			    /* Assure that video area is within accepted ranges */
  if (newWidget->xVideoIn.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
  {								    /* Yes, do the necessary math */
    newWidget->core.height = (newWidget->core.width * 3) / 4;	    /* Also resize the widget to conform to the aspect ratio */
  }
  SetColorValues(currentWidget, newWidget);			    /* Set hue, saturation, brightness, contrast, if necessary */
  SaveOrRestoreColorSettings(currentWidget, newWidget);		    /* Load/save to/from factory defaults, or user0, user1 NVRAM */
  if (newWidget->xVideoIn.videoParameterString !=		    /* Did app. programmer request a new "special effect"? */
      currentWidget->xVideoIn.videoParameterString)		    /* Yes, see if it's a valid one. */
    if (CheckForValidVideoParameterString(newWidget->xVideoIn.videoParameterString) == False)
    {
      XtWarning("XVideoIn Widget: videoParameterString was not an acceptable value.");
      newWidget->xVideoIn.videoParameterString =
	currentWidget->xVideoIn.videoParameterString;
    }
    else
    {
      if ((strcmp(newWidget->xVideoIn.videoParameterString,	    /* Was the requested video parameter an output function? */
		  "GENLOCK") == 0) ||
	  (strcmp(newWidget->xVideoIn.videoParameterString,
		  "FREERUN") == 0))
      {								    /* Yes, format the correct string */
	sprintf(tempString, "output_config %s ",
		newWidget->xVideoIn.videoParameterString);
      }
      else							    /* No, the requested video parameter was an input function */
      {
	sprintf(tempString, "input_config %s ",
		newWidget->xVideoIn.videoParameterString);
      }
      XPlxSetVideoParameterString(XtDisplay(newWidget),
				  newWidget->xVideoIn.videoWindow,
				  newWidget->xVideoIn.gc,
				  tempString);
      redisplayNeeded = True;
    }
  if (newWidget->xVideoIn.cropped !=
      currentWidget->xVideoIn.cropped)
    redisplayNeeded = True;
  
  if (newWidget->xVideoIn.live !=				    /* Did the programmer change the widget from still to... */
      currentWidget->xVideoIn.live)				    /* ...live (or from live to still, either way)? */
  {								    /* Yes, check which and take appropriate action */
    if (newWidget->xVideoIn.live == False)			    /* Change was made from live to still */
    {
      DestroyVideoStill(newWidget);				    /* Free resources for any still image that might be in use */
    } /* ...be created. */
    redisplayNeeded = True;					    /* In any case, it's time to redraw the widget's window */
  }
  
  if ((newWidget->xVideoIn.newFrame !=				    /* Was the newFrame resource set? */
      currentWidget->xVideoIn.newFrame) &&
      (newWidget->xVideoIn.newFrame == True))
  {
    newWidget->xVideoIn.newFrame = False;			    /* Reset this resource so it can be called again */
    if (newWidget->xVideoIn.live == True)			    /* Setting the "newFrame" resource to grab a new frame of... */
    {								    /* ...video makes no sense if live video is already running. */
      XtWarning("XVideoIn Widget: newFrame cannot be set unless live is set to False.");
    }
    else							    /* Still video is in place as expected; get rid of the... */
    {								    /* ...old Pixmap so a new still frame will be grabbed. */
      DestroyVideoStill(newWidget);				    /* Free resources for any still image that might be in use */
      Redisplay((Widget)newWidget, (XEvent*)NULL, (Region)NULL);    /* ...image immediately. */
    }
  }								    /* end if newFrame resource was set */
  if ((newWidget->xVideoIn.resizeParent !=			    /* Did the application programmer specify that this... */
       currentWidget->xVideoIn.resizeParent) &&			    /* ...widget should try to resize its parent whenever... */
      (newWidget->xVideoIn.resizeParent == True))		    /* ...it is resized? */
    redisplayNeeded = True;					    /* Yes, it is probably necessary to redraw the widget */
  if ((newWidget->xVideoIn.stillImage !=			    /* Did the application programmer request a copy of the... */
       currentWidget->xVideoIn.stillImage) &&			    /* ...widget's currently-stored still image?  And, is... */
      (newWidget->xVideoIn.stillImage != (XImage**)NULL) &&	    /* ...the widget currently in still mode? */
      (True == newWidget->xVideoIn.live))
    redisplayNeeded = True;					    /* Yes, indicate that redisplay is needed so image can... */
								    /* ...be copied. */
  if (redisplayNeeded == True)					    /* Was a parameter set that requires an Expose event to... */
  {
    return(True);						    /* ...be generated?  If so, return True. */
  }
  else
  {
    return(False);						    /* No need to cause an Expose event for these changes */
  }
}								    /* end function SetValues */



static XtGeometryResult QueryGeometry(Widget		w,
				      XtWidgetGeometry*	proposed,
				      XtWidgetGeometry*	answer)
{
  XVideoInWidget	widget = (XVideoInWidget)w;
  XRectangle*		theRect = (XRectangle*)NULL;
  
  if (widget->xVideoIn.sourceRect != (XRectangle*)NULL)		    /* Base geometry management decisions on the size of the... */
    theRect = widget->xVideoIn.sourceRect;			    /* ...rectangle for the video input window, whether... */
  else								    /* ...specified by the application programmer or using... */
    theRect = &widget->xVideoIn.defaultSourceRect;		    /* ...the widget-specific default rectangle. */
  answer->request_mode = (CWWidth | CWHeight);			    /* Only concerned about width and height parameters */
  if (widget->xVideoIn.normalAspectRatio == True)		    /* Account for possible 4:3 aspect ratio when determining... */
  {								    /* ...the preferred size of this widget. */
    answer->height = (answer->width * 3) / 4;
  }
  else
    answer->height = theRect->height;
  answer->width = theRect->width;
  if (((proposed->request_mode & (CWWidth | CWHeight)) ==	    /* Was a request made only for a width or height change,... */
       (CWWidth | CWHeight)) &&					    /* ...and was the requested width and/or height exactly... */
      (proposed->width == answer->width) &&			    /* ...what this widget would prefer anyway? */
      (proposed->height == answer->height))
    return(XtGeometryYes);					    /* Yes, return an exact "yes" result */
  
  if ((answer->width == widget->core.width) &&
      (answer->height == widget->core.height))
    return(XtGeometryNo);					    /* No changes necessary; the request was fine as-is */
  else
    return(XtGeometryAlmost);
}								    /* end function QueryGeometry */


static void XVideoInAction(Widget	w,
			   XEvent*	event,
			   String*	params,
			   Cardinal*	numParams)
{
  return;							    /* Right now this function does nothing */
}								    /* end function XVideoInAction */



static Boolean AutoSearchVideoSources(XVideoInWidget	theWidget,
				      plx_IO_list*	inputList)
{
  plx_format_list*	formatList = (plx_format_list*) NULL;
  plx_signal_list*	signalList = (plx_signal_list*) NULL;
  plx_signal*		theSignal;
  int			numValidMatches = 0;
  int			descriptorTableSize = 0;
  struct timeval	waitTime;				    /* Used to tell select() how long to wait until it returns */
  
  descriptorTableSize = sysconf(_SC_OPEN_MAX);			    /* Get size of descriptor table for later use by select() */
  waitTime.tv_sec = 0;						    /* Set up a timeval to wait for 500 usecs, giving the... */
  waitTime.tv_usec = 500000;					    /* ...hardware time to lock in the new signal. */
  for (signalList = inputList->signal_list;			    /* Search all configurations for the widget's current... */
       signalList != (plx_signal_list*) NULL;			    /* ...channel to try and find an active video signal. */
       signalList = signalList->next_signal)
  {
    for (formatList = inputList->format_list;
	 formatList != (plx_format_list*) NULL;
	 formatList = formatList->next_format)
    {
      if (True == theWidget->xVideoIn.diagnostics)
      {
	printf("AutoSearchVideoSources: 1st Pass: Selecting channel std fmt inst %d %d %d %d.\n",
	       inputList->channel,
	       signalList->signal->standard,
	       formatList->format,
	       formatList->instance);
      }
      XPlxVideoInputSelect(XtDisplay(theWidget),		    /* Try to re-configure the video input hardware to show... */
			   theWidget->xVideoIn.videoWindow,	    /* ...the newly-requested video signal. */
			   theWidget->xVideoIn.gc,
			   inputList->channel,
			   signalList->signal->standard,
			   formatList->format,
			   formatList->instance);
      theSignal = XPlxQueryVideo(XtDisplay(theWidget),		    /* Is there is active video on this configured... */
				 theWidget->xVideoIn.videoWindow,   /* ...video channel? */
				 theWidget->xVideoIn.gc);
      if (True == theSignal->sync_ok)				    /* Was a valid sync signal detected? */
      {
	if (True == theWidget->xVideoIn.diagnostics)
	  printf("\t\tVoted Yes on sync.\n");
	numValidMatches++;					    /* Yes, (probably), "vote" once for positive sync */
      }
      else
      {
	if (True == theWidget->xVideoIn.diagnostics)
	  printf("\t\tVoted No on sync.\n");
      }
      if (True == theSignal->sync_ok)				    /* If a positive result, vote again */
      {
	select(descriptorTableSize,				    /* Perform the wait (not really listening for input on... */
	       0, 0, 0,						    /* ...any file descriptors). */
	       &waitTime);
	if (True == theWidget->xVideoIn.diagnostics)
	{
	  printf("AutoSearchVideoSources: 2nd Pass: channel std fmt inst %d %d %d %d.\n",
		 inputList->channel,
		 signalList->signal->standard,
		 formatList->format,
		 formatList->format);
	}
	free(theSignal);					    /* Free memory allocated for signal information */
	theSignal = XPlxQueryVideo(XtDisplay(theWidget),	    /* Is there is active video on this configured... */
				   theWidget->xVideoIn.videoWindow, /* ...video channel? */
				   theWidget->xVideoIn.gc);
	if (True == theSignal->sync_ok)				    /* Was a valid sync signal detected? */
	{
	  if (True == theWidget->xVideoIn.diagnostics)
	    printf("\t\tVoted Yes on sync.\n");
	  numValidMatches++;					    /* Yes, (probably), "vote" once for positive sync */
	}
	else
	{
	  if (True == theWidget->xVideoIn.diagnostics)
	    printf("\t\tVoted No on sync.\n");
	}
	if (True == theSignal->sync_ok)				    /* If a positive result, vote a third time */
	{
	  select(descriptorTableSize,				    /* Perform the wait (not really listening for input on... */
		 0, 0, 0,					    /* ...any file descriptors). */
		 &waitTime);
	  if (True == theWidget->xVideoIn.diagnostics)
	  {
	    printf("AutoSearchVideoSources: 3rd Pass: channel std fmt inst %d %d %d %d.\n",
		   inputList->channel,
		   signalList->signal->standard,
		   formatList->format,
		   formatList->format);
	  }
	  free(theSignal);					    /* Free memory allocated for signal information */
	  theSignal = XPlxQueryVideo(XtDisplay(theWidget),	    /* Is there is active video on this configured... */
				     theWidget->xVideoIn.videoWindow, /* ...video channel? */
				     theWidget->xVideoIn.gc);
	  if (True == theSignal->sync_ok)			    /* Was a valid sync signal detected? */
	  {
	    if (True == theWidget->xVideoIn.diagnostics)
	      printf("\t\tVoted Yes on sync.\n");
	    numValidMatches++;					    /* Yes, (probably), "vote" once for positive sync */
	  }
	  else
	  {
	    if (True == theWidget->xVideoIn.diagnostics)
	      printf("\t\tVoted No on sync.\n");
	  }
	} /* end if 2nd pass returned positive */
      }	/* end if 1st pass returned positive */
      if (numValidMatches > 2)					    /* If all 3 out of 3 votes were "yes", guess that sync... */
	theSignal->sync_ok = True;				    /* ...found on this input was really there. */
      else
	theSignal->sync_ok = False;
      numValidMatches = 0;					    /* Regardless, reset the voting tabulator for next signal */
      if (theSignal->sync_ok == True) 				    /* Yes, there is active video on this channel */
      {								    /* Set up the widget's new configuration */
	theWidget->xVideoIn.channel = inputList->channel;	    /* Update the channel on which a valid signal was found */
	theWidget->xVideoIn.sync = theSignal->sync_ok;
	theWidget->xVideoIn.interlaced = signalList->signal->interlaced;
	theWidget->xVideoIn.instance = formatList->instance;
	theWidget->xVideoIn.signalWidth =
	  (Dimension)signalList->signal->w;
	theWidget->xVideoIn.signalHeight =			    /* Height of viewable portion of signal is total height... */
	  (Dimension)(signalList->signal->h -			    /* ...minus height of blanking interval. */
		      signalList->signal->b);
	theWidget->xVideoIn.blankIntervalHeight =
	  (Dimension)signalList->signal->b;
	theWidget->xVideoIn.signalStandard =			    /* Tell the widget to change its notion of the current... */
	  signalList->signal->standard;				    /* ...signal standard being used (NTSC, PAL, SECAM). */
	theWidget->xVideoIn.signalFormat = formatList->format;	    /* Tell the widget the format found on this channel */
	theWidget->xVideoIn.defaultSourceRect.width =		    /* Now that a signal has been found, set appropriate... */
	  theWidget->xVideoIn.signalWidth;			    /* ...width and height for the default input... */
	if (theWidget->xVideoIn.showBlankInterval == True)	    /* ...rectangle (it is not the policy of this widget to... */
	  theWidget->xVideoIn.defaultSourceRect.height =	    /* ...adjust any sourceRect rectangle that may have been... */
	    theWidget->xVideoIn.signalHeight +			    /* ...specified by the application programmer). */
	      theWidget->xVideoIn.blankIntervalHeight;
	else
	  theWidget->xVideoIn.defaultSourceRect.height =
	    theWidget->xVideoIn.signalHeight;
	PrintWidgetDiagnostics(theWidget, "AutoSearchVideoSources");
	free(theSignal);					    /* Free space allocated for signal information */
	return(True);						    /* Return a successful result */
      }	/* end (if theSignal...) */
      free(theSignal);						    /* Free space allocated for signal information */
    } /* end (for formatList...) */
  } /* end (for signalList...) */
  return(False);
} /* end function AutoSearchVideoSources */



static Boolean AutoSearchInputs(XVideoInWidget	widget)		    /* This function looks at all possible inputs for the... */
{								    /* ...hardware being used. */
  plx_IO_list*	inputList;
  Boolean	result;
  
  if (True == widget->xVideoIn.diagnostics)
  {
    printf("XVideoIn Widget: Entered AutoSearchInputs.\n");
  }
  if ((plx_IO*)NULL == widget->xVideoIn.hardwareParams)		    /* If hardwareParams is NULL, no hardware was found, so... */
    return(False);						    /* indicate that no valid signal was found. */
  for (inputList = widget->xVideoIn.hardwareParams->inputs;	    /* Look at each digitizer or cable group, one-by-one */
       inputList != NULL;
       inputList = inputList->next_IO)
  {
    if (True == widget->xVideoIn.diagnostics)
    {
      printf("Channel: %d\n", inputList->channel);
    }
    if ((AutoSearchInput0 == widget->xVideoIn.autoSearch) &&	    /* Is autoSearch set to look only for the 1st channel? */
	(inputList->channel != PLX_INPUT_0))			    /* And, if so, is this the wrong channel to be searching? */
      continue;							    /* Yes, move on to the next input on the list */
    if ((AutoSearchInput1 == widget->xVideoIn.autoSearch) &&	    /* Is autoSearch set to look only for the 2nd channel? */
	(inputList->channel != PLX_INPUT_1))			    /* And, if so, is this the wrong channel to be searching? */
      continue;							    /* Yes, move on to the next input on the list */
    result = AutoSearchVideoSources(widget, inputList);		    /* Determine if there's a valid signal on this digitizer */
    if (True == result)						    /* Is there a valid signal? */
      return(True);						    /* Yes, don't bother looking anymore */
  }
  return(False);
} /* end function AutoSearchInputs */



static void FindBestVideoSourceMatch(XVideoInWidget	theWidget,
				     plx_IO_list*	inputList)
{
  int			descriptorTableSize = 0;
  Bool			signalExists;
  plx_signal*		theSignal;
  struct timeval	waitTime;				    /* Used to tell select() how long to wait until it returns */
  int			numValidMatches = 0;
  
  signalExists =						    /* See if the signal being requested actually exists on... */
    XPlxFindMatchingSignal(theWidget->xVideoIn.hardwareParams,	    /* ...this hardware. */
			   theWidget->xVideoIn.channel,
			   theWidget->xVideoIn.signalStandard,
			   theWidget->xVideoIn.signalFormat,
			   theWidget->xVideoIn.instance);
  if (signalExists == True)					    /* Does the requested signal exist? */
  {								    /* Yes, try to select it and see what happens */
    XPlxVideoInputSelect(XtDisplay(theWidget),			    /* Yes, try to re-configure the video input hardware to... */
			 theWidget->xVideoIn.videoWindow,	    /* ...show the newly-requested video signal. */
			 theWidget->xVideoIn.gc,
			 theWidget->xVideoIn.channel,
			 theWidget->xVideoIn.signalStandard,
			 theWidget->xVideoIn.signalFormat,
			 theWidget->xVideoIn.instance);
    XFlush(XtDisplay(theWidget));				    /* Make sure the input selection request goes through */
    descriptorTableSize = sysconf(_SC_OPEN_MAX);		    /* Get size of descriptor table for later use by select() */
    waitTime.tv_sec = 0;					    /* Set up a timeval to wait for 500 msecs, giving the... */
    waitTime.tv_usec = 500000;					    /* ...hardware time to lock in the new signal. */
    select(descriptorTableSize,					    /* Perform the wait (not really listening for input on... */
	   0, 0, 0,						    /* ...any file descriptors). */
	   &waitTime);
    theSignal = XPlxQueryVideo(XtDisplay(theWidget),		    /* Is there is active video on this configured... */
			       theWidget->xVideoIn.videoWindow,	    /* ...video channel? */
			       theWidget->xVideoIn.gc);
    if (True == theSignal->sync_ok)				    /* Was a valid sync signal detected? */
    {
      if (True == theWidget->xVideoIn.diagnostics)
	printf("\t\tFindBest...  Voted Yes on sync.\n");
      numValidMatches++;					    /* Yes, (probably), "vote" once for positive sync */
    }
    else
    {
      if (True == theWidget->xVideoIn.diagnostics)
	printf("\t\tFindBest...  Voted No on sync.\n");
    }
    if (True == theSignal->sync_ok)				    /* If a positive result, vote a second time */
    {
      XFlush(XtDisplay(theWidget));				    /* Make sure the input selection request goes through */
      descriptorTableSize = sysconf(_SC_OPEN_MAX);		    /* Get size of descriptor table for later use by select() */
      waitTime.tv_sec = 0;					    /* Set up a timeval to wait for 500 msecs, giving the... */
      waitTime.tv_usec = 500000;				    /* ...hardware time to lock in the new signal. */
      select(descriptorTableSize,				    /* Perform the wait (not really listening for input on... */
	     0, 0, 0,						    /* ...any file descriptors). */
	     &waitTime);
      free(theSignal);						    /* Free memory allocated for signal information */
      theSignal = XPlxQueryVideo(XtDisplay(theWidget),		    /* Is there is active video on this configured... */
				 theWidget->xVideoIn.videoWindow,   /* ...video channel? */
				 theWidget->xVideoIn.gc);
      if (True == theSignal->sync_ok)				    /* Was a valid sync signal detected? */
      {
	if (True == theWidget->xVideoIn.diagnostics)
	  printf("\t\tFindBest...  Voted Yes on sync.\n");
	numValidMatches++;					    /* Yes, (probably), "vote" once for positive sync */
      }
      else
      {
	if (True == theWidget->xVideoIn.diagnostics)
	  printf("\t\tFindBest...  Voted No on sync.\n");
      }
      if (True == theSignal->sync_ok)				    /* If a positive result, vote a third time */
      {
	XFlush(XtDisplay(theWidget));				    /* Make sure the input selection request goes through */
	select(descriptorTableSize,				    /* Perform the wait (not really listening for input on... */
	       0, 0, 0,						    /* ...any file descriptors). */
	       &waitTime);
	free(theSignal);					    /* Free memory allocated for signal information */
	theSignal = XPlxQueryVideo(XtDisplay(theWidget),	    /* Is there is active video on this configured... */
				   theWidget->xVideoIn.videoWindow, /* ...video channel? */
				   theWidget->xVideoIn.gc);
	if (True == theSignal->sync_ok)				    /* Was a valid sync signal detected? */
	{
	  if (True == theWidget->xVideoIn.diagnostics)
	    printf("\t\tFindBest...  Voted Yes on sync.\n");
	  numValidMatches++;					    /* Yes, (probably), "vote" once for positive sync */
	}
	else
	{
	  if (True == theWidget->xVideoIn.diagnostics)
	    printf("\t\tFindBest...  Voted No on sync.\n");
	}
      } /* end if 2nd pass returned positive */
    } /* end if 1st pass returned positive */
    if (numValidMatches > 2)					    /* If all 3 out of 3 votes were "yes", guess that sync... */
      theSignal->sync_ok = True;				    /* ...found on this input was really there. */
    else
      theSignal->sync_ok = False;
    if ((theSignal->sync_ok == True) ||				    /* Yes, there is active video on this channel */
	(theWidget->xVideoIn.autoSearch == False))		    /* Or the widget is told to forcibly show what was selected */
    {								    /* Set up the widget's new configuration */
      theWidget->xVideoIn.sync = theSignal->sync_ok;
      theWidget->xVideoIn.interlaced = theSignal->interlaced;
      theWidget->xVideoIn.signalWidth = theSignal->w;
      theWidget->xVideoIn.signalHeight =			    /* Height of viewable portion of signal is total height... */
	(Dimension)(theSignal->h - theSignal->b);		    /* ...minus height of blanking interval */
      theWidget->xVideoIn.blankIntervalHeight =
	(Dimension)theSignal->b;
      theWidget->xVideoIn.defaultSourceRect.width =		    /* Now that a signal has been found, set appropriate... */
	theWidget->xVideoIn.signalWidth;			    /* ...width and height for the default input... */
      if (theWidget->xVideoIn.showBlankInterval == True)	    /* ...rectangle (it is not the policy of this widget to... */
	theWidget->xVideoIn.defaultSourceRect.height =		    /* ...adjust any sourceRect rectangle that may have been... */
	  theWidget->xVideoIn.signalHeight +			    /* ...specified by the application programmer). */
	    theWidget->xVideoIn.blankIntervalHeight;
      else
	theWidget->xVideoIn.defaultSourceRect.height =
	  theWidget->xVideoIn.signalHeight;
      PrintWidgetDiagnostics(theWidget, "FindBestVideoSourceMatch, Succeeding");
      free(theSignal);						    /* Free space allocated for signal information */
      return;							    /* Leave this function without doing anymore work */
    } /* end (if theSignal...) */
    free(theSignal);						    /* Free space allocated for signal information */
  } /* end (if signalExists...) */
  if (theWidget->xVideoIn.autoSearch != False)			    /* Should the widget try to find an alternate valid input... */
  {								    /* ...signal? */
    if (AutoSearchInputs(theWidget) == True)			    /* Yes, requested signal failed, try to find *any* signal... */
      return;							    /* ...that has video; if one is found, return immediately. */
  }
  theWidget->xVideoIn.sync = False;				    /* If this stmt is reached, no good video signal was found */
  theWidget->xVideoIn.defaultSourceRect.width =			    /* Now that a signal has been found, set appropriate... */
    theWidget->xVideoIn.signalWidth;				    /* ...width and height for the default input... */
  if (theWidget->xVideoIn.showBlankInterval == True)		    /* ...rectangle (it is not the policy of this widget to... */
    theWidget->xVideoIn.defaultSourceRect.height =		    /* ...adjust any sourceRect rectangle that may have been... */
      theWidget->xVideoIn.signalHeight +			    /* ...specified by the application programmer). */
      theWidget->xVideoIn.blankIntervalHeight;
  else
    theWidget->xVideoIn.defaultSourceRect.height =
      theWidget->xVideoIn.signalHeight;
  return;
}								    /* end function FindBestVideoSourceMatch */



static void AdjustSourceWidthAndHeight(XVideoInWidget widget)
{
  Dimension	videoHeight = 0;
  XRectangle*	theRect;
  
  if (widget->xVideoIn.sourceRect != (XRectangle*)NULL)
    theRect = widget->xVideoIn.sourceRect;
  else
    theRect = &(widget->xVideoIn.defaultSourceRect);
  
  if ((theRect->x + theRect->width) >				    /* Make sure that the rectangle to be digitized still... */
      widget->xVideoIn.signalWidth)				    /* ...falls into the area denoted by "signalWidth" */
    theRect->width =						    /* Clip the digitized rectangle if necessary to make it fit */
      widget->xVideoIn.signalWidth - theRect->x;
  
  videoHeight = widget->xVideoIn.signalHeight +			    /* Height here is signal height plus height of blanking... */
    widget->xVideoIn.blankIntervalHeight;			    /* ...interval information */
  
  if (widget->xVideoIn.showBlankInterval == False)		    /* Determine the height of the video signal; it depends... */
    widget->xVideoIn.startingY =				    /* Adjust starting digitization point to take into... */
      theRect->y +						    /* ...account the height of the blanking interval, but... */
	widget->xVideoIn.blankIntervalHeight;			    /* ...make this correction invisible to the app. programmer. */
  
  if ((widget->xVideoIn.startingY + theRect->height) >		    /* Make sure that the rectangle to be digitized still... */
      videoHeight)						    /* ...falls into the area denoted by the total height... */
    theRect->height = videoHeight -				    /* ...of the video rectangle to be digitized, taking into... */
      theRect->y;						    /* ...account whether to show blanking interval info or not. */
  
  if (widget->xVideoIn.showBlankInterval == False)		    /* Reset "videoHeight" if necessary, to represent the... */
    videoHeight = widget->xVideoIn.signalHeight;		    /* ...height of the incoming video signal. */
  
  if (widget->xVideoIn.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
    theRect->height = (theRect->width * 3) / 4;			    /* Yes, do the necessary math */
  return;
}								    /* end function AdjustSourceWidthAndHeight */




static void AdjustDisplayWidthAndHeight(XVideoInWidget widget)
{
  XRectangle*	sourceRect;
  XRectangle*	displayRect;
  
  if (widget->xVideoIn.displayRect != (XRectangle*)NULL)
    displayRect = widget->xVideoIn.displayRect;
  else
    displayRect = &(widget->xVideoIn.defaultDisplayRect);
  if (widget->xVideoIn.sourceRect != (XRectangle*)NULL)
    sourceRect = widget->xVideoIn.sourceRect;
  else
    sourceRect = &(widget->xVideoIn.defaultSourceRect);
  
  if (displayRect->width > sourceRect->width)			    /* Make sure that the destination rectangle (the one that... */
    displayRect->width = sourceRect->width;			    /* ...describes where to display video) is smaller than... */
  if (displayRect->height > sourceRect->height)			    /* ...the source video rectangle, as per the constraints... */
    displayRect->height = sourceRect->height;			    /* ...defined by the XPlxVideoSqueezeLive() function. */
  if (widget->xVideoIn.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
  {
    displayRect->height = (displayRect->width * 3) / 4;		    /* Yes, do the necessary math */
  }
  return;
}								    /* end function AdjustDisplayWidthAndHeight */



static void AdjustCoreWidthAndHeight(XVideoInWidget	widget,
				     Dimension*		preferredWidth,
				     Dimension*		preferredHeight)
{
  XRectangle*		sourceRect;
  unsigned short	width;
  unsigned short	height;
  
  width = widget->core.width;
  height = widget->core.height;
  if (widget->xVideoIn.sourceRect != (XRectangle*)NULL)
    sourceRect = widget->xVideoIn.sourceRect;
  else
    sourceRect = &(widget->xVideoIn.defaultSourceRect);
  if (width > sourceRect->width)				    /* Make sure that the widget is smaller than the source... */
    width = sourceRect->width;					    /* ...the source video rectangle, as per the constraints... */
  if (height > sourceRect->height)				    /* ...defined by the XPlxVideoSqueezeLive() function. */
    height = sourceRect->height;
  if (widget->xVideoIn.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
  {								    /* Yes, do the necessary math */
    height = (width * 3) / 4;
  }
  *preferredWidth = (Dimension)width;
  *preferredHeight = (Dimension)height;
  PrintWidgetDiagnostics(widget, "AdjustCoreWidthAndHeight.");
  return;
}								    /* end function AdjustCoreWidthAndHeight */



static void SetColorValues(XVideoInWidget currentWidget,
			   XVideoInWidget newWidget)
{
  Boolean	colorValuesChanged = False;
  
  if (newWidget->xVideoIn.brightness !=				    /* Was a new value requested?  And, is it valid? */
      currentWidget->xVideoIn.brightness)
  {
    if ((newWidget->xVideoIn.brightness < 0) ||
	(newWidget->xVideoIn.brightness > 255))
    {
      XtWarning("XVideoIn Widget: brightness must be between 0 and 255.  \
		Using currently-stored value.");
      newWidget->xVideoIn.brightness = currentWidget->xVideoIn.brightness;
    }
    else
      colorValuesChanged = True;
  }
  
  if (newWidget->xVideoIn.contrast !=				    /* Was a new value requested?  And, is it valid? */
      currentWidget->xVideoIn.contrast)
  {
    if ((newWidget->xVideoIn.contrast < 0) ||
	(newWidget->xVideoIn.contrast > 255))
    {
      XtWarning("XVideoIn Widget: contrast must be between 0 and 255.  \
		Using currently-stored value.");
      newWidget->xVideoIn.contrast = currentWidget->xVideoIn.contrast;
    }
    else
      colorValuesChanged = True;
  }
  
  if (newWidget->xVideoIn.saturation !=				    /* Was a new value requested?  And, is it valid? */
      currentWidget->xVideoIn.saturation)
  {
    if ((newWidget->xVideoIn.saturation < 0) ||
	(newWidget->xVideoIn.saturation > 255))
    {
      XtWarning("XVideoIn Widget: saturation must be between 0 and 255.  \
		Using currently-stored value.");
      newWidget->xVideoIn.saturation = currentWidget->xVideoIn.saturation;
    }
    else
      colorValuesChanged = True;
  }
  
  if (newWidget->xVideoIn.hue !=				    /* Was a new value requested?  And, is it valid? */
      currentWidget->xVideoIn.hue)
  {
    if ((newWidget->xVideoIn.hue < 0) ||
	(newWidget->xVideoIn.hue > 255))
    {
      XtWarning("XVideoIn Widget: hue must be between 0 and 255.  \
		Using currently-stored value.");
      newWidget->xVideoIn.hue = currentWidget->xVideoIn.hue;
    }
    else
      colorValuesChanged = True;
  }
  if (colorValuesChanged == True)				    /* Did any color values *really* change? */
    XPlxVideoValueLoad(XtDisplay(newWidget),			    /* Set video control values according to new widget settings */
		       newWidget->xVideoIn.videoWindow,
		       newWidget->xVideoIn.gc,
		       newWidget->xVideoIn.saturation,
		       newWidget->xVideoIn.contrast,
		       newWidget->xVideoIn.hue,
		       newWidget->xVideoIn.brightness);
  return;
}								    /* end function SetColorValues */



static void PrintWidgetDiagnostics(XVideoInWidget	widget,
				   char*		calledFromWhere)
{
  XRectangle*	theRect = (XRectangle*)NULL;
  
  if (widget->xVideoIn.diagnostics == False)			    /* Was the appropriate resource set to print diagnostics... */
    return;							    /* ...about this widget?  If not, return right now. */
  printf("XVideoIn Widget: In function %s, values are these:\n",
	 calledFromWhere);
  printf("Channel:\t\t");
  switch (widget->xVideoIn.channel)
  {
   case PLX_IO_OFF:
    printf("Off.\n");
    break;
   case PLX_INPUT_0:
    printf("Input 0.\n");
    break;
   case PLX_INPUT_1:
    printf("Input 1.\n");
    break;
   default:
    printf("%d (error).\n", widget->xVideoIn.channel);
    break;
  }
  printf("Format:\t\t\t");
  switch (widget->xVideoIn.signalFormat)
  {
   case PLX_COMP:
    printf("Composite %d.\n", widget->xVideoIn.instance);
    break;
   case PLX_YC:
    printf("YC %d.\n", widget->xVideoIn.instance);
    break;
   case PLX_YUV:
    printf("YUV %d.\n", widget->xVideoIn.instance);
    break;
   case PLX_RGB:
    printf("RGB %d.\n", widget->xVideoIn.instance);
    break;
   default:
    printf("%d (error).\n", widget->xVideoIn.signalFormat);
    break;
  }
  printf("Standard:\t\t");
  switch (widget->xVideoIn.signalStandard)
  {
   case PLX_NTSC:
    printf("NTSC.\n");
    break;
   case PLX_PAL:
    printf("PAL.\n");
    break;
   case PLX_SECAM:
    printf("SECAM.\n");
    break;
   default:
    printf("%d (error).\n", widget->xVideoIn.signalStandard);
    break;
  }
  if (widget->xVideoIn.sourceRect != (XRectangle*)NULL)
  {
    printf("Using programmer-supplied source rectangle.\n");
    theRect = widget->xVideoIn.sourceRect;
  }
  else
  {
    printf("Using default source rectangle.\n");
    theRect = &(widget->xVideoIn.defaultSourceRect);
  }
  printf("sourceX:\t\t%d\nsourceY:\t\t%d\nsourceWidth:\t\t%d\nsourceHeight:\t\t%d\n",
	 theRect->x,
	 theRect->y,
	 theRect->width,
	 theRect->height);
  printf("blankIntervalHeight:\t%d\nsignalWidth:\t\t%d\nsignalHeight:\t\t%d\n",
	 widget->xVideoIn.blankIntervalHeight,
	 widget->xVideoIn.signalWidth,
	 widget->xVideoIn.signalHeight);
  printf("widgetWidth:\t\t%d\nwidgetHeight:\t\t%d.\n",
	 widget->core.width, widget->core.height);
  if (widget->xVideoIn.cropped == True)
    printf("cropped:\t\tTrue.\n");
  else
    printf("cropped:\t\tFalse.\n");
  if (widget->xVideoIn.sync == True)
    printf("sync:\t\tTrue.\n");
  else
    printf("sync:\t\tFalse.\n");
  return;
}								    /* end function PrintWidgetDiagnostics */



static void PreventIllegalWidgetChanges(XVideoInWidget current,
					XVideoInWidget new)
{
  if (current->xVideoIn.hardwareType !=
      new->xVideoIn.hardwareType)
  {
    XtWarning("XVideoIn Widget: \"hardwareType\" is a read-only resource.");
    new->xVideoIn.hardwareType = current->xVideoIn.hardwareType;
  }
  if (current->xVideoIn.interlaced !=
      new->xVideoIn.interlaced)
  {
    XtWarning("XVideoIn Widget: \"interlaced\" is a read-only resource.");
    new->xVideoIn.interlaced = current->xVideoIn.interlaced;
  }
  if (current->xVideoIn.sync !=
      new->xVideoIn.sync)
  {
    XtWarning("XVideoIn Widget: \"sync\" is a read-only resource.");
    new->xVideoIn.sync = current->xVideoIn.sync;
  }
  if (current->xVideoIn.signalHeight !=
      new->xVideoIn.signalHeight)
  {
    XtWarning("XVideoIn Widget: \"signalHeight\" is a read-only resource.");
    new->xVideoIn.signalHeight = current->xVideoIn.signalHeight;
  }
  if (current->xVideoIn.signalWidth !=
      new->xVideoIn.signalWidth)
  {
    XtWarning("XVideoIn Widget: \"signalWidth\" is a read-only resource.");
    new->xVideoIn.signalWidth = current->xVideoIn.signalWidth;
  }
  if (current->xVideoIn.blankIntervalHeight !=
      new->xVideoIn.blankIntervalHeight)
  {
    XtWarning("XVideoIn Widget: \"blankIntervalHeight\" is a read-only resource.");
    new->xVideoIn.blankIntervalHeight = current->xVideoIn.blankIntervalHeight;
  }
  if (current->xVideoIn.hardwareParams !=
      new->xVideoIn.hardwareParams)
  {
    XtWarning("XVideoIn Widget: \"hardwareParams\" is a read-only resource.");
    new->xVideoIn.hardwareParams = current->xVideoIn.hardwareParams;
  }
  if (current->xVideoIn.gc !=
      new->xVideoIn.gc)
  {
    XtWarning("XVideoIn Widget: \"gc\" is a read-only resource.");
    new->xVideoIn.gc = current->xVideoIn.gc;
  }
  return;
}								    /* end function PreventIllegalWidgetChanges */



static void SaveOrRestoreColorSettings(XVideoInWidget current,
				       XVideoInWidget new)
{
  if (current->xVideoIn.saveColorSettings !=			    /* Did the programmer request to save current color... */
      new->xVideoIn.saveColorSettings)				    /* ...settings to NVRAM? */
  {								    /* Yep, check if a valid NVRAM value was specified */
    if ((new->xVideoIn.saveColorSettings == USER_0) ||		    /* Was a valid NVRAM value specified? */
	(new->xVideoIn.saveColorSettings == USER_1))
    {
      XPlxVideoValueSave(XtDisplay(current),
			 current->xVideoIn.videoWindow,
			 new->xVideoIn.gc,
			 new->xVideoIn.saveColorSettings);
      new->xVideoIn.saveColorSettings = NoNVRAM;		    /* Set this resource to a dummy value to reset it; that... */
      return;							    /* ...way, next time SetValues() is called, this resource... */
    }								    /* ...will almost surely be different from what the... */
								    /* ...programmer will set it to.  This enables the... */
    else							    /* ...programmer to do things like repeatedly press a... */
    {								    /* ...button to restore color settings from the same NVRAM. */
      XtWarning("XVideoIn Widget: Invalid NVRAM was specified for saving color settings.   Use USER_0 or USER_1.\n");
      new->xVideoIn.saveColorSettings =
	current->xVideoIn.saveColorSettings;
    }
  }
  if (current->xVideoIn.restoreColorSettings !=			    /* Did the programmer request to restore current color... */
      new->xVideoIn.restoreColorSettings)			    /* ...settings from NVRAM or factory defaults? */
  {								    /* Yep, check if a valid NVRAM valid was specified */
    if ((new->xVideoIn.restoreColorSettings == USER_0) ||	    /* Same comments as above for saving. */
	(new->xVideoIn.restoreColorSettings == USER_1) ||
	(new->xVideoIn.restoreColorSettings == FACTORY))
    {
      XPlxVideoValueRecall(XtDisplay(current),
			   current->xVideoIn.videoWindow,
			   new->xVideoIn.gc,
			   new->xVideoIn.restoreColorSettings);
      new->xVideoIn.restoreColorSettings = NoNVRAM;
      XPlxVideoValueQuery(XtDisplay(current),			    /* Update color setting resources for this widget to... */
			  current->xVideoIn.videoWindow,	    /* ...reflect those just restored from NVRAM. */
			  new->xVideoIn.gc,
			  &(new->xVideoIn.saturation),
			  &(new->xVideoIn.contrast),
			  &(new->xVideoIn.hue),
			  &(new->xVideoIn.brightness));
    }
    else
    {
      XtWarning("XVideoIn Widget: Invalid NVRAM was specified for restoring color settings.  Use USER_0, USER_1, or FACTORY.\n");
      new->xVideoIn.restoreColorSettings =
	current->xVideoIn.restoreColorSettings;
    }
  }
  return;
}								    /* end function SaveOrRestoreColorSettings */



static Boolean CheckForValidVideoParameterString(String theString)
{
  if ((strcmp(theString, "INVERTED") == 0) ||			    /* These are the only currently valid parameter strings... */
      (strcmp(theString, "NONINVERTED") == 0) ||		    /* ...acceptable to this widget. */
      (strcmp(theString, "BW") == 0) ||
      (strcmp(theString, "COLOR") == 0) ||
      (strcmp(theString, "GENLOCK") == 0) ||
      (strcmp(theString, "FREERUN") == 0))
    return(True);
  else
    return(False);
}								    /* end function CheckForValidVideoParameterString */



static void GetVideoPixmap(XVideoInWidget 	widget,		    /* Used to grab a still image when integrated framebuffer... */
			   XRectangle*		displayRect)	    /* ...is available.  Faster than using XImages. */
{
  DestroyVideoStill(widget);					    /* Free resources for any still image that might be in use */
  widget->xVideoIn.stillPixmap =				    /* Create a new Pixmap for this widget */
    XCreatePixmap(XtDisplay(widget),
		  widget->core.window,
		  (unsigned int)widget->core.width,
		  (unsigned int)widget->core.height,
		  (unsigned int)widget->core.depth);
  widget->xVideoIn.pixmapExists = True;				    /* Tell the widget that a Pixmap has now been created */
  XCopyArea(XtDisplay(widget),					    /* Copy the video still image to the widget's pixmap */
	    XtWindow(widget),
	    widget->xVideoIn.stillPixmap,
	    widget->xVideoIn.gc,
	    0, 0, displayRect->width, displayRect->height,
	    0, 0);
  return;
} /* end function GetVideoPixmap */



static void GetVideoImage(XVideoInWidget	widget,		    /* Used to grab a still image when using an overlay... */
			  XRectangle*		displayRect)	    /* ...product.  Slower than using Pixmaps. */
{
  DestroyVideoStill(widget);					    /* Free resources for any still image that might be in use */
  widget->xVideoIn.stillImagePixmap =				    /* Create a new XImage for this widget */
    XGetImage(XtDisplay(widget),
	      widget->core.window,
	      0, 0, displayRect->width, displayRect->height,
	      AllPlanes, ZPixmap);
  widget->xVideoIn.pixmapExists = True;				    /* Tell the widget that a Pixmap has now been created */
  return;
} /* end function GetVideoImage */



static void CreateVideoPixmap(XVideoInWidget widget)
{
  XRectangle		displayRect;
  
  if (widget->xVideoIn.displayRect != (XRectangle*)NULL)	    /* Did the application programmer choose to explicitly... */
  {								    /* ...describe a display rectangle for showing video? */
    displayRect.x = widget->xVideoIn.displayRect->x;		    /* Yes, use the application programmer's values for... */
    displayRect.y = widget->xVideoIn.displayRect->y;		    /* display of the video signal in the X window, instead... */
    displayRect.width = widget->xVideoIn.displayRect->width;	    /* ...of the default display values. */
    displayRect.height = widget->xVideoIn.displayRect->height;
  }
  else								    /* No, the application programmer did not explicitly... */
  {								    /* ...choose display coordinates; use widget-specific... */
    displayRect.x = 0;						    /* ...defaults instead. */
    displayRect.y = 0;
    displayRect.width = (unsigned short)widget->core.width;
    displayRect.height = (unsigned short)widget->core.height;
  }
  if ((plx_IO*)NULL == widget->xVideoIn.hardwareParams)		    /* If no hardware was available, return immediately */
    return;
  if (0 != widget->xVideoIn.hardwareParams->overlay)		    /* Is this server an overlay product? */
    GetVideoImage(widget, &displayRect);			    /* ...hardware in use. */
  else
    GetVideoPixmap(widget, &displayRect);
  if (widget->xVideoIn.stillImage != (XImage**)NULL)		    /* Did the app programmer request a copy of the image? */
  {								    /* Yes, attempt to return one */
    *widget->xVideoIn.stillImage =				    /* Create a new XImage for the application programmer */
      XGetImage(XtDisplay(widget),
		widget->core.window,
		0, 0, displayRect.width, displayRect.height,
		AllPlanes, ZPixmap);
    widget->xVideoIn.stillImage = (XImage**)NULL;		    /* Reset the widget's still image pointer so it can be... */
  }								    /* ...ready to hand another one to the app. pgmr. */
  return;
} /* end function CreateVideoPixmap */




static void  RegisterXVideoInWidgetEventHandler(XVideoInWidget newWidget) /* Look for window move events if on... */
{								    /* ...an overlay product. */
  Widget	childWidget;
  Widget	parentWidget;
  
  if ((plx_IO*)NULL == newWidget->xVideoIn.hardwareParams)	    /* If no hardware was available, return immediately */
    return;
  if (0 == newWidget->xVideoIn.hardwareParams->overlay)		    /* Is this server an overlay product? */
    return;							    /* No, no need to track ConfigureNotify events */
  childWidget = (Widget)newWidget;
  parentWidget = XtParent((Widget)newWidget);
  while ((XtIsShell(parentWidget) == False) &&			    /* Keep looking up the widget tree until a Shell widget... */
	 (parentWidget != (Widget)NULL))			    /* ... is found or the top of the widget tree is reached;... */
  {								    /* ...a Shell widget will surely receive... */
    childWidget = parentWidget;					    /* ...ConfigureNotify events. */
    parentWidget = XtParent(childWidget);
  }
  XtAddEventHandler(parentWidget,				    /* ...Seek StructureNotify events, and redraw when received. */
		    StructureNotifyMask,
		    False,					    /* Don't bother looking for non-maskable events */
		    XVideoInWidgetEventHandler,
		    (XtPointer)newWidget);
  newWidget->xVideoIn.eventHandlerInstalled = True;
  newWidget->xVideoIn.eventHandlerWidget = parentWidget;
  return;
} /* end function RegisterXVideoInWidgetEventHandler */



static void XVideoInWidgetEventHandler(Widget		widget,
				       XtPointer	clientData,
				       XEvent*		theEvent,
				       Boolean*		dispatchOtherEventHandlers)
{
  if (theEvent->type != ConfigureNotify)			    /* Don't bother doing anything if the window doesn't move */
    return;
  Redisplay((Widget)clientData, (XEvent*)NULL, (Region)NULL);	    /* Force a redisplay of widget, bypassing normal mechanisms */
  return;
} /* end function XVideoInWidgetEventHandler */



static void ShowStillImage(XVideoInWidget	theWidget,	    /* Display a still image in the method that best fits the... */
			   XRectangle*		displayRect)	    /* ...available hardware. */
{
  if ((plx_IO*)NULL == theWidget->xVideoIn.hardwareParams)	    /* If no hardware was available, return immediately */
    return;
  if (0 != theWidget->xVideoIn.hardwareParams->overlay)		    /* Is this server an overlay product? */
  {
    XPlxXPutImage(XtDisplay(theWidget), XtWindow(theWidget),	    /* Yes, use XImages to put the still into widget's window */
		  theWidget->xVideoIn.gc,
		  theWidget->xVideoIn.stillImagePixmap,
		  0, 0, 0, 0,
		  displayRect->width,
		  displayRect->height);
  }
  else								    /* No, use Pixmaps to put the still (faster than XImages) */
  {
    XCopyArea(XtDisplay(theWidget),
	      theWidget->xVideoIn.stillPixmap,
	      XtWindow(theWidget),
	      theWidget->xVideoIn.gc,
	      0, 0,
	      displayRect->width, displayRect->height,
	      0, 0);
  }
  return;
} /* end function ShowStillImage */



static void DestroyVideoStill(XVideoInWidget widget)		    /* Free resources taken to store the widget's still... */
{								    /* ...image, whether it be a Pixmap or XImage. */
  if (False == widget->xVideoIn.pixmapExists)			    /* If there's no still image to be destroyed, don't... */
    return;							    /* ...bother to try to destroy one. */
  if ((plx_IO*)NULL == widget->xVideoIn.hardwareParams)		    /* If no hardware was available, return immediately */
    return;
  if (0 == widget->xVideoIn.hardwareParams->overlay)		    /* Is this server an overlay product? */
  {								    /* No, this widget is using Pixmaps to store still... */
    XFreePixmap(XtDisplay(widget), widget->xVideoIn.stillPixmap);   /* Yes, free up its resources so a new one can be allocated */
  }
  else
  {								    /* Yes, this widget is using XImages to store still... */
    XDestroyImage(widget->xVideoIn.stillImagePixmap);		    /* ...images; destroy the XImage being used. */
  }
  widget->xVideoIn.pixmapExists = False;			    /* Note that the widget no longer has a still image stored */
  return;
} /* end function DestroyVideoStill */
