/*
 *   Copyright (c) 1993 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* $Header: /plx/source/Xclients/generic/RCS/XVideoOut.c,v 0.24 1996/01/10 09:31:05 drapeau Exp $ */
/* $Log: XVideoOut.c,v $
 * Revision 0.24  1996/01/10  09:31:05  drapeau
 * Fixed lots of minor compiler warnings to sanity-check the module.
 *
 * Revision 0.23  1996/01/10  09:15:28  drapeau
 * The widget now correctly checks if hardwareParams and associated info
 * is non-NULL.  If NULL, that means there is no Plx hardware available
 * on the Display* being used for this connection.
 *
 * Revision 0.22  1995/10/05  04:28:39  drapeau
 * Added significant support for tracking a non-Root window.  The new
 * functionality makes it simple for an application programmer to write
 * an application that can specify any window on the screen and limit the
 * output to the contents of that window.  If the source window is moved
 * or resized, the output widget now moves itself to track the change in
 * geometry.
 *
 * Revision 0.21  1995/09/28  21:56:16  drapeau
 * Enhanced diagnostic messages used in debugging the widget and Xlib
 * layer.
 *
 * Revision 0.20  1995/04/01  01:41:24  drapeau
 * Added support for better diagnostics printing.
 *
 * Revision 0.19  1994/09/14  19:13:41  drapeau
 * Removed the output display generation code from the Realize() method.
 * Output should not actually be generated until the Redisplay() method
 * is called.  It's okay to *select* an output format, just not to
 * display it until Redisplay().
 *
 * Revision 0.18  1994/09/14  03:18:54  drapeau
 * Fixed minor error in the Destroy() method: the widget should call
 * XPlxVideoStop() before changing the output format.
 *
 * Revision 0.17  1994/09/13  08:00:21  drapeau
 * Added significant changes to the widget to handle aspect ratio
 * management.
 *
 * Revision 0.16  1994/08/06  01:16:25  drapeau
 * Added diagnostic functions for printing the list of signal attributes
 * supplied by the hardware.  These diagnostic functions are useful in
 * debugging the widget.
 *
 * Revision 0.15  1994/03/14  04:02:53  drapeau
 * Major changes to fix video output and redisplay.
 *
 * Added a new resource, "outputWindow", a simple X window used solely
 * for the purpose of passing to Plx extension calls.  It is not clear to
 * me that the root window can be used for these calls; John
 * Malleo-Roach's XView-based video output package has a comment that
 * says using the root window does not work.
 *
 * Also, the Redisplay() method was changed; it now checks the output
 * channel used by the widget; if it is PLX_IO_OFF, Redisplay() does not
 * bother calling XPlxVideoScaleOutput().  It seems to me that calling
 * XPlxVideoScaleOutput() should produce no result if a previous call to
 * XPlxVideoOutputSelect() set the channel to PLX_IO_OFF (it seems to me
 * that those semantics mean to disable video output), but evidently the
 * implementation does not follow this rule.  The new workaround does the
 * trick, though.
 *
 * Revision 0.14  1993/12/15  23:28:14  drapeau
 * Several changes all having to do with the blanking interval for the
 * video signal being produced.  The widget had previously been doing bad
 * math, not taking into account the height of the blanking interval for
 * the video signal being produced.  The widget now does this without
 * exposing this accommodation to the application programmer.
 *
 * Also, diagnostic messages were augmented to account for the blanking
 * interval, and an old debugging message was removed now that it is no
 * longer needed.
 *
 * Revision 0.13  1993/08/16  09:51:53  drapeau
 * Minor cosmetic change: removed an extraneous diagnostic message.
 *
 * Revision 0.12  93/08/10  17:57:47  drapeau
 * Several changes:
 * * Fixed errors in code for default source and destination rectangles.
 *   It turned out not to be enough to make the two variables
 *   "defaultSourceRect" and "defaultDestRect" statics (this is the same
 *   problem as with the VideoIn widget; see comments there).  The fix was to
 *   create two new fields for the ViedoOut widget which hold the space for
 *   the two XRectangle's.
 * * Added code to more correctly turn on and off video output.  However, this
 *   code does not seem to be enough.
 * * Minor cosmetic changes to eliminate excess diagnostic reports.
 * 
 * Revision 0.11  93/06/29  15:03:15  drapeau
 * A number of changes were made, both important and cosmetic in nature:
 * * The definition of the default source and destination rectangles were made
 *   static so as to avoid modules outside this one from accidentally modifying
 *   them.
 * * The "signalWidth" and "signalHeight" resources were added.
 * * The Redisplay() method uses XTranslateCoordinates() to determine the
 *   absolute X & Y values when creating a video signal from part of the
 *   framebuffer.  This is as opposed to using coordinates that are relative to
 *   the widget's parent, which is usually not correct.
 * * Extra checks were added to prevent illegal widget changes from being made.
 * * Extraneous printf statements were removed.
 * 
 * Revision 0.10  93/06/10  17:07:38  drapeau
 * This is the first working version of the widget.
 *  */

static char xVideoOutRcsid[] = "$Header: /plx/source/Xclients/generic/RCS/XVideoOut.c,v 0.24 1996/01/10 09:31:05 drapeau Exp $";

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "XVideoOutP.h"

#define offset(field) XtOffsetOf(XVideoOutRec, xVideoOut.field)

static XtResource resources[] =
{				/* {name, class, type, size, offset, default_type, default_addr}, */
{
  XtNnormalAspectRatio,
  XtCNormalAspectRatio,
  XtRBoolean,
  sizeof(Boolean),
  offset(normalAspectRatio),
  XtRImmediate,
  (XtPointer) True
  },
{
  XtNsourceWin,
  XtCSourceWin,
  XtRWindow,
  sizeof(Window),
  offset(sourceWin),
  XtRImmediate,
  (XtPointer) 0
  },
{
  XtNsourceRect,
  XtCSourceRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(sourceRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNdestRect,
  XtCDestRect,
  XtRPointer,
  sizeof(XRectangle*),
  offset(destRect),
  XtRImmediate,
  (XtPointer) NULL
  },
{
  XtNsignalStandard,
  XtCSignalStandard,
  XtRInt,
  sizeof(int),
  offset(signalStandard),
  XtRImmediate,
  (XtPointer) PLX_NTSC
  },
{
  XtNsignalFormat,
  XtCSignalFormat,
  XtRInt,
  sizeof(int),
  offset(signalFormat),
  XtRImmediate,
  (XtPointer) PLX_COMP
  },
{
  XtNhardwareType,
  XtCHardwareType,
  XtRInt,
  sizeof(int),
  offset(hardwareType),
  XtRImmediate,
  (XtPointer) PLX_XVIDEO_24
  },
{
  XtNchannel,
  XtCChannel,
  XtRInt,
  sizeof(int),
  offset(channel),
  XtRImmediate,
  (XtPointer) PLX_OUTPUT_0
  },
{
  XtNsignalWidth,
  XtCSignalWidth,
  XtRInt,
  sizeof(int),
  offset(signalWidth),
  XtRImmediate,
  (XtPointer) 640
  },
{
  XtNsignalHeight,
  XtCSignalHeight,
  XtRInt,
  sizeof(int),
  offset(signalHeight),
  XtRImmediate,
  (XtPointer) 480
  },
{
  XtNdiagnostics,
  XtCDiagnostics,
  XtRBoolean,
  sizeof(Boolean),
  offset(diagnostics),
  XtRImmediate,
  (XtPointer) False
  },
};

#undef offset

static void XVideoOutAction(/* Widget, XEvent*, String*, Cardinal* */);

static XtActionsRec actions[] =
{
{"xVideoOut",	XVideoOutAction},				    /* {name, procedure}, */
};

static char translations[] =
  "<Key>:		xVideoOut()	\n\
";

/* Declaration of methods */

static void 	Initialize(Widget, Widget, ArgList, Cardinal*);
static void	Realize(Widget, XtValueMask*, XSetWindowAttributes*);
static void 	Redisplay(Widget, XEvent*, Region);
static void 	Destroy(Widget);
static void 	Resize(Widget);
static Boolean	SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void	FindBestVideoSourceMatch(XVideoOutWidget, plx_IO_list*);
static void	AdjustSourceWidthAndHeight(XVideoOutWidget);
static void	AdjustDisplayWidthAndHeight(XVideoOutWidget);
static void	AdjustCoreWidthAndHeight(XVideoOutWidget, Dimension*, Dimension*);
static void	PrintWidgetDiagnostics(XVideoOutWidget, char*);
static void	PreventIllegalWidgetChanges(XVideoOutWidget, XVideoOutWidget);
static Boolean	DoesSourceWindowExist(XVideoOutWidget, Boolean*, XWindowAttributes*);
int		XVideoOutErrorHandler(Display*, XErrorEvent*);
static void	TrackWindow(XVideoOutWidget);
static void	DisableOutputWindowTracking(XVideoOutWidget);
Boolean		XVideoOutWorkProc(XtPointer);
static long	ElapsedTime(struct timeval*, struct timeval*);
static void	XVideoOutGeometryRequest(XVideoOutWidget, XtWidgetGeometry*);


XVideoOutClassRec xVideoOutClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"XVideoOut",
    /* widget_size		*/	sizeof(XVideoOutRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* xVideoOut fields */
    /* empty			*/	0
  }
};

WidgetClass xVideoOutWidgetClass = (WidgetClass)&xVideoOutClassRec;
     

static void XVideoOutWidgetDebug(void)
{
  return;
} /* end function XVideoOutWidgetDebug */



void PrintPlxOutputList(plx_IO_list*	outputs)
{
  plx_signal_list*	signalList = (plx_signal_list*) NULL;
  plx_format_list*	formatList = (plx_format_list*) NULL;
  
  printf("PrintOutputList(): Channel is %d.\n", outputs->channel);
  for (signalList = outputs->signal_list;
       signalList != NULL;
       signalList = signalList->next_signal)
  {
    printf("Signal standard: %d\tWidth:%d\tHeight:%d\n",
	   signalList->signal->standard,
	   signalList->signal->w,
	   signalList->signal->h);
    printf("Signal standard: ");
    switch (signalList->signal->standard)
    {
     case PLX_NTSC:
      printf("NTSC (%d)\t", signalList->signal->standard);
      break;
     case PLX_PAL:
      printf("PAL (%d)\t", signalList->signal->standard);
      break;
     case PLX_SECAM:
      printf("SECAM (%d)\t", signalList->signal->standard);
      break;
    } /* end switch on the signal standard */
    printf("Width:%d\tHeight:%d\n",
	   signalList->signal->w,
	   signalList->signal->h);
  }
  for (formatList = outputs->format_list;
       formatList != NULL;
       formatList = formatList->next_format)
  {
    printf("Signal format: ");
    switch (formatList->format)
    {
     case PLX_COMP:
      printf("Composite (%d)\n", formatList->format);
      break;
     case PLX_YC:
      printf("YC (%d)\n", formatList->format);
      break;
     case PLX_YUV:
      printf("YUV (%d)\n", formatList->format);
      break;
     case PLX_RGB:
      printf("RGB (%d)\n", formatList->format);
      break;
    } /* end switch format */
  }
  return;
} /* end function PrintPlxOutputList */



static void PrintPlxHardwareList(plx_IO* ioList)
{
  plx_IO_list*	outputs;
  
  printf("Entered PrintPlxHardwareList.\n");
  if ((plx_IO*)NULL == ioList)
  {
    printf("The hardware list is empty; no Plx hardware was found.\n");
    return;
  }
  for (outputs = ioList->outputs;
       outputs != NULL;
       outputs = outputs->next_IO)
  {
    printf("Channel: %d\n", outputs->channel);
    PrintPlxOutputList(outputs);
  }
  printf("Exiting (output widget) PrintPlxHardwareList.\n\n");
  return;
} /* end function PrintPlxHardwareList */



static void Initialize(Widget		treq,
  Widget		tnew,
  ArgList		args,
  Cardinal*	numArgs)
{
  int			screen;
  Display*		display = XtDisplay(tnew);
  XVideoOutWidget	newWidget = (XVideoOutWidget) tnew;
  plx_IO_list*		outputList = (plx_IO_list*) NULL;
  
  XVideoOutWidgetDebug();					    /* Call the dummy debugging function */
  screen = XDefaultScreen(display);
  newWidget->xVideoOut.trackWorkProc = (XtWorkProcId)NULL;	    /* By default, no window tracking is done */
  newWidget->xVideoOut.errorHandler = XVideoOutErrorHandler;	    /* Point to the dummy error handler for this widget */
  newWidget->xVideoOut.outputWindow = XCreateSimpleWindow(display,  /* Create a window; this is necessary for subsequent... */
				   RootWindow(display, screen),	    /* ...calls to XPlx extensions, which take a Drawable... */
				   0, 0,			    /* ...as argument.  This window will be released when... */
				   1, 1,			    /* ...this function exits, since it is not needed for... */
				   1,				    /* ...anything outside the scope of this function. */
				   BlackPixel(display, screen),
				   BlackPixel(display, screen));
  newWidget->xVideoOut.gc = XDefaultGC(display,			    /* Create a GC necessary for calls to Xplx extensions... */
		      XDefaultScreen(display));			    /* ...only for this method. */
  newWidget->xVideoOut.hardwareParams =				    /* Determine the type of hardware being used and get... */
    XPlxQueryConfig(display, newWidget->xVideoOut.outputWindow,	    /* ...other relevant information about it. */
		    newWidget->xVideoOut.gc);
/*
  PrintPlxHardwareList(newWidget->xVideoOut.hardwareParams);
*/
  if ((plx_IO*)NULL == newWidget->xVideoOut.hardwareParams)	    /* If hardwareParams is NULL, no hardware was found, so... */
    outputList = (plx_IO_list*)NULL;				    /* ...indicate that no outputs were found. */
  else								    /* Outputs were found */
  {
    for (outputList = newWidget->xVideoOut.hardwareParams->outputs; /* Search through all available outputs for this... */
	 outputList != (plx_IO_list*) NULL;			    /* ...hardware, looking for the output channel being... */
	 outputList = outputList->next_IO)			    /* ...requested. */
    {
      if (outputList->channel == newWidget->xVideoOut.channel)	    /* Was the requested channel found? */
      {								    /* Yes, determine of this channel can supply requested... */
	FindBestVideoSourceMatch(newWidget, outputList);	    /* ...video format */
	break;							    /* No need to search any longer */
      }
    }
  } /* end else (outputs were found) */
  if (outputList == (plx_IO_list*) NULL)			    /* Was the requested channel found? */
  {								    /* No, reset output channel resource */
    newWidget->xVideoOut.signalFormat = PLX_NTSC;
    newWidget->xVideoOut.defaultSourceRect.x = 0;		    /* Set default size of rectangle to digitize from the... */
    newWidget->xVideoOut.defaultSourceRect.y = 0;		    /* ...frame buffer. */
    newWidget->xVideoOut.defaultSourceRect.width = 640;
    newWidget->xVideoOut.defaultSourceRect.height = 480;
    newWidget->xVideoOut.defaultDestRect.x = 0;			    /* Set default size of rectangle to generate as a video... */
    newWidget->xVideoOut.defaultDestRect.y = 0;			    /* ...signal. */
    newWidget->xVideoOut.defaultDestRect.width = 640;
    newWidget->xVideoOut.defaultDestRect.height = 480;
    newWidget->core.width = newWidget->xVideoOut.signalWidth;	    /* Set a default size for this widget */
    newWidget->core.height = newWidget->xVideoOut.signalHeight;
  }
  else								    /* Now that everything is set up, try to accommodate... */
  {								    /* ...user preferences */
    if (newWidget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the application programmer specify a sourceRect... */
    {								    /* ...from which to digitize a portion of the framebuffer? */
      if (newWidget->xVideoOut.sourceRect->x < 0)		    /* Yes, assure that the values specified are valid */
	newWidget->xVideoOut.sourceRect->x = 0;
      if (newWidget->xVideoOut.sourceRect->y < 0)
	newWidget->xVideoOut.sourceRect->y = 0;
    if ((newWidget->xVideoOut.sourceRect->width <= 0) ||	    /* Make sure that width is valid: not less than zero, and... */
	(newWidget->xVideoOut.sourceRect->width >		    /* ...not greater than the height of the video signal... */
	 newWidget->xVideoOut.signalWidth))			    /* ...being produced */
      newWidget->xVideoOut.sourceRect->width =
	newWidget->xVideoOut.signalWidth;
    if ((newWidget->xVideoOut.sourceRect->height <= 0) ||	    /* Make sure that height is valid: not less than zero,... */
	(newWidget->xVideoOut.sourceRect->height >		    /* ...and not greater than the height of the video signal... */
	 newWidget->xVideoOut.signalHeight))			    /* ...being produced. */
      newWidget->xVideoOut.sourceRect->height =
	newWidget->xVideoOut.signalHeight;
    }
    else
    {
    newWidget->xVideoOut.defaultSourceRect.x = 0;		    /* Set default size of rectangle to digitize from the... */
    newWidget->xVideoOut.defaultSourceRect.y = 0;		    /* ...frame buffer. */
    newWidget->xVideoOut.defaultSourceRect.width =
      newWidget->xVideoOut.signalWidth;
    newWidget->xVideoOut.defaultSourceRect.height =
      newWidget->xVideoOut.signalHeight;
    }
    newWidget->xVideoOut.defaultDestRect.x = 0;			    /* Set default size of rectangle to generate as a video... */
    newWidget->xVideoOut.defaultDestRect.y = 0;			    /* ...signal. */
    newWidget->xVideoOut.defaultDestRect.width =
      newWidget->xVideoOut.signalWidth;
    newWidget->xVideoOut.defaultDestRect.height =
      newWidget->xVideoOut.signalHeight;
    newWidget->core.width = newWidget->xVideoOut.signalWidth;
    newWidget->core.height = newWidget->xVideoOut.signalHeight;	    /* Set the correct initial size for this widget */
  }
  if (newWidget->xVideoOut.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
  {								    /* Yes, do the necessary math */
    if (newWidget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the application programmer supply a source rectangle? */
      newWidget->xVideoOut.sourceRect->height =			    /* Yes, apply the math to this rectangle */
	(newWidget->xVideoOut.sourceRect->width * 3) / 4;
    else							    /* No, apply the math to the default source rectangle */
      newWidget->xVideoOut.defaultSourceRect.height =
	(newWidget->xVideoOut.defaultSourceRect.width * 3) / 4;
  }
  if (newWidget->xVideoOut.sourceWin == (Window)0)		    /* Was a source window specified from which to digitize? */
    newWidget->xVideoOut.sourceWin =				    /* No, digitize from the root window */
      RootWindow(display, screen);
  return;
}								    /* end function Initialize */



static void Realize(Widget			w,
		    XtValueMask*		valueMask,
		    XSetWindowAttributes*	attributes)
{
  int			screenNumber;
  Display*		display = XtDisplay(w);
  XGCValues		gcValues;
  XVideoOutWidget	widget = (XVideoOutWidget) w;
  
  screenNumber = XScreenNumberOfScreen(XtScreen(widget));
  widget->core.border_width = 0;				    /* InputOnly windows must have a border width of zero. */
  widget->core.depth = 0;
  attributes->win_gravity = NorthWestGravity;
  *valueMask = (CWWinGravity);
  XtCreateWindow((Widget)widget, InputOnly,			    /* Create a dummy window that's not used for display at all */
		 XDefaultVisual(display, screenNumber),
		 *valueMask, attributes);
  XFlush(display);
  gcValues.function = GXxor;
  widget->xVideoOut.gc = XCreateGC(XtDisplay(widget),		    /* Create a Graphics Context to use for the video output... */
				   widget->xVideoOut.outputWindow,  /* ...widget functionality. */
				   (GCFunction),
				   &gcValues);
  XPlxVideoOutputSelect(display,				    /* Configure the video output hardware to show the... */
			widget->xVideoOut.outputWindow,		    /* ...requested video signal */
			widget->xVideoOut.gc,
			widget->xVideoOut.channel,
			widget->xVideoOut.signalStandard,
			widget->xVideoOut.signalFormat,
			PLX_RGB24);
  PrintWidgetDiagnostics(widget, "Just before leaving Realize()");
  return;
}								    /* end function Realize */



static void Redisplay(Widget	w,
		      XEvent*	event,
		      Region	region)
{
  Boolean		isRootWindow;
  Boolean		result;
  int			screenNumber;
  XVideoOutWidget	widget = (XVideoOutWidget) w;
  Display*		display = XtDisplay(w);
  XRectangle*		sourceRect;
  XRectangle*		destRect = (XRectangle*)NULL;
  XWindowAttributes	unused;
  
  if (!XtIsRealized(w))						    /* Don't bother trying to draw if the widget is not... */
    return;							    /* ...or has not been realized */
  if (widget->xVideoOut.channel == PLX_IO_OFF)			    /* Don't draw output if the current channel is... */
    return;							    /* ...video off. */
  PrintWidgetDiagnostics(widget, "Redisplay()");
  result = DoesSourceWindowExist(widget, &isRootWindow,		    /* Determine if window to be used for output is still there */
				 &unused);
  if (False == result)						    /* Is the source window still in existence? */
    return;							    /* No, don't bother trying to do an output operation */
  if (widget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the programmer specify a rectangle to use for input? */
  {								    /* Yes, use it */
    sourceRect = widget->xVideoOut.sourceRect;
  }
  else								    /* No, use the default source rectangle */
  {
    sourceRect = &(widget->xVideoOut.defaultSourceRect);
    if (True == isRootWindow)
    {
      sourceRect->x = widget->core.x;				    /* By default, grab output from the location of this widget */
      sourceRect->y = widget->core.y;
    }
    else							    /* In this case, the widget is tracking a window, so... */
    {								    /* ...begin tracking from the source window's origin. */
      sourceRect->x = 0;
      sourceRect->y = 0;
    }
  }
  if (widget->xVideoOut.destRect != (XRectangle*)NULL)		    /* Did the programmer specify a rectangle to use for output? */
  {								    /* Yes, use it */
    destRect = widget->xVideoOut.destRect;
  }
  else								    /* No, use the default destination rectangle */
  {
    destRect = &(widget->xVideoOut.defaultDestRect);
  }
  screenNumber = XScreenNumberOfScreen(XtScreen(widget));
  XPlxVideoScaleOutput(display,					    /* Begin digitizing from the frame buffer and generating... */
		       widget->xVideoOut.sourceWin,		    /* ...a video signal */
		       widget->xVideoOut.gc,
		       sourceRect->x,
		       sourceRect->y,
		       sourceRect->width,
		       sourceRect->height,
		       destRect->x,
		       destRect->y +				    /* Don't draw video into the blanking interval portion of... */
		       widget->xVideoOut.signalBlankInterval,	    /* ...the video signal being generated. */
		       destRect->width,
		       destRect->height);
  return;
}								    /* end function Redisplay */



static void Destroy(Widget w)
{
  Boolean		isRootWindow;
  Boolean		result;
  XVideoOutWidget	theWidget = (XVideoOutWidget) w;
  XWindowAttributes	unused;
  
  result = DoesSourceWindowExist(theWidget, &isRootWindow,	    /* Determine if window to be used for output is still there */
				 &unused);
  if (True == result)						    /* Is the source window still in existence? */
  {								    /* Yes, it's okay to turn off output for it */
    XPlxVideoStop(XtDisplay(theWidget),				    /* Forcibly turn off output the correct way */
		  theWidget->xVideoOut.sourceWin,
		  theWidget->xVideoOut.gc);
  }
  XFlush(XtDisplay(theWidget));					    /* Make sure output is turned off before continuing */
  XPlxVideoOutputSelect(XtDisplay(theWidget),			    /* Forcibly turn off generation of the output video signal */
			theWidget->xVideoOut.outputWindow,
			theWidget->xVideoOut.gc,
			PLX_IO_OFF,
			theWidget->xVideoOut.signalStandard,
			theWidget->xVideoOut.signalFormat,
			PLX_RGB24);
  XFreeGC(XtDisplay(theWidget), theWidget->xVideoOut.gc);	    /* Destroy the GC created for this video widget */
  XDestroyWindow(XtDisplay(theWidget),				    /* Free up dummy window no longer needed */
		 theWidget->xVideoOut.outputWindow);
  return;
}								    /* end function Destroy */



static void Resize(Widget w)
{
  XVideoOutWidget	theWidget = (XVideoOutWidget) w;
  Dimension		desiredWidth, desiredHeight;
  XtWidgetGeometry	geometry;
  XRectangle*		sourceRect = (XRectangle*)NULL;
  
  AdjustSourceWidthAndHeight(theWidget);			    /* Adjust to account for possible 4-to-3 aspect ratio */
  AdjustDisplayWidthAndHeight(theWidget);			    /* Make sure that display parameters are also acceptable */
  AdjustCoreWidthAndHeight(theWidget, &desiredWidth, &desiredHeight);
  if ((desiredWidth != theWidget->core.width) ||		    /* Was width or height of this widget modified after... */
      (desiredHeight != theWidget->core.height))		    /* ...conforming to size constraints? */
  {
    geometry.width = (Dimension)desiredWidth;
    geometry.height = (Dimension)desiredHeight;
    geometry.request_mode = (CWWidth | CWHeight);		    /* Only interested in size changes, not positional changes */
    XVideoOutGeometryRequest(theWidget, &geometry);		    /* Try to make the widget size change */
  }								    /* end (if widthOrHeightChanged == True...) */
  if (theWidget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the programmer specify a source rectangle? */
    sourceRect = theWidget->xVideoOut.sourceRect;		    /* If so, point to it; if not, use the... */
  else								    /* ...default display rectangle. */
  {
    sourceRect = &(theWidget->xVideoOut.defaultSourceRect);
    sourceRect->width = theWidget->core.width;			    /* If the default sourceRect is being used, make the... */
    sourceRect->height = theWidget->core.height;		    /* ...sourceRect's size tracks the size of the widget. */
  }
  Redisplay(w, (XEvent*)NULL, (Region)NULL);			    /* Force the widget to re-digitize at the new geometry */
  return;
}								    /* end function Resize */



static Boolean SetValues(Widget		current,
			 Widget		request,
			 Widget		new,
			 ArgList	args,
			 Cardinal*	numArgs)
{
  XVideoOutWidget	currentWidget = (XVideoOutWidget) current;
  XVideoOutWidget	newWidget = (XVideoOutWidget) new;
  Boolean		redisplayNeeded = False;
  XRectangle*		theRect = (XRectangle*)NULL;
  plx_IO_list*		outputList = (plx_IO_list*) NULL;
  
  PrintWidgetDiagnostics(currentWidget, "Entering SetValues()");
  PreventIllegalWidgetChanges(currentWidget, newWidget);	    /* Don't let read-only fields be modified by the application */
  if (currentWidget->xVideoOut.hardwareType !=			    /* Prevent any illegal widget changes */
      newWidget->xVideoOut.hardwareType)
  {
    XtWarning("XVideoOut Widget: \"hardwareType\" is a read-only resource.");
    newWidget->xVideoOut.hardwareType = currentWidget->xVideoOut.hardwareType;
    redisplayNeeded = False;
  }
  if ((currentWidget->xVideoOut.sourceWin !=			    /* Was a new source window specified?  */
       newWidget->xVideoOut.sourceWin) &&			    /* And, is the window valid? */
      (newWidget->xVideoOut.sourceWin == (Window)0))
  {								    /* Window was different but not valid */
    XtWarning("XVideoOut Widget: invalid Window specified.  Using currently-stored value.");
    newWidget->xVideoOut.sourceWin = currentWidget->xVideoOut.sourceWin;
    redisplayNeeded = False;
  }
  if (currentWidget->xVideoOut.sourceWin !=			    /* Did the programmer specify a new window to track? */
      newWidget->xVideoOut.sourceWin)
  {
    TrackWindow(newWidget);					    /* Yes, determine whether to track that window for output */
  }
  if ((newWidget->xVideoOut.channel !=				    /* Was a new output requested? */
       currentWidget->xVideoOut.channel) ||
      (newWidget->xVideoOut.signalStandard !=			    /* Was a new signal standard requested? */
       currentWidget->xVideoOut.signalStandard) ||
      (newWidget->xVideoOut.signalFormat !=			    /* Was a new signal format requested? */
       currentWidget->xVideoOut.signalFormat))
  {								    /* Yes, see if the request can be satisfied */
    if ((newWidget->xVideoOut.channel == PLX_IO_OFF) &&		    /* Should this widget turn off video? */
	(newWidget->xVideoOut.sourceWin != 0))			    /* And, is the window to be turned off non-zero? */
    {
      XPlxVideoStop(XtDisplay(newWidget),			    /* Yes, forcibly disable producing video output */
		    newWidget->xVideoOut.sourceWin,
		    newWidget->xVideoOut.gc);
      if (newWidget->xVideoOut.trackWorkProc != (XtWorkProcId)NULL) /* Does the widget have a WorkProc registered already? */
      {								    /* Yes, unregister the WorkProc */
	XtRemoveWorkProc(newWidget->xVideoOut.trackWorkProc);	    /* But don't re-initialize the workProcId, noting that...  */
      }								    /* ...it's still active when output is turned on again. */
    }
    else							    /* Off was not selected, new output configuration chosen */
    {
      if ((plx_IO*)NULL == newWidget->xVideoOut.hardwareParams)	    /* If hardwareParams is NULL, no hardware was found, so... */
	outputList = (plx_IO_list*)NULL;			    /* ...indicate that no outputs were found. */
      else							    /* Outputs were found */
      {
	for (outputList = newWidget->xVideoOut.hardwareParams->outputs; /* Search all available outputs for this hardware,... */
	     outputList != (plx_IO_list*) NULL;			    /* ...looking for the output channel being requested. */
	     outputList = outputList->next_IO)
	{
	  if (outputList->channel == newWidget->xVideoOut.channel)  /* Was the requested channel found? */
	  {							    /* Yes, determine if this channel can supply requested... */
	    FindBestVideoSourceMatch(newWidget, outputList);	    /* ...video format */
	    break;						    /* No need to search any longer */
	  }
	}
      }	/* end else (outputs were found) */
      if ((outputList == (plx_IO_list*) NULL) &&		    /* Was the requested channel found? */
	  (newWidget->xVideoOut.channel != PLX_IO_OFF))		    /* (it won't be found if the requested channel is "off") */
      {
	newWidget->xVideoOut.channel =				    /* No, don't bother changing the channel... */
	  currentWidget->xVideoOut.channel;
	newWidget->xVideoOut.signalStandard =			    /* ...or the signal standard */
	  currentWidget->xVideoOut.signalStandard;
	newWidget->xVideoOut.signalFormat =			    /* ...or the signal format */
	  currentWidget->xVideoOut.signalFormat;
	redisplayNeeded = False;
      }
      else							    /* Yes, the display should probably be redrawn */
      {
	PrintWidgetDiagnostics(newWidget, "SetValues");
	XPlxVideoOutputSelect(XtDisplay(newWidget),		    /* Configure the video output hardware to show the... */
			      newWidget->xVideoOut.outputWindow,    /* ...requested video signal */
			      newWidget->xVideoOut.gc,
			      newWidget->xVideoOut.channel,
			      newWidget->xVideoOut.signalStandard,
			      newWidget->xVideoOut.signalFormat,
			      PLX_RGB24);
	if (newWidget->xVideoOut.trackWorkProc != (XtWorkProcId)NULL) /* Should the widget be tracking a specific window when... */
	{							    /* ...output is turned back on? */
	  TrackWindow(newWidget);				    /* Yes, enable the output tracking WorkProc again. */
	}
      } /* end (if outputList is NULL or sync is False) */
    } /* end (videoOut.channel was not PLX_IO_OFF) */
  } /* end (if channel, standard, or format where changed) */
  Redisplay((Widget)newWidget, (XEvent*)NULL, (Region)NULL);	    /* Force the widget to digitize the new signal */
  if (newWidget->xVideoOut.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
  {								    /* Yes, do the necessary math */
    if (newWidget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the programmer supply a source rectangle? */
      theRect = newWidget->xVideoOut.sourceRect;		    /* Yes, apply the math to the programmer's rectangle */
    else							    /* No, apply the math to the default source rectangle */
      theRect = &(newWidget->xVideoOut.defaultSourceRect);
    newWidget->core.height = (newWidget->core.width * 3) / 4;	    /* Also resize the widget to conform to the aspect ratio */
    theRect->height = (theRect->width * 3) / 4;
    Redisplay((Widget)newWidget, (XEvent*)NULL, (Region)NULL);	    /* Force the widget to digitize the new signal */
  }
  return(False);						    /* The XtSetValues() function does an XClearArea() if the... */
								    /* ...method returns True.  Since this widget's window is... */
								    /* ...an InputOnly window, XClearArea() should never be... */
								    /* ...called for this window.  So, always return "False". */
}								    /* end function SetValues */



static void XVideoOutAction(Widget	w,
			   XEvent*	event,
			   String*	params,
			   Cardinal*	numParams)
{
  return;							    /* Right now this function does nothing */
}								    /* end function XVideoOutAction */



static void FindBestVideoSourceMatch(XVideoOutWidget	theWidget,
				     plx_IO_list*	outputList)
{
  plx_signal_list*	signalList = (plx_signal_list*) NULL;
  plx_format_list*	formatList = (plx_format_list*) NULL;
  XRectangle*		destRect = (XRectangle*)NULL;
  
  if ((plx_IO_list*)NULL == outputList)				    /* If outputList is NULL, no hardware was found, so... */
    return;							    /* ...don't bother trying to do any work. */
  if (theWidget->xVideoOut.destRect != (XRectangle*)NULL)	    /* Did the programmer supply a destination rectangle? */
    destRect = theWidget->xVideoOut.destRect;			    /* Yes, use it */
  else								    /* No, use the default destination rectangle */
    destRect = &(theWidget->xVideoOut.defaultDestRect);
  for (signalList = outputList->signal_list;			    /* Search through all signal formats available for this... */
       signalList != (plx_signal_list*) NULL;			    /* ...channel to find the format requested, if possible */
       signalList = signalList->next_signal)
  {
    if (signalList->signal->standard ==				    /* Was the requested signal standard found? */
	theWidget->xVideoOut.signalStandard)
    {								    /* Yes, use the associated structure to fill in... */
      destRect->width = signalList->signal->w;
      destRect->height =					    /* Height of viewable portion of signal is total height... */
	signalList->signal->h -	signalList->signal->b;		    /* ...minus height of blanking interval */
      theWidget->xVideoOut.signalWidth = signalList->signal->w;
      theWidget->xVideoOut.signalHeight =
	signalList->signal->h - signalList->signal->b;
      theWidget->xVideoOut.signalBlankInterval =
	signalList->signal->b;
      break;							    /* No need to search further */
    }
  }
  if (signalList == (plx_signal_list*) NULL)			    /* Did the search fail? */
  {								    /* Yes, set default values */
    if (outputList->signal_list == (plx_signal_list*) NULL)	    /* Did this channel have *any* default signal list? */
    {								    /* Evidently not, set poor default values */
      destRect->width = 640;
      destRect->height = 525 - 43;				    /* NTSC height minus NTSC blanking interval height */
      theWidget->xVideoOut.channel = PLX_IO_OFF;		    /* Turn off output */
      theWidget->xVideoOut.signalBlankInterval = 43;
    }
    else							    /* Yes, this channel had at least some default signal... */
    {								    /* ...list; use the first list found for default values */
      signalList = outputList->signal_list;
      destRect->width = signalList->signal->w;
      destRect->height =					    /* Height of viewable portion of signal is total height... */
	signalList->signal->h -	signalList->signal->b;		    /* ...minus height of blanking interval */
      theWidget->xVideoOut.signalWidth = signalList->signal->w;
      theWidget->xVideoOut.signalHeight =
	signalList->signal->h - signalList->signal->b;
      printf("(Output):FindBestVideoSourceMatch(): not found: blankIntervalHeight is %d.\n",
	     signalList->signal->b);
      theWidget->xVideoOut.signalBlankInterval =
	signalList->signal->b;
    }
  }
  for (formatList = outputList->format_list;			    /* Search through all format types available for this... */
       formatList != (plx_format_list*) NULL;			    /* ...channel to find the type requested, if possible */
       formatList = formatList->next_format)
  {
    if (formatList->format == theWidget->xVideoOut.signalFormat)    /* Was the requested signal format found? */
      break;							    /* Yes, no need to search any further */
  }
  if (formatList == (plx_format_list*) NULL)			    /* Did the search fail? */
  {
    if (outputList->format_list == (plx_format_list*) NULL)	    /* Did this channel have *any* default format list? */
      theWidget->xVideoOut.signalFormat = PLX_NTSC;		    /* Evidently not, set a poor default value */
    else							    /* Yes, use the first format found */
      theWidget->xVideoOut.signalFormat =
	outputList->format_list->format;
  }
  return;
} /* end function FindBestVideoSourceMatch */



static void AdjustSourceWidthAndHeight(XVideoOutWidget widget)
{
  Dimension	videoHeight = 0;
  XRectangle*	sourceRect = (XRectangle*)NULL;
  
  if (widget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the programmer supply a source rectangle? */
    sourceRect = widget->xVideoOut.sourceRect;			    /* Yes, use it */
  else								    /* No, use the default source rectangle */
    sourceRect = &(widget->xVideoOut.defaultSourceRect);
  if ((sourceRect->x + sourceRect->width) >			    /* Make sure that the rectangle to be digitized still... */
      widget->xVideoOut.signalWidth)				    /* ...falls into the area denoted by "signalWidth" */
    sourceRect->width =
      widget->xVideoOut.signalWidth - sourceRect->x;		    /* Clip the digitized rectangle if necessary to make it fit */
  
  videoHeight = widget->xVideoOut.signalHeight;			    /* Make sure that the rectangle to be digitized still... */
  if ((sourceRect->y + sourceRect->height) > videoHeight)	    /* ...falls into the area denoted by the total height... */
    sourceRect->y = videoHeight - sourceRect->y;		    /* ...of the video rectangle to be digitized, taking into... */
  if (widget->xVideoOut.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
    sourceRect->height = (sourceRect->width * 3) / 4;		    /* Yes, do the necessary math */
  return;							    /* ...account whether to show blanking interval info or not. */
}								    /* end function AdjustSourceWidthAndHeight */



static void AdjustDisplayWidthAndHeight(XVideoOutWidget widget)
{
  widget->xVideoOut.defaultDestRect.x = widget->core.x;		    /* Re-adjust the default output rectangle to fit the... */
  widget->xVideoOut.defaultDestRect.y = widget->core.y;		    /* ...size of the current video output signal standard. */
  widget->xVideoOut.defaultDestRect.width = widget->xVideoOut.signalWidth;
  widget->xVideoOut.defaultDestRect.height = widget->xVideoOut.signalHeight;
  return;
}								    /* end function AdjustDisplayWidthAndHeight */



static void AdjustCoreWidthAndHeight(XVideoOutWidget	widget,
				     Dimension*		preferredWidth,
				     Dimension*		preferredHeight)
{
  XRectangle*		sourceRect;
  unsigned short	width;
  unsigned short	height;
  
  width = widget->core.width;
  height = widget->core.height;
  if (widget->xVideoOut.sourceRect != (XRectangle*)NULL)
    sourceRect = widget->xVideoOut.sourceRect;
  else
    sourceRect = &(widget->xVideoOut.defaultSourceRect);
  if (width > widget->xVideoOut.signalWidth)			    /* Make sure the widget's smaller than the video output... */
    width = widget->xVideoOut.signalWidth;			    /* ...signal to be producted. */
  if (height > widget->xVideoOut.signalHeight)
    height = widget->xVideoOut.signalHeight;
  if (widget->xVideoOut.normalAspectRatio == True)		    /* Must this widget conform to a 4:3 aspect ratio? */
  {								    /* Yes, do the necessary math */
    height = (width * 3) / 4;
  }
  *preferredWidth = (Dimension)width;
  *preferredHeight = (Dimension)height;
  PrintWidgetDiagnostics(widget, "AdjustCoreWidthAndHeight.");
  return;
}								    /* end function AdjustCoreWidthAndHeight */


static void PrintWidgetDiagnostics(XVideoOutWidget widget,
				   char* calledFromWhere)
{
  XRectangle*		sourceRect = (XRectangle*)NULL;
  
  if (widget->xVideoOut.diagnostics == False)			    /* Was the appropriate resource set to print diagnostics... */
    return;							    /* Don't print any diagnostics right now */
  
  
  printf("XVideoOut Widget: In function %s, values are these:\n",
	 calledFromWhere);
  printf("Channel:\t");
  switch (widget->xVideoOut.channel)
  {
   case PLX_IO_OFF:
    printf("Off.\n");
    break;
   case PLX_INPUT_0:
    printf("Input 0.\n");
    break;
   case PLX_INPUT_1:
    printf("Input 1.\n");
    break;
   case PLX_OUTPUT_0:
    printf("Output 1.\n");
    break;
   default:
    printf("%d (error).\n", widget->xVideoOut.channel);
    break;
  }
  printf("Format:\t");
  switch (widget->xVideoOut.signalFormat)
  {
   case PLX_COMP:
    printf("Composite.\n");
    break;
   case PLX_YC:
    printf("YC.\n");
    break;
   case PLX_YUV:
    printf("YUV.\n");
    break;
   case PLX_RGB:
    printf("RGB.\n");
    break;
   default:
    printf("%d (error).\n", widget->xVideoOut.signalFormat);
    break;
  }
  printf("Standard:\t");
  switch (widget->xVideoOut.signalStandard)
  {
   case PLX_NTSC:
    printf("NTSC.\n");
    break;
   case PLX_PAL:
    printf("PAL.\n");
    break;
   case PLX_SECAM:
    printf("SECAM.\n");
    break;
   default:
    printf("%d (error).\n", widget->xVideoOut.signalStandard);
    break;
  }
  if (widget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the programmer supply a source rectangle? */
    sourceRect = widget->xVideoOut.sourceRect;			    /* Yes, use it */
  else								    /* No, use the default source rectangle */
    sourceRect = &(widget->xVideoOut.defaultSourceRect);
  printf("sourceX: %d\nsourceY: %d\nsourceWidth: %d\nsourceHeight: %d\n",
	 sourceRect->x,
	 sourceRect->y,
	 sourceRect->width,
	 sourceRect->height);
  printf("signalWidth: %d\nsignalHeight: %d\nsignalBlankInterval: %d\n",
	 widget->xVideoOut.signalWidth,
	 widget->xVideoOut.signalHeight,
	 widget->xVideoOut.signalBlankInterval);
  printf("sourceWindow: %d\n", (int)widget->xVideoOut.sourceWin);
  return;
}								    /* end function PrintWidgetDiagnostics */



static void PreventIllegalWidgetChanges(XVideoOutWidget current,
					XVideoOutWidget new)
{
  if (current->xVideoOut.signalHeight !=
      new->xVideoOut.signalHeight)
  {
    XtWarning("XVideoOut Widget: \"signalHeight\" is a read-only resource.");
    new->xVideoOut.signalHeight = current->xVideoOut.signalHeight;
  }
  if (current->xVideoOut.signalWidth !=
      new->xVideoOut.signalWidth)
  {
    XtWarning("XVideoOut Widget: \"signalWidth\" is a read-only resource.");
    new->xVideoOut.signalWidth = current->xVideoOut.signalWidth;
  }
  if (current->xVideoOut.signalBlankInterval !=
      new->xVideoOut.signalBlankInterval)
  {
    XtWarning("XVideoOut Widget: \"signalBlankInterval\" is a read-only resource.");
    new->xVideoOut.signalBlankInterval =
      current->xVideoOut.signalBlankInterval;
  }
  return;
}								    /* end function PreventIllegalWidgetChanges */



static Boolean DoesSourceWindowExist(XVideoOutWidget	widget,	    /* This function determines whether the "sourceWin"... */
				     Boolean*		isRootWindowReturn, /* ...window for the widget passed in as argument... */
				     XWindowAttributes*	winAttrs)   /* ...exists, whether it is the RootWindow, and also... */
{								    /* ...info about the window if it exists. */
  XErrorHandler		oldErrorHandlerFunction;
  Status		status;
  XWindowAttributes	windowAttributes;
  
  oldErrorHandlerFunction =					    /* Store the error handler currently in use; it'll be... */
    XSetErrorHandler(widget->xVideoOut.errorHandler);		    /* ...restored after this call is done. */
  status = XGetWindowAttributes(XtDisplay(widget),		    /* Try to get information about the window in question */
				widget->xVideoOut.sourceWin,
				&windowAttributes);
  XSetErrorHandler(oldErrorHandlerFunction);			    /* Restore the previously-installed error handler */
  if (0 == status)						    /* Did the query fail for some reason? */
  {								    /* Yes, assume the window doesn't exist, return a value... */
    return(False);						    /* ...indicating this. */
  }
  if (windowAttributes.root == widget->xVideoOut.sourceWin)	    /* Is the window to be tracked the RootWindow? */
    *isRootWindowReturn = True;					    /* Yes, indicate so before returning from this function */
  else
    *isRootWindowReturn = False;
  memcpy(winAttrs, &windowAttributes, sizeof(XWindowAttributes));   /* Copy the window attributes into the argument passed in */
  return(True);							    /* The query succeeded, so the sourceWin must exist */
} /* end function DoesSourceWindowExist */



int XVideoOutErrorHandler(Display*	display,		    /* Dummy error handler just to catch X errors before the... */
			  XErrorEvent*	error)			    /* ...application can crash. */
{
  return(0);
} /* end function XVideoOutErrorHandler */



static void TrackWindow(XVideoOutWidget	widget)
{
  XtAppContext		appContext;
  XtWidgetGeometry	geometry;
  Boolean		isRootWindow;
  Boolean		result;
  XRectangle*		sourceRect = (XRectangle*)NULL;
  Dimension		width;
  Dimension		height;
  Dimension		unused1;
  Dimension		unused2;
  XWindowAttributes	winAttrs;
  
  DisableOutputWindowTracking(widget);				    /* Disable any existing WorkProc for this widget */
  result = DoesSourceWindowExist(widget, &isRootWindow,		    /* Does the window in question exist?  Is it the RootWindow? */
				 &winAttrs);
  if (False == result)						    /* The window in question does not exist; don't bother... */
  {								    /* ...trying to track a non-existent window */
    return;
  }
  if (True == isRootWindow)					    /* The window in question is the RootWindow; don't bother... */
  {								    /* ...trying to track it. */
    return;
  }
  if (widget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the programmer supply a source rectangle? */
    sourceRect = widget->xVideoOut.sourceRect;			    /* Yes, use it */
  else								    /* No, use the default source rectangle */
    sourceRect = &(widget->xVideoOut.defaultSourceRect);
  sourceRect->x = 0;						    /* Adjust the output rectangle to enclose the window... */
  sourceRect->y = 0;						    /* ...being tracked. */
  sourceRect->width = width;
  sourceRect->height = height;
  geometry.x = (Position)winAttrs.x;				    /* Set up a Geometry Request to move and resize this... */
  geometry.y = (Position)winAttrs.y;				    /* ... widget along with the source window being tracked. */
  geometry.width =						    /* Don't allow the widget's width or height to fall...  */
    (winAttrs.width > widget->xVideoOut.signalWidth?		    /* ...outside of legal values. */
     (Dimension)widget->xVideoOut.signalWidth :
     (Dimension)winAttrs.width);
  geometry.height =
    (winAttrs.height > widget->xVideoOut.signalHeight?
     (Dimension)widget->xVideoOut.signalHeight :
     (Dimension)winAttrs.height);
  geometry.request_mode = (CWX | CWY | CWWidth | CWHeight);
  XVideoOutGeometryRequest(widget, &geometry);			    /* Attempt to change this widget's geometry */
  AdjustCoreWidthAndHeight(widget, &unused1, &unused2);
  AdjustSourceWidthAndHeight(widget);				    /* Make sure the new rectangle fits within legal boundaries */
  XSelectInput(XtDisplay(widget),				    /* Look for window changes that would affect the output... */
	       widget->xVideoOut.sourceWin,			    /* ...operation on this window. */
	       (StructureNotifyMask));
  appContext = XtWidgetToApplicationContext((Widget)widget);	    /* Get the Application Context for this widget */
  widget->xVideoOut.trackWorkProc =				    /* Start a WorkProc for the widget to begin tracking the... */
    XtAppAddWorkProc(appContext,				    /* ...window to be output. */
		     XVideoOutWorkProc,
		     (XtPointer)widget);
  return;
} /* end function TrackWindow */



static void DisableOutputWindowTracking(XVideoOutWidget widget)
{
  if (widget->xVideoOut.trackWorkProc != (XtWorkProcId)NULL)	    /* Does the widget have a WorkProc registered already? */
  {								    /* Yes, unregister the WorkProc */
    XtRemoveWorkProc(widget->xVideoOut.trackWorkProc);
    widget->xVideoOut.trackWorkProc = (XtWorkProcId)NULL;
  }
  return;
} /* end function DisableOutputWindowTracking */



Boolean XVideoOutWorkProc(XtPointer	widgetPtr)
{
  Boolean		eventIsWaiting;
  XtWidgetGeometry	geometry;
  static struct timeval	lastTimeUpdated;
  struct timeval	now;
  static Boolean	resetLastTimeUpdated = True;
  XRectangle*		sourceRect = (XRectangle*)NULL;
  XConfigureEvent	theEvent;
  XVideoOutWidget	widget = (XVideoOutWidget)widgetPtr;
  
  if (PLX_IO_OFF == widget->xVideoOut.channel)			    /* Don't bother checking for anything if output is off */
    return(True);
  if (True == resetLastTimeUpdated)				    /* Should the last time this clock was recorded be updated? */
  {
    gettimeofday(&lastTimeUpdated, (struct timezone*)NULL);	    /* Yes, get the current time */
    resetLastTimeUpdated = False;				    /* Note that the clock was updated properly */
  }
  gettimeofday(&now, (struct timezone*)NULL);
  if (ElapsedTime(&lastTimeUpdated, &now) < 200)		    /* Has at least one fifth of a second passed? */
    return(False);						    /* No, it's too soon to try to consume CPU cycles */
  gettimeofday(&lastTimeUpdated, (struct timezone*)NULL);	    /* Yes, enough time has passed, so reset the timer */
  XSync(XtDisplay(widget), False);				    /* Make sure any incoming events are now delivered */
  eventIsWaiting = XCheckTypedWindowEvent(XtDisplay(widget),	    /* Did the source window (i.e., the window to be tracked)... */
					  widget->xVideoOut.sourceWin, /* ...change state somehow? */
					  ConfigureNotify,
					  (XEvent*)&theEvent);
  if (False == eventIsWaiting)					    /* No, don't bother doing anything else this time around */
  {
    return(False);
  }
  if (DestroyNotify == theEvent.type)				    /* Was the window being tracked destroyed? */
  {								    /* Yes, no reason to try to track it anymore */
    DisableOutputWindowTracking(widget);
    return(True);						    /* Done tracking the source window */
  }
  if (widget->xVideoOut.sourceRect != (XRectangle*)NULL)	    /* Did the programmer supply a source rectangle? */
    sourceRect = widget->xVideoOut.sourceRect;			    /* Yes, use it */
  else								    /* No, use the default source rectangle */
    sourceRect = &(widget->xVideoOut.defaultSourceRect);
  geometry.x = (Position)theEvent.x;				    /* Set up a Geometry Request to move this widget along... */
  geometry.y = (Position)theEvent.y;				    /* ...with the source window being tracked. */
  geometry.request_mode = (CWX | CWY);
  XVideoOutGeometryRequest(widget, &geometry);			    /* Attempt to move this widget to the new coordinates */
  sourceRect->x = 0;						    /* Adjust the output rectangle to enclose the window... */
  sourceRect->y = 0;						    /* ...being tracked. */
  sourceRect->width = theEvent.width;
  sourceRect->height = theEvent.height;
  AdjustSourceWidthAndHeight(widget);				    /* Make sure the new rectangle fits within legal boundaries */
  Redisplay((Widget)widget, (XEvent*)NULL, (Region)NULL);	    /* Force the widget to digitize the new signal */
  return(False);						    /* Done with 1 iteration, but call me again */
} /* end function XVideoOutWorkProc */




static long	ElapsedTime(struct timeval*	before,
			    struct timeval*	after)
{
  return(((after->tv_sec - before->tv_sec) * 1000) +		    /* Convert the seconds part to milliseconds */
	 ((after->tv_usec - before->tv_usec) / 1000));		    /* Convert the microseconds part to milliseconds */
} /* end function ElapsedTime */



static void XVideoOutGeometryRequest(XVideoOutWidget	widget,	    /* This function is called when the widget attempts to... */
				     XtWidgetGeometry*	geometry)   /* ...move or resize itself.  This is the proper way to... */
{								    /* ...make a geometry change within the widget. */
  XtWidgetGeometry	compromiseGeometry;
  XtGeometryResult	result;
  
  result = XtMakeGeometryRequest((Widget)widget,		    /* Attempt to change the geometry of this widget */
				 geometry,
				 &compromiseGeometry);
  if (XtGeometryAlmost == result)				    /* Was a compromise suggested?  */
  {
    result = XtMakeGeometryRequest((Widget)widget,		    /* Yes, take that compromise */
				   &compromiseGeometry,
				   NULL);			    /* Not interested in compromises at this point */
  }
  return;
} /* end function XVideoOutGeometryRequest */
