/****************************************************************************
 *
 *	VideoOut.c - Define XView interface for Parallax Graphics XLIB extensions
 *
 *	(C) Copyright 1991 by Paradise Software, Inc.
 *
 *	Written by John Malleo-Roach
 *
 *	Version %G% - %I%
 *
 ****************************************************************************/

/* #define TRACE */

/* INCLUDE FILES */

#include	<sys/types.h>
#include	<X11/Xlib.h>
#include	<X11/Xlibint.h>
#include	"XPlxExt.h"
#include	"VideoImpl.h"

/* CONSTANTS */

#define CHAN_INUSE			0xff00

#define DEFAULT_CHANNEL		PLX_OUTPUT_0
#define DEFAULT_STANDARD	PLX_NTSC
#define DEFAULT_FORMAT		PLX_COMP
#define DEFAULT_TYPE		PLX_RGB24

#define DEFAULT_INTERVAL	30
#define DEFAULT_WIDTH		640
#define DEFAULT_HEIGHT		512

#define VIDEO_KEY			0xfe7a

#define VIDEO_MIN_WIDTH		40
#define VIDEO_MIN_HEIGHT	30

/* TYPEDEFS */

/* STATIC VARIABLES */

static int		 VideoOutInit();
static Xv_opaque VideoOutSet();
static Xv_opaque VideoOutGet();
static int		 VideoOutDestroy();

/* GLOBAL VARIABLES */

Xv_pkg video_out_pkg = {
	"VideoOut",
	ATTR_PKG_VIDEO_OUT,
	sizeof(VideoOutPublic),
	XV_GENERIC_OBJECT,
	VideoOutInit,
	VideoOutSet,
	VideoOutGet,
	VideoOutDestroy,
	NULL
};

/* EXTERNAL VARIABLES */

extern plx_IO *PlxBoardSetup;			/* Pointer to hardware setup	*/
extern int Xv_VideoStatus;				/* Storage for error codes		*/

/* FUNCTIONS */

/* Verify good values for some parameters */

/* Translate VIDEO_OUT_FORMAT into a XPlxVideo format */

static int TranslateFormat(value)
int value;
{
	switch (value) {
	case VIDEO_COMPOSITE:
		return(PLX_COMP);
	case VIDEO_SUPER_VHS:
		return(PLX_YC);
	case VIDEO_YUV:
		return(PLX_YUV);
	case VIDEO_RGB:
		return(PLX_RGB);
	default:
		return(-1);
	}
}

/* Translate a XPlxVideo format into a VIDEO_OUT_FORMAT */

static int UnTranslateFormat(value)
int value;
{
	switch (value) {
	case PLX_COMP:
		return(VIDEO_COMPOSITE);
	case PLX_YC:
		return(VIDEO_SUPER_VHS);
	case PLX_YUV:
		return(VIDEO_YUV);
	case PLX_RGB:
		return(VIDEO_RGB);
	default:
		return(0);
	}
}

/* Translate VIDEO_OUT_STANDARD into a XPlxVideo standard */

static int TranslateStandard(value)
int value;
{
	switch (value) {
	case VIDEO_NTSC:
		return(PLX_NTSC);
	case VIDEO_PAL:
		return(PLX_PAL);
	case VIDEO_SECAM:
		return(PLX_SECAM);
	default:
		return(-1);
	}
}

/* Translate a XPlxVideo standard into VIDEO_OUT_STANDARD values */

static int UnTranslateStandard(value)
int value;
{
	switch (value) {
	case PLX_NTSC:
		return(VIDEO_NTSC);
	case PLX_PAL:
		return(VIDEO_PAL);
	case PLX_SECAM:
		return(VIDEO_SECAM);
	default:
		return(0);
	}
}

/* Translate a VIDEO_OUTPUT_TYPE into a XPlxVIdeo output type */

static int TranslateType(value)
int value;
{
	switch (value) {
	case VIDEO_YUV7:
		return(PLX_YUV7);
	case VIDEO_RGB24:
		return(PLX_RGB24);
	case VIDEO_RGB8:
		return(PLX_RGB8);
	case VIDEO_GRAY7:
		return(PLX_GRAY7);
	case VIDEO_GRAY8:
		return(PLX_GRAY8);
	default:
		return(-1);
	}
}

/* Translate a XPlxVideo output type into a VIDEO_OUTPUT_TYPE value */

static int UnTranslateType(value)
int value;
{
	switch (value) {
	case PLX_YUV7:
		return(VIDEO_YUV7);
	case PLX_RGB24:
		return(VIDEO_RGB24);
	case PLX_RGB8:
		return(VIDEO_RGB8);
	case PLX_GRAY7:
		return(VIDEO_GRAY7);
	case PLX_GRAY8:
		return(VIDEO_GRAY8);
	default:
		return(0);
	}
}

/* Translate a VIDEO_IO into a XPlxVideo channel */

static int TranslateChannel(value)
int value;
{
	switch (value) {
	case VIDEO_IO_OFF:
		return(PLX_IO_OFF);
	case VIDEO_IN_ONE:
		return(PLX_INPUT_0);
	case VIDEO_IN_TWO:
		return(PLX_INPUT_1);
	case VIDEO_OUT_ONE:
		return(PLX_OUTPUT_0);
	default:
		return(-1);
	}
}

/* Translate a XPlxVideo channel into a VIDEO_IO value */

static int UnTranslateChannel(value)
int value;
{
	switch (value) {
	case PLX_IO_OFF:
		return(VIDEO_IO_OFF);
	case PLX_INPUT_0:
		return(VIDEO_IN_ONE);
	case PLX_INPUT_1:
		return(VIDEO_IN_TWO);
	case PLX_OUTPUT_0:
		return(VIDEO_OUT_ONE);
	default:
		return(0);
	}
}

/* UnTranslate a XPlxVideo value into a VIDEO_HARDWARE value */

static int UnTranslateHardware(value)
int value;
{
	switch (value) {
	case PLX_1280:
		return(VIDEO_1280);
	case PLX_VIPER:
		return(VIDEO_VIPER);
	case PLX_VIDEOVIEW:
		return(VIDEO_VIDEOVIEW);
	case PLX_XVIDEO_24:
		return(VIDEO_XVIDEO_24);
	case PLX_XVIDEO_8:
		return(VIDEO_XVIDEO_8);
/* This case should never happen */
	default:
		return(0);
	}
}

/* Specific video output functions supported in the XLIB extensions */

static void ShowVideoOutput(vp)
VideoOutPrivate *vp;
{
	switch (PlxBoardSetup->hardware) {
	case PLX_VIPER:
		/* NOP */
		break;
	case PLX_VIDEOVIEW:
		if (vp->vw > vp->width)  vp->vw = vp->width;
		if (vp->vh > vp->height) vp->vh = vp->height;
		if (vp->vx > (vp->width - vp->vw))  vp->vx = (vp->width - vp->vw);	
		if (vp->vy > (vp->height - vp->vh)) vp->vy = (vp->height - vp->vh);
		break;
	case PLX_XVIDEO_8:
	case PLX_XVIDEO_24:
		vp->vx = 0;
		vp->vy = vp->interval;
		vp->vw = vp->width;
		vp->vh = vp->height - vp->interval;
		break;
	}
#ifdef TRACE
	fprintf(stderr,"TRACE: XPlxVideoScaleOutput(sx=%d,sy=%d,vx=%d,vy=%d,vw=%d,vh=%d)\n",
		vp->sx,vp->sy,vp->vx,vp->vy,vp->vw,vp->vh);
#endif
	XPlxVideoScaleOutput(vp->dpy,vp->root_xid,vp->root_gc,vp->sx,vp->sy,vp->vw,vp->vh,
		vp->vx,vp->vy,vp->vw,vp->vh);
}

static void ShowVideoScaleOutput(vp)
VideoOutPrivate *vp;
{

	switch (PlxBoardSetup->hardware) {
	case PLX_VIPER:
		/* NOP */
		break;
	case PLX_VIDEOVIEW:
		if (vp->sw > vp->width) vp->sw = vp->width;
		if (vp->sh > (vp->height + vp->interval)) vp->sh = (vp->height + vp->interval);
		if (vp->vx > (vp->width - vp->sw))  vp->vx = (vp->width - vp->sw);
		if (vp->vy > (vp->height + vp->interval - vp->sh))
			vp->vy = (vp->height + vp->interval - vp->sh);
		break;
	case PLX_XVIDEO_8:
	case PLX_XVIDEO_24:
		vp->vx = 0;
		vp->vy = vp->interval;
		vp->vw = vp->width;
		vp->vh = vp->height - vp->interval;
		if (vp->sw < 1) vp->sw = 1;
		if (vp->sh < 1) vp->sh = 1;
		if (vp->sw > vp->width ) vp->sw = vp->width;
		if (vp->sh > vp->height) vp->sh = vp->height;
		break;
	}
#ifdef TRACE
	fprintf(stderr,"TRACE: XPlxVideoScaleOutput(sx=%d,sy=%d,sw=%d,sh=%d,vx=%d,vy=%d,vw=%d,vh=%d)\n",
		vp->sx,vp->sy,vp->sw,vp->sh,vp->vx,vp->vy,vp->vw,vp->vh);
#endif
	XPlxVideoScaleOutput(vp->dpy,vp->root_xid,vp->root_gc,vp->sx,vp->sy,vp->sw,vp->sh,
		vp->vx,vp->vy,vp->vw,vp->vh);
}

static void VideoOutput(vp)
VideoOutPrivate *vp;
{
	int i,j;

/* Check for normal aspect */

	if (vp->flags & VideoAspectFlag) {
		i = vp->sw / 4;
		j = vp->sh / 3;
		if (i != j) {
			if (j < i) i = j;
			vp->sw = i * 4;
			vp->sh = i * 3;
		}
	}

/* Enforce maximum output size */

	if (vp->sw > vp->width) vp->sw = vp->width;
	if (vp->sh > vp->height) vp->sh = vp->height;

	if (vp->flags & VideoOutFlag) {
		if ((vp->flags & VideoAdjustFlag) == 0) {
			vp->flags |= VideoAdjustFlag;

/* Turn on video if not already on */

			XPlxVideoOutputSelect(vp->dpy,vp->root_xid,vp->root_gc,vp->channel,
				vp->standard,vp->format,vp->output_type);

		}
		if (vp->flags & VideoScaleFlag) {
			ShowVideoScaleOutput(vp);
		}
		else {
			ShowVideoOutput(vp);
		}
		vp->lx = vp->sx; vp->ly = vp->sy; vp->lw = vp->sw; vp->lh = vp->sh;
	}
	else {
		if (vp->flags & VideoAdjustFlag) {
			vp->flags &= ~VideoAdjustFlag;
#ifdef TRACE
			fprintf(stderr,"TRACE: XPlxVideoOutputSelect(PLX_IO_OFF)\n");
#endif
/* Why not this call?
			XPlxVideoStop(vp->dpy,vp->root_xid,vp->root_gc);
*/
			XPlxVideoOutputSelect(vp->dpy,vp->root_xid,vp->root_gc,PLX_IO_OFF,
				vp->standard,vp->format,vp->output_type);
		}
	}
	vp->flags |= VideoCalledFlag;
}

static int VideoOutInit(owner,video_public,avlist)
Xv_opaque     owner;
VideoOutPublic *video_public;
Attr_avlist   avlist;
{
	XGCValues gcv;
	VideoOutPrivate *video_private = xv_alloc(VideoOutPrivate);

	if (!video_private) {
		xv_error((Xv_object) video_public,
			ERROR_STRING,	"Unable to allocate video data structure",
			ERROR_SEVERITY,	ERROR_RECOVERABLE,
			ERROR_PKG,		VIDEO_OUT,
			NULL);
		return(XV_ERROR);
	}

/* Setup inheritance links */

	video_public->private_data = (Xv_opaque) video_private;
	video_private->public_self = (Xv_opaque) video_public;

/* Initialize the defaults for the private variables */

	video_private->flags			= VideoCreateFlag;
	video_private->sx				= 0;
	video_private->sy				= 0;
	video_private->sw				= DEFAULT_WIDTH;
	video_private->sh				= DEFAULT_HEIGHT - DEFAULT_INTERVAL;
	video_private->vx				= 0;
	video_private->vy				= DEFAULT_INTERVAL;
	video_private->vw				= DEFAULT_WIDTH;
	video_private->vh				= DEFAULT_HEIGHT;
	video_private->lx				= 0;
	video_private->ly				= 0;
	video_private->lw				= 0;
	video_private->lh				= 0;
	video_private->width			= DEFAULT_WIDTH;
	video_private->height			= DEFAULT_HEIGHT;
	video_private->interval			= DEFAULT_INTERVAL;
	video_private->standard			= DEFAULT_STANDARD;
	video_private->p_standard		= DEFAULT_STANDARD;
	video_private->format			= DEFAULT_FORMAT;
	video_private->p_format			= DEFAULT_FORMAT;
	video_private->output_type		= DEFAULT_TYPE;
	video_private->p_output_type	= DEFAULT_TYPE;
	video_private->channel			= DEFAULT_CHANNEL;
	video_private->p_channel		= DEFAULT_CHANNEL;
	video_private->dpy				= (Display *) xv_get(xv_default_server,XV_DISPLAY);

/*  Because of Parallax Server requirements, can`t used the root window in
	the X video extension calls, so we must create an unmapped window and
	drag it around with us everywhere (UGLY!).
*/
	video_private->root_xid			= RootWindow(video_private->dpy,
										DefaultScreen(video_private->dpy));
	video_private->root_xid			=
		XCreateSimpleWindow(video_private->dpy,video_private->root_xid,0,0,1,1,1,0,1);

/* Set up our graphics context for a video bounding box */

	gcv.function			= GXinvert;
	gcv.line_width			= 2;
	gcv.line_style			= LineDoubleDash;
	gcv.subwindow_mode		= IncludeInferiors;
	video_private->root_gc	= XCreateGC(video_private->dpy,video_private->root_xid,
		GCFunction | GCLineWidth | GCLineStyle | GCSubwindowMode,&gcv);
	Xv_VideoStatus = VIDEO_OKAY;
	return(XV_OK);
}

/* ReAdjust the requested board IO parameter */

static int AdjustBoardIO(vp)
VideoOutPrivate *vp;
{
	plx_IO_list     *a,*f;
	plx_signal_list *b;
	plx_format_list *c;
	plx_type_list   *d;
	plx_signal		*e;

	Xv_VideoStatus = VIDEO_OKAY;

	if (vp->p_channel != vp->channel) {
		a = PlxBoardSetup->outputs;
		while (a) {
			if (a->channel == vp->channel) break;
			a = a->next_IO;
		}
		if (a == NULL) {
			vp->channel = vp->p_channel;
			Xv_VideoStatus = VIDEO_CHANNEL_IN_USE;
			return(XV_ERROR);
		}
		a->channel |= CHAN_INUSE;
		f = PlxBoardSetup->outputs;
		while (f) {
			if ((f->channel & 0xff) == vp->p_channel) break;
			f = f->next_IO;
		}
		if (f) f->channel &= 0xff;
		vp->p_channel = vp->channel;
	}
	else {
		a = PlxBoardSetup->outputs;
		while (a) {
			if ((a->channel & 0xff) == vp->channel) break;
			a = a->next_IO;
		}
		if (a == NULL) {
			Xv_VideoStatus = VIDEO_NO_MORE_CHANNELS;
			return(XV_ERROR);
		}
	}
	for (b=a->signal_list; b!=NULL; b=b->next_signal) {
		if (b->signal->standard == vp->standard) break;
	}
	if (b == NULL) {
		vp->standard = vp->p_standard;
		Xv_VideoStatus = VIDEO_BAD_STANDARD;
		return(XV_ERROR);
	}
	else vp->p_standard = vp->standard;
	for (c=a->format_list; c!=NULL; c=c->next_format) {
		if (c->format == vp->format) break;
	}
	if (c == NULL) {
		vp->format = vp->p_format;
		Xv_VideoStatus = VIDEO_BAD_FORMAT;
		return(XV_ERROR);
	}
	else vp->p_format = vp->format;
	for (d=a->type_list; d!=NULL; d=d->next_type) {
		if (d->type == vp->output_type) break;
	}
	if (d == NULL) {
		vp->output_type = vp->p_output_type;
		Xv_VideoStatus = VIDEO_BAD_OUTPUT_TYPE;
		return(XV_ERROR);
	}
	else vp->p_output_type = vp->output_type;

/* Set the desired output parameters we want (and should get) */

	XPlxVideoOutputSelect(vp->dpy,vp->root_xid,vp->root_gc,vp->channel,
		vp->standard,vp->format,vp->output_type);

/* See what we really got off the board */

	e = (plx_signal *) XPlxQueryVideo(vp->dpy,vp->root_xid,vp->root_gc);

	vp->sw       = e->w;
	vp->sh       = e->h;
	vp->vw	     = e->w;
	vp->vh	     = e->h;
	vp->width    = e->w;
	vp->height   = e->h;
	vp->interval = e->b;
	vp->standard = e->standard;

	return(XV_OK);
}

/* Called at the end of the initialization process */

static int FinishInit(vp)
VideoOutPrivate *vp;
{
	Xv_Window		 win;
	plx_IO_list     *a;
	plx_signal_list *b;
	plx_format_list *c;
	plx_type_list   *d;
	plx_signal		*e;

	Xv_VideoStatus = VIDEO_OKAY;

/* Determine video configuration and setup if necessary */

	if (PlxBoardSetup == NULL)
		PlxBoardSetup = (plx_IO *) XPlxQueryConfig(vp->dpy,vp->root_xid,vp->root_gc);

	if (PlxBoardSetup->hardware == PLX_UNKNOWN) {
		Xv_VideoStatus = VIDEO_NO_HARDWARE;
		return(XV_ERROR);
	}

/* Setup the channels and format parameters from the QueryConfig call */

start:
	a = PlxBoardSetup->outputs;
loop1:
	if (a == NULL) {
		if (vp->flags & VideoIOFlag) {
			vp->flags &= ~VideoIOFlag;
			a = PlxBoardSetup->outputs;
			while (a) {
				if ((a->channel & 0xff) == vp->channel) break;
				a = a->next_IO;
			}
			if (a) Xv_VideoStatus = VIDEO_CHANNEL_IN_USE;
			else Xv_VideoStatus = VIDEO_NO_MORE_CHANNELS;
			vp->channel = -1;
			return(XV_ERROR);
		}
		else {
			a = PlxBoardSetup->outputs;
			while (a) {
				if ((a->channel & CHAN_INUSE) == 0) break;
				a = a->next_IO;
			}
			if (a) {
				vp->channel = a->channel;
				goto start;
			}
			else {
				Xv_VideoStatus = VIDEO_NO_MORE_CHANNELS;
				vp->channel = -1;
				return(XV_ERROR);
			}
		}
	}
	if (vp->channel == a->channel) {
		for (b=a->signal_list; b!=NULL; b=b->next_signal) {
			if (b->signal->standard == vp->standard) break;

		}
		if (b == NULL) {
			if (vp->flags & VideoIOFlag) {
				vp->flags &= ~VideoIOFlag;
				Xv_VideoStatus = VIDEO_BAD_STANDARD;
				vp->channel = -1;
				return(XV_ERROR);
			}
			else b = a->signal_list;
		}
		for (c=a->format_list; c!=NULL; c=c->next_format) {
			if (c->format == vp->format) break;
		}
		if (c == NULL) {
			if (vp->flags & VideoIOFlag) {
				vp->flags &= ~VideoIOFlag;
				Xv_VideoStatus = VIDEO_BAD_FORMAT;
				vp->channel = -1;
				return(XV_ERROR);
			}
			else c = a->format_list;
		}
		for (d=a->type_list; d!=NULL; d=d->next_type) {
			if (d->type == vp->output_type) break;
		}
		if (d == NULL) {
			if (vp->flags & VideoIOFlag) {
				vp->flags &= ~VideoIOFlag;
				Xv_VideoStatus = VIDEO_BAD_OUTPUT_TYPE;
				vp->channel = -1;
				return(XV_ERROR);
			}
			else d = a->type_list;
		}
	}
	else {
		a = a->next_IO;
		goto loop1;
	}
done:
	vp->channel      = a->channel;
	a->channel      |= CHAN_INUSE;
	vp->standard     = b->signal->standard;
	vp->format       = c->format;
	vp->output_type  = d->type;

/* Set the desired output parameters we want (and should get) */

	XPlxVideoOutputSelect(vp->dpy,vp->root_xid,vp->root_gc,vp->channel,
		vp->standard,vp->format,vp->output_type);

/* See what we really got off the board */

	e = (plx_signal *) XPlxQueryVideo(vp->dpy,vp->root_xid,vp->root_gc);

	vp->sw       = e->w;
	vp->sh       = e->h;
	vp->vw	     = e->w;
	vp->vh	     = e->h;
	vp->width    = e->w;
	vp->height   = e->h;
	vp->standard = e->standard;
	vp->interval = e->b;

	return(XV_OK);
}

static Xv_opaque VideoOutSet(video_public,avlist)
VideoOutPublic *video_public;
Attr_avlist   avlist;
{
	VideoOutPrivate *video_private = VIDEO_OUT_PRIVATE(video_public);
	Attr_attribute *attr;
	int save;

	Xv_VideoStatus = VIDEO_OKAY;

	for (attr = avlist; *attr; attr = attr_next(attr)) {
		switch ((int) attr[0]) {
		case VIDEO_OUTPUT:
			if (attr[1]) video_private->flags |= VideoOutFlag;
			else video_private->flags &= ~VideoOutFlag;
			video_private->flags |= VideoRepaintFlag;
			break;
		case VIDEO_OUT_SCALE:
			if (attr[1]) video_private->flags |= VideoScaleFlag;
			else video_private->flags &= ~VideoScaleFlag;
			video_private->flags |= VideoRepaintFlag;
			break;
		case VIDEO_OUT_ASPECT:
			if (attr[1]) video_private->flags |= VideoAspectFlag;
			else video_private->flags &= ~VideoAspectFlag;
			video_private->flags |= VideoRepaintFlag;
			break;
		case VIDEO_OUT_CHANNEL:
			save = TranslateChannel((int) attr[1]);
			if (save == -1) {
				xv_error((Xv_object) video_public,
					ERROR_STRING,		"Bad VIDEO_OUT_CHANNEL value supplied",
					ERROR_BAD_VALUE,	attr,
					ERROR_PKG,			VIDEO_OUT,
					NULL);
				break;
			}
			video_private->p_channel = video_private->channel;
			video_private->channel   = save;
			if (video_private->p_channel == video_private->channel) break;
			video_private->flags |= VideoIOFlag;
			break;
		case VIDEO_OUT_STANDARD:
			save = TranslateStandard((int) attr[1]);
			if (save == -1) {
				xv_error((Xv_object) video_public,
					ERROR_STRING,		"Bad VIDEO_OUT_STANDARD value supplied",
					ERROR_BAD_VALUE,	attr,
					ERROR_PKG,			VIDEO_OUT,
					NULL);
				break;
			}
			video_private->p_standard = video_private->standard;
			video_private->standard = save;
			if (video_private->p_standard == video_private->standard) break;
			video_private->flags |= VideoIOFlag;
			break;
		case VIDEO_OUT_FORMAT:
			save = TranslateFormat((int) attr[1]);
			if (save == -1) {
				xv_error((Xv_object) video_public,
					ERROR_STRING,		"Bad VIDEO_OUT_FORMAT value supplied",
					ERROR_BAD_VALUE,	attr,
					ERROR_PKG,			VIDEO_OUT,
					NULL);
				break;
			}
			video_private->p_format = video_private->format;
			video_private->format = save;
			if (video_private->p_format == video_private->format) break;
			video_private->flags |= VideoIOFlag;
			break;
		case VIDEO_OUTPUT_TYPE:
			save = TranslateType((int) attr[1]);
			if (save == -1) {
				xv_error((Xv_object) video_public,
					ERROR_STRING,		"Bad VIDEO_OUTPUT_TYPE value supplied",
					ERROR_BAD_VALUE,	attr,
					ERROR_PKG,			VIDEO_OUT,
					NULL);
				break;
			}
			video_private->p_output_type = video_private->output_type;
			video_private->output_type = save;
			if (video_private->p_output_type == video_private->output_type) break;
			video_private->flags |= VideoIOFlag;
			break;
		case VIDEO_OUT_X:
			video_private->sx = (int) attr[1];
			video_private->flags |= VideoRepaintFlag;
			break;
		case VIDEO_OUT_Y:
			video_private->sy = (int) attr[1];
			video_private->flags |= VideoRepaintFlag;
			break;
		case VIDEO_OUT_WIDTH:
			video_private->sw = (int) attr[1];
			video_private->flags |= VideoRepaintFlag;
			break;
		case VIDEO_OUT_HEIGHT:
			video_private->sh = (int) attr[1];
			video_private->flags |= VideoRepaintFlag;
			break;
		case XV_END_CREATE:
			video_private->flags &= ~VideoCreateFlag;
			return(FinishInit(video_private));
			break;

/* These attribute(s) are not settable */

		case VIDEO_HARDWARE:
			xv_error((Xv_object) video_public,
				ERROR_STRING,		"Video hardware type is read only!",
				ERROR_CANNOT_SET,	attr,
				ERROR_PKG,			VIDEO_OUT,
				NULL);
			break;
		default:
			xv_check_bad_attr(VIDEO_OUT,attr[0]);
			break;
		}
	}

/* If we are still in the create pass, then skip the rest */

	if (video_private->flags & VideoCreateFlag) return(XV_OK);

/* If the output parameters have changed, try to do it */

	if (video_private->flags & VideoIOFlag) {
		video_private->flags &= ~VideoIOFlag;
		save = AdjustBoardIO(video_private);
		if (save != VIDEO_OKAY) return(save);
	}

/* If any parameter has changed that requires a redislay, do it */

	if (video_private->flags & VideoRepaintFlag) {
		video_private->flags &= ~VideoRepaintFlag;
		VideoOutput(video_private);
	}

	return(XV_OK);
}

static Xv_opaque VideoOutGet(video_public,status,attr,avlist)
VideoOutPublic  *video_public;
int 		  *status;
Attr_attribute attr;
Attr_avlist    avlist;
{
	plx_signal *s;
	VideoOutPrivate *video_private = VIDEO_OUT_PRIVATE(video_public);

	Xv_VideoStatus = VIDEO_OKAY;

	switch ((int) attr) {
	case VIDEO_OUTPUT:
		if (video_private->flags & VideoOutFlag) return(TRUE);
		else return(FALSE);
		break;
	case VIDEO_OUT_SCALE:
		if (video_private->flags & VideoScaleFlag) return(TRUE);
		else return(FALSE);
		break;
	case VIDEO_OUT_ASPECT:
		if (video_private->flags & VideoAspectFlag) return(TRUE);
		else return(FALSE);
		break;
	case VIDEO_OUT_STANDARD:
		return(UnTranslateStandard(video_private->standard));
		break;
	case VIDEO_OUT_FORMAT:
		return(UnTranslateFormat(video_private->format));
		break;
	case VIDEO_OUTPUT_TYPE:
		return(UnTranslateType(video_private->output_type));
		break;
	case VIDEO_OUT_CHANNEL:
		return(UnTranslateChannel(video_private->channel));
		break;
	case VIDEO_OUT_X:
		return(video_private->sx);
		break;
	case VIDEO_OUT_Y:
		return(video_private->sy);
		break;
	case VIDEO_OUT_WIDTH:
		return(video_private->sw);
		break;
	case VIDEO_OUT_HEIGHT:
		return(video_private->sh);
		break;
	case VIDEO_HARDWARE:
		return(UnTranslateHardware(PlxBoardSetup->hardware));
		break;

/* These attributes can't be gotten */

	default:
		*status = xv_check_bad_attr(VIDEO_OUT,attr);
		return(XV_OK);
		break;
	}
}

static int VideoOutDestroy(video_public,status)
VideoOutPublic  *video_public;
Destroy_status status;
{
	plx_IO_list *a;
	VideoOutPrivate *video_private = VIDEO_OUT_PRIVATE(video_public);

	if ((status == DESTROY_CLEANUP) || (status == DESTROY_PROCESS_DEATH)) {
#ifdef TRACE
		fprintf(stderr,"TRACE: video_destroy called on video output\n");
#endif
		if (video_private->flags & VideoAdjustFlag) {
			;
		}
		if (video_private->flags & VideoCalledFlag) {
/* Why not this call?
			XPlxVideoStop(video_private->dpy,video_private->root_xid,
				video_private->root_gc);
*/
			XPlxVideoOutputSelect(video_private->dpy,video_private->root_xid,
				video_private->root_gc,PLX_IO_OFF,video_private->standard,
				video_private->format,video_private->output_type);
			XFreeGC(video_private->dpy,video_private->root_gc);
		}
		a = PlxBoardSetup->outputs;
		while (a) {
			if ((a->channel & 0xff) == video_private->channel) {
				a->channel &= 0xff;
				break;
			}
			a = a->next_IO;
		}
		free(video_private);
	}
	return(XV_OK);
}

/* end of VideoOut.c */
